
AC_DEFUN(LAC_LDAP_ARGS,
[

AC_CACHE_CHECK(LDAP Root Host, ldap_cv_root_host, [dnl
    ldap_cv_root_host="www.globus.org"
])

AC_CACHE_CHECK(LDAP Root Port, ldap_cv_root_port, [dnl
    ldap_cv_root_port="389"
])

AC_CACHE_CHECK(LDAP Root DN, ldap_cv_root_dn, [dnl
    ldap_cv_root_dn="o=Globus, c=US"
])


AC_CACHE_CHECK(LDAP Host, ldap_cv_local_host, [dnl
AC_ARG_WITH(ldap-host,
	[  --with-ldap-host=value           Name of the local LDAP host],
	[ldap_cv_local_host="$withval"],
	[ldap_cv_local_host="$ldap_cv_root_host"])
])


AC_CACHE_CHECK(LDAP Port, ldap_cv_local_port, [dnl
AC_ARG_WITH(ldap-port,
	[  --with-ldap-port=value           Port number for the local LDAP host],
	[ldap_cv_local_port="$withval"],
	[ldap_cv_local_port="$ldap_cv_root_port"])
])


AC_CACHE_CHECK(LDAP Local Base DN, ldap_cv_local_base_dn, [dnl
AC_ARG_WITH(ldap-base-dn,
	[  --with-ldap-base-dn=value        DN of the local site],
	[ldap_cv_local_base_dn="$withval"],
	[ldap_cv_local_base_dn="$mds_cv_root_dn"])
])






AC_CACHE_CHECK(LDAP Debugging, nx_cv_ldap_debug, [dnl
AC_ARG_ENABLE(ldap-debug,
	[  --enable-ldap-debug     enable ldap debugging output],
	[nx_cv_ldap_debug="$enableval"],
	[nx_cv_ldap_debug="yes"])
])

AC_CACHE_CHECK(U of M, nx_cv_uofm, [dnl
AC_ARG_ENABLE(uofm,
	[  --enable-uofm           enable U of Michigan Extensions LDAP],
	[nx_cv_uofm="$enableval"],
	[nx_cv_uofm="no"])
])

AC_CACHE_CHECK(User Interface, nx_cv_userinterface, [dnl
AC_ARG_ENABLE(userinterface,
	[  --enable-userinterface  enable user interface from LDAP library],
	[nx_cv_userinterface="$enableval"],
	[nx_cv_userinterface="yes"])
])

AC_CACHE_CHECK(Connectionless LDAP, nx_cv_cldap, [dnl
AC_ARG_ENABLE(ldap-referrals,
	[  --enable-cldap          enable connectionless LDAP],
	[nx_cv_cldap="$enableval"],
	[nx_cv_cldap="no"])
])

AC_CACHE_CHECK(LDAP referrals, nx_cv_ldap_referrals, [dnl
AC_ARG_ENABLE(ldap-referrals,
	[  --enable-ldap-referrals enable referrals in LDAP],
	[nx_cv_ldap_referrals="$enableval"],
	[nx_cv_ldap_referrals="no"])
])

AC_CACHE_CHECK(LDAP Cache, nx_cv_cache, [dnl
AC_ARG_ENABLE(ldap-cache,
	[  --enable-cache          enable local caching in LDAP library],
	[nx_cv_cache="$enableval"],
	[nx_cv_cache="yes"])
])

])



AC_DEFUN(LAC_LDAP_COMP_OPTIONS,
[

if test "x$prefix" = xNONE; then
   prefix=$ac_default_prefix
   eval etcdir=$sysconfdir
   prefix=NONE
else
   eval etcdir=$sysconfdir
fi


dnl NO -64 if its an IRIX

case "$target" in
  *sunos4* | *solaris1* ) PLATFORMCFLAGS="-Dsunos4" ;;
  *sunos5* | *solaris2* ) PLATFORMCFLAGS="-Dsunos5 -D_REENTRANT" ;;
  *-ibm-aix4* )           PLATFORMCFLAGS="-Daix"  ;;
  *-*-hpux* )             PLATFORMCFLAGS="-Dhpux" ;;
  mips-sgi-irix6* )       PLATFORMCFLAGS="-DUSE_WAITPID"  
        CFLAGS="`echo $CFLAGS | sed -e 's/-64//'` -n32"
        LDFLAGS="`echo $LDFLAGS | sed -e 's/-64//'` -n32"
    ;;
  * )
	echo "platform not configured"
  ;;
esac

ALLFLAGS="-DLDAP_DEBUG -DLDAP_REFERRALS -DLDAP_LDBM -DLDAP_SHELL -DLDAP_PASSWD -DLDBM_USE_NDBM"

PLATFORMCFLAGS="$PLATFORMCFLAGS $ALLFLAGS $CFLAGS"
PLATFORMLDFLAGS="$LDFLAGS"

AC_SUBST(PLATFORMCFLAGS)
AC_SUBST(PLATFORMLDFLAGS)

])


AC_DEFUN(LAC_LDAP_PROGS,
[

AC_PATH_PROG(BASENAME, basename)
AC_PATH_PROG(DIRNAME, dirname)
AC_PATH_PROG(PWD, pwd)
AC_PATH_PROG(LN, ln)

])





AC_DEFUN(LAC_LDAP,
[

LAC_LDAP_ARGS
LAC_LDAP_PROGS


dnl Checks for header files.
AC_HEADER_STDC
AC_HEADER_TIME
AC_CHECK_HEADERS(fcntl.h malloc.h sys/file.h sys/ioctl.h sys/time.h unistd.h sysexits.h)

dnl Checks for typedefs, structures, and compiler characteristics.
AC_TYPE_SIZE_T
AC_STRUCT_TM
 
dnl Checks for library functions.
AC_FUNC_MEMCMP
AC_CHECK_FUNCS(re_comp select socket strdup)




dnl LDAP library specific stuff.
AC_DEFINE_UNQUOTED(LDAPHOST,"$ldap_cv_local_host")
AC_DEFINE_UNQUOTED(DEFAULT_BASE,"$ldap_cv_local_base_dn")
AC_DEFINE_UNQUOTED(ETCDIR, "$etcdir")


if test "$nx_cv_ldap_debug" = "yes"; then 
  AC_DEFINE(LDAP_DEBUG)
fi

if test "$nx_cv_uofm" = "yes"; then
  AC_DEFINE(UOFM)
fi

if test "$nx_cv_userinterface" = "no"; then
  AC_DEFINE(NO_USERINTERFACE)
fi

if test "$nx_cv_cldap" = "yes"; then
  AC_DEFINE(CLDAP)
fi

if test "$nx_cv_ldap_referrals" = "yes"; then
  AC_DEFINE(LDAP_REFERRALS)
fi

if test "$nx_cv_cache" = "no"; then
  AC_DEFINE(NO_CACHE)
fi


LAC_LDAP_COMP_OPTIONS

])





