/*
************************************************************************
**                                                                    **
**  hbm_monitor_proc.c - source file for external                     **
**                       heartbeat client routines                    **
**                                                                    **
************************************************************************
*/
#define HBM_MONITOR_PROC_C

#include "hbm.h"
#include "hbm_client.h"

int         hbm_flags;

struct hbm_parms
            hbm_parm_info;


void    usage(  char   *pgm_name)
{
    fprintf(stderr, "Usage: %s", pgm_name);
    fprintf(stderr, " [-r | -u | -ua | -un] [-n xxxx]\n");
    fprintf(stderr, " [-d 999.999.999.999 9999 | -h xxx.xxx.xxx 9999] 99999\n");
    fprintf(stderr, "       where: -r      = register\n");
    fprintf(stderr, "              -u      = unregister normal\n");
    fprintf(stderr, "              -ua     = unregister abnormal\n");
    fprintf(stderr, "              -un     = unregister normal\n");
    fprintf(stderr, "              -n xxxx = name xxxx to be used by LM when reporting to DC\n");
    fprintf(stderr, "              -d 999.999.999.999 9999 =\n");
    fprintf(stderr, "                        data collector address, with:\n");
    fprintf(stderr, "                            999.999.999.999: IP Number of host\n");
    fprintf(stderr, "                            9999: port number\n");
    fprintf(stderr, "              -h xxx.xxx.xxx 9999 =\n");
    fprintf(stderr, "                        data collector address, with:\n");
    fprintf(stderr, "                            xxx.xxx.xxx: fully qualified host name\n");
    fprintf(stderr, "                            9999: port number\n");
    fprintf(stderr, "              99999   = decimal pid\n");
    fprintf(stderr, "\n");
    exit(1);
}


main(   int     argc,
        char   *argv[])
{
    int     ck_argc,
            error;
    char    reg_option,
            shutdown_option;

/*
**  Set hbm_HBClient_host_name_ptr for error messages.
*/
    if  (   (argc < 2)
         || (argc > 8))
    {
        usage(*argv);
        return 0;
    }

    reg_option = ' ';
    shutdown_option = ' ';
    hbm_flags = 0;
    bzero((char *) &hbm_parm_info, sizeof(struct hbm_parms));

    ck_argc = 1;
    while (ck_argc < argc)
    {
/*
**      Get and validate parameters.
*/
        if  (*(argv[ck_argc]) == '-')
        {
/*
**          Optional parameters.
*/
            if  (argv[ck_argc][1] == '\0')
            {
                usage(*argv);
                return -1;
            }
            if  (argv[ck_argc][1] == 'r')
            {
                if  (reg_option != ' ')
                {
                    usage(*argv);
                    return -1;
                }
                reg_option = 'R';
            }
            else if  (argv[ck_argc][1] == 'u')
            {
                if  (reg_option != ' ')
                {
                    usage(*argv);
                    return -1;
                }
                reg_option = 'U';
                if  (   (argv[ck_argc][2] != '\0')
                     && (argv[ck_argc][2] != 'a')
                     && (argv[ck_argc][2] != 'n'))
                {
                    usage(*argv);
                    return -1;
                }
                if  (   (argv[ck_argc][2] == '\0')
                     || (argv[ck_argc][2] == 'n'))
                {
                    shutdown_option = 'N';
                }
                else
                {
                    shutdown_option = 'A';
                }
            }
            else if  (argv[ck_argc][1] == 'n')
            {
                if  (ck_argc + 1 >= argc)
                {
                    usage(*argv);
                    return -1;
                }
                else if  (hbm_flags & HBM_CLFLAG_SET_CL_RPTNAME)
                {
                    usage(*argv);
                    return -1;
                }
                hbm_flags |= HBM_CLFLAG_SET_CL_RPTNAME;
                hbm_parm_info.cl_procname_rpt = argv[ck_argc + 1];
                ck_argc++;
            }
            else if  (argv[ck_argc][1] == 'd')
            {
                if  (ck_argc + 2 >= argc)
                {
                    usage(*argv);
                    return -1;
                }
                else if  (   (hbm_flags & HBM_CLFLAG_SET_DC_HOSTNAME)
                          || (hbm_flags & HBM_CLFLAG_SET_DC_HOSTIPNUM))
                {
                    usage(*argv);
                    return -1;
                }
                hbm_flags |= HBM_CLFLAG_SET_DC_HOSTIPNUM;
                hbm_flags |= HBM_CLFLAG_SET_DC_PORT;
                hbm_parm_info.dc_hostipnum = argv[ck_argc + 1];
                error = sscanf(argv[ck_argc + 2], "%u", &(hbm_parm_info.dc_port));
                if  (   (error != 1)
                     || (hbm_parm_info.dc_port == 0))
                {
                    usage(*argv);
                    return -1;
                }
                ck_argc += 2;
            }
            else if  (argv[ck_argc][1] == 'h')
            {
                if  (ck_argc + 2 >= argc)
                {
                    usage(*argv);
                    return -1;
                }
                else if  (   (hbm_flags & HBM_CLFLAG_SET_DC_HOSTNAME)
                          || (hbm_flags & HBM_CLFLAG_SET_DC_HOSTIPNUM))
                {
                    usage(*argv);
                    return -1;
                }
                hbm_flags |= HBM_CLFLAG_SET_DC_HOSTNAME;
                hbm_flags |= HBM_CLFLAG_SET_DC_PORT;
                hbm_parm_info.dc_hostname = argv[ck_argc + 1];
                error = sscanf(argv[ck_argc + 2], "%u", &(hbm_parm_info.dc_port));
                if  (   (error != 1)
                     || (hbm_parm_info.dc_port == 0))
                {
                    usage(*argv);
                    return -1;
                }
                ck_argc += 2;
            }
            else
            {
                usage(*argv);
                return -1;
            }
        }
        else
        {
/*
**          Get the (mandatory) parameter -- pid.
*/
            if  (hbm_flags & HBM_CLFLAG_SET_CL_PID)
            {
                usage(*argv);
                return -1;
            }
            hbm_flags |= HBM_CLFLAG_SET_CL_PID;
            error = sscanf(argv[ck_argc], "%d", &(hbm_parm_info.cl_pid));
            if  (   (error != 1)
                 || (hbm_parm_info.cl_pid == 0))
            {
                usage(*argv);
                return -1;
            }
        }
        ck_argc++;
    }

    if  (!(hbm_flags & HBM_CLFLAG_SET_CL_PID))
    {
        usage(*argv);
        return -1;
    }
    if  (reg_option == ' ')
    {
        reg_option = 'R';
    }

    if  (reg_option == 'R')
    {
        error = hbm_register(*argv,
                             hbm_flags,
                             &hbm_parm_info,NULL);
    }
    else
    {
    	if  (shutdown_option == 'N')
    	{
            error = hbm_shutdown(*argv,
                                 hbm_flags,
                                 &hbm_parm_info,
                                 HBM_MSGTYPE_SHUTDOWN_NORMAL);
    	}
    	else
    	{
            error = hbm_shutdown(*argv,
                                 hbm_flags,
                                 &hbm_parm_info,
                                 HBM_MSGTYPE_SHUTDOWN_ABNORMAL);
    	}
    }

    if  (error)
    {
        fprintf(stderr,
                "\n%s: Heartbeat processing not done\n",
                *argv);
        return (-4000 + error);
    }

    return 0;
}

/*
************************************************************************
**                                                                    **
**  hbm_monitor_proc.c - end                                          **
**                                                                    **
************************************************************************
*/
