/*
************************************************************************
**                                                                    **
**  hbmcl.c - source file for heartbeat client routines               **
**                                                                    **
************************************************************************
*/
#define HBM_CL_C

#include "hbm.h"
#include "hbm_cllm.h"
#include "hbm_common.h"
#include "hbm_parms.h"

#ifndef SYSCONFDIR
#define SYSCONFDIR "."
#endif  /*  SYSCONFDIR  */

/* STUFF I'M ADDING TO MAKE THIS LOSER COMPILE */
/* Commented out to compile under solaris
typedef int ssize_t;
*/

int     hbm_doing_heartbeat_processing = 1;

int     hbm_HBClient_host_name_len;
char   *hbm_HBClient_host_name_ptr;

char   *hbm_error_procname_str;

char   *hbm_client_procname_str;

char   *hbm_client_procname_rpt_str;

int     hbm_HBClient_PID;

int     hbm_initialization_done = 0;

struct sockaddr_in
        hbm_HBMLM_addr_reg;

char    hbm_HBMDC_host_addr_str[HBM_IP_ADDRESS_SIZE + 1];
int     hbm_HBMDC_portNum;

int     hbm_Client_reg_HBMLM_time;

/***** NETSOLVE STUFF *****/
char agent_hostname[256];


char   *hbm_copy_procname(  char   *procname_str)
{
    char   *procname_cpy_str;

    if  (procname_str == NULL)
    {
        return NULL;
    }
    procname_cpy_str = (char *)malloc(strlen(procname_str) + 1);
    if  (procname_cpy_str == NULL)
    {
        return NULL;
    }
    strcpy(procname_cpy_str, procname_str);
    return procname_cpy_str;
}


char   *hbm_get_procname_fr_pid(    int     procpid)
{
    int     pipe_fd[2];

    int     pid_fork,
            pid_wait,
            status;

    int     ps_output_pid;
    char   *ps_output_procname_str;

    int     ps_read;
    char   *ps_buff_str,
           *ps_buff_ptr_wk;

    int     retcd;

    if  (retcd = pipe(pipe_fd))
    {
        fprintf(stderr,
                "\n%s: Error %d creating pipe for ps to get HBMCL process name\n",
                hbm_error_procname_str,
                retcd);
        perror(hbm_error_procname_str);
        return NULL;
    }
    if  ((pid_fork = fork()) < 0)
    {
        fprintf(stderr,
                "\n%s: Error %d forking child for ps to get HBMCL process name",
                hbm_error_procname_str,
                pid_fork);
        perror(hbm_error_procname_str);
        close(pipe_fd[0]);
        close(pipe_fd[1]);
        return NULL;
    }
    if  (pid_fork == 0)
    {
/*
**      Child Code
**
**      Close stdout and the read pipe fd,
**      then dup the write pipe fd to 1 (stdout) and close it.
**      Finally, exec ps with the appropriate parms.
*/
        close(1);
        close(pipe_fd[0]);
        if  ((retcd = dup2(pipe_fd[1],1)) < 0)
        {
            fprintf(stderr,
                    "\nChild of %s: Error %d duping write pipe for ps to get HBMCL process name",
                    hbm_error_procname_str,
                    retcd);
            perror(hbm_error_procname_str);
            close(pipe_fd[1]);
            exit(0);
        }
        close(pipe_fd[1]);
        ps_buff_str = (char *) malloc(32);
        retcd = sprintf(ps_buff_str, "%d", procpid);

#if (   (defined TARGET_ARCH_AIX) \
     || (defined TARGET_ARCH_IRIX) \
     || (defined TARGET_ARCH_SOLARIS))

        retcd = execl("/bin/ps",
                      "ps",
                      "-p",
                      ps_buff_str,
                      "-o",
                      "pid comm",
                      (char *)0);
#elif  (defined TARGET_ARCH_HPUX)
        retcd = execl("/bin/ps",
                      "ps",
                      "-p",
                      ps_buff_str,
                      (char *)0);
#elif  ((defined TARGET_ARCH_SUNOS)) 
        retcd = execl("/bin/ps",
                      "ps",
                      ps_buff_str,
                      (char *)0);
#else
        retcd = execl("/bin/ps",
                      "ps",
                      "-p",
                      ps_buff_str,
                      "-o",
                      "pid comm",
                      (char *)0);
#endif  /*  TARGET_ARCH_xxxx defined  */

        fprintf(stderr,
                "\nChild of %s: Error %d on execl of \"ps\" failed\n",
                hbm_error_procname_str,
                retcd);
        perror(hbm_error_procname_str);
        exit(0);
/*
**      End of Child Code.
*/
    }
/*
**  Parent Code from here...
*/
    close(pipe_fd[1]);

    ps_buff_str = (char *)malloc(HBM_BUFF_SIZE_PS);
    ps_output_procname_str = (char *)malloc(256);

    if  (hbm_read_line(pipe_fd[0],
                       ps_buff_str,
                       HBM_BUFF_SIZE_PS) > 0)
    {
        while ((ps_read = hbm_read_line(pipe_fd[0],
                                        ps_buff_str,
                                        HBM_BUFF_SIZE_PS)) > 0)
        {
            ps_output_pid = 0;
            ps_output_procname_str[0] = '\0';

/*
**          Parse the status line:
**              PID first
*/
/*
**          ps_oputput_status_ptr, cpu seconds, minutes, ps_output_procname
*/
#if (   (defined TARGET_ARCH_AIX) \
     || (defined TARGET_ARCH_IRIX) \
     || (defined TARGET_ARCH_SOLARIS))
            sscanf(ps_buff_str,
                   "%d %[ -z]\n",
                   &ps_output_pid,
                   ps_output_procname_str);
#elif  (   (defined TARGET_ARCH_HPUX) \
        || (defined TARGET_ARCH_SUNOS))
/*
**          Get the PID, and skip the pointer to the command.
*/
            ps_buff_ptr_wk = ps_buff_str;
            sscanf(ps_buff_ptr_wk, "%d", &ps_output_pid);

            if  ((ps_buff_ptr_wk = index(ps_buff_ptr_wk, ':')) == NULL)
            {
                ps_buff_ptr_wk = ps_buff_str;
                *ps_buff_ptr_wk = '\0';
            }
            while (   (*ps_buff_ptr_wk != '\0')
                   && (!isspace(*ps_buff_ptr_wk)))
            {
                ps_buff_ptr_wk++;
            }
            while (   (*ps_buff_ptr_wk != '\0')
                   && (isspace(*ps_buff_ptr_wk)))
            {
                ps_buff_ptr_wk++;
            }
/*
**              Get the command.
*/
            sscanf(ps_buff_ptr_wk,
                   "%[ -z]",
                   ps_output_procname_str);
#else
            sscanf(ps_buff_str,
                   "%d %[ -z]\n",
                   &ps_output_pid,
                   ps_output_procname_str);
#endif  /* TARGET_ARCH_xxxx defined */
            if  (ps_output_pid == procpid)
            {
                break;
            }
        }
        if  (ps_read)
        {
            while ((ps_read = hbm_read_line(pipe_fd[0],
                                            ps_buff_str,
                                            HBM_BUFF_SIZE_PS)) > 0);
        }
    }
    close(pipe_fd[0]);
    pid_wait = wait(&status);

    if  (ps_output_pid == 0)
    {
        free(ps_buff_str);
        free(ps_output_procname_str);
        return NULL;
    }
    ps_buff_ptr_wk = (char *)malloc(strlen(ps_output_procname_str) + 1);
    strcpy(ps_buff_ptr_wk, ps_output_procname_str);
    free(ps_buff_str);
    free(ps_output_procname_str);

    return ps_buff_ptr_wk;
}


/*
************************************************************************
**                                                                    **
**  hbmcl_get_machine_info                                            **
**                                                                    **
************************************************************************
*/
int     hbmcl_get_machine_info( int                    *host_name_len_ptr,
                                char                  **host_name_str_ptr,
                                struct sockaddr_in     *host_addr_reg_ptr)
{
    int     retcd;
    char   *str_ptr1,
           *str_ptr2;

/*
**  Get machine-specific information:
*/
/*
**  Get host name (and name length) and IP Number.
*/
    *host_name_str_ptr = (char *) malloc(MAXHOSTNAMELEN + 256);
    bzero(*host_name_str_ptr, MAXHOSTNAMELEN + 256);
    if  (gethostname(*host_name_str_ptr, MAXHOSTNAMELEN + 256) < 0)
    {
        fprintf(stderr,
                "\n%s: HBM: Error getting host name\n",
                hbm_error_procname_str);
        perror(hbm_error_procname_str);
        return -1;
    }
    retcd = hbm_get_IPNum_fr_name(&(host_addr_reg_ptr->sin_addr),
                                  *host_name_str_ptr,
                                  host_name_str_ptr);

    if  (host_addr_reg_ptr->sin_addr.s_addr == 0)
    {
        fprintf(stderr,
                "\n%s: HBM: Bad host address\n",
                hbm_error_procname_str);
        perror(hbm_error_procname_str);
        return -3;
    }

    *host_name_len_ptr = strlen(*host_name_str_ptr);

    return 0;
}


/*
************************************************************************
**                                                                    **
**  hbm_get_HBClient_parm_val                                         **
**                                                                    **
************************************************************************
*/
int     hbm_get_HBClient_parm_val(  char   *parm_str,
                                    char   *HBMDC_host_addr_str,
                                    int    *HBMDC_port_num_ptr)
{
    int     error;
    int     got_val;

    struct sockaddr_in
            test_addr;
    char   *addr_str,
           *parm_ptr;

    unsigned int    read_u_val;

    error = 0;
    if  (!strncmp(parm_str, "HBMDCHostname", 13))
    {
/*      Host name of HBMDC that the HBMLM should report our status to   */
        parm_str += 13;
            /****************************************************/
            /******                                         *****/
            /******  ADDED BY H. CASANOVA                   *****/
            /******  UGLY, UGLY, UGLY                       *****/
            /******  MUST BE A BETTER WAY                   *****/
            /******                                         *****/
            /****************************************************/
        if (agent_hostname != NULL)
          parm_str = agent_hostname;

        if  (*parm_str == ':')
        {
            parm_str++;
        }
        while (isspace(*parm_str))
        {
            parm_str++;
        }
        if  (*parm_str == '\0')
        {
            return -1;
        }
        parm_ptr = parm_str;
        while (   (!isspace(*parm_ptr))
               && (*parm_ptr != '\0'))
        {
            parm_ptr++;
        }
        *parm_ptr = '\0';
        hbm_get_IPNum_fr_name(&(test_addr.sin_addr),
                              parm_str,
                              NULL);
        if  (test_addr.sin_addr.s_addr == 0)
        {
            fprintf(stderr,
                    "\n%s: HBM: HBMDCHostname parameter invalid/not found\n",
                    hbm_error_procname_str);
            return -1;
        }
        else
        {
            addr_str = inet_ntoa(test_addr.sin_addr);
            strncpy(HBMDC_host_addr_str, addr_str, HBM_IP_ADDRESS_SIZE);
        }
    }
    else if  (!strncmp(parm_str, "HBMDCIPNum", 10))
    {
/*      IPNum of HBMDC that the HBMLM should report our status to   */
        parm_str += 10;
        if  (*parm_str == ':')
        {
            parm_str++;
        }
        while (isspace(*parm_str))
        {
            parm_str++;
        }
        if  (*parm_str == '\0')
        {
            return -1;
        }
        test_addr.sin_addr.s_addr = inet_addr(parm_str);
        if  (test_addr.sin_addr.s_addr == 0)
        {
            fprintf(stderr,
                    "\n%s: HBM: HBMDCIPNum parameter invalid\n",
                    hbm_error_procname_str);
            perror(hbm_error_procname_str);
            return -1;
        }
        else
        {
            addr_str = inet_ntoa(test_addr.sin_addr);
            strncpy(HBMDC_host_addr_str, addr_str, HBM_IP_ADDRESS_SIZE);
        }
    }
    else if  (!strncmp(parm_str, "HBMDCPortNum", 12))
    {
/*      Port Number of HBMDC that the HBMLM should report our status to */
        parm_str += 12;
        if  (*parm_str == ':')
        {
            parm_str++;
        }
        got_val = sscanf(parm_str, "%u", &read_u_val);
        if  (   (!got_val)
             || (read_u_val == 0))
        {
            fprintf(stderr,
                    "\n%s: HBM: HBMDCPortNumCmd parameter invalid\n",
                    hbm_error_procname_str);
            return -1;
        }
        *HBMDC_port_num_ptr = ntohs((unsigned short) read_u_val);
        
        if  (*HBMDC_port_num_ptr == 0)
        {
            fprintf(stderr,
                    "\n%s: HBM: HBMDCPortNumCmd parameter invalid\n",
                    hbm_error_procname_str);
            perror(hbm_error_procname_str);
            return 1;
        }
    }

    return 0;
}


/*
************************************************************************
**                                                                    **
**  hbm_get_HBClient_parms                                            **
**                                                                    **
************************************************************************
*/
int     hbm_get_HBClient_parms( char   *parm_fname_str,
                                char   *HBMDC_host_addr_str,
                                int    *HBMDC_port_num_ptr)
{
    FILE   *parm_fp_ptr;

    char   *parm_fname_dflt1_str = "./hbmcl.conf";
    char   *parm_fname_dflt2_str = SYSCONFDIR "/hbmcl.conf";

    char   *parms_str,
           *parms_wk_str;
    int     any_error,
            error;

    any_error = 0;

    parms_str = (char *) malloc(HBM_BUFF_SIZE_PARM);
    *(parms_str + HBM_BUFF_SIZE_PARM - 1) = '\0';

    if  (parm_fname_str != NULL)
    {
        if  ((parm_fp_ptr = fopen(parm_fname_str, "r")) == NULL)
        {
            fprintf(stderr,
                    "\n%s: Error opening hbmcl parameter file: %s\n",
                    hbm_error_procname_str,
                    parm_fname_str);
            fprintf(stderr,
                    "\n%s:                                   : %s\n",
                    hbm_error_procname_str,
                    parm_fname_str);
            return -10;
        }
    }
    else
    {
        if  ((parm_fp_ptr = fopen(parm_fname_dflt1_str, "r")) == NULL)
        {
            if  ((parm_fp_ptr = fopen(parm_fname_dflt2_str, "r")) == NULL)
            {
                fprintf(stderr,
                        "\n%s: Error opening heartbeat parameter files: %s",
                        hbm_error_procname_str,
                        parm_fname_dflt1_str);
                fprintf(stderr,
                        "\n%s:                              : %s\n",
                        hbm_error_procname_str,
                        parm_fname_dflt2_str);
                return -20;
            }
        }
    }

    while (fgets(parms_str, HBM_BUFF_SIZE_PARM, parm_fp_ptr) != NULL)
    {
        if  ((parms_wk_str = index(parms_str, '#')) != NULL)
        {
            *parms_wk_str = '\0';
        }
        parms_wk_str = parms_str;
        while (isspace(*parms_wk_str))
        {
            parms_wk_str++;
        }
        if  (*parms_wk_str != '\0')
        {
            if  (error = hbm_get_HBClient_parm_val(parms_wk_str,
                                          HBMDC_host_addr_str,
                                          HBMDC_port_num_ptr))
            {
                any_error = -30 + error;
            }
        }
    }

    fclose(parm_fp_ptr);
    free(parm_fp_ptr);

    free(parms_str);

    if  (*HBMDC_port_num_ptr == 0)
    {
        fprintf(stderr,
                "\n%s: Heartbeat client parameters not present or invalid\n",
                hbm_error_procname_str);
        return -40;
    }

    return any_error;
}


/*
************************************************************************
**                                                                    **
**  hbm_get_HBMLM_parm_val                                            **
**                                                                    **
************************************************************************
*/
int     hbm_get_HBMLM_parm_val( char               *parm_str,
                                struct sockaddr_in *HBMLM_addr_reg_ptr)
{
    int             error;
    int             got_val;

    unsigned int    read_u_val;

    error = 0;
    if  (!strncmp(parm_str, "LocalMonitorPortNumReg", 22))
    {
/*      Port Number for Command Comm from client to HBMLM     */
        parm_str += 22;
        if  (*parm_str == ':')
        {
            parm_str++;
        }
        while (isspace(*parm_str))
        {
            parm_str++;
        }
        sscanf(parm_str, "%u", &read_u_val);
        HBMLM_addr_reg_ptr->sin_port = ntohs((unsigned short) read_u_val);
        
        if  (HBMLM_addr_reg_ptr->sin_port == 0)
        {
            fprintf(stderr,
                    "\n%s: HBM: LMPortNumReg parameter invalid\n",
                    hbm_error_procname_str);
            perror(hbm_error_procname_str);
            return -1;
        }
    }

    return 0;
}


/*
************************************************************************
**                                                                    **
**  hbm_get_HBMLM_parms                                               **
**                                                                    **
************************************************************************
*/
int     hbm_get_HBMLM_parms(    struct sockaddr_in *HBMLM_addr_reg_ptr)
{
    FILE   *parm_fp_ptr;

    char   *parm_fname_dflt1_str = "./hbmlm.conf",
           *parm_fname_dflt2_str = SYSCONFDIR "/hbmlm.conf";

    char   *parms_str,
           *parms_wk_str;
    int     any_error,
            error;

    any_error = 0;

    parms_str = (char *) malloc(HBM_BUFF_SIZE_PARM);
    *(parms_str + HBM_BUFF_SIZE_PARM - 1) = '\0';

    if  ((parm_fp_ptr = fopen(parm_fname_dflt1_str, "r")) == NULL)
    {
        if  ((parm_fp_ptr = fopen(parm_fname_dflt2_str, "r")) == NULL)
        {
            fprintf(stderr,
                    "\n%s: Error opening parameter files: %s",
                    hbm_error_procname_str,
                    parm_fname_dflt1_str);
            fprintf(stderr,
                    "\n%s:                              : %s\n",
                    hbm_error_procname_str,
                    parm_fname_dflt2_str);
            return -2;
        }
    }

    while (fgets(parms_str, HBM_BUFF_SIZE_PARM, parm_fp_ptr) != NULL)
    {
        if  ((parms_wk_str = index(parms_str, '#')) != NULL)
        {
            *parms_wk_str = '\0';
        }
        parms_wk_str = parms_str;
        while (isspace(*parms_wk_str))
        {
            parms_wk_str++;
        }
        if  (*parms_wk_str != '\0')
        {
            if  (error = hbm_get_HBMLM_parm_val(parms_wk_str,
                                                HBMLM_addr_reg_ptr))
            {
                any_error = error;
            }
        }
    }

    fclose(parm_fp_ptr);
    free(parm_fp_ptr);

    free(parms_str);

    if  (HBMLM_addr_reg_ptr->sin_port == 0)
    {
        fprintf(stderr,
                "\n%s: HBM: Heartbeat parameters not present or invalid\n",
                hbm_error_procname_str);
        return -2;
    }

    if  (any_error)
    {
        return (-10 + any_error);
    }

    return 0;
}


/*
************************************************************************
**                                                                    **
**  hbm_client_init                                                   **
**                                                                    **
************************************************************************
*/
int    hbm_client_init( unsigned int        hbm_flags,
                        struct hbm_parms   *hbm_parm_info_ptr)
{
    char               *addr_str;

    int                 error;

    struct sockaddr_in  test_addr;

    if  (!hbm_initialization_done)
    {
        hbm_initialization_done = 1;

/*
**      Get the PID and client process name.
*/
        if  (hbm_flags & HBM_CLFLAG_SET_CL_PID)
        {
            if  (hbm_parm_info_ptr->cl_pid == 0)
            {
                fprintf(stderr,
                        "\n%s: HBM: Invalid pid input parameter",
                        hbm_error_procname_str);
                fprintf(stderr,
                        "\n%s: HBM: Continuing without heartbeat processing\n",
                        hbm_error_procname_str);
                hbm_doing_heartbeat_processing = 0;
                return(-100);
            }
            hbm_HBClient_PID = hbm_parm_info_ptr->cl_pid;
        }
        else
        {
            hbm_HBClient_PID = getpid();
            if  (hbm_HBClient_PID == 0)
            {
                fprintf(stderr,
                        "\n%s: HBM: Error getting pid",
                        hbm_error_procname_str);
                fprintf(stderr,
                        "\n%s: HBM: Continuing without heartbeat processing\n",
                        hbm_error_procname_str);
                hbm_doing_heartbeat_processing = 0;
                return(-200);
            }
        }

        hbm_client_procname_str = hbm_get_procname_fr_pid(hbm_HBClient_PID);        
        if  (hbm_client_procname_str == NULL)
        {
            fprintf(stderr,
                    "\n%s: HBM: Error getting client process name",
                    hbm_error_procname_str);
            fprintf(stderr,
                    "\n%s: HBM: Continuing without heartbeat processing\n",
                    hbm_error_procname_str);
            hbm_doing_heartbeat_processing = 0;
            return(-300);
        }

        if  (hbm_flags & HBM_CLFLAG_SET_CL_RPTNAME)
        {
            hbm_client_procname_rpt_str = hbm_copy_procname(hbm_parm_info_ptr->cl_procname_rpt);
            if  (hbm_client_procname_rpt_str == NULL)
            {
                fprintf(stderr,
                        "\n%s: HBM: Invalid procname for reporting input parameter",
                        hbm_error_procname_str);
                fprintf(stderr,
                        "\n%s: HBM: Continuing without heartbeat processing\n",
                        hbm_error_procname_str);
                hbm_doing_heartbeat_processing = 0;
                return(-100);
            }
        }
        else
        {
            hbm_client_procname_rpt_str = hbm_client_procname_str;
        }

/*
**      Get the (hbclient) host IP Number (for HBMLM communication).
*/
        bzero((char *) &hbm_HBMLM_addr_reg,
              sizeof(hbm_HBMLM_addr_reg));
        hbm_HBMLM_addr_reg.sin_family = AF_INET;

        if  (error = hbmcl_get_machine_info(&hbm_HBClient_host_name_len,
                                            &hbm_HBClient_host_name_ptr,
                                            &hbm_HBMLM_addr_reg))
        {
            fprintf(stderr,
                    "\n%s: HBM: Error getting machine information",
                    hbm_error_procname_str);
            fprintf(stderr,
                    "\n%s: HBM: Continuing without heartbeat processing\n",
                    hbm_error_procname_str);
            hbm_doing_heartbeat_processing = 0;
            return (-300 + error);
        }

/*
**      Get the HBMLM port number for registration messages
**      (from the HBMLM parameter file).
**
*/
        if  (error = hbm_get_HBMLM_parms(&hbm_HBMLM_addr_reg))
        {
            fprintf(stderr,
                    "\n%s: HBM: Continuing without heartbeat processing\n",
                    hbm_error_procname_str);
            hbm_doing_heartbeat_processing = 0;
            return (-400 + error);
        }

/*
**      Initialize the Data Collector fields (IP address and
**      port number) (from the HBClient parameter file).
*/
        bzero(hbm_HBMDC_host_addr_str, HBM_IP_ADDRESS_SIZE);
        hbm_HBMDC_portNum = 0;

        if  (!(hbm_flags & (  HBM_CLFLAG_SET_DC_HOSTNAME
                            | HBM_CLFLAG_SET_DC_HOSTIPNUM
                            | HBM_CLFLAG_SET_DC_PORT)))
        {
            if  (error = hbm_get_HBClient_parms(NULL,
                                                hbm_HBMDC_host_addr_str,
                                                &hbm_HBMDC_portNum))
            {
                return (-500 + error);
            }
        }
        else
        {
            if  (   (!(hbm_flags & (  HBM_CLFLAG_SET_DC_HOSTNAME
                                    | HBM_CLFLAG_SET_DC_HOSTIPNUM)))
                 || (!(hbm_flags & HBM_CLFLAG_SET_DC_PORT)))
            {
                return -600;
            }
            if  (hbm_flags & HBM_CLFLAG_SET_DC_HOSTIPNUM)
            {
                test_addr.sin_addr.s_addr = inet_addr(hbm_parm_info_ptr->dc_hostipnum);
                if  (test_addr.sin_addr.s_addr <= 0)
                {
                    return -700;
                }
                strncpy(hbm_HBMDC_host_addr_str,
                        hbm_parm_info_ptr->dc_hostipnum,
                        HBM_IP_ADDRESS_SIZE);
            }
            else /*  => (hbm_flags & HBM_CLFLAG_SET_DC_HOSTNAME)  */
            {
                hbm_get_IPNum_fr_name(&(test_addr.sin_addr),
                                      hbm_parm_info_ptr->dc_hostname,
                                      NULL);
                if  (test_addr.sin_addr.s_addr <= 0)
                {
                    fprintf(stderr,
                            "\n%s: HBM: HBMDCHostname parameter invalid/not found\n",
                            hbm_error_procname_str);
                    return -800;
                }
                else
                {
                    addr_str = inet_ntoa(test_addr.sin_addr);
                    strncpy(hbm_HBMDC_host_addr_str,
                            addr_str,
                            HBM_IP_ADDRESS_SIZE);
                }
            }
            if  (hbm_parm_info_ptr->dc_port == 0)
            {
                return -900;
            }
            hbm_HBMDC_portNum = hbm_parm_info_ptr->dc_port;
        }
    }

    return 0;
}


/*
************************************************************************
**                                                                    **
**  hbm_format_reg_msg                                                **
**                                                                    **
************************************************************************
*/
void    hbm_format_reg_msg()
{
    int     str1_len,
            str2_len,
            str3_len;

/*
**  Initialize the registration data (packet) for
**  transmission to the HBMLM.
**    First get the size,
**    then allocate it and fill it in.
*/
/*
**  hbm_msg_RegCmsgLength = 5 + 4 + 1 + strlen(hbmcl_proc_name) + 1
**                        + strlen(hbmcl_proc_name_rpt) + 1
**                        + 4 + strlen(hbm_HBMDC_host_addr_str) + 1
**                        + 4;
*/
    str1_len = strlen(hbm_client_procname_str) + 1;
    str2_len = strlen(hbm_client_procname_rpt_str) + 1;
    str3_len = strlen(hbm_HBMDC_host_addr_str) + 1;

    hbm_msg_RegCmsgLength = 18 + str1_len + str2_len + str3_len;

    hbm_msg_RegCmsg = (char *) malloc(hbm_msg_RegCmsgLength);
    bzero((void *)hbm_msg_RegCmsg, hbm_msg_RegCmsgLength);

    strcpy(hbm_msg_RegCmsg, "HBMCL");

    hbm_msg_RegCmsgLength_ptr = hbm_msg_RegCmsg + 5;
    PackUInt32(hbm_msg_RegCmsgLength_ptr,
               (unsigned long) hbm_msg_RegCmsgLength);

    hbm_msg_RegCmsgType_ptr = hbm_msg_RegCmsgLength_ptr + 4;
    *hbm_msg_RegCmsgType_ptr = HBM_MSGTYPE_REGISTER;

    hbm_msg_RegCprocessPID_ptr = hbm_msg_RegCmsgType_ptr + 1;
    PackUInt32(hbm_msg_RegCprocessPID_ptr,
               (unsigned long) hbm_HBClient_PID);

    hbm_msg_RegCprocessName_str = hbm_msg_RegCprocessPID_ptr + 4;
    strcpy(hbm_msg_RegCprocessName_str, hbm_client_procname_str);

    hbm_msg_RegCprocessNameRpt_str = hbm_msg_RegCprocessName_str
                                + str1_len;
    strcpy(hbm_msg_RegCprocessNameRpt_str, hbm_client_procname_rpt_str);

    hbm_msg_RegCDChostIPNum_str = hbm_msg_RegCprocessNameRpt_str
                                + str2_len;
    strcpy(hbm_msg_RegCDChostIPNum_str, hbm_HBMDC_host_addr_str);

    hbm_msg_RegCDCportNum_ptr = hbm_msg_RegCDChostIPNum_str
                                + str3_len;
    PackUInt32(hbm_msg_RegCDCportNum_ptr,
               (unsigned long) hbm_HBMDC_portNum);

    return;
}


/*
************************************************************************
**                                                                    **
**  hbm_format_unreg_msg                                              **
**                                                                    **
************************************************************************
*/
void    hbm_format_unreg_msg(   char    shutdown_mode)
{
    int     str1_len;

/*
**  Initialize the unregistration data (packet) for
**  transmission to the HBMLM.
**    First get the size,
**    then allocate it and fill it in.
*/
/*
**  hbm_msg_UnRegCmsgLength = 5 + strlen(hbmcl_proc_name) + 1
**                          + 4;
*/
    str1_len = strlen(hbm_client_procname_str) + 1;
    hbm_msg_UnRegCmsgLength = 14 + str1_len;

    hbm_msg_UnRegCmsg = (char *) malloc(hbm_msg_UnRegCmsgLength);
    bzero((void *)hbm_msg_UnRegCmsg, hbm_msg_UnRegCmsgLength);

    strcpy(hbm_msg_UnRegCmsg, "HBMCL");

    hbm_msg_UnRegCmsgLength_ptr = hbm_msg_UnRegCmsg + 5;
    PackUInt32(hbm_msg_UnRegCmsgLength_ptr,
               (unsigned long) hbm_msg_UnRegCmsgLength);

    hbm_msg_UnRegCmsgType_ptr = hbm_msg_UnRegCmsgLength_ptr + 4;
    *hbm_msg_UnRegCmsgType_ptr = shutdown_mode;
    if  (   (*hbm_msg_UnRegCmsgType_ptr != HBM_MSGTYPE_SHUTDOWN_NORMAL)
         && (*hbm_msg_UnRegCmsgType_ptr != HBM_MSGTYPE_SHUTDOWN_ABNORMAL))
    {
        fprintf(stderr,
                "\n%s: HBM: Shutdown mode not NORMAL or ABNORMAL",
                hbm_error_procname_str);
        fprintf(stderr,
                "\n%s: HBM: Using ABNORMAL\n",
                hbm_error_procname_str);
        *hbm_msg_UnRegCmsgType_ptr = HBM_MSGTYPE_SHUTDOWN_ABNORMAL;
    }

    hbm_msg_UnRegCprocessPID_ptr = hbm_msg_UnRegCmsgType_ptr + 1;
    PackUInt32(hbm_msg_UnRegCprocessPID_ptr,
               (unsigned long) hbm_HBClient_PID);

    hbm_msg_UnRegCprocessName_str = hbm_msg_UnRegCprocessPID_ptr + 4;
    strcpy(hbm_msg_UnRegCprocessName_str, hbm_client_procname_str);
}


/*
************************************************************************
**                                                                    **
**  hbm_send_msg_save_resp                                            **
**                                                                    **
************************************************************************
*/
int     hbm_send_msg_save_resp( char   *hbmcl_msg_ptr,
                                int     hbmcl_msg_len)
{
    int             retcd;

    int             rcv_attempts,
                    send_attempts;

    int             nleft,
                    nrcvd,
                    nsent;

#if  (defined TARGET_ARCH_HPUX)
    int             fd_readset;
    size_t          max_fdp1;
#else
    fd_set          fd_readset;
    int             max_fdp1;
#endif  /*  TARGET_ARCH_HPUX  */

    int             sock_fd_reg;

    struct sockaddr_in
                    HBC_reg_addr;

    struct timeval  time_left;

#if  (   (defined TARGET_ARCH_CRAYT3E) \
      || (defined TARGET_ARCH_HPUX) \
      || (defined TARGET_ARCH_IRIX) \
      || (defined TARGET_ARCH_SOLARIS))
    int             from_addr_len;
#else
    size_t          from_addr_len;
#endif

    char           *msg_buff_ptr;


/*
**  Open the UDP connection to the local monitor and register.
*/
    if  ((sock_fd_reg = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
    {
        fprintf(stderr,
                "\n%s: HBM: Error opening socket to register with HBMLM\n",
                hbm_error_procname_str);
        perror(hbm_error_procname_str);
        fprintf(stderr,
                "\n%s: HBM: Continuing without heartbeat processing\n",
                hbm_error_procname_str);
        return -300;
    }

#if  (defined TARGET_ARCH_HPUX)
    if  (sock_fd_reg >= (8 * sizeof(int)))
    {
        fprintf(stderr,
                "\n%s: HBM: Funky HP-UX error giving fd number too large to process.\n",
                hbm_error_procname_str);
        return -400;
    }
#endif  /*  TARGET_ARCH_HPUX  */

    bzero((void *)&HBC_reg_addr, sizeof(HBC_reg_addr));
    HBC_reg_addr.sin_family = AF_INET;
    HBC_reg_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    HBC_reg_addr.sin_port = htons(0);

    if  (bind(sock_fd_reg,
              (struct sockaddr *) &HBC_reg_addr,
              sizeof(HBC_reg_addr)) < 0 )
    {
        fprintf(stderr,
                "\n%s: HBM: Error binding register port. (error=%d)\n",
                hbm_error_procname_str,
                errno);
        perror(hbm_error_procname_str);
        fprintf(stderr,
                "\n%s: HBM: Continuing without heartbeat processing\n",
                hbm_error_procname_str);
        close(sock_fd_reg);
        return -400;
    }

/*
**  Send the message.
*/
    DEBUGF2("Registering with LM at %d", hbm_HBMLM_addr_reg.sin_addr.S_un.S_un_b.s_b1);
    DEBUGF2(".%d", hbm_HBMLM_addr_reg.sin_addr.S_un.S_un_b.s_b2);
    DEBUGF2(".%d", hbm_HBMLM_addr_reg.sin_addr.S_un.S_un_b.s_b3);
    DEBUGF2(".%d", hbm_HBMLM_addr_reg.sin_addr.S_un.S_un_b.s_b4);
    DEBUGF2(" on port %d\n", hbm_HBMLM_addr_reg.sin_port);

    hbm_doing_heartbeat_processing = 0;
    send_attempts = 0;

    while  (   (!hbm_doing_heartbeat_processing)
            && (send_attempts < 5))
    {
        send_attempts++;

        msg_buff_ptr = hbmcl_msg_ptr;
        nleft = hbmcl_msg_len;
        while (nleft > 0)
        {
            nsent = sendto(sock_fd_reg,
                           msg_buff_ptr,
                           nleft,
                           0,
                           (struct sockaddr *) &hbm_HBMLM_addr_reg,
                           sizeof(hbm_HBMLM_addr_reg));
            if  (nsent <= 0)
            {
                fprintf(stderr,
                        "\n%s: HBM: Error sending hbm register request\n",
                        hbm_error_procname_str);
                perror(hbm_error_procname_str);
                fprintf(stderr,
                        "\n%s: HBM: Continuing without heartbeat processing\n",
                        hbm_error_procname_str);
                close(sock_fd_reg);
                return -500;
            }
            msg_buff_ptr += nsent;
            nleft -= nsent;
        }

/*
**      Get the response.
*/
        nleft = hbm_msg_RegCmsgResp_len;
        rcv_attempts = 0;
        while (   (nleft > 0)
               && (rcv_attempts < 5))
        {
            rcv_attempts++;

#ifdef  TARGET_ARCH_HPUX
            bzero((char *) &fd_readset, sizeof(int));
            fd_readset |= (1 << sock_fd_reg);
#else
            FD_ZERO(&fd_readset);
            FD_SET(sock_fd_reg, &fd_readset);
#endif
            max_fdp1 = sock_fd_reg + 1;
            time_left.tv_sec = 1;
            time_left.tv_usec = 0;

            retcd = select(max_fdp1, &fd_readset, NULL, NULL, &time_left);

            if  (   (retcd < 0)
                 && (errno != EINTR))
            {
                if  (   (errno == EBADF)
                     || (errno == EFAULT))
                {
                    fprintf(stderr,
                            "\n%s: Error in select() -- trying again\n",
                            hbm_error_procname_str);
                    perror(hbm_error_procname_str);
                }
            }
            else if  (retcd == 0)
            {
                fprintf(stderr,
                        "\n%s: No response from HBMLM for reg msg -- trying again",
                        hbm_error_procname_str);
                fprintf(stderr,
                        "\n%s: Number registration attempts: %d  Number receive attempts: %d\n",
                        hbm_error_procname_str,
                        send_attempts,
                        rcv_attempts);
            }
            else if (retcd > 0)
            {
                msg_buff_ptr = hbm_msg_RegCmsgResp;
                nleft = hbm_msg_RegCmsgResp_len;

                bzero((void *)&HBC_reg_addr, sizeof(HBC_reg_addr));
                from_addr_len = sizeof(HBC_reg_addr);
                strncpy(hbm_msg_RegCmsgResp, "          ", hbm_msg_RegCmsgResp_len);

                nrcvd = recvfrom(sock_fd_reg,
                                 msg_buff_ptr,
                                 nleft,
                                 0,
                                 (struct sockaddr *)&HBC_reg_addr,
                                 &from_addr_len);

                if  (nrcvd <= 0)
                {
                    fprintf(stderr,
                            "\n%s: HBM: Error receiving hbm register response",
                            hbm_error_procname_str);
                    perror(hbm_error_procname_str);
                }
                else
                {
                    if  (   (HBC_reg_addr.sin_addr.s_addr !=
                                        hbm_HBMLM_addr_reg.sin_addr.s_addr)
                         || (HBC_reg_addr.sin_port !=
                                        hbm_HBMLM_addr_reg.sin_port))
                    {
                        nleft = hbm_msg_RegCmsgResp_len;
                    }
                    else
                    {
                        nleft -= nrcvd;
                        if  (   (nleft)
                             || (strncmp(hbm_msg_RegCmsgResp,"HBMLM", 5))
                             || (   (hbmcl_msg_ptr[9] == HBM_MSGTYPE_REGISTER)
                                 && (hbm_msg_RegCmsgResp[5] != HBM_MSGTYPE_RESPONSE_REGISTER))
                             || (   (   (hbmcl_msg_ptr[9] == HBM_MSGTYPE_SHUTDOWN_NORMAL)
                                     || (hbmcl_msg_ptr[9] == HBM_MSGTYPE_SHUTDOWN_ABNORMAL))
                                 && (hbm_msg_RegCmsgResp[5] != HBM_MSGTYPE_RESPONSE_SHUTDOWN)))
                        {
                            fprintf(stderr,
                                    "\n%s: HBM: Invalid hbm register response\n",
                                    hbm_error_procname_str);
                            fprintf(stderr,
                                    "\n%s: HBM: Accepting as valid\n",
                                    hbm_error_procname_str);
                            nleft = hbm_msg_RegCmsgResp_len;
                        }
                        else
                        {
/*
**                          Save the response time.
*/
                            UnpackUInt32(hbm_msg_RegCmsgResp, hbm_Client_reg_HBMLM_time);
                            hbm_doing_heartbeat_processing = 1;

    DEBUGF2("\n%s: HBM: Heartbeat registration successfully completed.\n", hbm_error_procname_str);
                        }
                    }
                }
            }            
        }  /*  while for recvfrom  */
    }  /*  while for sendto/recvfrom  */

    close(sock_fd_reg);

    if  (nleft)
    {
        return -100;
    }

    return 0;
}


int     hbm_register(   char               *hbm_procname_str,
                        unsigned int        hbm_flags,
                        struct hbm_parms   *hbm_parm_fields_ptr,
                        char               *hostname)
{
    int     error;

    /* NETSOLVE STUFF */
    strcpy(agent_hostname,hostname);
    strcat(agent_hostname,"\n");
/*
**  Set hbm_HBClient_host_name_ptr for error messages.
*/
    if  (hbm_procname_str == NULL)
    {
        fprintf(stderr,
                "\nHBM: Program name pointer is null");
        fprintf(stderr,
                "\nHBM: Continuing without heartbeat processing\n");
        hbm_doing_heartbeat_processing = 0;
        return -100;
    }
    hbm_error_procname_str = hbm_procname_str;

    error = hbm_client_init(hbm_flags, hbm_parm_fields_ptr);
    if  (error)
    {
        return (-1000 + error);
    }

    hbm_format_reg_msg();
    error = hbm_send_msg_save_resp(hbm_msg_RegCmsg,
                                   hbm_msg_RegCmsgLength);
    if  (error)
    {
        return (-2000 + error);
    }

    return 0;
}


int     hbm_shutdown(   char               *hbm_procname_str,
                        unsigned int        hbm_flags,
                        struct hbm_parms   *hbm_parm_fields_ptr,
                        char                shutdown_mode)
{
    int     shutdown_sock_fd;

    struct sockaddr_in
            HBC_unreg_addr;

    int     error;

    size_t  nleft;

    ssize_t nsent;

    char   *msg_buff_ptr;

    if  (!hbm_doing_heartbeat_processing)
    {
        return 0;
    }

    DEBUGF1("Starting HBM detach\n");

    if  (hbm_procname_str == NULL)
    {
        fprintf(stderr,
                "\nHBM: Program name pointer is null");
        fprintf(stderr,
                "\nHBM: Continuing without heartbeat processing\n");
        hbm_doing_heartbeat_processing = 0;
        return -100;
    }
    hbm_error_procname_str = hbm_procname_str;

    error = hbm_client_init(hbm_flags, hbm_parm_fields_ptr);
    if  (error)
    {
        return (-1000 + error);
    }

    hbm_format_unreg_msg(shutdown_mode);

    error = hbm_send_msg_save_resp(hbm_msg_UnRegCmsg,
                                   hbm_msg_UnRegCmsgLength);
    if  (error)
    {
        return (-2000 + error);
    }

    return 0;
}
/*
************************************************************************
**                                                                    **
**  hbm_cl.c - end                                                    **
**                                                                    **
************************************************************************
*/
