/*
************************************************************************
**  GUSTO Heartbeat Monitor Data Collector (HBM_DC) program           **
************************************************************************
*/
#define HBM_DC_C

#include "hbm.h"
#include "hbm_common.h"
#include "hbm_list_macros.h"
#include "hbm_lmdc.h"

#ifndef SYSCONFDIR
#define SYSCONFDIR "."
#endif /*  defined SYSCONFDIR  */

    int                 quit_now = 0;

    int                 sock_fd_hbmsg;

    char               *hbmdc_ckpt_fname_str = NULL;

    char               *hbmdc_proc_name_str;

    long                time_conv_diff;

    struct timeval      time_curr,
                        time_to_monitor_write;

struct hbmdc_dc_fields {
  /*  Machine Information fields  */
    struct in_addr      DC_hostIPNum;
    char               *DC_hostIPNum_str;
    int                 DC_hostNameLen;
    char               *DC_hostName_str;
  /*  DC Parameter Fields  */
    u_short             DC_portNumHBMsg;
    struct sockaddr_in  DC_addr;
    unsigned int        DC_MonitorInterval;
  /*  DC Run time/Control fields  */
    u_long              DC_LocalMonitorsCt;
    u_long              DC_ClientsCt;
    u_long              DC_ClientsLiveCt;
  /*  DC Parameter Field  */
    char               *DC_ckptFileName_str;
    char               *DC_ckptFileNameWk_str;
  /*  DC Run time/Control field  */
    UTCtime             DC_ckptTime;
};

    struct hbmdc_dc_fields
        hbmdc_dc_info;

struct hbmdc_lm_fields {
    struct hbmdc_lm_fields *prev;
    struct hbmdc_lm_fields *next;
    struct in_addr          LM_hostIPNum_hostfmt;
    char                   *LM_hostIPNum_str;
    char                   *LM_hostName_str;
    u_short                 LM_portNumRpt_hostfmt;
    struct sockaddr_in      LM_addr_rpt;
    u_long                  LM_ClientsCt;
    u_long                  LM_ClientsLiveCt;
    u_long                  LM_ClientsActiveCt;
    unsigned int            LM_SummaryStatus;
    UTCtime                 LM_rptTimeLast;
};

    struct hbmdc_lm_fields
        hbmdc_lm_info;

struct hbmdc_lm_list_head {
    struct hbmdc_lm_fields *head;
    struct hbmdc_lm_fields *tail;
};

    struct hbmdc_lm_list_head   hbmdc_lm_list;

struct hbmdc_cl_fields {
    struct hbmdc_cl_fields *prev;
    struct hbmdc_cl_fields *next;
    struct hbmdc_lm_fields *CL_LocalMonitor_ptr;
    struct in_addr          CL_hostIPNum_hostfmt;
    char                   *CL_hostIPNum_str;
    u_short                 CL_LMportNumRpt_hostfmt;
    char                   *CL_procName_str;
    unsigned int            CL_procPID;
    unsigned int            CL_procStatus;
    UTCtime                 CL_regTime;
    UTCtime                 CL_blockedTime;
    unsigned int            CL_cpuSecs;
    UTCtime                 CL_shutdownTime;
    UTCtime                 CL_deleteTime;
    unsigned int            CL_rptNumber;
    UTCtime                 CL_rptTimeLast;
    unsigned int            CL_rptInterval;
    UTCtime                 CL_rptTimeNext;
    int                     CL_updated;
};

struct hbmdc_cl_list_head {
    struct hbmdc_cl_fields *head;
    struct hbmdc_cl_fields *tail;
};

    struct hbmdc_cl_list_head   hbmdc_cl_list;


/*
************************************************************************
**  get_DC_machine_info                                               **
**      Parameters: hbmdc_dc_info_ptr   pointer to struct of hbmdc    **
**                                      host and process fields       **
**                                                                    **
**      Returns:    integer return code -- 0 if ok, error otherwise   **
**                                                                    **
**      This routine sets the HBMDC host information fields that are  **
**      dependent on the address/configuration of the host.           **
**      Those fields are:                                             **
**              hostIPNum               IP Number (network format)    **
**              hostIPNum_str           IP Number (string in '.'      **
**                                      format)                       **
**              hostNameLen             length of the host name (int) **
**              hostName_str            host name (string)            **
**                                                                    **
************************************************************************
*/
int     get_DC_machine_info()
{
    int     retcd;
    char   *str_ptr1,
           *str_ptr2;

/*
**  Get machine-specific information:
*/
/*
**  Get host name (and name length) and IP Number.
*/
    hbmdc_dc_info.DC_hostName_str =
                                (char *) malloc(MAXHOSTNAMELEN + 256);
    bzero(hbmdc_dc_info.DC_hostName_str, MAXHOSTNAMELEN + 256);
    if  (gethostname(hbmdc_dc_info.DC_hostName_str,
                     MAXHOSTNAMELEN + 256) < 0)
    {
        fprintf(stderr,
                "\n%s: Error getting host name\n",
                hbmdc_proc_name_str);
        perror(hbmdc_proc_name_str);
        return -1;
    }
    retcd = hbm_get_IPNum_fr_name(&(hbmdc_dc_info.DC_hostIPNum),
                                  hbmdc_dc_info.DC_hostName_str,
                                  &(hbmdc_dc_info.DC_hostName_str));

    if  (retcd)
    {
        fprintf(stderr,
                "\n%s: Error getting host information\n",
                hbmdc_proc_name_str);
        return -2;
    }

/*  DC_hostIPNum field, hbmdc_dc_info                        */
    if  (hbmdc_dc_info.DC_hostIPNum.s_addr == 0)
    {
        fprintf(stderr,
                "\n%s: Bad host address\n",
                hbmdc_proc_name_str);
        perror(hbmdc_proc_name_str);
        return -3;
    }

/*  DC_hostIPNum_str field, hbmdc_dc_info                    */
    str_ptr1 = inet_ntoa(hbmdc_dc_info.DC_hostIPNum);
    hbmdc_dc_info.DC_hostIPNum_str =
                                (char *) malloc(strlen(str_ptr1) + 1);
    strcpy(hbmdc_dc_info.DC_hostIPNum_str, str_ptr1);
    str_ptr1 = NULL;

/*  DC_hostNameLen field, hbmdc_dc_info                      */
    hbmdc_dc_info.DC_hostNameLen =
                            strlen(hbmdc_dc_info.DC_hostName_str);

    hbmdc_dc_info.DC_hostIPNum.s_addr =
                            ntohl(hbmdc_dc_info.DC_hostIPNum.s_addr);

    return 0;
}


/*
************************************************************************
**  get_hbmdc_parmfile_name                                           **
**      Parameters: program parameters (argc, argv)                   **
**      Returns:    pointer to string with parameter filename         **
**                  (NULL if error)                                   **
**                                                                    **
**      This routine gets the name of the file with the parameter     **
**      information in it (parameter is delimited by `-p'.            **
************************************************************************
*/
char   *get_hbmdc_parmfile_name(    int     pargc,
                                    char   *pargv[])
{
    int         i = 1;

    for (i = 1; i + 2 <= pargc; i++)
    {
        if  (   (pargv[i][0] == '-')
             && (pargv[i][1] == 'p'))
        {
            return pargv[i+1];
        }
    }
    return NULL;
}


/*
************************************************************************
**  get_hbmdc_parm_val                                                **
**      Parameters: parm_str              line from the parm file     **
**                                                                    **
**      Returns:    integer return code                               **
**                      0   OK                                        **
**                      1   Error                                     **
**                                                                    **
**      This routine:                                                 **
**          scans the value pointed to by parms_str                   **
**          for a parameter name and                                  **
**              if it finds one                                       **
**                  gets the value and                                **
**                  sets the appropriate variable                     **
**                                                                    **
**          valid parameter names and parameter formats are:          **
**              one parameter per line                                **
**              <PARMNAME>: <parmvalue>                               **
**              Colon (:) is optional                                 **
**              white space before <PARMNAME> is ignored              **
**              Lines beginning with pound sign (#) are ignored       **
************************************************************************
*/
int     get_hbmdc_parm_val( char   *parm_str)
{
    int             error;
    int             got_val;
    int             i,
                    j;
    unsigned int    u_int_val;
    u_short         u_short_val;

    error = 0;
    while (isspace(*parm_str))
    {
        parm_str++;
    }
    if  (!strncmp(parm_str, "DCPortNumHBMsg", 14))
    {
/*      Port to which HBMLMs are to send report messages  */
        parm_str += 14;
        if  (*parm_str == ':')
        {
            parm_str++;
        }
        got_val = sscanf(parm_str, " %hu", &u_short_val);
        if  (   (!got_val)
             || (u_short_val < HBM_PORT_MIN_AVAILABLE)
             || (u_short_val > HBM_PORT_MAX_AVAILABLE))
        {
            fprintf(stderr,
                    "\n%s: Invalid Parameter: HB Message Port",
                    hbmdc_proc_name_str);
            fprintf(stderr,
                    "\n%s:                    must be > %d and < %d\n",
                    hbmdc_proc_name_str,
                    (int) HBM_PORT_MIN_AVAILABLE,
                    (int) HBM_PORT_MAX_AVAILABLE);
            return -1;
        }
/*      portNumReg field, hbmdc_dc_info           */
        hbmdc_dc_info.DC_portNumHBMsg = u_short_val;
/*      addr_reg.sin_port field, hbmdc_dc_info    */
        hbmdc_dc_info.DC_addr.sin_port = htons(u_short_val);
    }
    else if  (!strncmp(parm_str, "DCMonitorInterval", 17))
    {
/*      DC Monitor Interval        */
        parm_str += 17;
        if  (*parm_str == ':')
        {
            parm_str++;
        }
/*      MonitorInterval field, hbmdc_dc_info        */
        got_val = sscanf(parm_str, " %du", &(u_int_val));
        if  (   (!got_val)
             || (u_int_val < HBMDC_LIM_MONITOR_INTERVAL_MIN)
             || (u_int_val > HBMDC_LIM_MONITOR_INTERVAL_MAX))
        {
            fprintf(stderr,
                    "\n%s: Invalid Parameter: MonitorInterval",
                    hbmdc_proc_name_str);
            fprintf(stderr,
                    "\n%s:                    must be > %lu and < %lu\n",
                    hbmdc_proc_name_str,
                    HBMDC_LIM_MONITOR_INTERVAL_MIN,
                    HBMDC_LIM_MONITOR_INTERVAL_MAX);
            return -2;
        }
        else
        {
            hbmdc_dc_info.DC_MonitorInterval = u_int_val;
        }
    }
    else if  (!strncmp(parm_str, "DCCheckpointFileName", 20))
    {
/*
**      Name of checkpoint file.
**      Initially will check this file for old checkpoint information,
**      later will save checkpoint information after each change.
*/
        parm_str += 20;
        if  (*parm_str == ':')
        {
            parm_str++;
        }
        while (isspace(*parm_str))
        {
            parm_str++;
        }
        if  (*parm_str == '\0')
        {
            fprintf(stderr,
                    "\n%s: Missing/Null Parameter: CheckpointFileName",
                    hbmdc_proc_name_str);
            return -3;
        }
        i = 0;
        while (   (parm_str[i] != '\0')
               && (!isspace(parm_str[i])))
        {
            i++;
        }
        free(hbmdc_dc_info.DC_ckptFileName_str);
        hbmdc_dc_info.DC_ckptFileName_str = (char *) malloc(i+1);
        (hbmdc_dc_info.DC_ckptFileName_str)[i] = '\0';
        strncpy(hbmdc_dc_info.DC_ckptFileName_str, parm_str, i);

        free(hbmdc_dc_info.DC_ckptFileNameWk_str);
        hbmdc_dc_info.DC_ckptFileNameWk_str = (char *) malloc(i+4);
        strcpy(hbmdc_dc_info.DC_ckptFileNameWk_str,
               hbmdc_dc_info.DC_ckptFileName_str);
        strcat(hbmdc_dc_info.DC_ckptFileNameWk_str, "_wk");        
    }

    return 0;
}


/*
************************************************************************
**  get_hbmdc_parms                                                   **
**      Parameters: parm_str              line from the parm file     **
**                                                                    **
**                                                                    **
**      Returns:    integer return code                               **
**                      0   OK                                        **
**                      1   Error                                     **
**                                                                    **
**      This routine:                                                 **
**          sets all default parameter values                         **
**          opens the parameter file and:                             **
**              reads it one line at a time,                          **
**              passing the read line to a routine that extracts      **
**                parameter values and sets the parameters.           **
**                                                                    **
**          parameter file format is:                                 **
**              one parameter per line                                **
**              <PARMNAME>: <parmvalue>                               **
**              Colon (:) is optional                                 **
**              white space before <PARMNAME> is ignored              **
**              Lines beginning with pound sign (#) are ignored       **
************************************************************************
*/
int     get_hbmdc_parms(    char   *parm_fname_str)
{
    FILE   *parm_fp;

    char   *parm_fname_dflt1_str = "./hbmdc.conf",
           *parm_fname_dflt2_str = SYSCONFDIR "/hbmdc.conf";

    char   *parms_read_str,
           *parms_read_str_wk,
           *parms_read_str_end;
    char   *parms_str;
    int     error;

    parms_read_str = (char *) malloc(HBM_BUFF_SIZE_PARM);
    parms_read_str_end = parms_read_str + HBM_BUFF_SIZE_PARM - 1;
    *parms_read_str_end = '\0';
    parms_str = (char *) malloc(HBM_BUFF_SIZE_PARM);
    parms_str[HBM_BUFF_SIZE_PARM - 1] = '\0';

    hbmdc_dc_info.DC_portNumHBMsg = HBMDC_DEFAULT_PORT_HBMSG;
    hbmdc_dc_info.DC_addr.sin_port = htons(HBMDC_DEFAULT_PORT_HBMSG);
    hbmdc_dc_info.DC_MonitorInterval = HBMDC_DEFAULT_MONITOR_INTERVAL;
    hbmdc_dc_info.DC_ckptFileName_str =
                (char *) malloc(strlen(HBMDC_DEFAULT_CHKPTFILE) + 1);
    strcpy(hbmdc_dc_info.DC_ckptFileName_str,
           HBMDC_DEFAULT_CHKPTFILE);
    hbmdc_dc_info.DC_ckptFileNameWk_str =
                (char *) malloc(strlen(hbmdc_dc_info.DC_ckptFileName_str) + 4);
    strcpy(hbmdc_dc_info.DC_ckptFileNameWk_str,
           hbmdc_dc_info.DC_ckptFileName_str);
    strcat(hbmdc_dc_info.DC_ckptFileNameWk_str, "_wk");

    if  (parm_fname_str != NULL)
    {
        if  ((parm_fp = fopen(parm_fname_str, "r")) == NULL)
        {
            fprintf(stderr,
                    "\n%s: Error opening parameter file: %s\n",
                    hbmdc_proc_name_str,
                    parm_fname_str);
            return -10;
        }
    }
    else
    {
        if  ((parm_fp = fopen(parm_fname_dflt1_str, "r")) == NULL)
        {
            if  ((parm_fp = fopen(parm_fname_dflt2_str, "r")) == NULL)
            {
                fprintf(stderr,
                        "\n%s: Error opening parameter files: %s",
                        hbmdc_proc_name_str,
                        parm_fname_dflt1_str);
                fprintf(stderr,
                        "\n%s:                              : %s\n",
                        hbmdc_proc_name_str,
                        parm_fname_dflt2_str);
                return -20;
            }
        }
    }

    while (fgets(parms_read_str, HBM_BUFF_SIZE_PARM, parm_fp) != NULL)
    {
        parms_read_str_wk = parms_read_str;
        while (isspace(*parms_read_str_wk))
        {
            parms_read_str_wk++;
        }
        if  (   (parms_read_str_wk <= parms_read_str_end)
             && (*parms_read_str_wk != '#')
             && (*parms_read_str_wk != '\n')
             && (*parms_read_str_wk != '\0'))
        {
            strcpy(parms_str, parms_read_str_wk);
            error = get_hbmdc_parm_val(parms_str);
            if  (error)
            {
                return (-30 + error);
            }
        }
    }

    fclose(parm_fp);

    free(parms_read_str);
    free(parms_str);

    return 0;
}


/*
************************************************************************
**  find_lm_entry                                                     **
**      Parameters: hostIPNum_hostfmt   pointer to struct in_addr     **
**                                      with IP Number of host the    **
**                                      Local Monitor is running on   **
**                                      (in host format).             **
**                  portNumRpt_hostfmt  unsigned short with the       **
**                                      number of the port used by    **
**                                      the Local Monitor for report  **
**                                      messages (in host format).    **
**                                                                    **
**      Returns:    pointer to struct hbmdc_lm_fields with matching   **
**                                      hostIPNum_hostfmt and         **
**                                      portNumRpt_hostfmt values.    **
**                                      (NULL if no such entry)       **
**                                                                    **
**      This routine checks for a Local Monitor entry with the        **
**      passed parameter values and returns a pointer to it or        **
**      NULL if no such entry was found.                              **
************************************************************************
*/
struct hbmdc_lm_fields *find_lm_entry(  struct in_addr *lm_hostIPNum_hostfmt_ptr,
                                        u_short         lm_portNumRpt_hostfmt)
{
    int                     i;
    struct hbmdc_lm_fields *lm_entry_ptr;

    for (lm_entry_ptr = hbmdc_lm_list.head;
         lm_entry_ptr;
         lm_entry_ptr = lm_entry_ptr->next)
    {
        if  (   (lm_entry_ptr->LM_hostIPNum_hostfmt.s_addr ==
                                lm_hostIPNum_hostfmt_ptr->s_addr)
             && (lm_entry_ptr->LM_portNumRpt_hostfmt ==
                                lm_portNumRpt_hostfmt))
        {
            return lm_entry_ptr;
        }
    }

    return NULL;
}


/*
************************************************************************
**  add_lm_entry                                                      **
**      Parameters: new_lm_entry_ptr    pointer to struct             **
**                                      hbmdc_lm_fields entry to be   **
**                                      added.                        **
**                                                                    **
**      Returns:    pointer to struct hbmdc_lm_fields with matching   **
**                                      hostIPNum_hostfmt and         **
**                                      portNumRpt_hostfmt values.    **
**                                      (NULL if no such entry)       **
**                                                                    **
**      This routine checks for a Local Monitor entry with the        **
**      same hostIPNum_hostfmt and portNumRpt_hostfmt.                **
**      If finds one it updates it, otherwise it adds the new one.    **
**      It returns a pointer to the new/updated entry.                **
************************************************************************
*/
struct hbmdc_lm_fields *add_lm_entry(   struct hbmdc_lm_fields
                                               *lm_entry_new_ptr)
{
    struct hbmdc_lm_fields *lm_entry_ptr;

    int                     i,
                            str1_comp_str2;

    if  ((lm_entry_ptr = find_lm_entry(&(lm_entry_new_ptr->LM_hostIPNum_hostfmt),
                                       lm_entry_new_ptr->LM_portNumRpt_hostfmt))
                                != NULL)
    {
        fprintf(stderr,
                "\n%s: Attempt to add existing LM entry",
                hbmdc_proc_name_str);
        fprintf(stderr,
                "\n%s:   Original entry updated",
                hbmdc_proc_name_str);

        free(lm_entry_new_ptr->LM_hostIPNum_str);
        free(lm_entry_ptr->LM_hostName_str);
        lm_entry_ptr->LM_hostName_str = lm_entry_new_ptr->LM_hostName_str;
        lm_entry_new_ptr->LM_hostName_str = NULL;

        free(lm_entry_new_ptr);

        return lm_entry_ptr;
    }

    for (lm_entry_ptr = hbmdc_lm_list.head;
         lm_entry_ptr;
         lm_entry_ptr = lm_entry_ptr->next)
    {
        if  (   (lm_entry_ptr->LM_hostIPNum_hostfmt.s_addr >
                        lm_entry_new_ptr->LM_hostIPNum_hostfmt.s_addr)
             || (   (lm_entry_ptr->LM_hostIPNum_hostfmt.s_addr ==
                        lm_entry_new_ptr->LM_hostIPNum_hostfmt.s_addr)
                 && (lm_entry_ptr->LM_portNumRpt_hostfmt >
                            lm_entry_new_ptr->LM_portNumRpt_hostfmt)))
        {
            break;
        }

    }
    ListInsertBefore(&hbmdc_lm_list, lm_entry_new_ptr, lm_entry_ptr);

    hbmdc_dc_info.DC_LocalMonitorsCt++;

    return lm_entry_new_ptr;    
}


/*
************************************************************************
**  set_lm_summary_status                                             **
**      Parameters: new_lm_entry_ptr    pointer to struct             **
**                                      hbmdc_lm_fields entry for     **
**                                      which the LM_SummaryStatus    **
**                                      field is to be updated.       **
**                                                                    **
**      Returns:    (Nothing)                                         **
**                                                                    **
**      This routine updates the LM_Summary_Status field for the      **
**      Local Monitor entry pointed to by the passed pointer based    **
**      on the fields in the corresponding Client entries.            **
************************************************************************
*/
void    set_lm_summary_status(  struct hbmdc_lm_fields
                                               *lm_entry_ptr)
{
    struct hbmdc_cl_fields *cl_entry_ptr;

    int                     i,
                            have_active,
                            have_blocked,
                            have_overdue,
                            str1_comp_str2;

    if  (lm_entry_ptr->LM_ClientsLiveCt == 0)
    {
        lm_entry_ptr->LM_SummaryStatus =
                                    HBM_LMSTATUS_CLIENTS_NONE_LIVE;
    }
    else if  (lm_entry_ptr->LM_ClientsLiveCt ==
                                lm_entry_ptr->LM_ClientsActiveCt)
    {
        lm_entry_ptr->LM_SummaryStatus =
                                    HBM_LMSTATUS_CLIENTS_ACTIVE_ALL;
    }
    else
    {
        have_active = have_blocked = have_overdue = 0;
        for (cl_entry_ptr = hbmdc_cl_list.head;
             cl_entry_ptr;
             cl_entry_ptr = cl_entry_ptr->next)
        {
            if  (cl_entry_ptr->CL_LocalMonitor_ptr == lm_entry_ptr)
            {
                if  (cl_entry_ptr->CL_procStatus ==
                                HBM_PROCSTATUS_ACTIVE)
                {
                    have_active = 1;
                }
                else if  (cl_entry_ptr->CL_procStatus ==
                                HBM_PROCSTATUS_BLOCKED)
                {
                    have_blocked = 1;
                }
                else if  (cl_entry_ptr->CL_procStatus ==
                                HBM_PROCSTATUS_OVERDUE)
                {
                    have_overdue = 1;
                }
            }
        }
        if  (have_overdue)
        {
            if  (   (have_active)
                 || (have_blocked))
            {
                lm_entry_ptr->LM_SummaryStatus =
                                        HBM_LMSTATUS_CLIENTS_LATE_SOME;
            }
            else
            {
                lm_entry_ptr->LM_SummaryStatus =
                                        HBM_LMSTATUS_CLIENTS_LATE_ALL;
            }
        }
        else if  (have_blocked)
        {
            if  (have_active)
            {
                lm_entry_ptr->LM_SummaryStatus =
                                        HBM_LMSTATUS_CLIENTS_ACTIVE_SOME;
            }
            else
            {
                lm_entry_ptr->LM_SummaryStatus =
                                        HBM_LMSTATUS_CLIENTS_BLOCKED;
            }
        }
        else
        {
            lm_entry_ptr->LM_SummaryStatus =
                                        HBM_LMSTATUS_CLIENTS_UNKNOWN;
        }
    }

    return;    
}


/*
************************************************************************
**  find_cl_entry                                                     **
**      Parameters: hostIPNum_hostfmt   pointer to struct in_addr     **
**                                      with IP Number of host the    **
**                                      Local Monitor for this client **
**                                      process (and hence the        **
**                                      client process itself) is     **
**                                      running on (in host format).  **
**                  procPID             unsigned int with the PID     **
**                                      of the client process.        **
**                                      the Local Monitor for report  **
**                                      messages.                     **
**                                                                    **
**      Returns:    pointer to struct hbmdc_cl_fields with matching   **
**                                      hostIPNum_hostfmt and         **
**                                      procPID values.               **
**                                      (NULL if no such entry)       **
**                                                                    **
**      This routine checks for a Client process entry with the       **
**      passed parameter values and returns a pointer to it or        **
**      NULL if no such entry was found.                              **
************************************************************************
*/
struct hbmdc_cl_fields *find_cl_entry(  struct in_addr *cl_hostIPNum_hostfmt_ptr,
                                        unsigned int    cl_procPID)
{
    int                     i;
    struct hbmdc_cl_fields *curr_cl_entry_ptr;

    for (curr_cl_entry_ptr = hbmdc_cl_list.head;
         curr_cl_entry_ptr;
         curr_cl_entry_ptr = curr_cl_entry_ptr->next)
    {
        if  (   (curr_cl_entry_ptr->CL_hostIPNum_hostfmt.s_addr ==
                                cl_hostIPNum_hostfmt_ptr->s_addr)
             && (curr_cl_entry_ptr->CL_procPID == cl_procPID))
        {
            return curr_cl_entry_ptr;
        }
    }

    return NULL;
}


/*
************************************************************************
**  add_cl_entry                                                      **
**      Parameters: new_cl_entry_ptr    pointer to struct             **
**                                      hbmdc_cl_fields entry to be   **
**                                      added.                        **
**                                                                    **
**      Returns:    pointer to struct hbmdc_cl_fields with matching   **
**                                      hostIPNum_hostfmt and         **
**                                      procPID values.               **
**                                      (NULL if no such entry)       **
**                                                                    **
**      This routine checks for a Client process entry with the       **
**      same hostIPNum_hostfmt and procPID.                           **
**      If finds one it updates it, otherwise it adds the new one.    **
**      It returns a pointer to the new/updated entry.                **
************************************************************************
*/
struct hbmdc_cl_fields *add_cl_entry(   struct hbmdc_cl_fields
                                               *new_cl_entry_ptr)
{
    struct hbmdc_cl_fields *cl_entry_ptr;
    struct hbmdc_lm_fields *lm_entry_ptr,
                           *lm_entry2_ptr;

    int                     i,
                            str1_comp_str2;

    lm_entry_ptr = find_lm_entry(&(new_cl_entry_ptr->CL_hostIPNum_hostfmt),
                                 new_cl_entry_ptr->CL_LMportNumRpt_hostfmt);
    if  (!lm_entry_ptr)
    {
        free(new_cl_entry_ptr->CL_hostIPNum_str);
        free(new_cl_entry_ptr->CL_procName_str);
        free(new_cl_entry_ptr);

        return NULL;
    }
    lm_entry2_ptr = NULL;

    if  ((cl_entry_ptr = find_cl_entry(&(new_cl_entry_ptr->CL_hostIPNum_hostfmt),
                                       new_cl_entry_ptr->CL_procPID))
                                != NULL)
    {
        fprintf(stderr,
                "\n%s: Attempt to add existing CL entry",
                hbmdc_proc_name_str);
        fprintf(stderr,
                "\n%s:   Original entry updated",
                hbmdc_proc_name_str);

        free(new_cl_entry_ptr->CL_hostIPNum_str);
        free(cl_entry_ptr->CL_procName_str);
        cl_entry_ptr->CL_procName_str = new_cl_entry_ptr->CL_procName_str;
        new_cl_entry_ptr->CL_procName_str = NULL;

/*
**      Save the old Local Monitor.
**      (old Local Monitor may be the same as the new one).
*/
        lm_entry2_ptr = cl_entry_ptr->CL_LocalMonitor_ptr;
/*
**      Back out the counts for the old Local Monitor.
**      (old Local Monitor may be the same as the new one).
*/
        hbmdc_dc_info.DC_ClientsCt--;
        lm_entry2_ptr->LM_ClientsCt--;
        if  (cl_entry_ptr->CL_procStatus ==
                                        HBM_PROCSTATUS_ACTIVE)
        {
            hbmdc_dc_info.DC_ClientsLiveCt--;
            lm_entry2_ptr->LM_ClientsLiveCt--;
            lm_entry2_ptr->LM_ClientsActiveCt--;
        }
        else if  (   (cl_entry_ptr->CL_procStatus ==
                                    HBM_PROCSTATUS_BLOCKED)
                  || (cl_entry_ptr->CL_procStatus ==
                                    HBM_PROCSTATUS_OVERDUE))
        {
            hbmdc_dc_info.DC_ClientsLiveCt--;
            lm_entry2_ptr->LM_ClientsLiveCt--;
        }

/*
**      Update the client fields with the newer information.
*/
        cl_entry_ptr->CL_procStatus = new_cl_entry_ptr->CL_procStatus;

        cl_entry_ptr->CL_regTime = new_cl_entry_ptr->CL_regTime;
        cl_entry_ptr->CL_blockedTime = new_cl_entry_ptr->CL_blockedTime;
        cl_entry_ptr->CL_cpuSecs = new_cl_entry_ptr->CL_cpuSecs;
        cl_entry_ptr->CL_shutdownTime = new_cl_entry_ptr->CL_shutdownTime;
        cl_entry_ptr->CL_deleteTime = new_cl_entry_ptr->CL_deleteTime;
        cl_entry_ptr->CL_rptNumber = new_cl_entry_ptr->CL_rptNumber;
        cl_entry_ptr->CL_rptTimeLast = new_cl_entry_ptr->CL_rptTimeLast;
        if  (new_cl_entry_ptr->CL_rptTimeLast >
                                        lm_entry_ptr->LM_rptTimeLast)
        {
            lm_entry_ptr->LM_rptTimeLast =
                                        new_cl_entry_ptr->CL_rptTimeLast;
        }
        cl_entry_ptr->CL_rptInterval = new_cl_entry_ptr->CL_rptInterval;
        cl_entry_ptr->CL_rptTimeNext = new_cl_entry_ptr->CL_rptTimeNext;
        cl_entry_ptr->CL_updated = new_cl_entry_ptr->CL_updated;

        free(new_cl_entry_ptr);
        new_cl_entry_ptr = cl_entry_ptr;
        cl_entry_ptr = NULL;
        if  (lm_entry2_ptr == lm_entry_ptr)
        {
            lm_entry2_ptr = NULL;
        }
    }
    else
    {
/*
**      Add the client entry to the list.
*/
        for (cl_entry_ptr = hbmdc_cl_list.head;
             cl_entry_ptr;
             cl_entry_ptr = cl_entry_ptr->next)
        {
            if  (   (cl_entry_ptr->CL_hostIPNum_hostfmt.s_addr >
                            new_cl_entry_ptr->CL_hostIPNum_hostfmt.s_addr)
                 || (   (cl_entry_ptr->CL_hostIPNum_hostfmt.s_addr ==
                            new_cl_entry_ptr->CL_hostIPNum_hostfmt.s_addr)
                     && (cl_entry_ptr->CL_procPID >
                            new_cl_entry_ptr->CL_procPID)))
            {
                break;
            }
        }
        ListInsertBefore(&hbmdc_cl_list, new_cl_entry_ptr, cl_entry_ptr);
    }

/*
**  Set the Local Monitor fields.
*/
    new_cl_entry_ptr->CL_LocalMonitor_ptr = lm_entry_ptr;

    hbmdc_dc_info.DC_ClientsCt++;
    lm_entry_ptr->LM_ClientsCt++;
    if  (new_cl_entry_ptr->CL_procStatus ==
                                        HBM_PROCSTATUS_ACTIVE)
    {
        hbmdc_dc_info.DC_ClientsLiveCt++;
        lm_entry_ptr->LM_ClientsLiveCt++;
        lm_entry_ptr->LM_ClientsActiveCt++;
    }
    else if  (   (new_cl_entry_ptr->CL_procStatus ==
                                        HBM_PROCSTATUS_BLOCKED)
              || (new_cl_entry_ptr->CL_procStatus ==
                                        HBM_PROCSTATUS_OVERDUE))
    {
        hbmdc_dc_info.DC_ClientsLiveCt++;
        lm_entry_ptr->LM_ClientsLiveCt++;
    }
    if  (new_cl_entry_ptr->CL_rptTimeLast >
                                        lm_entry_ptr->LM_rptTimeLast)
    {
        lm_entry_ptr->LM_rptTimeLast = new_cl_entry_ptr->CL_rptTimeLast;
    }

    set_lm_summary_status(lm_entry_ptr);

    if  (lm_entry2_ptr)
    {
        set_lm_summary_status(lm_entry2_ptr);
    }

    return new_cl_entry_ptr;    
}


/*
************************************************************************
**  get_lm_record_data                                                **
**      Parameters: buff_str            pointer to character string   **
**                                      with local monitor data from  **
**                                      the checkpoint file.          **
**                                                                    **
**      Returns:    (Nothing)                                         **
**                                                                    **
**      This routine validates the checkpointed Local Monitor data    **
**      and loads it.                                                 **
**      Fields are delimited by semi-colons (';').                    **
**      If the entry is found to be invalid then it is thrown out.    **
************************************************************************
*/
void    get_lm_record_data( char   *buff_str)
{
    char                   *buff_str_wk1,
                           *buff_str_wk2;

    char                   *date_decode_str = "%Y/%m/%d %T GMT";

    int                     got_val,
                            valid_entry;

    struct hbmdc_lm_fields *hbmdc_lm_entry_new_ptr;

    struct tm               time_wk;
    struct tm              *time_wk_ptr;

/*
**  Validate and load this local monitor record data
*/
    time_wk_ptr = &time_wk;

    valid_entry = 1;
    buff_str_wk1 = buff_str;

    while (isspace(*buff_str_wk1))
    {
        buff_str_wk1++;
    }
    if  (*buff_str_wk1 == '\0')
    {
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: Local Monitor checkpoint data incomplete, "
                "entry ommitted:\n",
                hbmdc_proc_name_str);
        return;
    }

    hbmdc_lm_entry_new_ptr =
                        (struct hbmdc_lm_fields *)
                            malloc(sizeof(struct hbmdc_lm_fields));
    bzero((char *) hbmdc_lm_entry_new_ptr,
          sizeof(struct hbmdc_lm_fields));

/*
**  Get the LM host IP Number
*/
    if  (   (*buff_str_wk1 == ';')
         || (*buff_str_wk1 == '\0'))
    {
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: Local Monitor checkpoint data incomplete, "
                "entry ommitted:\n",
                hbmdc_proc_name_str);
    }
    else
    {
        buff_str_wk2 = index(buff_str_wk1, ';');
        if  (!buff_str_wk2)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: HBMLM checkpoint data incomplete, "
                    "record skipped.\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            hbmdc_lm_entry_new_ptr->LM_hostIPNum_str =
                                (char *) malloc(buff_str_wk2 
                                              - buff_str_wk1 + 1);
            *buff_str_wk2 = '\0';
            strcpy(hbmdc_lm_entry_new_ptr->LM_hostIPNum_str,
                   buff_str_wk1);
            hbmdc_lm_entry_new_ptr->LM_hostIPNum_hostfmt.s_addr =
                                        ntohl(inet_addr(buff_str_wk1));
            if  (hbmdc_lm_entry_new_ptr->LM_hostIPNum_hostfmt.s_addr <= 0)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: HBMLM checkpoint data incomplete, "
                        "record skipped.\n",
                        hbmdc_proc_name_str);
            }
            else
            {
                *buff_str_wk2 = ';';
                buff_str_wk1 = buff_str_wk2 + 1;
            }
        }
    }

/*
**  Get the LM host name.
*/
    if  (valid_entry)
    {
        while (isspace(*buff_str_wk1))
        {
            buff_str_wk1++;
        }
        buff_str_wk2 = index(buff_str_wk1, ';');
        if  (   (!buff_str_wk2)
             || (buff_str_wk2 == buff_str_wk1))
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: HBMLM checkpoint data incomplete, "
                    "record skipped.\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            hbmdc_lm_entry_new_ptr->LM_hostName_str =
                    (char *) malloc(buff_str_wk2 - buff_str_wk1 + 1);
            *buff_str_wk2 = '\0';
            strcpy(hbmdc_lm_entry_new_ptr->LM_hostName_str,
                   buff_str_wk1);
            *buff_str_wk2 = ';';
            buff_str_wk1 = buff_str_wk2 + 1;
        }
    }

/*
**  Get LM port number used for sending report messages.
*/
    if  (valid_entry)
    {
        got_val = sscanf(buff_str_wk1, " %hu",
                         &(hbmdc_lm_entry_new_ptr->LM_portNumRpt_hostfmt));
        if  (   (!got_val)
             || (hbmdc_lm_entry_new_ptr->LM_portNumRpt_hostfmt == 0))
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: LM Report PortNum missing/invalid, "
                    "entry ommitted:\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            buff_str_wk1 = index(buff_str_wk1, ';');
            if  (!buff_str_wk1)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: LM process checkpoint data incomplete, "
                        "entry ommitted:\n",
                        hbmdc_proc_name_str);
            }
            else
            {
                buff_str_wk1++;
            }
        }
    }

/*
**  Set up the LM_addr_rpt fields.
*/
    if  (valid_entry)
    {
        hbmdc_lm_entry_new_ptr->LM_addr_rpt.sin_family = AF_INET;
        hbmdc_lm_entry_new_ptr->LM_addr_rpt.sin_addr.s_addr =
                    htonl(hbmdc_lm_entry_new_ptr->LM_hostIPNum_hostfmt.s_addr);
        hbmdc_lm_entry_new_ptr->LM_addr_rpt.sin_port =
                    htons(hbmdc_lm_entry_new_ptr->LM_portNumRpt_hostfmt);
        if  (   (hbmdc_lm_entry_new_ptr->LM_addr_rpt.sin_addr.s_addr <= 0)
             || (hbmdc_lm_entry_new_ptr->LM_addr_rpt.sin_port == 0))
        {
            valid_entry = 0;
        }
    }

/*
**  Skip the client count, setting it to 0.
*/
    if  (valid_entry)
    {
        buff_str_wk1 = index(buff_str_wk1, ';');
        if  (!buff_str_wk1)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: HBMLM checkpoint data incomplete, "
                    "record skipped.\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            hbmdc_lm_entry_new_ptr->LM_ClientsCt = 0;
            buff_str_wk1++;
        }
    }

/*
**  Skip the live client count, setting it to 0.
*/
    if  (valid_entry)
    {
        buff_str_wk1 = index(buff_str_wk1, ';');
        if  (!buff_str_wk1)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: HBMLM checkpoint data incomplete, "
                    "record skipped.\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            hbmdc_lm_entry_new_ptr->LM_ClientsLiveCt = 0;
            buff_str_wk1++;
        }
    }

/*
**  Skip the active client count, setting it to 0.
*/
    if  (valid_entry)
    {
        buff_str_wk1 = index(buff_str_wk1, ';');
        if  (!buff_str_wk1)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: HBMLM checkpoint data incomplete, "
                    "record skipped.\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            hbmdc_lm_entry_new_ptr->LM_ClientsActiveCt = 0;
            buff_str_wk1++;
        }
    }

/*
**  Skip the summary status, setting it to HBM_LMSTATUS_CLIENTS_NONELIVE.
*/
    if  (valid_entry)
    {
        buff_str_wk1 = index(buff_str_wk1, ';');
        if  (!buff_str_wk1)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: HBMLM checkpoint data incomplete, "
                    "record skipped.\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            hbmdc_lm_entry_new_ptr->LM_SummaryStatus =
                                        HBM_LMSTATUS_CLIENTS_NONE_LIVE;
            buff_str_wk1++;
        }
    }

/*
**  Get the last report message time.
*/
    if  (valid_entry)
    {
        bzero((char *) &time_wk, sizeof(time_wk));

        buff_str_wk1 = (char *) strptime(buff_str_wk1,
                                         date_decode_str,
                                         &time_wk);
        if  (buff_str_wk1 == NULL)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: CL_regTime missing or invalid, "
                    "entry omitted:\n",
                    hbmdc_proc_name_str);
        }
        else
        {

#ifdef HAVE_MKTIME
            time_wk.tm_isdst = 0;
            hbmdc_lm_entry_new_ptr->LM_rptTimeLast =
                        (UTCtime) ((long) mktime(&time_wk) + time_conv_diff);
#else
            hbmdc_lm_entry_new_ptr->LM_rptTimeLast =
                        (UTCtime) ((long) timegm(&time_wk) + time_conv_diff);
            if  (hbmdc_lm_entry_new_ptr->LM_rptTimeLast > 0)
            {
                hbmdc_lm_entry_new_ptr->LM_rptTimeLast +=
                        time_wk.tm_gmtoff;
            }
#endif
            if  (hbmdc_lm_entry_new_ptr->LM_rptTimeLast == 0)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: CL_regTime missing or invalid, "
                        "entry omitted:\n",
                        hbmdc_proc_name_str);
            }
/* ******** Omit code to go to next field when it is the last field.
            else
            {
                buff_str_wk1 = index(buff_str_wk1, ';');
                if  (!buff_str_wk1)
                {
                    valid_entry = 0;
                    fprintf(stderr,
                            "\n%s: Client process checkpoint data incomplete, "
                            "entry ommitted:\n",
                            hbmdc_proc_name_str);
                }
                else
                {
                    file_buff_str_wk1++;
                }
            }
*/
        }
    }


/*
**  Increment the Local Monitor counter and
**  add the Local Monitor record to the table.
*/
    if  (valid_entry)
    {
        add_lm_entry(hbmdc_lm_entry_new_ptr);
    }
    else
    {
        if  (hbmdc_lm_entry_new_ptr->LM_hostIPNum_str != NULL)
        {
            free(hbmdc_lm_entry_new_ptr->LM_hostIPNum_str);
            hbmdc_lm_entry_new_ptr->LM_hostIPNum_str = NULL;
        }
        if  (hbmdc_lm_entry_new_ptr->LM_hostName_str != NULL)
        {
            free(hbmdc_lm_entry_new_ptr->LM_hostName_str);
            hbmdc_lm_entry_new_ptr->LM_hostName_str = NULL;
        }
        free(hbmdc_lm_entry_new_ptr);
        hbmdc_lm_entry_new_ptr = NULL;
    }

    return;
}


/*
************************************************************************
**  get_cl_record_data                                                **
**      Parameters: buff_str            pointer to character string   **
**                                      with client data from the     **
**                                      checkpoint file.              **
**                                                                    **
**      Returns:    (Nothing)                                         **
**                                                                    **
**      This routine validates the checkpointed Client data and       **
**      loads it.                                                     **
**      Fields are delimited by semi-colons (';').                    **
**      If the entry is found to be invalid then it is thrown out.    **
************************************************************************
*/
void    get_cl_record_data( char   *buff_str)
{
    char                   *buff_str_wk1,
                           *buff_str_wk2;

    char                   *date_decode_str = "%Y/%m/%d %T GMT";

    int                     got_val,
                            valid_entry;

    int                     int_wk,
                            wk_cpu_days,
                            wk_cpu_hours,
                            wk_cpu_mins,
                            wk_cpu_secs;

    unsigned int            read_u_int;

    struct hbmdc_cl_fields *hbmdc_cl_entry_new_ptr;

    struct tm               time_wk;

/*
**  Validate and load this client record data
*/
    valid_entry = 1;
    buff_str_wk1 = buff_str;

    while (isspace(*buff_str_wk1))
    {
        buff_str_wk1++;
    }
    if  (*buff_str_wk1 == '\0')
    {
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: Client checkpoint data incomplete, "
                "entry ommitted:\n",
                hbmdc_proc_name_str);
        return;
    }

    hbmdc_cl_entry_new_ptr =
                        (struct hbmdc_cl_fields *)
                            malloc(sizeof(struct hbmdc_cl_fields));
    bzero((char *) hbmdc_cl_entry_new_ptr,
          sizeof(struct hbmdc_cl_fields));

/*
**  Get the Client host IP Number
*/
    if  (   (*buff_str_wk1 == ';')
         || (*buff_str_wk1 == '\0'))
    {
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: Local Monitor checkpoint data incomplete, "
                "entry ommitted:\n",
                hbmdc_proc_name_str);
    }
    else
    {
        buff_str_wk2 = index(buff_str_wk1, ';');
        if  (!buff_str_wk2)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: Client checkpoint data incomplete, "
                    "record skipped.\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            hbmdc_cl_entry_new_ptr->CL_hostIPNum_str =
                                (char *) malloc(buff_str_wk2 
                                              - buff_str_wk1 + 1);
            *buff_str_wk2 = '\0';
            strcpy(hbmdc_cl_entry_new_ptr->CL_hostIPNum_str,
                   buff_str_wk1);
            hbmdc_cl_entry_new_ptr->CL_hostIPNum_hostfmt.s_addr =
                                ntohl(inet_addr(buff_str_wk1));
            if  (hbmdc_cl_entry_new_ptr->CL_hostIPNum_hostfmt.s_addr <= 0)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client checkpoint data incomplete, "
                        "record skipped.\n",
                        hbmdc_proc_name_str);
            }
            else
            {
                *buff_str_wk2 = ';';
                buff_str_wk1 = buff_str_wk2 + 1;
            }
        }
    }

/*
**  Skip the host name.
*/
    if  (valid_entry)
    {
        buff_str_wk2 = index(buff_str_wk1, ';');
        if  (   (!buff_str_wk2)
             || (buff_str_wk2 == buff_str_wk1))
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: Client checkpoint data incomplete, record skipped.\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            buff_str_wk1 = buff_str_wk2 + 1;
        }
    }

/*
**  Get LM port number used for sending report messages.
*/
    if  (valid_entry)
    {
        got_val = sscanf(buff_str_wk1, " %hu",
                         &(hbmdc_cl_entry_new_ptr->CL_LMportNumRpt_hostfmt));
        if  (   (!got_val)
             || (hbmdc_cl_entry_new_ptr->CL_LMportNumRpt_hostfmt == 0))
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: Client LM Report PortNum missing/invalid, "
                    "entry ommitted:\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            buff_str_wk1 = index(buff_str_wk1, ';');
            if  (!buff_str_wk1)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client process checkpoint data incomplete, "
                        "entry ommitted:\n",
                        hbmdc_proc_name_str);
            }
            else
            {
                buff_str_wk1++;
            }
        }
    }

/*
**  Get the Client process name.
*/
    if  (valid_entry)
    {
        while (isspace(*buff_str_wk1))
        {
            buff_str_wk1++;
        }
        buff_str_wk2 = index(buff_str_wk1, ';');
        if  (   (!buff_str_wk2)
             || (buff_str_wk2 == buff_str_wk1))
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: Client checkpoint data incomplete, "
                    "record skipped.\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            hbmdc_cl_entry_new_ptr->CL_procName_str =
                    (char *) malloc(buff_str_wk2 - buff_str_wk1 + 1);
            *buff_str_wk2 = '\0';
            strcpy(hbmdc_cl_entry_new_ptr->CL_procName_str,
                   buff_str_wk1);
            *buff_str_wk2 = ';';
            buff_str_wk1 = buff_str_wk2 + 1;
        }
    }

/*
**  Get Client PID.
*/
    if  (valid_entry)
    {
        got_val = sscanf(buff_str_wk1, " %u",
                         &(hbmdc_cl_entry_new_ptr->CL_procPID));
        if  (   (!got_val)
             || (hbmdc_cl_entry_new_ptr->CL_procPID == 0))
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: Client PID missing/invalid, "
                    "entry ommitted:\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            buff_str_wk1 = index(buff_str_wk1, ';');
            if  (!buff_str_wk1)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client process checkpoint data incomplete, "
                        "entry ommitted:\n",
                        hbmdc_proc_name_str);
            }
            else
            {
                buff_str_wk1++;
            }
        }
    }

/*
**  Get the Client Process status.
*/
    if  (valid_entry)
    {
        got_val = sscanf(buff_str_wk1, " %u", &read_u_int);
        if  (   (!got_val)
             || (read_u_int < HBM_PROCSTATUS_MINVAL)
             || (read_u_int > HBM_PROCSTATUS_MAXVAL))
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: CL_procStatus missing or invalid, "
                    "entry omitted:\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            hbmdc_cl_entry_new_ptr->CL_procStatus = read_u_int;
        }
        buff_str_wk1 = index(buff_str_wk1, ';');
        if  (!buff_str_wk1)
        {
/*          processStatus last or invalid -- skip this one    */
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: Client process checkpoint data incomplete, "
                    "entry ommitted:\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            buff_str_wk1++;
        }
    }

/*
**  Get the Client registration time.
*/
    if  (valid_entry)
    {
        bzero((char *) &time_wk, sizeof(time_wk));

        buff_str_wk1 = (char *) strptime(buff_str_wk1,
                                         date_decode_str,
                                         &time_wk);
        if  (buff_str_wk1 == NULL)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: CL_regTime missing or invalid, "
                    "entry omitted:\n",
                    hbmdc_proc_name_str);
        }
        else
        {

#ifdef HAVE_MKTIME
            time_wk.tm_isdst = 0;
            hbmdc_cl_entry_new_ptr->CL_regTime =
                        (UTCtime) ((long) mktime(&time_wk) + time_conv_diff);
#else
            hbmdc_cl_entry_new_ptr->CL_regTime =
                        (UTCtime) ((long) timegm(&time_wk) + time_conv_diff);
            if  (hbmdc_cl_entry_new_ptr->CL_regTime > 0)
            {
                hbmdc_cl_entry_new_ptr->CL_regTime +=
                        time_wk.tm_gmtoff;
            }
#endif
            if  (hbmdc_cl_entry_new_ptr->CL_regTime == 0)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: CL_regTime missing or invalid, "
                        "entry omitted:\n",
                        hbmdc_proc_name_str);
            }
            else
            {
                buff_str_wk1 = index(buff_str_wk1, ';');
                if  (!buff_str_wk1)
                {
                    valid_entry = 0;
                    fprintf(stderr,
                            "\n%s: Client process checkpoint data incomplete, "
                            "entry ommitted:\n",
                            hbmdc_proc_name_str);
                }
                else
                {
                    buff_str_wk1++;
                }
            }
        }
    }

/*
**  Get the Client blocked (since) time.
*/
    if  (valid_entry)
    {
        bzero((char *) &time_wk, sizeof(time_wk));

        buff_str_wk1 = (char *) strptime(buff_str_wk1,
                                         date_decode_str,
                                         &time_wk);
        if  (buff_str_wk1 == NULL)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: CL_blockedTime missing or invalid, "
                    "entry omitted:\n",
                    hbmdc_proc_name_str);
        }
        else
        {

#ifdef HAVE_MKTIME
            time_wk.tm_isdst = 0;
            hbmdc_cl_entry_new_ptr->CL_blockedTime =
                        (UTCtime) ((long) mktime(&time_wk) + time_conv_diff);
#else
            hbmdc_cl_entry_new_ptr->CL_blockedTime =
                        (UTCtime) ((long) timegm(&time_wk) + time_conv_diff);
            if  (hbmdc_cl_entry_new_ptr->CL_blockedTime > 0)
            {
                hbmdc_cl_entry_new_ptr->CL_blockedTime +=
                        time_wk.tm_gmtoff;
            }
#endif
            if  (   (   (hbmdc_cl_entry_new_ptr->CL_blockedTime == 0)
                     && (hbmdc_cl_entry_new_ptr->CL_procStatus ==
                                        HBM_PROCSTATUS_BLOCKED))
                 || (   (hbmdc_cl_entry_new_ptr->CL_blockedTime != 0)
                     && (hbmdc_cl_entry_new_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_BLOCKED)
                     && (hbmdc_cl_entry_new_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_OVERDUE)
                     && (hbmdc_cl_entry_new_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_NO_RPT)))
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: CL_blockedTime missing or invalid, "
                        "entry omitted:\n",
                        hbmdc_proc_name_str);
            }
            else
            {
                buff_str_wk1 = index(buff_str_wk1, ';');
                if  (!buff_str_wk1)
                {
                    valid_entry = 0;
                    fprintf(stderr,
                            "\n%s: Client process checkpoint data incomplete, "
                            "entry ommitted:\n",
                            hbmdc_proc_name_str);
                }
                else
                {
                    buff_str_wk1++;
                }
            }
        }
    }

/*
**      Get the Client process CPU time.
*/
    if  (valid_entry)
    {
        buff_str_wk2 = index(buff_str_wk1, ':');
        if  (buff_str_wk2)
        {
            if  (*(buff_str_wk2 - 3) == '-')
            {
                got_val = sscanf(buff_str_wk1,
                                 "%d-%d:%d:%d",
                                 &wk_cpu_days,
                                 &wk_cpu_hours,
                                 &wk_cpu_mins,
                                 &wk_cpu_secs);
            }
            else
            {
                wk_cpu_days = 0;
                if  (*(buff_str_wk2 + 3) == ':')
                {
                    got_val = sscanf(buff_str_wk1,
                                     "%d:%d:%d\n",
                                     &wk_cpu_hours,
                                     &wk_cpu_mins,
                                     &wk_cpu_secs);
                    got_val++;
                }
                else
                {
                    wk_cpu_hours = 0;
                    got_val = sscanf(buff_str_wk1,
                                     "%d:%d\n",
                                     &wk_cpu_mins,
                                     &wk_cpu_secs);
                    got_val += 2;
                }
            }
        }
        if  (got_val != 4)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: CL_cpuSecs missing or invalid, "
                    "entry omitted:\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            wk_cpu_secs += (wk_cpu_days  * HBM_TIME_SECS_PER_DAY);
            wk_cpu_secs += (wk_cpu_hours * HBM_TIME_SECS_PER_HOUR);
            wk_cpu_secs += (wk_cpu_mins  * HBM_TIME_SECS_PER_MIN);
            hbmdc_cl_entry_new_ptr->CL_cpuSecs = wk_cpu_secs;

            buff_str_wk1 = index(buff_str_wk1, ';');
            if  (!buff_str_wk1)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client process checkpoint data incomplete, "
                        "entry ommitted:\n",
                        hbmdc_proc_name_str);
            }
            else
            {
                buff_str_wk1++;
            }
        }
    }

/*
**      Get the Client shutdown time.
*/
    if  (valid_entry)
    {
        bzero((char *) &time_wk, sizeof(time_wk));

        buff_str_wk1 = (char *) strptime(buff_str_wk1,
                                         date_decode_str,
                                         &time_wk);
        if  (buff_str_wk1 == NULL)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: shutdownTime missing or invalid, "
                    "entry omitted:\n",
                    hbmdc_proc_name_str);
        }
        else
        {
#ifdef HAVE_MKTIME
            time_wk.tm_isdst = 0;
            hbmdc_cl_entry_new_ptr->CL_shutdownTime =
                    (UTCtime) ((long) mktime(&time_wk) + time_conv_diff);
#else
            hbmdc_cl_entry_new_ptr->CL_shutdownTime =
                    (UTCtime) ((long) timegm(&time_wk) + time_conv_diff);
            if  (hbmdc_cl_entry_new_ptr->CL_shutdownTime > 0)
            {
                hbmdc_cl_entry_new_ptr->CL_shutdownTime +=
                    time_wk.tm_gmtoff;
            }
#endif
            if  (   (hbmdc_cl_entry_new_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_NORMAL)
                 && (hbmdc_cl_entry_new_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_ABNORMAL)
                 && (hbmdc_cl_entry_new_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_DIED))
            {
                hbmdc_cl_entry_new_ptr->CL_shutdownTime = 0;
            }
            else if  (hbmdc_cl_entry_new_ptr->CL_shutdownTime == 0)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: shutdownTime missing or invalid, "
                        "entry omitted:\n",
                        hbmdc_proc_name_str);
            }
            buff_str_wk1 = index(buff_str_wk1, ';');
            if  (!buff_str_wk1)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client process checkpoint data incomplete, "
                        "entry ommitted:\n",
                        hbmdc_proc_name_str);
            }
            else
            {
                buff_str_wk1++;
            }
        }
    }

/*
**      Get the Client delete time.
*/
    if  (valid_entry)
    {
        bzero((char *) &time_wk, sizeof(time_wk));

        buff_str_wk1 = (char *) strptime(buff_str_wk1,
                                         date_decode_str,
                                         &time_wk);
        if  (buff_str_wk1 == NULL)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: deleteTime missing or invalid, "
                    "entry omitted:\n",
                    hbmdc_proc_name_str);
        }
        else
        {
#ifdef HAVE_MKTIME
            time_wk.tm_isdst = 0;
            hbmdc_cl_entry_new_ptr->CL_deleteTime =
                    (UTCtime) ((long) mktime(&time_wk) + time_conv_diff);
#else
            hbmdc_cl_entry_new_ptr->CL_deleteTime =
                    (UTCtime) ((long) timegm(&time_wk) + time_conv_diff);
            if  (hbmdc_cl_entry_new_ptr->CL_deleteTime > 0)
            {
                hbmdc_cl_entry_new_ptr->CL_deleteTime +=
                    time_wk.tm_gmtoff;
            }
#endif
            if  (   (hbmdc_cl_entry_new_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_NORMAL)
                 && (hbmdc_cl_entry_new_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_ABNORMAL)
                 && (hbmdc_cl_entry_new_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_DIED))
            {
                hbmdc_cl_entry_new_ptr->CL_deleteTime = 0;
            }
            else if  (hbmdc_cl_entry_new_ptr->CL_deleteTime == 0)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: deleteTime missing or invalid, "
                        "entry omitted:\n",
                        hbmdc_proc_name_str);
            }
            buff_str_wk1 = index(buff_str_wk1, ';');
            if  (!buff_str_wk1)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client process checkpoint data incomplete, "
                        "entry ommitted:\n",
                        hbmdc_proc_name_str);
            }
            else
            {
                buff_str_wk1++;
            }
        }
    }

/*
**  Get the Client report (message) number.
*/
    if  (valid_entry)
    {
        got_val = sscanf(buff_str_wk1, " %d", &int_wk);
        if  (!got_val)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: CL_rptNumber missing or invalid, "
                    "entry omitted:\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            if  (int_wk <= 0)
            {
                int_wk = 0;
            }
            hbmdc_cl_entry_new_ptr->CL_rptNumber = (unsigned int) int_wk;

            buff_str_wk1 = index(buff_str_wk1, ';');
            if  (!buff_str_wk1)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client process checkpoint data incomplete, "
                        "entry ommitted:\n",
                        hbmdc_proc_name_str);
            }
            else
            {
                buff_str_wk1++;
            }
        }
    }

/*
**  Get the Client last report time.
*/
    if  (valid_entry)
    {
        bzero((char *) &time_wk, sizeof(time_wk));

        buff_str_wk1 = (char *) strptime(buff_str_wk1,
                                         date_decode_str,
                                         &time_wk);
        if  (buff_str_wk1 == NULL)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: rptTimeLast missing or invalid, "
                    "entry omitted:\n",
                    hbmdc_proc_name_str);
        }
        else
        {

#ifdef HAVE_MKTIME
            time_wk.tm_isdst = 0;
            hbmdc_cl_entry_new_ptr->CL_rptTimeLast =
                        (UTCtime) ((long) mktime(&time_wk) + time_conv_diff);
#else
            hbmdc_cl_entry_new_ptr->CL_rptTimeLast =
                        (UTCtime) ((long) timegm(&time_wk) + time_conv_diff);
            if  (hbmdc_cl_entry_new_ptr->CL_rptTimeLast > 0)
            {
                hbmdc_cl_entry_new_ptr->CL_rptTimeLast +=
                time_wk.tm_gmtoff;
            }
#endif
            if  (hbmdc_cl_entry_new_ptr->CL_rptTimeLast < 0)
            {
                hbmdc_cl_entry_new_ptr->CL_rptTimeLast = 0;
            }
            buff_str_wk1 = index(buff_str_wk1, ';');
            if  (!buff_str_wk1)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client process checkpoint data incomplete, "
                        "entry ommitted:\n",
                        hbmdc_proc_name_str);
            }
            else
            {
                buff_str_wk1++;
            }
        }
    }

/*
**  Get the Client report interval (seconds).
*/
    if  (valid_entry)
    {
        got_val = sscanf(buff_str_wk1, " %d", &int_wk);
        if  (!got_val)
        {
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: Client Report Interval missing or invalid, "
                    "entry omitted:\n",
                    hbmdc_proc_name_str);
        }
        else
        {
            if  (int_wk <= 0)
            {
                int_wk = HBMLM_DEFAULT_RPT_INTERVAL;
            }
            hbmdc_cl_entry_new_ptr->CL_rptInterval = int_wk;

/* ******** Omit code to go to next field when it is the last field.
            buff_str_wk1 = index(buff_str_wk1, ';');
            if  (!buff_str_wk1)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client process checkpoint data incomplete, "
                        "entry ommitted:\n",
                        hbmdc_proc_name_str);
            }
            else
            {
                buff_str_wk1++;
            }
*/
        }
    }

/*
**  Set the Client next report time and updated flag.
*/
    if  (valid_entry)
    {
        hbmdc_cl_entry_new_ptr->CL_rptTimeNext =
                                    hbmdc_cl_entry_new_ptr->CL_rptTimeLast;
        while (hbmdc_cl_entry_new_ptr->CL_rptTimeNext <= time_curr.tv_sec)
        {
            hbmdc_cl_entry_new_ptr->CL_rptTimeNext +=
                                    hbmdc_cl_entry_new_ptr->CL_rptInterval;
        }

        hbmdc_cl_entry_new_ptr->CL_updated = 1;
    }


    if  (valid_entry)
    {
/*
**      Add the Client entry.
*/
        add_cl_entry(hbmdc_cl_entry_new_ptr);
    }
    else
    {
        if  (hbmdc_cl_entry_new_ptr->CL_hostIPNum_str != NULL)
        {
            free(hbmdc_cl_entry_new_ptr->CL_hostIPNum_str);
            hbmdc_cl_entry_new_ptr->CL_hostIPNum_str = NULL;
        }
        if  (hbmdc_cl_entry_new_ptr->CL_procName_str != NULL)
        {
            free(hbmdc_cl_entry_new_ptr->CL_procName_str);
            hbmdc_cl_entry_new_ptr->CL_procName_str = NULL;
        }
        free(hbmdc_cl_entry_new_ptr);
        hbmdc_cl_entry_new_ptr = NULL;
    }

    return;
}


/*
************************************************************************
**  read_hbmdc_ckpointfile                                            **
**      Parameters: <None>                                            **
**      Returns:    <None>                                            **
**                                                                    **
**      This routine validates the DC host IP Number and loads the    **
**      checkpointed local monitor and client data.                   **
**      Entries are separated by newline characters ('\n').           **
**      Fields within entries are delimited by semi-colons (';').     **
**      If an entry is found to be invalid then that entry is         **
**      skipped.                                                      **
************************************************************************
*/
void    read_hbmdc_ckpointfile()
{
    char                   *file_buff_str,
                           *file_buff_str_end,
                           *file_buff_str_wk1,
                           *file_buff_str_wk2;

    FILE                   *ckpt_fd;
    int                     ckpt_fd_eof = 0;

    char                   *date_decode_str = "%Y/%m/%d %T GMT";

    u_long                  read_ul_val;
    unsigned int            read_u_int;
    u_short                 read_us_val;

    int                     got_val,
                            retcd,
                            valid_entry;

    int                     i,
                            j,
                            k,
                            int_wk,
                            wk_cpu_days,
                            wk_cpu_hours,
                            wk_cpu_mins,
                            wk_cpu_secs;

    int                     cl_entry_ct,
                            cl_entry_live_ct,
                            lm_entry_ct;

    struct hbmdc_lm_fields *hbmdc_lm_entry_new_ptr;

    struct hbmdc_cl_fields *hbmdc_cl_entry_new_ptr;

    struct in_addr          in_addr_wk;

    time_t                  curr_time,
                            curr_time_conv;

    struct tm               time_wk;

    struct tm              *time_wk_ptr;

    if  ((ckpt_fd = fopen(hbmdc_dc_info.DC_ckptFileName_str, "r")) == NULL)
    {
/*
**      There is no table file -- initialize table to be empty.
*/
        fclose(ckpt_fd);
        return;
    }

/*
**  Set up the buffer areas and the time conversion values.
*/
    file_buff_str = (char *) malloc(HBM_BUFF_SIZE_CKPT);
    file_buff_str_end = file_buff_str + HBM_BUFF_SIZE_CKPT - 1;
    bzero(file_buff_str, HBM_BUFF_SIZE_CKPT);

    time_wk_ptr = &time_wk;
    bzero((char *) time_wk_ptr, sizeof(time_wk));

    retcd = gettimeofday(&time_curr, NULL);

    curr_time = (time_t) time_curr.tv_sec;
#ifdef HAVE_GMTIME_R
    gmtime_r(&curr_time,
             time_wk_ptr);
#else
    time_wk_ptr = gmtime(&curr_time);
#endif

#ifdef HAVE_MKTIME
    curr_time_conv = mktime(time_wk_ptr);
#else
    curr_time_conv = timegm(time_wk_ptr);
#endif
#ifndef HAVE_GMTIME_R
    free(time_wk_ptr);
#endif

    time_conv_diff = (long) curr_time - curr_time_conv;

/*
**  The first line has a literal "HBMDC CHECKPOINT" for file
**  identification.
**  We don't need/use that, just verify that it is an HBMDC checkpoint
**  file and throw it away.
*/
    if  (fgets(file_buff_str, HBM_BUFF_SIZE_CKPT, ckpt_fd) == NULL)
    {
/*      The file is empty -- close it, clean up, and return.          */
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    if  (strncasecmp(file_buff_str, "HBMDC CHECKPOINT", 16))
    {
/*
**      The file is not an HBMDC Checkpoint file --
**      close it, clean up, and return.
*/
        fprintf(stderr,
                "\n%s: Invalid HBMDC Checkpoint file, file skipped.\n",
                hbmdc_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }

/*
**  Read the next line and extract the hbmdc host information.
*/
    if  (fgets(file_buff_str, HBM_BUFF_SIZE_CKPT, ckpt_fd) == NULL)
    {
        fprintf(stderr,
                "\n%s: HBMDC checkpoint data incomplete, file skipped.\n",
                hbmdc_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }

/*
**  Verify that the first line is a DC data record.
*/
    if  (strncasecmp(file_buff_str, "DC Data:", 8))
    {
/*
**      The second record does not have DC data --
**      close it, clean up, and return.
*/
        fprintf(stderr,
                "\n%s: Invalid HBMDC Checkpoint file, file skipped.\n",
                hbmdc_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    file_buff_str_wk1 = file_buff_str + 8;
    while (isspace(*file_buff_str_wk1))
    {
        file_buff_str_wk1++;
    }

/*
**  Check the DC_hostIPNum
*/
    if  (   (*file_buff_str_wk1 == ';')
         || (*file_buff_str_wk1 == '\0'))
    {
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: HBMDC checkpoint data incomplete, file skipped.\n",
                hbmdc_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    file_buff_str_wk2 = index(file_buff_str_wk1, ';');
    if  (!file_buff_str_wk2)
    {
        valid_entry = 0;
        fprintf(stderr,
            "\n%s: HBMDC checkpoint data incomplete, file skipped.\n",
            hbmdc_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    *file_buff_str_wk2 = '\0';
    if  (strcmp(file_buff_str_wk1, hbmdc_dc_info.DC_hostIPNum_str))
    {
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: DC_hostIPNum for different host, file skipped:\n",
                hbmdc_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    *file_buff_str_wk2 = ';';
    file_buff_str_wk1 = file_buff_str_wk2 + 1;

/*
**  Skip the DC_hostName
*/
    file_buff_str_wk2 = index(file_buff_str_wk1, ';');
    if  (!file_buff_str_wk1)
    {
/*      DC_hostName missing or invalid -- skip the file   */
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: HBMDC checkpoint data incomplete, file skipped:\n",
                hbmdc_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    *file_buff_str_wk2 = '\0';
    if  (strcmp(file_buff_str_wk1, hbmdc_dc_info.DC_hostName_str))
    {
/*      DC_hostName missing or for a different host -- skip the file   */
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: HBMDC checkpoint data incomplete, file skipped:\n",
                hbmdc_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    *file_buff_str_wk2 = ';';
    file_buff_str_wk1 = file_buff_str_wk2 + 1;

/*
**  Verify the DC_portNumHBMsg (port number for heartbeat messages)
*/
    got_val = sscanf(file_buff_str_wk1, "%d", &int_wk);
    if  (   (!got_val)
         || ((u_short) int_wk != hbmdc_dc_info.DC_portNumHBMsg))
    {
        fprintf(stderr,
                "\n%s: Warning -- DC_portNumHBMsg missing or different from current",
                hbmdc_proc_name_str);
    } 
    file_buff_str_wk1 = index(file_buff_str_wk1, ';');
    if  (!file_buff_str_wk1)
    {
/*      DC_portNumHBMsg last field -- skip the file     */
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: HBMDC checkpoint data incomplete, file skipped:\n",
                hbmdc_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    file_buff_str_wk1++;

/*
**  Skip the DC_MonitorInterval (Interval in seconds for monitorring clients)
*/
    file_buff_str_wk1 = index(file_buff_str_wk1, ';');
    if  (!file_buff_str_wk1)
    {
/*      MonitorInterval last field -- skip the file    */
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: HBMDC checkpoint data incomplete, file skipped:\n",
                hbmdc_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    file_buff_str_wk1++;

/*
**  Get the DC_LocalMonitorsCt (number of local monitor processes)
*/
    got_val = sscanf(file_buff_str_wk1, "%d", &lm_entry_ct);
    if  (   (!got_val)
         || (lm_entry_ct < 0))
    {
        fprintf(stderr,
                "\n%s: Warning -- local monitor count < 0, using 0",
                hbmdc_proc_name_str);
        lm_entry_ct = 0;
    }
    file_buff_str_wk1 = index(file_buff_str_wk1, ';');
    file_buff_str_wk1++;

/*
**  Get the DC_ClientsCt (number of client processes)
*/
    got_val = sscanf(file_buff_str_wk1, "%d", &cl_entry_ct);
    if  (   (!got_val)
         || (cl_entry_ct < 0))
    {
        fprintf(stderr,
                "\n%s: Warning -- client count < 0, using 0",
                hbmdc_proc_name_str);
        cl_entry_ct = 0;
    }
    file_buff_str_wk1 = index(file_buff_str_wk1, ';');
    file_buff_str_wk1++;

/*
**  Get the DC_ClientsLiveCt (number of live client processes)
*/
    got_val = sscanf(file_buff_str_wk1, "%d", &cl_entry_live_ct);
    if  (   (!got_val)
         || (cl_entry_live_ct < 0))
    {
        fprintf(stderr,
                "\n%s: Warning -- live client count < 0, using 0",
                hbmdc_proc_name_str);
        cl_entry_live_ct = 0;
    }
/*  No need to skip to next field after the last field.
    file_buff_str_wk1 = index(file_buff_str_wk1, ';');
    file_buff_str_wk1++;
*/

/*
**  Read Local Monitor and Client data one line at a time
**  (up to HBM_BUFF_SIZE_CKPT characters) and extract the
**  data for the table entry.
*/
    ckpt_fd_eof = 0;
    while (fgets(file_buff_str, HBM_BUFF_SIZE_CKPT, ckpt_fd) != NULL)
    {
        file_buff_str_wk1 = file_buff_str;

/*
**      Verify that this is a Client data record.
*/
        valid_entry = 1;
        if  (!strncasecmp(file_buff_str_wk1, "LM Data:", 8))
        {
/*
**          Get the local monitor data.
*/
            file_buff_str_wk1 += 8;
            get_lm_record_data(file_buff_str_wk1);
        }
        else if  (!strncasecmp(file_buff_str_wk1, "Client Data:", 12))
        {
/*
**          Get the client data.
*/
            file_buff_str_wk1 += 12;
            get_cl_record_data(file_buff_str_wk1);
        }
        else
        {
/*
**          This record is of an invalid type -- skip it.
*/
            fprintf(stderr,
                    "\n%s: Invalid record -- client data expected, "
                    "record skipped.\n",
                    hbmdc_proc_name_str);
            valid_entry = 0;
        }
    }
    if  (lm_entry_ct != hbmdc_dc_info.DC_LocalMonitorsCt)
    {
        fprintf(stderr,
                    "\n%s: Warning -- Local Monitor count != count from "
                    "checkpoint file.\n",
                    hbmdc_proc_name_str);
    }
    if  (cl_entry_ct != hbmdc_dc_info.DC_ClientsCt)
    {
        fprintf(stderr,
                    "\n%s: Warning -- Client count != count from "
                    "checkpoint file.\n",
                    hbmdc_proc_name_str);
    }
    if  (cl_entry_live_ct != hbmdc_dc_info.DC_ClientsLiveCt)
    {
        fprintf(stderr,
                    "\n%s: Warning -- live Client count != count from "
                    "checkpoint file.\n",
                    hbmdc_proc_name_str);
    }

    fclose(ckpt_fd);
    free(file_buff_str);
    return;
}


/*
************************************************************************
**  set_up_udp_port                                                   **
**      Parameters: pointer to fd field to use for the socket         **
**                  pointer to the (already filled in) port address   **
**      Returns:    integer value: 0 - socket successfully set up     **
**                                <0 - error setting up the port      **
**                                                                    **
**      This routine sets up a udp port for receiving registration    **
**      and unregistration messages from clients, or for              **
**      sending report messages to data collectors.                   **
************************************************************************
*/
int     set_up_udp_port(    int                *sock_fd_udp_ptr,
                            struct sockaddr_in *udpaddr_in_ptr)
{
    if  ((*sock_fd_udp_ptr = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
    {
        fprintf(stderr,
                "\n%s: socket() call for heartbeat port failed\n",
                hbmdc_proc_name_str);
        perror(hbmdc_proc_name_str);
        return -1;
    }

    if  (bind(*sock_fd_udp_ptr,
              (struct sockaddr *) udpaddr_in_ptr,
              sizeof(*udpaddr_in_ptr)) < 0)
    {
        fprintf(stderr,
                "\n%s: bind() failed.\n",
                hbmdc_proc_name_str);
        perror(hbmdc_proc_name_str);
        return -2;
    }

    return 0;
}
/* MYPOINTER */

/*
************************************************************************
**  write_hbmdc_ckpointfile                                           **
**      Parameters: (None)                                            **
**      Returns:    (Nothing)                                         **
**                                                                    **
**      This routine writes the checkpoint file based on the data     **
**      in the hbmdc_dc_info structure and the hbmdc_lm_info and      **
**      hbmdc_cl_info tables.                                         **
**      The fields are separated by semi-colons (';'),                **
**      and entries are separated by new-lines ('\n').                **
************************************************************************
*/
void    write_hbmdc_ckpointfile()
{
    FILE                   *ckpt_fd;

    struct in_addr          ipaddr_wk;

    struct hbmdc_cl_fields *cl_entry_ptr;

    struct hbmdc_lm_fields *lm_entry_ptr;

    struct tm               tm_struct_wk;
    struct tm              *tm_struct_wk_ptr;

    char                    date_time_str[HBM_DATE_TIME_LEN];
    char                   *date_time_fmt_out_str = "%Y/%m/%d %T GMT";

    int                     i;
    int                     retcd;

    unsigned int            secs_left;

    tm_struct_wk_ptr = &tm_struct_wk;

/*  First get the current time (and convert to UTC)     */

    retcd = gettimeofday(&time_curr, NULL);

/*
**  Now write the table file.                                         **
*/
    if  ((ckpt_fd = fopen(hbmdc_dc_info.DC_ckptFileNameWk_str, "w")) == NULL)
    {
        fprintf(stderr,
                "\n%s: Error opening hbmdc checkpoint file",
                hbmdc_proc_name_str);
        fprintf(stderr,
                "\n%s: Continuing without writing file\n",
                hbmdc_proc_name_str);
        return;
    }

/*
**  First output the header literal and the current date and time     **
**  for identification and "version control".                         **
*/
#ifdef HAVE_GMTIME_R
    gmtime_r((time_t *) &(time_curr.tv_sec), tm_struct_wk_ptr);
#else
    tm_struct_wk_ptr = gmtime((time_t *) &(time_curr.tv_sec));
#endif
    strftime(date_time_str,
             HBM_DATE_TIME_LEN,
             date_time_fmt_out_str,
             tm_struct_wk_ptr);
#ifndef HAVE_GMTIME_R
    free(tm_struct_wk_ptr);
#endif
    fprintf(ckpt_fd, "HBMDC CHECKPOINT:%s\n",
            date_time_str);
/*
**  Now write the Data Collector data to the file.                     **
*/
    fprintf(ckpt_fd, "DC Data:%s;%s;%hu;%u;%lu;%lu;%lu\n",
            hbmdc_dc_info.DC_hostIPNum_str,
            hbmdc_dc_info.DC_hostName_str,
            hbmdc_dc_info.DC_portNumHBMsg,
            hbmdc_dc_info.DC_MonitorInterval,
            hbmdc_dc_info.DC_LocalMonitorsCt,
            hbmdc_dc_info.DC_ClientsCt,
            hbmdc_dc_info.DC_ClientsLiveCt);
/*
**  Now write the Local Monitor data (for each Local Monitor).
*/
    for (lm_entry_ptr = hbmdc_lm_list.head;
         lm_entry_ptr;
         lm_entry_ptr = lm_entry_ptr->next)
    {
        fprintf(ckpt_fd,
                "LM Data:%s;%s;%hu;%lu;%lu;%lu;%u;",
                lm_entry_ptr->LM_hostIPNum_str,
                lm_entry_ptr->LM_hostName_str,
                lm_entry_ptr->LM_portNumRpt_hostfmt,
                lm_entry_ptr->LM_ClientsCt,
                lm_entry_ptr->LM_ClientsLiveCt,
                lm_entry_ptr->LM_ClientsActiveCt,
                lm_entry_ptr->LM_SummaryStatus);

#ifdef HAVE_GMTIME_R
        gmtime_r(&(lm_entry_ptr->LM_rptTimeLast),
                 tm_struct_wk_ptr);
#else
        tm_struct_wk_ptr = gmtime(&(lm_entry_ptr->LM_rptTimeLast));
#endif
        strftime(date_time_str,
                 HBM_DATE_TIME_LEN,
                 date_time_fmt_out_str,
                 tm_struct_wk_ptr);
#ifndef HAVE_GMTIME_R
        free(tm_struct_wk_ptr);
#endif
        fprintf(ckpt_fd, "%s\n",
                date_time_str);
    }

/*
**  Now write the client data (for each client process).
*/
/* MYPOINTER */
    for (cl_entry_ptr = hbmdc_cl_list.head;
         cl_entry_ptr;
         cl_entry_ptr = cl_entry_ptr->next)
    {

        fprintf(ckpt_fd, "Client Data:%s;%s;%hu;%s;%u;%u;",
                cl_entry_ptr->CL_hostIPNum_str,
                cl_entry_ptr->CL_LocalMonitor_ptr->LM_hostName_str,
                cl_entry_ptr->CL_LMportNumRpt_hostfmt,
                cl_entry_ptr->CL_procName_str,
                cl_entry_ptr->CL_procPID,
                cl_entry_ptr->CL_procStatus);
                    
#ifdef HAVE_GMTIME_R
        gmtime_r(&(cl_entry_ptr->CL_regTime),
                 tm_struct_wk_ptr);
#else
        tm_struct_wk_ptr = gmtime(&(cl_entry_ptr->CL_regTime));
#endif
        strftime(date_time_str,
                 HBM_DATE_TIME_LEN,
                 date_time_fmt_out_str,
                 tm_struct_wk_ptr);
#ifndef HAVE_GMTIME_R
        free(tm_struct_wk_ptr);
#endif
        fprintf(ckpt_fd, "%s;",
                date_time_str);

#ifdef HAVE_GMTIME_R
        gmtime_r(&(cl_entry_ptr->CL_blockedTime),
                 tm_struct_wk_ptr);
#else
        tm_struct_wk_ptr = gmtime(&(cl_entry_ptr->CL_blockedTime));
#endif
        strftime(date_time_str,
                 HBM_DATE_TIME_LEN,
                 date_time_fmt_out_str,
                 tm_struct_wk_ptr);
#ifndef HAVE_GMTIME_R
        free(tm_struct_wk_ptr);
#endif
        fprintf(ckpt_fd, "%s;",
                date_time_str);

        secs_left = cl_entry_ptr->CL_cpuSecs;
        if  (secs_left >= HBM_TIME_SECS_PER_DAY)
        {
            fprintf(ckpt_fd, "%u-",
                    secs_left / HBM_TIME_SECS_PER_DAY);
            secs_left %= HBM_TIME_SECS_PER_DAY;
            fprintf(ckpt_fd, "%2.2u:",
                    secs_left / HBM_TIME_SECS_PER_HOUR);
            secs_left %= HBM_TIME_SECS_PER_HOUR;
        }
        else if  (secs_left >= HBM_TIME_SECS_PER_HOUR)
        {
            fprintf(ckpt_fd, "%2.2u:",
                    secs_left / HBM_TIME_SECS_PER_HOUR);
            secs_left %= HBM_TIME_SECS_PER_HOUR;
        }
        fprintf(ckpt_fd, "%2.2u:%2.2u;",
                (secs_left / HBM_TIME_SECS_PER_MIN),
                (secs_left % HBM_TIME_SECS_PER_MIN));

#ifdef HAVE_GMTIME_R
        gmtime_r(&(cl_entry_ptr->CL_shutdownTime),
                 tm_struct_wk_ptr);
#else
        tm_struct_wk_ptr = gmtime(&(cl_entry_ptr->CL_shutdownTime));
#endif
        strftime(date_time_str,
                 HBM_DATE_TIME_LEN,
                 date_time_fmt_out_str,
                 tm_struct_wk_ptr);
#ifndef HAVE_GMTIME_R
        free(tm_struct_wk_ptr);
#endif
        fprintf(ckpt_fd, "%s;",
                date_time_str);

#ifdef HAVE_GMTIME_R
        gmtime_r(&(cl_entry_ptr->CL_deleteTime),
                 tm_struct_wk_ptr);
#else
        tm_struct_wk_ptr = gmtime(&(cl_entry_ptr->CL_deleteTime));
#endif
        strftime(date_time_str,
                 HBM_DATE_TIME_LEN,
                 date_time_fmt_out_str,
                 tm_struct_wk_ptr);
#ifndef HAVE_GMTIME_R
        free(tm_struct_wk_ptr);
#endif
        fprintf(ckpt_fd, "%s;%u;",
                date_time_str,
                cl_entry_ptr->CL_rptNumber);

#ifdef HAVE_GMTIME_R
        gmtime_r(&(cl_entry_ptr->CL_rptTimeLast),
                 tm_struct_wk_ptr);
#else
        tm_struct_wk_ptr = gmtime(&(cl_entry_ptr->CL_rptTimeLast));
#endif
        strftime(date_time_str,
                 HBM_DATE_TIME_LEN,
                 date_time_fmt_out_str,
                 tm_struct_wk_ptr);
#ifndef HAVE_GMTIME_R
        free(tm_struct_wk_ptr);
#endif
        fprintf(ckpt_fd, "%s;%u\n",
                date_time_str,
                cl_entry_ptr->CL_rptInterval);
    }
/*
**  Close the file and return.
*/
    fclose(ckpt_fd);
    unlink(hbmdc_dc_info.DC_ckptFileName_str);
    link(hbmdc_dc_info.DC_ckptFileNameWk_str,
         hbmdc_dc_info.DC_ckptFileName_str);
    unlink(hbmdc_dc_info.DC_ckptFileNameWk_str);

    return;
}

/*
************************************************************************
**  process_heartbeat_msg                                             **
**      Parameters: sock_fd_hbmsg           the socket fd to get the  **
**                                          message from              **
**      Returns:    (Nothing)                                         **
**                                                                    **
**      This routine reads in a message from the specified socket,    **
**      verifies that it is a heartbeat message, and                  **
**      processes the message.                                        **
**                                                                    **
**      LM and Client entries are created or updated as appropriate   **
**      if the message is valid.                                      **
************************************************************************
*/
void    process_heartbeat_msg(  int     sock_fd_hbmsg)
{
    struct hbmdc_cl_fields *cl_entry_ptr,
                           *cl_entry_new_ptr;

    struct hbmdc_lm_fields *lm_entry_ptr,
                           *lm_entry_old_ptr;

    struct sockaddr_in      from_addr;

#if  (   (defined TARGET_ARCH_CRAYT3E) \
      || (defined TARGET_ARCH_HPUX) \
      || (defined TARGET_ARCH_IRIX) \
      || (defined TARGET_ARCH_SOLARIS))
   int                     from_addr_len;
#else
    size_t                  from_addr_len;
#endif

    char                   *str_ptr;

    int                     msg_len;
    int                     nleft;
    int                     nread;

    int                     retcd;

/*
**  Heartbeat port was selected,                                        **
**  so get the message and process the report.                          **
*/
    lm_entry_old_ptr = NULL;
    retcd = gettimeofday(&time_curr, NULL);

    bzero((void *)&from_addr, sizeof(from_addr));
    from_addr.sin_family = AF_INET;
    from_addr_len = sizeof(from_addr);
    msg_len = recvfrom(sock_fd_hbmsg,
                       hbm_msg_Rptbuff,
                       HBM_BUFF_SIZE_MSG,
                       0,
                       (struct sockaddr *) &from_addr,
                       &from_addr_len);

    if  (errno == EAGAIN)
    {
        return;
    }
    if  (strncmp(hbm_msg_Rptbuff, "HBMLM", 5))
    {
        return;
    }

    hbm_msg_RptmsgLength_ptr = hbm_msg_Rptbuff + 5;
    UnpackUInt32(hbm_msg_RptmsgLength_ptr, hbm_msg_RptmsgLength);
    if  (msg_len != hbm_msg_RptmsgLength)
    {
        return;
    }

    hbm_msg_RptLMhostName_ptr = hbm_msg_RptmsgLength_ptr + 4;
    if  (*hbm_msg_RptLMhostName_ptr == '\0')
    {
        return;
    }

    hbm_msg_RptCLprocessName_ptr = hbm_msg_RptLMhostName_ptr;
    while (*hbm_msg_RptCLprocessName_ptr != '\0')
    {
        hbm_msg_RptCLprocessName_ptr++;
    }
    hbm_msg_RptCLprocessName_ptr++;
    if  (*hbm_msg_RptCLprocessName_ptr == '\0')
    {
        return;
    }

    hbm_msg_RptCLprocessPID_ptr = hbm_msg_RptCLprocessName_ptr;
    while (*hbm_msg_RptCLprocessPID_ptr != '\0')
    {
        hbm_msg_RptCLprocessPID_ptr++;
    }
    hbm_msg_RptCLprocessPID_ptr++;

    hbm_msg_RptCLstatus_ptr = hbm_msg_RptCLprocessPID_ptr + 4;
    hbm_msg_RptCLrptNumber_ptr = hbm_msg_RptCLstatus_ptr + 4;
    hbm_msg_RptCLrptTime_ptr = hbm_msg_RptCLrptNumber_ptr + 4;
    hbm_msg_RptCLrptInterval_ptr = hbm_msg_RptCLrptTime_ptr + 4;
    hbm_msg_RptCLregTime_ptr = hbm_msg_RptCLrptInterval_ptr + 4;
    hbm_msg_RptCLblockedTime_ptr = hbm_msg_RptCLregTime_ptr + 4;
    hbm_msg_RptCLcpuSecs_ptr = hbm_msg_RptCLblockedTime_ptr + 4;
    hbm_msg_RptCLshutdownTime_ptr = hbm_msg_RptCLcpuSecs_ptr + 4;
    hbm_msg_RptCLshutdownMsgCt_ptr = hbm_msg_RptCLshutdownTime_ptr + 4;

/*
**  Create a new Client entry,
**  fill in the fields,
**  check for Local Monitor entry and create if not there.
*/
    cl_entry_new_ptr = (struct hbmdc_cl_fields *)
                                malloc(sizeof(struct hbmdc_cl_fields));

    cl_entry_new_ptr->CL_LocalMonitor_ptr = NULL;
    cl_entry_new_ptr->CL_hostIPNum_hostfmt.s_addr = ntohl(from_addr.sin_addr.s_addr);
    cl_entry_new_ptr->CL_hostIPNum_str = NULL;
/*
**  CL_hostIPNum_str will be set later if necessary.
*/
    cl_entry_new_ptr->CL_LMportNumRpt_hostfmt = ntohs(from_addr.sin_port);
    cl_entry_new_ptr->CL_procName_str = NULL;
/*
**  CL_procName_str will be set later if necessary.
*/
    UnpackUInt32(hbm_msg_RptCLprocessPID_ptr,
                 cl_entry_new_ptr->CL_procPID);
    UnpackUInt32(hbm_msg_RptCLstatus_ptr,
                 cl_entry_new_ptr->CL_procStatus);
    UnpackUInt32(hbm_msg_RptCLregTime_ptr,
                 cl_entry_new_ptr->CL_regTime);
    UnpackUInt32(hbm_msg_RptCLblockedTime_ptr,
                 cl_entry_new_ptr->CL_blockedTime);
    UnpackUInt32(hbm_msg_RptCLcpuSecs_ptr,
                 cl_entry_new_ptr->CL_cpuSecs);
    if  (   (cl_entry_new_ptr->CL_procStatus ==
                                    HBM_PROCSTATUS_ACTIVE)
         || (cl_entry_new_ptr->CL_procStatus ==
                                    HBM_PROCSTATUS_BLOCKED))
    {
        cl_entry_new_ptr->CL_shutdownTime = 0;
        cl_entry_new_ptr->CL_deleteTime = 0;
    }
    else
    {
        UnpackUInt32(hbm_msg_RptCLshutdownTime_ptr,
                     cl_entry_new_ptr->CL_shutdownTime);
        cl_entry_new_ptr->CL_deleteTime = cl_entry_new_ptr->CL_shutdownTime
                                        + HBMDC_CL_SHUTDOWN_RETENTION_SECS;
    }
    UnpackUInt32(hbm_msg_RptCLrptNumber_ptr,
                 cl_entry_new_ptr->CL_rptNumber);
    cl_entry_new_ptr->CL_rptTimeLast = time_curr.tv_sec;
    UnpackUInt32(hbm_msg_RptCLrptInterval_ptr,
                 cl_entry_new_ptr->CL_rptInterval);
    if  (   (cl_entry_new_ptr->CL_procStatus ==
                                    HBM_PROCSTATUS_ACTIVE)
         || (cl_entry_new_ptr->CL_procStatus ==
                                    HBM_PROCSTATUS_BLOCKED))
    {
        cl_entry_new_ptr->CL_rptTimeNext = cl_entry_new_ptr->CL_rptTimeLast
                                        + cl_entry_new_ptr->CL_rptInterval;
    }
    else
    {
        cl_entry_new_ptr->CL_rptTimeNext = 0;
    }

    lm_entry_ptr = find_lm_entry(&(cl_entry_new_ptr->CL_hostIPNum_hostfmt),
                                 cl_entry_new_ptr->CL_LMportNumRpt_hostfmt);

    if  (!lm_entry_ptr)
    {
        lm_entry_ptr = (struct hbmdc_lm_fields *)
                                malloc(sizeof(struct hbmdc_lm_fields));

        lm_entry_ptr->LM_hostIPNum_hostfmt =
                                cl_entry_new_ptr->CL_hostIPNum_hostfmt;

        str_ptr = inet_ntoa(from_addr.sin_addr);
        lm_entry_ptr->LM_hostIPNum_str =
                                (char *) malloc(strlen(str_ptr) + 1);
        strcpy(lm_entry_ptr->LM_hostIPNum_str, str_ptr);
        str_ptr = NULL;

        lm_entry_ptr->LM_hostName_str =
                                (char *) malloc(strlen(hbm_msg_RptLMhostName_ptr) + 1);
        strcpy(lm_entry_ptr->LM_hostName_str,
               hbm_msg_RptLMhostName_ptr);

        lm_entry_ptr->LM_portNumRpt_hostfmt = ntohs(from_addr.sin_port);

        lm_entry_ptr->LM_addr_rpt.sin_family = from_addr.sin_family;
        lm_entry_ptr->LM_addr_rpt.sin_addr.s_addr =
                                from_addr.sin_addr.s_addr;
        lm_entry_ptr->LM_addr_rpt.sin_port = from_addr.sin_port;

        lm_entry_ptr->LM_ClientsCt = 0;
        lm_entry_ptr->LM_ClientsLiveCt = 0;
        lm_entry_ptr->LM_ClientsActiveCt = 0;
        lm_entry_ptr->LM_SummaryStatus = HBM_LMSTATUS_CLIENTS_NONE_LIVE;
        lm_entry_ptr->LM_rptTimeLast = time_curr.tv_sec;

        lm_entry_ptr = add_lm_entry(lm_entry_ptr);
    }

    cl_entry_new_ptr->CL_LocalMonitor_ptr = lm_entry_ptr;

    cl_entry_ptr = find_cl_entry(&(cl_entry_new_ptr->CL_hostIPNum_hostfmt),
                                 cl_entry_new_ptr->CL_procPID);
    if  (!cl_entry_ptr)
    {
        cl_entry_new_ptr->CL_hostIPNum_str =
                        (char *) malloc(strlen(lm_entry_ptr->LM_hostIPNum_str) + 1);
        strcpy(cl_entry_new_ptr->CL_hostIPNum_str,
               lm_entry_ptr->LM_hostIPNum_str);

        cl_entry_new_ptr->CL_procName_str =
                        (char *) malloc(hbm_msg_RptCLprocessPID_ptr
                                      - hbm_msg_RptCLprocessName_ptr + 1);

        strcpy(cl_entry_new_ptr->CL_procName_str,
               hbm_msg_RptCLprocessName_ptr);

        cl_entry_ptr = add_cl_entry(cl_entry_new_ptr);
    }
    else
    {
        if  (cl_entry_new_ptr->CL_rptNumber > cl_entry_ptr->CL_rptNumber)
        {
            lm_entry_old_ptr = cl_entry_ptr->CL_LocalMonitor_ptr;
            hbmdc_dc_info.DC_ClientsCt--;
            lm_entry_old_ptr->LM_ClientsCt--;
            if  (cl_entry_ptr->CL_procStatus ==
                                    HBM_PROCSTATUS_ACTIVE)
            {
                hbmdc_dc_info.DC_ClientsLiveCt--;
                lm_entry_old_ptr->LM_ClientsLiveCt--;
                lm_entry_old_ptr->LM_ClientsActiveCt--;
            }
            else if  (   (cl_entry_ptr->CL_procStatus ==
                                    HBM_PROCSTATUS_BLOCKED)
                      || (cl_entry_ptr->CL_procStatus ==
                                    HBM_PROCSTATUS_OVERDUE))
            {
                hbmdc_dc_info.DC_ClientsLiveCt--;
                lm_entry_old_ptr->LM_ClientsLiveCt--;
            }

            cl_entry_ptr->CL_LocalMonitor_ptr = lm_entry_ptr;
            cl_entry_ptr->CL_procStatus = cl_entry_new_ptr->CL_procStatus;
            hbmdc_dc_info.DC_ClientsCt++;
            lm_entry_ptr->LM_ClientsCt++;
            if  (cl_entry_ptr->CL_procStatus == HBM_PROCSTATUS_ACTIVE)
            {
                hbmdc_dc_info.DC_ClientsLiveCt++;
                lm_entry_ptr->LM_ClientsLiveCt++;
                lm_entry_ptr->LM_ClientsActiveCt++;
            }
            if  (cl_entry_ptr->CL_procStatus == HBM_PROCSTATUS_BLOCKED)
            {
                hbmdc_dc_info.DC_ClientsLiveCt++;
                lm_entry_ptr->LM_ClientsLiveCt++;
            }

            if  (strcmp(hbm_msg_RptLMhostName_ptr,
                        lm_entry_ptr->LM_hostName_str))
            {
                free(lm_entry_ptr->LM_hostName_str);
                lm_entry_ptr->LM_hostName_str =
                                (char *) malloc(strlen(hbm_msg_RptLMhostName_ptr) + 1);
                strcpy(lm_entry_ptr->LM_hostName_str,
                       hbm_msg_RptLMhostName_ptr);
            }
            if  (strcmp(hbm_msg_RptCLprocessName_ptr,
                        cl_entry_ptr->CL_procName_str))
            {
                free(cl_entry_ptr->CL_procName_str);
                cl_entry_ptr->CL_procName_str =
                                (char *) malloc(strlen(hbm_msg_RptCLprocessName_ptr) + 1);
                strcpy(cl_entry_ptr->CL_procName_str,
                       hbm_msg_RptCLprocessName_ptr);
            }

            cl_entry_ptr->CL_regTime = cl_entry_new_ptr->CL_regTime;
            cl_entry_ptr->CL_blockedTime = cl_entry_new_ptr->CL_blockedTime;
            cl_entry_ptr->CL_cpuSecs = cl_entry_new_ptr->CL_cpuSecs;
            cl_entry_ptr->CL_shutdownTime = cl_entry_new_ptr->CL_shutdownTime;
            cl_entry_ptr->CL_deleteTime = cl_entry_new_ptr->CL_deleteTime;
            cl_entry_ptr->CL_rptNumber = cl_entry_new_ptr->CL_rptNumber;
            cl_entry_ptr->CL_rptTimeLast = cl_entry_new_ptr->CL_rptTimeLast;
            cl_entry_ptr->CL_rptInterval = cl_entry_new_ptr->CL_rptInterval;
            cl_entry_ptr->CL_rptTimeNext = cl_entry_new_ptr->CL_rptTimeNext;

            free(cl_entry_new_ptr);
            if  (lm_entry_old_ptr == lm_entry_ptr)
            {
                lm_entry_old_ptr = NULL;
            }
        }
    }

    if  (lm_entry_old_ptr)
    {
        set_lm_summary_status(lm_entry_old_ptr);
    }
    lm_entry_ptr->LM_rptTimeLast = time_curr.tv_sec;
    set_lm_summary_status(lm_entry_ptr);
}


/*
************************************************************************
**  update_client_and_lm_data                                         **
**      Parameters: (None)                                            **
**                                                                    **
**      Returns:    (Nothing)                                         **
**                                                                    **
**      This routine checks all Client and Local Monitor entries,     **
**      updating those missing report messages as appropriate.        **
************************************************************************
*/
void    update_client_and_lm_data()
{
    int                     num_rpt_overdue;

    int                     retcd;

    struct hbmdc_cl_fields *cl_entry_ptr,
                           *cl_entry_next_ptr;

    struct hbmdc_lm_fields *lm_entry_ptr,
                           *lm_entry_next_ptr;

    struct timeval          time_next_rpt_overdue,
                            time_next_rpt_shutdown;

    retcd = gettimeofday(&time_curr, NULL);
    time_next_rpt_overdue.tv_sec = time_curr.tv_sec
                                    - HBMDC_MSG_DELAY_OVERDUE;
    time_next_rpt_overdue.tv_usec = 0;

    cl_entry_ptr = hbmdc_cl_list.head;
    while (cl_entry_ptr)
    {
        cl_entry_next_ptr = cl_entry_ptr->next;

        if  (cl_entry_ptr->CL_procStatus == HBM_PROCSTATUS_ACTIVE)
        {
            if  (time_next_rpt_overdue.tv_sec > cl_entry_ptr->CL_rptTimeNext)
            {
                cl_entry_ptr->CL_procStatus = HBM_PROCSTATUS_OVERDUE;
                cl_entry_ptr->CL_LocalMonitor_ptr->LM_ClientsActiveCt--;

                num_rpt_overdue = ((time_next_rpt_overdue.tv_sec
                                    - cl_entry_ptr->CL_rptTimeNext)
                                   / cl_entry_ptr->CL_rptInterval)
                                  +1;
                if  (num_rpt_overdue >= HBMDC_MSG_NUM_OVERDUE_SHUTDOWN)
                {
                    cl_entry_ptr->CL_procStatus =
                                    HBM_PROCSTATUS_SHUTDOWN_NO_RPT;
                    cl_entry_ptr->CL_shutdownTime =
                                    time_curr.tv_sec;
                    cl_entry_ptr->CL_deleteTime =
                                    time_curr.tv_sec
                                    + HBMDC_CL_SHUTDOWN_RETENTION_SECS;
                    hbmdc_dc_info.DC_ClientsLiveCt--;
                    cl_entry_ptr->CL_LocalMonitor_ptr->LM_ClientsLiveCt--;
                }
            }
        }
        else if  (cl_entry_ptr->CL_procStatus == HBM_PROCSTATUS_BLOCKED)
        {
            if  (time_next_rpt_overdue.tv_sec > cl_entry_ptr->CL_rptTimeNext)
            {
                cl_entry_ptr->CL_procStatus = HBM_PROCSTATUS_OVERDUE;

                num_rpt_overdue = ((time_next_rpt_overdue.tv_sec
                                    - cl_entry_ptr->CL_rptTimeNext)
                                   / cl_entry_ptr->CL_rptInterval)
                                  +1;
                if  (num_rpt_overdue >= HBMDC_MSG_NUM_OVERDUE_SHUTDOWN)
                {
                    cl_entry_ptr->CL_procStatus =
                                    HBM_PROCSTATUS_SHUTDOWN_NO_RPT;
                    cl_entry_ptr->CL_shutdownTime =
                                    time_curr.tv_sec;
                    cl_entry_ptr->CL_deleteTime =
                                    time_curr.tv_sec
                                    + HBMDC_CL_SHUTDOWN_RETENTION_SECS;
                    hbmdc_dc_info.DC_ClientsLiveCt--;
                    cl_entry_ptr->CL_LocalMonitor_ptr->LM_ClientsLiveCt--;
                }
            }
        }
        else if  (cl_entry_ptr->CL_procStatus == HBM_PROCSTATUS_OVERDUE)
        {
            num_rpt_overdue = ((time_next_rpt_overdue.tv_sec
                                - cl_entry_ptr->CL_rptTimeNext)
                               / cl_entry_ptr->CL_rptInterval)
                              +1;
            if  (num_rpt_overdue >= HBMDC_MSG_NUM_OVERDUE_SHUTDOWN)
            {
                cl_entry_ptr->CL_procStatus =
                                    HBM_PROCSTATUS_SHUTDOWN_NO_RPT;
                cl_entry_ptr->CL_shutdownTime =
                                    time_curr.tv_sec;
                cl_entry_ptr->CL_deleteTime =
                                    time_curr.tv_sec
                                    + HBMDC_CL_SHUTDOWN_RETENTION_SECS;
                hbmdc_dc_info.DC_ClientsLiveCt--;
                cl_entry_ptr->CL_LocalMonitor_ptr->LM_ClientsLiveCt--;
            }
        }
        else
        {
            if  (time_curr.tv_sec > cl_entry_ptr->CL_deleteTime)
            {
                hbmdc_dc_info.DC_ClientsCt--;
                cl_entry_ptr->CL_LocalMonitor_ptr->LM_ClientsCt--;
                ListRemove(&hbmdc_cl_list, cl_entry_ptr);
                free(cl_entry_ptr->CL_procName_str);
                free(cl_entry_ptr);
            }
        }

        cl_entry_ptr = cl_entry_next_ptr;
    }

    lm_entry_ptr = hbmdc_lm_list.head;
    while (lm_entry_ptr)
    {
        lm_entry_next_ptr = lm_entry_ptr->next;

        if  (lm_entry_ptr->LM_ClientsCt)
        {
            set_lm_summary_status(lm_entry_ptr);
        }
        else
        {
            ListRemove(&hbmdc_lm_list, lm_entry_ptr);
            free(lm_entry_ptr->LM_hostIPNum_str);
            free(lm_entry_ptr->LM_hostName_str);
            free(lm_entry_ptr);
            hbmdc_dc_info.DC_LocalMonitorsCt--;
        }

        lm_entry_ptr = lm_entry_next_ptr;
    }
}


/*
************************************************************************
**  shutdown_quit                                                     **
**      Parameters: dummy       integer parameter that is ignored.    **
**                                                                    **
**      Returns:    (Nothing)                                         **
**                                                                    **
**      This routine sets the "quit_now" flag to 1.                   **
************************************************************************
*/
#ifdef  TARGET_ARCH_HPUX
void    shutdown_quit()
#else
void    shutdown_quit(  int     dummy)
#endif
{
    quit_now = 1;
}


/*
************************************************************************
**  main                                                              **
**      Parameters: program parameters (argc, argv)                   **
**      Returns:    (void)                                            **
**                                                                    **
************************************************************************
*/
main(   int     argc,
        char   *argv[])
{
    char                   *parm_fname_str = NULL;

#if  (defined TARGET_ARCH_HPUX)
    int                     fd_readset;
    size_t                  max_fdp1;
#else
    fd_set                  fd_readset;
    int                     max_fdp1;
#endif  /*  TARGET_ARCH_HPUX  */

    struct timeval          time_left;

    struct hbmdc_lm_fields *lm_entry_ptr;
    struct hbmdc_cl_fields *cl_entry_ptr;

    int                     i,
                            error,
                            retcd,
                            retcd_time;

/*
************************************************************************
**  get the parameter file, then get the parameters:                  **
**                                                                    **
**      dflt_rpt_interval --  default report (to HBMDCs) interval     **
**      port_num          --  Port (socket) number for recieving      **
**                            registration/unregistration messages    **
**      ckpt_fname_str    --  Name of checkpoint file                 **
************************************************************************
*/
    hbmdc_proc_name_str = *argv;

    bzero((char *) &hbmdc_dc_info, sizeof(struct hbmdc_dc_fields));

/*
**  Get the HBMDC host name and IP Number.
*/
    if  (error = get_DC_machine_info())
    {
        fprintf(stderr,
                "\n%s: Error getting machine information",
                hbmdc_proc_name_str);
        fprintf(stderr,
                "\n%s: Continuing without heartbeat processing\n",
                hbmdc_proc_name_str);
        return (-100 + error);
    }

/*
**  Initialize the Internet address fields.
*/
    hbmdc_dc_info.DC_addr.sin_family = AF_INET;
    hbmdc_dc_info.DC_addr.sin_addr.s_addr = htonl(INADDR_ANY);

/*
**  Get the hbmdc parameter file name,
**  then get the parameter values.
*/
    parm_fname_str = get_hbmdc_parmfile_name(argc, argv);

    if  (error = get_hbmdc_parms(parm_fname_str))
    {
        return (-200 + error);
    }

    DEBUGF1("Parameter file parsed\n");
    DEBUGF2("Local Monitor hostname: %s\n", hbmdc_dc_info.DC_hostName_str);
    DEBUGF2("Local Monitor IP addr:  %s\n", hbmdc_dc_info.DC_hostIPNum_str);

/*
**  Check for and read the checkpoint file for previous run
**  information, including information about client processes
**  requiring continuing (resumed) monitoring.
*/
    hbmdc_dc_info.DC_LocalMonitorsCt = 0;
    hbmdc_dc_info.DC_ClientsCt = 0;
    hbmdc_dc_info.DC_ClientsLiveCt = 0;
    hbmdc_lm_list.head = hbmdc_lm_list.tail = NULL;
    hbmdc_cl_list.head = hbmdc_cl_list.tail = NULL;
    
    read_hbmdc_ckpointfile();

/*
************************************************************************
**  Now we have all the parameters, so we set up the sockets.         **
************************************************************************
*/
    if  (retcd = set_up_udp_port(&sock_fd_hbmsg,
                                 &(hbmdc_dc_info.DC_addr)))
    {
        fprintf(stderr,
                "\n%s: Error setting up UDP (reg/unreg) port\n",
                hbmdc_proc_name_str);
        close(sock_fd_hbmsg);
        return retcd;
    }
#ifdef  TARGET_ARCH_HPUX
    if  (sock_fd_hbmsg >= (8 * sizeof(int)))
    {
        fprintf(stderr,
                "\n%s: HBM: Funky HP-UX error giving fd number too large to process.\n",
                hbmdc_proc_name_str);
        return -300;
    }
#endif  /*  TARGET_ARCH_HPUX defined  */


    DEBUGF2("Listening for heartbeat messages on port %d\n", hbmdc_dc_info.DC_addr.sin_port);

    max_fdp1 = sock_fd_hbmsg;
    max_fdp1++;

    signal(SIGINT, shutdown_quit);
    signal(SIGQUIT, shutdown_quit);

    retcd_time = gettimeofday(&time_curr, NULL);
    if  (hbmdc_dc_info.DC_ClientsLiveCt == 0)
    {
        time_to_monitor_write.tv_sec = 0;
        time_to_monitor_write.tv_usec = 0;
    }
    else
    {
/*
**      Go through all Client entries, setting rptTimeNext to now.
*/
        for (cl_entry_ptr = hbmdc_cl_list.head;
             cl_entry_ptr;
             cl_entry_ptr = cl_entry_ptr->next)
        {
            cl_entry_ptr->CL_rptTimeNext = time_curr.tv_sec;
        }
        time_to_monitor_write.tv_sec = time_curr.tv_sec
                                        + hbmdc_dc_info.DC_MonitorInterval;
        time_to_monitor_write.tv_usec = 0;
    }
    write_hbmdc_ckpointfile();

    /***********************/
    /*   Main Select Loop  */
    /***********************/

    DEBUGF1("Entering main select loop\n");

    while (!quit_now)
    {
#ifdef  TARGET_ARCH_HPUX
        bzero((char *) &fd_readset, sizeof(int));
        fd_readset |= (1 << sock_fd_hbmsg);
#else
        FD_ZERO(&fd_readset);
        FD_SET(sock_fd_hbmsg, &fd_readset);
#endif  /*  TARGET_ARCH_HPUX defined  */

        if  (hbmdc_dc_info.DC_ClientsLiveCt == 0)
        {
            if  (!quit_now)
            {
                retcd = select(max_fdp1, &fd_readset, NULL, NULL, NULL);
                DEBUGF1("select completed\n");
            }
            else
            {
                retcd = 0;
            }
        }
        else
        {
            retcd_time = gettimeofday(&time_curr, NULL);
            if  (time_to_monitor_write.tv_sec <= time_curr.tv_sec)
            {
                time_left.tv_sec = time_left.tv_usec = 0;
            }
            else
            {
                time_left.tv_sec = time_to_monitor_write.tv_sec
                                        - time_curr.tv_sec;
/*              (time_to_monitor_write.tv_usec is 0,        */
/*               so always <= time_curr.tv_usec)            */
                if  (time_curr.tv_usec)
                {
                    time_left.tv_sec--;
                    time_left.tv_usec = 1000000 - time_curr.tv_usec;
                }
            }

            if  (!quit_now)
            {
                retcd = select(max_fdp1, &fd_readset, NULL, NULL, &time_left);
                DEBUGF1("timed select completed\n");
            }
            else
            {
                retcd = 0;
            }
        }

        if  (   (retcd < 0)
             && (errno != EINTR))
        {
            if  (errno == EBADF)
            {
                fprintf(stderr,
                        "\n%s: Error in select(), invalid descriptor specified\n",
                        hbmdc_proc_name_str);
                perror(hbmdc_proc_name_str);
            }
            if  (errno == EFAULT)
            {
                fprintf(stderr,
                        "\n%s: Error in select(), invalid pointer parameter\n",
                        hbmdc_proc_name_str);
                perror(hbmdc_proc_name_str);
            }
            return 1;
        }

        if  (retcd > 0)
        {
#ifdef  TARGET_ARCH_HPUX
            if  (fd_readset & (1 << sock_fd_hbmsg))
#else
            if  (FD_ISSET(sock_fd_hbmsg, &fd_readset))
#endif  /*  TARGET_ARCH_HPUX defined  */
            {
                process_heartbeat_msg(sock_fd_hbmsg);
            }
            write_hbmdc_ckpointfile();
        }
        
        retcd_time = gettimeofday(&time_curr, NULL);
        if  (   (time_curr.tv_sec >= time_to_monitor_write.tv_sec)
             || (quit_now))
        {
            DEBUGF2("start performing timed updates, curr time is %d\n", time_curr.tv_sec);

            update_client_and_lm_data();
            write_hbmdc_ckpointfile();
            retcd_time = gettimeofday(&time_curr, NULL);
            while (time_to_monitor_write.tv_sec <= time_curr.tv_sec)
            {
                time_to_monitor_write.tv_sec +=
                                        hbmdc_dc_info.DC_MonitorInterval;
            }
        }
    }

    close(sock_fd_hbmsg);

    exit(0);
}
