/*
************************************************************************
**  GUSTO Heartbeat Monitor Local Monitor (HBMLM) program             **
************************************************************************
*/
#define HBM_LM_C
/*  #define DEBUGMSGS  */

#include "hbm.h"
#include "hbm_cllm.h"
#include "hbm_common.h"
#include "hbm_list_macros.h"
#include "hbm_lmdc.h"

#define CHKPT_FILENAME "hbmlm_checkpoint"

#ifndef SYSCONFDIR
#define SYSCONFDIR "."
#endif  /*  SYSCONFDIR  */

#ifdef DEBUG_MSGS
#define DEBUGF1(x)      fprintf(stderr,(x))
#define DEBUGF2(x,y)    fprintf(stderr,(x),(y))
#define DEBUGF3(x,y,z)  fprintf(stderr,(x),(y),(z))
#else
#define DEBUGF1(x)
#define DEBUGF2(x,y)
#define DEBUGF3(x,y,z)
#endif

    char                   *hbmlm_proc_name_str;

    struct sockaddr_in      addr_rcv;

    int                     quit_now = 0;

    int                     sock_fd_reg,
                            sock_fd_rpt;


struct hbmlm_lm_fields {
    struct in_addr          LM_hostIPNum;
    char                   *LM_hostIPNum_str;
    int                     LM_hostNameLen;
    char                   *LM_hostName_str;
    u_short                 LM_portNumReg;
    u_short                 LM_portNumRpt;
    struct sockaddr_in      LM_addr_reg;
    struct sockaddr_in      LM_addr_rpt;
    unsigned int            LM_defaultReportInterval;
    u_long                  LM_ClientsCt;
    char                   *LM_ckptFileName_str;
    char                   *LM_ckptFileNameWk_str;
    UTCtime                 LM_ckptTime;
};

    struct hbmlm_lm_fields  hbmlm_lm_info;


struct hbmlm_cl_fields {
    struct hbmlm_cl_fields *prev;
    struct hbmlm_cl_fields *next;
    unsigned int            CL_procPID;
    char                   *CL_procName_str;
    char                   *CL_procNameRpt_str;
    unsigned int            CL_procStatus;
    UTCtime                 CL_regTime;
    char                   *CL_DCHostIPNum_str;
    u_short                 CL_DCPortNum;
    struct sockaddr_in      CL_DCaddrRpt;
    UTCtime                 CL_blockedTime;
    unsigned int            CL_cpuSecs;
    UTCtime                 CL_shutdownTime;
    unsigned int            CL_shutdownMsgCt;
    unsigned int            CL_rptNumber;
    UTCtime                 CL_rptTimeLast;
    unsigned int            CL_rptInterval;
    UTCtime                 CL_rptTimeNext;
    int                     CL_updated;
};

struct hbmlm_cl_list_head {
    struct hbmlm_cl_fields *head;
    struct hbmlm_cl_fields *tail;
};

    struct hbmlm_cl_list_head
                            hbmlm_cl_list;

    struct timeval          time_curr,
                            time_to_monitor_write;

/*
************************************************************************
**  get_LM_machine_info                                               **
**      Parameters: (None)                                            **
**                                                                    **
**      Returns:    integer return code -- 0 if ok, error otherwise   **
**                                                                    **
**      This routine sets the HBMLM host information fields that are  **
**      dependent on the address/configuration of the host.           **
**      Those fields are:                                             **
**              hostIPNum               IP Number (network format)    **
**              hostIPNum_str           IP Number (string in '.'      **
**                                      format)                       **
**              hostNameLen             length of the host name (int) **
**              hostName_str            host name (string)            **
**                                                                    **
************************************************************************
*/
int     get_LM_machine_info()
{
    int     retcd;
    char   *str_ptr1,
           *str_ptr2;

/*
**  Get machine-specific information:
*/
/*
**  Get host name (and name length) and IP Number.
*/
    hbmlm_lm_info.LM_hostName_str =
                                (char *) malloc(MAXHOSTNAMELEN + 256);
    bzero(hbmlm_lm_info.LM_hostName_str, MAXHOSTNAMELEN + 256);
    if  (gethostname(hbmlm_lm_info.LM_hostName_str,
                     MAXHOSTNAMELEN + 256) < 0)
    {
        fprintf(stderr,
                "\n%s: Error getting host name\n",
                hbmlm_proc_name_str);
        perror(hbmlm_proc_name_str);
        return -1;
    }
    retcd = hbm_get_IPNum_fr_name(&(hbmlm_lm_info.LM_hostIPNum),
                                  hbmlm_lm_info.LM_hostName_str,
                                  &(hbmlm_lm_info.LM_hostName_str));

    if  (retcd)
    {
        fprintf(stderr,
                "\n%s: Error getting host information\n",
                hbmlm_proc_name_str);
        return -2;
    }

/*  LM_hostIPNum field, hbmlm_lm_info                        */
    if  (hbmlm_lm_info.LM_hostIPNum.s_addr == 0)
    {
        fprintf(stderr,
                "\n%s: Bad host address\n",
                hbmlm_proc_name_str);
        perror(hbmlm_proc_name_str);
        return -3;
    }

/*  LM_hostIPNum_str field, hbmlm_lm_info                    */
    str_ptr1 = inet_ntoa(hbmlm_lm_info.LM_hostIPNum);
    hbmlm_lm_info.LM_hostIPNum_str =
                                (char *) malloc(strlen(str_ptr1) + 1);
    strcpy(hbmlm_lm_info.LM_hostIPNum_str, str_ptr1);
    str_ptr1 = NULL;

/*  LM_hostNameLen field, hbmlm_lm_info                      */
    hbmlm_lm_info.LM_hostNameLen =
                            strlen(hbmlm_lm_info.LM_hostName_str);

    return 0;
}


/*
************************************************************************
**  get_hbmlm_parmfile_name                                           **
**      Parameters: program parameters (argc, argv)                   **
**      Returns:    pointer to string with parameter filename         **
**                  (NULL if error)                                   **
**                                                                    **
**      This routine gets the name of the file with the parameter     **
**      information in it (parameter is delimited by `-p'.            **
************************************************************************
*/
char   *get_hbmlm_parmfile_name(    int     pargc,
                                    char   *pargv[])
{
    int         i = 1;

    for (i = 1; i + 2 <= pargc; i++)
    {
        if  (   (pargv[i][0] == '-')
             && (pargv[i][1] == 'p'))
        {
            return pargv[i+1];
        }
    }
    return NULL;
}


/*
************************************************************************
**  get_hbmlm_parm_val                                                **
**      Parameters: parm_str              line from the parm file     **
**                                                                    **
**      Returns:    integer return code                               **
**                      0   OK                                        **
**                      1   Error                                     **
**                                                                    **
**      This routine:                                                 **
**          scans the value pointed to by parms_str                   **
**          for a parameter name and                                  **
**              if it finds one                                       **
**                  gets the value and                                **
**                  sets the appropriate variable                     **
**                                                                    **
**          valid parameter names and parameter formats are:          **
**              one parameter per line                                **
**              <PARMNAME>: <parmvalue>                               **
**              Colon (:) is optional                                 **
**              white space before <PARMNAME> is ignored              **
**              Lines beginning with pound sign (#) are ignored       **
************************************************************************
*/
int     get_hbmlm_parm_val( char   *parm_str)
{
    int     error;
    int     got_val;
    int     i,
            j;
    u_short portNum;

    error = 0;
    while (isspace(*parm_str))
    {
        parm_str++;
    }
    if  (!strncmp(parm_str, "LocalMonitorPortNumReg", 22))
    {
/*      Port for HBM Clients to use for registering/unregistering  */
        parm_str += 22;
        if  (*parm_str == ':')
        {
            parm_str++;
        }
        got_val = sscanf(parm_str, " %hu", &portNum);
        if  (   (!got_val)
             || (portNum < HBM_PORT_MIN_AVAILABLE)
             || (portNum > HBM_PORT_MAX_AVAILABLE))
        {
            fprintf(stderr,
                    "\n%s: Invalid Parameter: Register/Unregister Port",
                    hbmlm_proc_name_str);
            fprintf(stderr,
                    "\n%s:                    must be > %d and < %d\n",
                    hbmlm_proc_name_str,
                    (int) HBM_PORT_MIN_AVAILABLE,
                    (int) HBM_PORT_MAX_AVAILABLE);
            return -1;
        }
/*      portNumReg field, hbmlm_lm_info           */
        hbmlm_lm_info.LM_portNumReg = portNum;
/*      addr_reg.sin_port field, hbmlm_lm_info    */
        hbmlm_lm_info.LM_addr_reg.sin_port = htons(portNum);
    }

    else if  (!strncmp(parm_str, "LocalMonitorPortNumRpt", 22))
    {
/*      Port for HBMLM to use for reporting to DCs        */
        parm_str += 22;
        if  (*parm_str == ':')
        {
            parm_str++;
        }
        got_val = sscanf(parm_str, " %hu", &portNum);
        if  (   (!got_val)
             || (portNum < HBM_PORT_MIN_AVAILABLE)
             || (portNum > HBM_PORT_MAX_AVAILABLE))
        {
            fprintf(stderr,
                    "\n%s: Invalid Parameter: Report Port",
                    hbmlm_proc_name_str);
            fprintf(stderr,
                    "\n%s:                    must be > %d and < %d\n",
                    hbmlm_proc_name_str,
                    (int) HBM_PORT_MIN_AVAILABLE,
                    (int) HBM_PORT_MAX_AVAILABLE);
            return -1;
        }
/*      portNumRpt field, hbmlm_lm_info           */
        hbmlm_lm_info.LM_portNumRpt = portNum;
/*      addr_reg.sin_port field, hbmlm_lm_info    */
        hbmlm_lm_info.LM_addr_rpt.sin_port = htons(portNum);
    }

    else if  (!strncmp(parm_str, "DefaultReportInterval", 21))
    {
/*      Default Report Interval        */
        parm_str += 21;
        if  (*parm_str == ':')
        {
            parm_str++;
        }
/*      defaultReportInterval field, hbmlm_lm_info        */
        got_val = sscanf(parm_str, " %u",
                         &(hbmlm_lm_info.LM_defaultReportInterval));
        if  (   (!got_val)
             || (hbmlm_lm_info.LM_defaultReportInterval <
                                          HBMLM_LIM_REPORT_INTERVAL_MIN)
             || (hbmlm_lm_info.LM_defaultReportInterval >
                                          HBMLM_LIM_REPORT_INTERVAL_MAX))
        {
            fprintf(stderr,
                    "\n%s: Invalid Parameter: MonitorInterval",
                    hbmlm_proc_name_str);
            fprintf(stderr,
                    "\n%s:                    must be > %lu and < %lu\n",
                    hbmlm_proc_name_str,
                    HBMLM_LIM_REPORT_INTERVAL_MIN,
                    HBMLM_LIM_REPORT_INTERVAL_MAX);
            return -2;
        }
    }
    else if  (!strncmp(parm_str, "CheckpointDirName", 17))
    {
/*
**      Name of checkpoint file.
**      Initially will check this file for old checkpoint information,
**      later will save checkpoint information after each change.
*/
        parm_str += 17;
        if  (*parm_str == ':')
        {
            parm_str++;
        }
        while (isspace(*parm_str))
        {
            parm_str++;
        }
        if  (*parm_str == '\0')
        {
            fprintf(stderr,
                    "\n%s: Missing/Null Parameter: CheckpointFileName\n",
                    hbmlm_proc_name_str);
            return -3;
        }
        i = 0;
        while (   (parm_str[i] != '\0')
               && (!isspace(parm_str[i])))
        {
            i++;
        }
        free(hbmlm_lm_info.LM_ckptFileName_str);
        hbmlm_lm_info.LM_ckptFileName_str = 
                        (char *) malloc(i + strlen(CHKPT_FILENAME)
                                        + hbmlm_lm_info.LM_hostNameLen + 3);
        strncpy(hbmlm_lm_info.LM_ckptFileName_str, parm_str, i);
        (hbmlm_lm_info.LM_ckptFileName_str)[i] = '\0';
        if  (hbmlm_lm_info.LM_ckptFileName_str[i-1] != '/')
            strcat(hbmlm_lm_info.LM_ckptFileName_str,"/");
        strcat(hbmlm_lm_info.LM_ckptFileName_str,CHKPT_FILENAME);
        strcat(hbmlm_lm_info.LM_ckptFileName_str,"_");
        strcat(hbmlm_lm_info.LM_ckptFileName_str, hbmlm_lm_info.LM_hostName_str);

        free(hbmlm_lm_info.LM_ckptFileNameWk_str);
        hbmlm_lm_info.LM_ckptFileNameWk_str = 
                (char *) malloc(strlen(hbmlm_lm_info.LM_ckptFileName_str) + 4);
        strcpy(hbmlm_lm_info.LM_ckptFileNameWk_str,
               hbmlm_lm_info.LM_ckptFileName_str);
        strcat(hbmlm_lm_info.LM_ckptFileNameWk_str, "_wk");
    }

    return 0;
}


/*
************************************************************************
**  get_hbmlm_parms                                                   **
**      Parameters: parm_str              line from the parm file     **
**                                                                    **
**                                                                    **
**      Returns:    integer return code                               **
**                      0   OK                                        **
**                      1   Error                                     **
**                                                                    **
**      This routine:                                                 **
**          sets all default parameter values                         **
**          opens the parameter file and:                             **
**              reads it one line at a time,                          **
**              passing the read line to a routine that extracts      **
**                parameter values and sets the parameters.           **
**                                                                    **
**          parameter file format is:                                 **
**              one parameter per line                                **
**              <PARMNAME>: <parmvalue>                               **
**              Colon (:) is optional                                 **
**              white space before <PARMNAME> is ignored              **
**              Lines beginning with pound sign (#) are ignored       **
************************************************************************
*/
int     get_hbmlm_parms(    char   *parm_fname_str)
{
    FILE   *parm_fp;

    char   *parm_fname_dflt1_str = "./hbmlm.conf",
           *parm_fname_dflt2_str = SYSCONFDIR "/hbmlm.conf";

    int     parm_buff_size = HBM_BUFF_SIZE_PARM;

    char   *parms_read_str,
           *parms_read_str_wk,
           *parms_read_str_end;
    char   *parms_str;
    int     error;

    parms_read_str = (char *) malloc(parm_buff_size);
    parms_read_str_end = parms_read_str + parm_buff_size - 1;
    *parms_read_str_end = '\0';
    parms_str = (char *) malloc(parm_buff_size);
    parms_str[parm_buff_size - 1] = '\0';

    hbmlm_lm_info.LM_portNumReg = HBMLM_DEFAULT_PORT_REG;
    hbmlm_lm_info.LM_addr_reg.sin_port = htons(HBMLM_DEFAULT_PORT_REG);
    hbmlm_lm_info.LM_portNumRpt = HBMLM_DEFAULT_PORT_RPT;
    hbmlm_lm_info.LM_addr_rpt.sin_port = htons(HBMLM_DEFAULT_PORT_RPT);
    hbmlm_lm_info.LM_defaultReportInterval = HBMLM_DEFAULT_RPT_INTERVAL;
    hbmlm_lm_info.LM_ckptFileName_str = NULL;
    hbmlm_lm_info.LM_ckptFileName_str =
                    (char *) malloc(strlen(HBMLM_DEFAULT_CHKPTFILE) + 1);
    strcpy(hbmlm_lm_info.LM_ckptFileName_str, HBMLM_DEFAULT_CHKPTFILE);
    hbmlm_lm_info.LM_ckptFileNameWk_str = NULL;
    hbmlm_lm_info.LM_ckptFileNameWk_str = 
                (char *) malloc(strlen(hbmlm_lm_info.LM_ckptFileName_str) + 4);
    strcpy(hbmlm_lm_info.LM_ckptFileNameWk_str,
           hbmlm_lm_info.LM_ckptFileName_str);
    strcat(hbmlm_lm_info.LM_ckptFileNameWk_str, "_wk");

    if  (parm_fname_str != NULL)
    {
        if  ((parm_fp = fopen(parm_fname_str, "r")) == NULL)
        {
            fprintf(stderr,
                    "\n%s: Error opening parameter file: %s\n",
                    hbmlm_proc_name_str,
                    parm_fname_str);
            return -10;
        }
    }
    else
    {
        if  ((parm_fp = fopen(parm_fname_dflt1_str, "r")) == NULL)
        {
            if  ((parm_fp = fopen(parm_fname_dflt2_str, "r")) == NULL)
            {
                fprintf(stderr,
                        "\n%s: Error opening parameter files: %s",
                        hbmlm_proc_name_str,
                        parm_fname_dflt1_str);
                fprintf(stderr,
                        "\n%s:                              : %s\n",
                        hbmlm_proc_name_str,
                        parm_fname_dflt2_str);
                return -20;
            }
        }
    }

    while (fgets(parms_read_str, parm_buff_size, parm_fp) != NULL)
    {
        parms_read_str_wk = parms_read_str;
        while (isspace(*parms_read_str_wk))
        {
            parms_read_str_wk++;
        }
        if  (   (parms_read_str_wk <= parms_read_str_end)
             && (*parms_read_str_wk != '#')
             && (*parms_read_str_wk != '\n')
             && (*parms_read_str_wk != '\0'))
        {
            strcpy(parms_str, parms_read_str_wk);
            error = get_hbmlm_parm_val(parms_str);
            if  (error)
            {
                return (-30 + error);
            }
        }
    }

    fclose(parm_fp);

    free(parms_read_str);
    free(parms_str);

    return 0;
}


struct hbmlm_cl_fields *find_cl_entry(  unsigned long   cl_proc_PID)
{
    int                     i;
    struct hbmlm_cl_fields *cl_entry_curr_ptr;

    for (cl_entry_curr_ptr = hbmlm_cl_list.head;
         cl_entry_curr_ptr;
         cl_entry_curr_ptr = cl_entry_curr_ptr->next)
    {
        if  (cl_entry_curr_ptr->CL_procPID == cl_proc_PID)
        {
            return cl_entry_curr_ptr;
        }
    }

    return NULL;
}


struct hbmlm_cl_fields *add_cl_entry(   struct hbmlm_cl_fields
                                               *cl_entry_new_ptr)
{
    struct hbmlm_cl_fields *cl_entry_ptr;

    int                     i,
                            str1_comp_str2;

    if  ((cl_entry_ptr = find_cl_entry(cl_entry_new_ptr->CL_procPID))
                                != NULL)
    {
        fprintf(stderr,
                "\n%s: Registration message received from registered process",
                hbmlm_proc_name_str);
        fprintf(stderr,
                "\n%s:   Original entry updated\n",
                hbmlm_proc_name_str);

        if  (cl_entry_ptr->CL_procNameRpt_str !=
                                    cl_entry_ptr->CL_procName_str)
        {
            free(cl_entry_ptr->CL_procNameRpt_str);
        }
        free(cl_entry_ptr->CL_procName_str);

        cl_entry_ptr->CL_procName_str = cl_entry_new_ptr->CL_procName_str;
        cl_entry_ptr->CL_procNameRpt_str =
                                    cl_entry_new_ptr->CL_procNameRpt_str;
        cl_entry_new_ptr->CL_procName_str = NULL;
        cl_entry_new_ptr->CL_procNameRpt_str = NULL;

        cl_entry_ptr->CL_procStatus = HBM_PROCSTATUS_ACTIVE;

        free(cl_entry_ptr->CL_DCHostIPNum_str);
        cl_entry_ptr->CL_DCHostIPNum_str =
                                cl_entry_new_ptr->CL_DCHostIPNum_str;
        cl_entry_new_ptr->CL_DCHostIPNum_str = NULL;

        cl_entry_ptr->CL_DCPortNum = cl_entry_new_ptr->CL_DCPortNum;
        cl_entry_ptr->CL_DCaddrRpt = cl_entry_new_ptr->CL_DCaddrRpt;
        cl_entry_ptr->CL_shutdownTime = 0;
        cl_entry_ptr->CL_shutdownMsgCt = 0;

        free(cl_entry_new_ptr);

        return cl_entry_ptr;
    }


    for (cl_entry_ptr = hbmlm_cl_list.head;
         cl_entry_ptr;
         cl_entry_ptr = cl_entry_ptr->next)
    {
        if  (cl_entry_ptr->CL_procPID > cl_entry_new_ptr->CL_procPID)
        {
            break;
        }

    }
    ListInsertBefore(&hbmlm_cl_list, cl_entry_new_ptr, cl_entry_ptr);

    hbmlm_lm_info.LM_ClientsCt++;

    return cl_entry_new_ptr;    
}


/*
************************************************************************
**  read_hbmlm_ckpointfile                                            **
**      Parameters: <None>                                            **
**      Returns:    <None>                                            **
**                                                                    **
**      This routine validates the LM host IP Number and loads the    **
**      checkpointed data collector and client data.                  **
**      Entries are separated by newline characters ('\n').           **
**      Fields within entries are delimited by semi-colons (';').     **
**      If an entry is found to be invalid then that entry is         **
**      skipped.                                                      **
************************************************************************
*/
void    read_hbmlm_ckpointfile()
{
    char                   *file_buff_str,
                           *file_buff_str_end,
                           *file_buff_str_wk1,
                           *file_buff_str_wk2;
    
    FILE                   *ckpt_fd;
    int                     ckpt_fd_eof = 0;

    char                   *date_decode_str = "%Y/%m/%d %T GMT";

    unsigned int            read_u_int;
    
    int                     got_val,
                            retcd,
                            valid_entry;

    int                     i,
                            j,
                            k,
                            int_wk,
                            wk_cpu_days,
                            wk_cpu_hours,
                            wk_cpu_mins,
                            wk_cpu_secs;

    int                     cl_entry_ct;

    struct hbmlm_cl_fields *hbmlm_cl_entry_new_ptr;

    struct in_addr          in_addr_wk;

    time_t                  curr_time,
                            curr_time_conv;

    long                    time_conv_diff;

    struct  tm              time_wk;

    struct  tm             *time_wk_ptr;

    if  ((ckpt_fd = fopen(hbmlm_lm_info.LM_ckptFileName_str, "r")) == NULL)
    {
/*
**      There is no table file -- initialize table to be empty.
*/
        fclose(ckpt_fd);
        return;
    }

/*
**  Set up the buffer areas and the time conversion values.
*/
    file_buff_str = (char *) malloc(HBM_BUFF_SIZE_CKPT);
    file_buff_str_end = file_buff_str + HBM_BUFF_SIZE_CKPT - 1;
    bzero(file_buff_str, HBM_BUFF_SIZE_CKPT);

    time_wk_ptr = &time_wk;
    bzero((char *) time_wk_ptr, sizeof(time_wk));

    retcd = gettimeofday(&time_curr, NULL);

    curr_time = (time_t) time_curr.tv_sec;
#ifdef HAVE_GMTIME_R
    gmtime_r(&curr_time,
             time_wk_ptr);
#else
    time_wk_ptr = gmtime(&curr_time);
#endif

#ifdef HAVE_MKTIME
    curr_time_conv = mktime(time_wk_ptr);
#else
    curr_time_conv = timegm(time_wk_ptr);
#endif
#ifndef HAVE_GMTIME_R
    free(time_wk_ptr);
#endif

    time_conv_diff = (long) curr_time - curr_time_conv;

/*
**  The first line has a literal "HBMLM CHECKPOINT" for file
**  identification.
**  We don't need/use that, just verify that it is an HBMLM checkpoint
**  file and throw it away.
*/
    if  (fgets(file_buff_str, HBM_BUFF_SIZE_CKPT, ckpt_fd) == NULL)
    {
/*      The file is empty -- close it, clean up, and return.          */
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    if  (strncasecmp(file_buff_str, "HBMLM CHECKPOINT", 16))
    {
/*
**      The file is not an HBMLM Checkpoint file --
**      close it, clean up, and return.
*/
        fprintf(stderr,
                "\n%s: Invalid HBMLM Checkpoint file, "
                "file skipped.\n",
                hbmlm_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }

/*
**  Read the next line and extract the hbmlm host information.
*/
    if  (fgets(file_buff_str, HBM_BUFF_SIZE_CKPT, ckpt_fd) == NULL)
    {
        fprintf(stderr,
                "\n%s: HBMLM checkpoint data incomplete, "
                "file skipped.\n",
                hbmlm_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }

/*
**  Verify that the first line is a LM data record.
*/
    if  (strncasecmp(file_buff_str, "LM Data:", 8))
    {
/*
**      The second record does not have LM data --
**      close it, clean up, and return.
*/
        fprintf(stderr,
                "\n%s: Invalid HBMLM Checkpoint file, "
                "file skipped.\n",
                hbmlm_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    file_buff_str_wk1 = file_buff_str + 8;
    while (isspace(*file_buff_str_wk1))
    {
        file_buff_str_wk1++;
    }

/*
**  Check the LM_hostIPNum
*/
    if  (   (*file_buff_str_wk1 == ';')
         || (*file_buff_str_wk1 == '\0'))
    {
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: HBMLM checkpoint data incomplete, "
                "file skipped.\n",
                hbmlm_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    file_buff_str_wk2 = index(file_buff_str_wk1, ';');
    if  (!file_buff_str_wk2)
    {
        valid_entry = 0;
        fprintf(stderr,
            "\n%s: HBMLM checkpoint data incomplete, "
            "file skipped.\n",
            hbmlm_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    *file_buff_str_wk2 = '\0';
    if  (strcmp(file_buff_str_wk1, hbmlm_lm_info.LM_hostIPNum_str))
    {
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: LM_hostIPNum for different host, "
                "file skipped:\n",
                hbmlm_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    *file_buff_str_wk2 = ';';
    file_buff_str_wk1 = file_buff_str_wk2 + 1;

/*
**  Verify the LM_hostName
*/
    file_buff_str_wk2 = index(file_buff_str_wk1, ';');
    if  (   (!file_buff_str_wk2)
         || (file_buff_str_wk2 == file_buff_str_wk1))
    {
/*      LM_hostName missing or invalid -- skip the file   */
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: HBMLM checkpoint data incomplete, "
                "file skipped:\n",
                hbmlm_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    *file_buff_str_wk2 = '\0';
    if  (strcmp(file_buff_str_wk1, hbmlm_lm_info.LM_hostName_str))
    {
/*      LM_hostName missing or invalid -- skip the file   */
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: HBMLM checkpoint data incomplete, "
                "file skipped:\n",
                hbmlm_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    *file_buff_str_wk2 = ';';
    file_buff_str_wk1 = file_buff_str_wk2 + 1;

/*
**  Verify the LM_portNumReg (port number for registration messages)
*/
    got_val = sscanf(file_buff_str_wk1, "%d", &int_wk);
    if  (   (!got_val)
         || ((u_short) int_wk != hbmlm_lm_info.LM_portNumReg))
    {
        fprintf(stderr,
                "\n%s: Warning -- LM_portNumReg missing or "
                "different from current\n",
                hbmlm_proc_name_str);
    } 
    file_buff_str_wk1 = index(file_buff_str_wk1, ';');
    if  (!file_buff_str_wk1)
    {
/*      LM_portNumReg last field -- skip the file     */
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: HBMLM checkpoint data incomplete, "
                "file skipped:\n",
                hbmlm_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    file_buff_str_wk1++;

/*
**  Skip the LM_portNumRpt (port number for reporting to DCs)
*/
    file_buff_str_wk1 = index(file_buff_str_wk1, ';');
    if  (!file_buff_str_wk1)
    {
/*      portNumRpt last field -- skip the file    */
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: HBMLM checkpoint data incomplete, "
                "file skipped:\n",
                hbmlm_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    file_buff_str_wk1++;

/*
**  Skip the LM_defaultReportInterval
*/
    file_buff_str_wk1 = index(file_buff_str_wk1, ';');
    if  (!file_buff_str_wk1)
    {
/*      LM_defaultReportInterval last field -- skip the file   */
        valid_entry = 0;
        fprintf(stderr,
                "\n%s: HBMLM checkpoint data incomplete, "
                "file skipped:\n",
                hbmlm_proc_name_str);
        fclose(ckpt_fd);
        free(file_buff_str);
        return;
    }
    file_buff_str_wk1++;

/*
**  Get the LM_ClientsCt (number of client processes)
*/
    got_val = sscanf(file_buff_str_wk1, "%d", &cl_entry_ct);
    if  (   (!got_val)
         || (cl_entry_ct < 0))
    {
        fprintf(stderr,
                "\n%s: Warning -- client count < 0, using 0\n",
                hbmlm_proc_name_str);
        cl_entry_ct = 0;
    }
    file_buff_str_wk1 = index(file_buff_str_wk1, ';');
    file_buff_str_wk1++;

/*
**  Read client data one line at a time (up to HBM_BUFF_SIZE_CKPT
**  characters) and extract the data for the table entry.
*/
    ckpt_fd_eof = 0;
    while (fgets(file_buff_str, HBM_BUFF_SIZE_CKPT, ckpt_fd) != NULL)
    {
        file_buff_str_wk1 = file_buff_str;

/*
**      Verify that this is a Client data record.
*/
        valid_entry = 1;
        if  (strncasecmp(file_buff_str_wk1, "Client Data:", 12))
        {
/*
**          This record does not have Client data -- skip it.
*/
            valid_entry = 0;
            fprintf(stderr,
                    "\n%s: Invalid record -- client data expected, "
                    " record skipped.\n",
                    hbmlm_proc_name_str);
        }
        else
        {
            file_buff_str_wk1 = file_buff_str_wk1 + 12;
            while (isspace(*file_buff_str_wk1))
            {
                file_buff_str_wk1++;
            }
            if  (*file_buff_str_wk1 == '\0')
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client process checkpoint data incomplete, "
                        "entry ommitted.\n",
                        hbmlm_proc_name_str);
            }
        }

/*
**      Validate and load this client record data
*/
        if  (valid_entry)
        {
            hbmlm_cl_entry_new_ptr =
                        (struct hbmlm_cl_fields *)
                            malloc(sizeof(struct hbmlm_cl_fields));
            bzero((char *) hbmlm_cl_entry_new_ptr,
                  sizeof(struct hbmlm_cl_fields));
        }

/*
**      Get the Client PID.
*/
        if  (valid_entry)
        {
            got_val = sscanf(file_buff_str_wk1, "%d", &int_wk);
            if  (   (!got_val)
                 || (int_wk <= 0))
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: CL_procPID missing or invalid, "
                        "entry omitted.\n",
                        hbmlm_proc_name_str);
            }
            else
            {
                hbmlm_cl_entry_new_ptr->CL_procPID = (unsigned long) int_wk;
                file_buff_str_wk1 = index(file_buff_str_wk1, ';');
                if  (!file_buff_str_wk1)
                {
                    valid_entry = 0;
                    fprintf(stderr,
                            "\n%s: Client process checkpoint data incomplete, "
                            "entry ommitted.\n",
                            hbmlm_proc_name_str);
                }
                else
                {
                    file_buff_str_wk1++;
                }
            }
        }
 
/*
**      Get the Client process name.
*/
        if  (valid_entry)
        {
            file_buff_str_wk2 = index(file_buff_str_wk1, ';');
            if  (   (!file_buff_str_wk2)
                 || (file_buff_str_wk2 == file_buff_str_wk1))
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client process name missing/invalid or",
                        hbmlm_proc_name_str);
                fprintf(stderr,
                        "\n%s: process checkpoint data incomplete, "
                        "entry ommitted.\n",
                        hbmlm_proc_name_str);
            }
            else
            {
                hbmlm_cl_entry_new_ptr->CL_procName_str =
                                (char *) malloc(file_buff_str_wk2 
                                              - file_buff_str_wk1 + 1);
                *file_buff_str_wk2 = '\0';
                strcpy(hbmlm_cl_entry_new_ptr->CL_procName_str,
                       file_buff_str_wk1);
                *file_buff_str_wk2 = ';';
                file_buff_str_wk1 = file_buff_str_wk2 + 1;
            }
        }

/*
**      Get the Client process report name.
*/
        if  (valid_entry)
        {
            file_buff_str_wk2 = index(file_buff_str_wk1, ';');
            if  (   (!file_buff_str_wk2)
                 || (file_buff_str_wk2 == file_buff_str_wk1))
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client process report name missing/invalid or",
                        hbmlm_proc_name_str);
                fprintf(stderr,
                        "\n%s: process checkpoint data incomplete, "
                        "entry ommitted.\n",
                        hbmlm_proc_name_str);
            }
            else
            {
                *file_buff_str_wk2 = '\0';
                if  (!strcmp(hbmlm_cl_entry_new_ptr->CL_procName_str,
                             file_buff_str_wk1))
                {
                    hbmlm_cl_entry_new_ptr->CL_procNameRpt_str =
                                    hbmlm_cl_entry_new_ptr->CL_procName_str;
                }
                else
                {
                    hbmlm_cl_entry_new_ptr->CL_procNameRpt_str =
                                (char *) malloc(file_buff_str_wk2 
                                              - file_buff_str_wk1 + 1);
                    strcpy(hbmlm_cl_entry_new_ptr->CL_procNameRpt_str,
                           file_buff_str_wk1);
                }
                *file_buff_str_wk2 = ';';
                file_buff_str_wk1 = file_buff_str_wk2 + 1;
            }
        }

/*
**      Get the Client Process status.
*/
        if  (valid_entry)
        {
            got_val = sscanf(file_buff_str_wk1, " %u", &read_u_int);
            if  (   (!got_val)
                 || (read_u_int < HBM_PROCSTATUS_MINVAL)
                 || (read_u_int > HBM_PROCSTATUS_MAXVAL))
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: CL_procStatus missing or invalid, "
                        "entry omitted.\n",
                        hbmlm_proc_name_str);
            }
            else
            {
                hbmlm_cl_entry_new_ptr->CL_procStatus = read_u_int;
            }
            file_buff_str_wk1 = index(file_buff_str_wk1, ';');
            if  (!file_buff_str_wk1)
            {
/*              processStatus last or invalid -- skip this one    */
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client process checkpoint data incomplete, "
                        "entry ommitted.\n",
                        hbmlm_proc_name_str);
            }
            else
            {
                file_buff_str_wk1++;
            }
        }

/*
**      Get the Client registration time.
*/
        if  (valid_entry)
        {
            bzero((char *) &time_wk, sizeof(time_wk));

            file_buff_str_wk1 = (char *) strptime(file_buff_str_wk1,
                                                  date_decode_str,
                                                  &time_wk);
            if  (file_buff_str_wk1 == NULL)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: CL_regTime missing or invalid, "
                        "entry omitted.\n",
                        hbmlm_proc_name_str);
            }
            else
            {

#ifdef HAVE_MKTIME
                time_wk.tm_isdst = 0;
                hbmlm_cl_entry_new_ptr->CL_regTime =
                        (UTCtime) ((long) mktime(&time_wk) + time_conv_diff);
#else
                hbmlm_cl_entry_new_ptr->CL_regTime =
                        (UTCtime) ((long) timegm(&time_wk) + time_conv_diff);
                if  (hbmlm_cl_entry_new_ptr->CL_regTime > 0)
                {
                    hbmlm_cl_entry_new_ptr->CL_regTime +=
                        time_wk.tm_gmtoff;
                }
#endif
                if  (hbmlm_cl_entry_new_ptr->CL_regTime == 0)
                {
                    valid_entry = 0;
                    fprintf(stderr,
                            "\n%s: CL_regTime missing or invalid, "
                            "entry omitted.\n",
                            hbmlm_proc_name_str);
                }
                else
                {
                    file_buff_str_wk1 = index(file_buff_str_wk1, ';');
                    if  (!file_buff_str_wk1)
                    {
                        valid_entry = 0;
                        fprintf(stderr,
                                "\n%s: Client process checkpoint data incomplete, "
                                "entry ommitted.\n",
                                hbmlm_proc_name_str);
                    }
                    else
                    {
                        file_buff_str_wk1++;
                    }
                }
            }
        }

/*
**      Get the Data Collector Reporting address -- CL_DChostIPNum.
*/
        if  (valid_entry)
        {
            while  (isspace(*file_buff_str_wk1))
            {
                file_buff_str_wk1++;
            }
            file_buff_str_wk2 = index(file_buff_str_wk1, ';');
            if  (   (!file_buff_str_wk2)
                 || (file_buff_str_wk2 == file_buff_str_wk1))
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client DCHostIPNum missing/invalid or",
                        hbmlm_proc_name_str);
                fprintf(stderr,
                        "\n%s: process checkpoint data incomplete, "
                        "entry ommitted.\n",
                        hbmlm_proc_name_str);
            }
            else
            {
                hbmlm_cl_entry_new_ptr->CL_DCHostIPNum_str =
                                (char *) malloc(file_buff_str_wk2 
                                              - file_buff_str_wk1 + 1);
                *file_buff_str_wk2 = '\0';
                strcpy(hbmlm_cl_entry_new_ptr->CL_DCHostIPNum_str,
                       file_buff_str_wk1);
                *file_buff_str_wk2 = ';';
                file_buff_str_wk1 = file_buff_str_wk2 + 1;
            }
        }

/*
**      Get the Data Collector Reporting address -- CL_DCportNum.
*/
        if  (valid_entry)
        {
            got_val = sscanf(file_buff_str_wk1, " %hu",
                             &(hbmlm_cl_entry_new_ptr->CL_DCPortNum));
            if  (hbmlm_cl_entry_new_ptr->CL_DCPortNum == 0)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client DCPortNum missing/invalid, "
                        "entry ommitted.\n",
                        hbmlm_proc_name_str);
            }
            else
            {
                file_buff_str_wk1 = index(file_buff_str_wk1, ';');
                if  (!file_buff_str_wk1)
                {
                    valid_entry = 0;
                    fprintf(stderr,
                            "\n%s: Client process checkpoint data incomplete, "
                            "entry ommitted.\n",
                            hbmlm_proc_name_str);
                }
                else
                {
                    file_buff_str_wk1++;
                }
            }
        }

/*
**      Set up the Data Collector address field.
*/
        if  (valid_entry)
        {
            hbmlm_cl_entry_new_ptr->CL_DCaddrRpt.sin_family = AF_INET;
            hbmlm_cl_entry_new_ptr->CL_DCaddrRpt.sin_addr.s_addr =
                    inet_addr(hbmlm_cl_entry_new_ptr->CL_DCHostIPNum_str);
            hbmlm_cl_entry_new_ptr->CL_DCaddrRpt.sin_port =
                    htons(hbmlm_cl_entry_new_ptr->CL_DCPortNum);
            if  (   (hbmlm_cl_entry_new_ptr->CL_DCaddrRpt.sin_addr.s_addr <= 0)
                 || (hbmlm_cl_entry_new_ptr->CL_DCaddrRpt.sin_port == 0))
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client DCHostIPNum invalid, "
                        "entry ommitted.\n",
                        hbmlm_proc_name_str);
            }
            
        }
        
/*
**      Get the Client blocked (since) time.
*/
        if  (valid_entry)
        {
            bzero((char *) &time_wk, sizeof(time_wk));

            file_buff_str_wk1 = (char *) strptime(file_buff_str_wk1,
                                              date_decode_str,
                                              &time_wk);
            if  (file_buff_str_wk1 == NULL)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: CL_blockedTime missing or invalid, "
                        "entry omitted:\n",
                        hbmlm_proc_name_str);
            }
            else
            {

#ifdef HAVE_MKTIME
                time_wk.tm_isdst = 0;
                hbmlm_cl_entry_new_ptr->CL_blockedTime =
                        (UTCtime) ((long) mktime(&time_wk) + time_conv_diff);
#else
                hbmlm_cl_entry_new_ptr->CL_blockedTime =
                        (UTCtime) ((long) timegm(&time_wk) + time_conv_diff);
                if  (hbmlm_cl_entry_new_ptr->CL_blockedTime > 0)
                {
                    hbmlm_cl_entry_new_ptr->CL_blockedTime +=
                        time_wk.tm_gmtoff;
                }
#endif
                if  (   (   (hbmlm_cl_entry_new_ptr->CL_procStatus ==
                                        HBM_PROCSTATUS_BLOCKED)
                         && (hbmlm_cl_entry_new_ptr->CL_blockedTime == 0))
                     || (   (hbmlm_cl_entry_new_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_BLOCKED)
                         && (hbmlm_cl_entry_new_ptr->CL_blockedTime != 0)))
                {
                    valid_entry = 0;
                    fprintf(stderr,
                            "\n%s: CL_blockedTime missing or invalid, "
                            "entry omitted:\n",
                            hbmlm_proc_name_str);
                }
                else
                {
                    file_buff_str_wk1 = index(file_buff_str_wk1, ';');
                    if  (!file_buff_str_wk1)
                    {
                        valid_entry = 0;
                        fprintf(stderr,
                                "\n%s: Client process checkpoint data incomplete, "
                                "entry ommitted:\n",
                                hbmlm_proc_name_str);
                    }
                    else
                    {
                        file_buff_str_wk1++;
                    }
                }
            }
        }

/*
**      Get the Client process CPU time.
*/
        if  (valid_entry)
        {
            file_buff_str_wk2 = index(file_buff_str_wk1, ':');
            if  (file_buff_str_wk2)
            {
                if  (*(file_buff_str_wk2 - 3) == '-')
                {
                    got_val = sscanf(file_buff_str_wk1,
                                     "%d-%d:%d:%d",
                                     &wk_cpu_days,
                                     &wk_cpu_hours,
                                     &wk_cpu_mins,
                                     &wk_cpu_secs);
                }
                else
                {
                    wk_cpu_days = 0;
                    if  (*(file_buff_str_wk2 + 3) == ':')
                    {
                        got_val = sscanf(file_buff_str_wk1,
                                         "%d:%d:%d\n",
                                         &wk_cpu_hours,
                                         &wk_cpu_mins,
                                         &wk_cpu_secs);
                        got_val++;
                    }
                    else
                    {
                        wk_cpu_hours = 0;
                        got_val = sscanf(file_buff_str_wk1,
                                         "%d:%d\n",
                                         &wk_cpu_mins,
                                         &wk_cpu_secs);
                        got_val += 2;
                    }
                }
            }
            if  (got_val != 4)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: CL_cpuSecs missing or invalid, "
                        "entry omitted:\n",
                        hbmlm_proc_name_str);
            }
            else
            {
                wk_cpu_secs += (wk_cpu_days  * HBM_TIME_SECS_PER_DAY);
                wk_cpu_secs += (wk_cpu_hours * HBM_TIME_SECS_PER_HOUR);
                wk_cpu_secs += (wk_cpu_mins  * HBM_TIME_SECS_PER_MIN);
                hbmlm_cl_entry_new_ptr->CL_cpuSecs = wk_cpu_secs;

                file_buff_str_wk1 = index(file_buff_str_wk1, ';');
                if  (!file_buff_str_wk1)
                {
                    valid_entry = 0;
                    fprintf(stderr,
                            "\n%s: Client process checkpoint data incomplete, "
                            "entry ommitted:\n",
                            hbmlm_proc_name_str);
                }
                else
                {
                    file_buff_str_wk1++;
                }                   
            }
        }

/*
**      Get the Client shutdown time.
*/
        if  (valid_entry)
        {
            bzero((char *) &time_wk, sizeof(time_wk));

            file_buff_str_wk1 = (char *) strptime(file_buff_str_wk1,
                                                  date_decode_str,
                                                  &time_wk);
            if  (file_buff_str_wk1 == NULL)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: shutdownTime missing or invalid, "
                        "entry omitted:\n",
                        hbmlm_proc_name_str);
            }
            else
            {
#ifdef HAVE_MKTIME
                time_wk.tm_isdst = 0;
                hbmlm_cl_entry_new_ptr->CL_shutdownTime =
                        (UTCtime) ((long) mktime(&time_wk) + time_conv_diff);
#else
                hbmlm_cl_entry_new_ptr->CL_shutdownTime =
                        (UTCtime) ((long) timegm(&time_wk) + time_conv_diff);
                if  (hbmlm_cl_entry_new_ptr->CL_shutdownTime > 0)
                {
                    hbmlm_cl_entry_new_ptr->CL_shutdownTime +=
                    time_wk.tm_gmtoff;
                }
#endif
                if  (   (hbmlm_cl_entry_new_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_NORMAL)
                     && (hbmlm_cl_entry_new_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_ABNORMAL)
                     && (hbmlm_cl_entry_new_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_DIED))
                {
                    hbmlm_cl_entry_new_ptr->CL_shutdownTime = 0;
                }
                else if  (hbmlm_cl_entry_new_ptr->CL_shutdownTime == 0)
                {
                    valid_entry = 0;
                    fprintf(stderr,
                            "\n%s: shutdownTime missing or invalid, "
                            "entry omitted:\n",
                            hbmlm_proc_name_str);
                }
                file_buff_str_wk1 = index(file_buff_str_wk1, ';');
                if  (!file_buff_str_wk1)
                {
                    valid_entry = 0;
                    fprintf(stderr,
                            "\n%s: Client process checkpoint data incomplete, "
                            "entry ommitted:\n",
                            hbmlm_proc_name_str);
                }
                else
                {
                    file_buff_str_wk1++;
                }
            }
        }

/*
**      Get the Client shutdown message count.
*/
        if  (valid_entry)
        {
            got_val = sscanf(file_buff_str_wk1, " %d", &int_wk);
            if  (!got_val)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: CL_shutdownMsgCt missing or invalid, "
                        "entry omitted:\n",
                        hbmlm_proc_name_str);
            }
            else
            {
                if  (int_wk <= 0)
                {
                    int_wk = 0;
                }
                hbmlm_cl_entry_new_ptr->CL_shutdownMsgCt = (unsigned int) int_wk;
                
                file_buff_str_wk1 = index(file_buff_str_wk1, ';');
                if  (!file_buff_str_wk1)
                {
                    valid_entry = 0;
                    fprintf(stderr,
                            "\n%s: Client process checkpoint data incomplete, "
                            "entry ommitted:\n",
                            hbmlm_proc_name_str);
                }
                else
                {
                    file_buff_str_wk1++;
                }
            }
        }

/*
**      Get the Client report (message) number.
*/
        if  (valid_entry)
        {
            got_val = sscanf(file_buff_str_wk1, " %d", &int_wk);
            if  (!got_val)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: CL_rptNumber missing or invalid, "
                        "entry omitted:\n",
                        hbmlm_proc_name_str);
            }
            else
            {
                if  (int_wk <= 0)
                {
                    int_wk = 0;
                }
                hbmlm_cl_entry_new_ptr->CL_rptNumber = (unsigned int) int_wk;
                
                file_buff_str_wk1 = index(file_buff_str_wk1, ';');
                if  (!file_buff_str_wk1)
                {
                    valid_entry = 0;
                    fprintf(stderr,
                            "\n%s: Client process checkpoint data incomplete, "
                            "entry ommitted:\n",
                            hbmlm_proc_name_str);
                }
                else
                {
                    file_buff_str_wk1++;
                }
            }
        }

/*
**      Get the Client last report time.
*/
        if  (valid_entry)
        {
            bzero((char *) &time_wk, sizeof(time_wk));

            file_buff_str_wk1 = (char *) strptime(file_buff_str_wk1,
                                                  date_decode_str,
                                                  &time_wk);
            if  (file_buff_str_wk1 == NULL)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: rptTimeLast missing or invalid, "
                        "entry omitted:\n",
                        hbmlm_proc_name_str);
            }
            else
            {

#ifdef HAVE_MKTIME
                time_wk.tm_isdst = 0;
                hbmlm_cl_entry_new_ptr->CL_rptTimeLast =
                        (UTCtime) ((long) mktime(&time_wk) + time_conv_diff);
#else
                hbmlm_cl_entry_new_ptr->CL_rptTimeLast =
                        (UTCtime) ((long) timegm(&time_wk) + time_conv_diff);
                if  (hbmlm_cl_entry_new_ptr->CL_rptTimeLast > 0)
                {
                    hbmlm_cl_entry_new_ptr->CL_rptTimeLast +=
                    time_wk.tm_gmtoff;
                }
#endif
                if  (hbmlm_cl_entry_new_ptr->CL_rptTimeLast < 0)
                {
                    hbmlm_cl_entry_new_ptr->CL_rptTimeLast = 0;
                }
                file_buff_str_wk1 = index(file_buff_str_wk1, ';');
                if  (!file_buff_str_wk1)
                {
                    valid_entry = 0;
                    fprintf(stderr,
                            "\n%s: Client process checkpoint data incomplete, "
                            "entry ommitted:\n",
                            hbmlm_proc_name_str);
                }
                else
                {

                    file_buff_str_wk1++;
                }
            }
        }

/*
**      Get the Client report interval (seconds).
*/
        if  (valid_entry)
        {
            got_val = sscanf(file_buff_str_wk1, " %d", &int_wk);
            if  (!got_val)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: Client Report Interval missing or invalid, "
                        "entry omitted:\n",
                        hbmlm_proc_name_str);
            }
            else
            {
                if  (int_wk <= 0)
                {
                    int_wk = HBMLM_DEFAULT_RPT_INTERVAL;
                }
                hbmlm_cl_entry_new_ptr->CL_rptInterval = int_wk;
                
                file_buff_str_wk1 = index(file_buff_str_wk1, ';');
                if  (!file_buff_str_wk1)
                {
                    valid_entry = 0;
                    fprintf(stderr,
                            "\n%s: Client process checkpoint data incomplete, "
                            "entry ommitted:\n",
                            hbmlm_proc_name_str);
                }
                else
                {
                    file_buff_str_wk1++;
                }
            }
        }

/*
**      Get the Client next report time.
*/
        if  (valid_entry)
        {
            bzero((char *) &time_wk, sizeof(time_wk));

            file_buff_str_wk1 = (char *) strptime(file_buff_str_wk1,
                                                  date_decode_str,
                                                  &time_wk);
            if  (file_buff_str_wk1 == NULL)
            {
                valid_entry = 0;
                fprintf(stderr,
                        "\n%s: rptTimeNext missing or invalid, "
                        "entry omitted:\n",
                        hbmlm_proc_name_str);
            }
            else
            {

#ifdef HAVE_MKTIME
                time_wk.tm_isdst = 0;
                hbmlm_cl_entry_new_ptr->CL_rptTimeNext =
                        (UTCtime) ((long) mktime(&time_wk) + time_conv_diff);
#else
                hbmlm_cl_entry_new_ptr->CL_rptTimeNext =
                        (UTCtime) ((long) timegm(&time_wk) + time_conv_diff);
                if  (hbmlm_cl_entry_new_ptr->CL_rptTimeNext > 0)
                {
                    hbmlm_cl_entry_new_ptr->CL_rptTimeNext +=
                    time_wk.tm_gmtoff;
                }
#endif
                if  (hbmlm_cl_entry_new_ptr->CL_rptTimeNext < 
                                        hbmlm_cl_entry_new_ptr->CL_rptTimeLast)
                {
                    hbmlm_cl_entry_new_ptr->CL_rptTimeNext =
                                        hbmlm_cl_entry_new_ptr->CL_rptTimeLast;
                }
                while  (   (*file_buff_str_wk1 != ';')
                        && (*file_buff_str_wk1 != '\0'))
                {
                    file_buff_str_wk1++;
                }
                if  (*file_buff_str_wk1 != '\0')
                {
                    fprintf(stderr,
                            "\n%s: Client process checkpoint data overrun\n",
                            hbmlm_proc_name_str);
                }
            }
        }

        if  (valid_entry)
        {
/*
**      Increment the Client counter and
**      add the client record to the table.
*/
            add_cl_entry(hbmlm_cl_entry_new_ptr);
        }
        else
        {
            if  (hbmlm_cl_entry_new_ptr->CL_procName_str ==
                                hbmlm_cl_entry_new_ptr->CL_procNameRpt_str)
            {
                if  (hbmlm_cl_entry_new_ptr->CL_procName_str != NULL)
                {
                    free(hbmlm_cl_entry_new_ptr->CL_procName_str);
                    hbmlm_cl_entry_new_ptr->CL_procName_str = NULL;
                    hbmlm_cl_entry_new_ptr->CL_procNameRpt_str = NULL;
                }
            }
            else
            {
                if  (hbmlm_cl_entry_new_ptr->CL_procName_str != NULL)
                {
                    free(hbmlm_cl_entry_new_ptr->CL_procName_str);
                    hbmlm_cl_entry_new_ptr->CL_procName_str = NULL;
                }
                if  (hbmlm_cl_entry_new_ptr->CL_procNameRpt_str != NULL)
                {
                    free(hbmlm_cl_entry_new_ptr->CL_procNameRpt_str);
                    hbmlm_cl_entry_new_ptr->CL_procNameRpt_str = NULL;
                }
            }
            if  (hbmlm_cl_entry_new_ptr->CL_DCHostIPNum_str != NULL)
            {
                free(hbmlm_cl_entry_new_ptr->CL_DCHostIPNum_str);
                hbmlm_cl_entry_new_ptr->CL_DCHostIPNum_str = NULL;
            }
            free(hbmlm_cl_entry_new_ptr);
        }
        hbmlm_cl_entry_new_ptr = NULL;
    }

/*
**  Have reached end of file -- close out and return.
*/
    if  (hbmlm_lm_info.LM_ClientsCt < cl_entry_ct)
    {
        fprintf(stderr,
                "\n%s: Warning -- client records missing or invalid",
                hbmlm_proc_name_str);
        fprintf(stderr,
                "\n%s: will use available data\n",
                hbmlm_proc_name_str);
    }
    free(file_buff_str);
    fclose(ckpt_fd);

    return;
}


/*
************************************************************************
**  set_up_udp_port                                                   **
**      Parameters: pointer to fd field to use for the socket         **
**                  pointer to the (already filled in) port address   **
**      Returns:    integer value: 0 - socket successfully set up     **
**                                <0 - error setting up the port      **
**                                                                    **
**      This routine sets up a udp port for receiving registration    **
**      and unregistration messages from clients, or for              **
**      sending report messages to data collectors.                   **
************************************************************************
*/
int     set_up_udp_port(    int                *sock_fd_udp_ptr,
                            struct sockaddr_in *udpaddr_in_ptr)
{
    if  ((*sock_fd_udp_ptr = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0)
    {
        fprintf(stderr,
                "\n%s: socket() call for heartbeat port failed\n",
                hbmlm_proc_name_str);
        perror(hbmlm_proc_name_str);
        return -1;
    }

    if  (bind(*sock_fd_udp_ptr,
              (struct sockaddr *) udpaddr_in_ptr,
              sizeof(*udpaddr_in_ptr)) < 0)
    {
        fprintf(stderr,
                "\n%s: bind() failed.\n",
                hbmlm_proc_name_str);
        perror(hbmlm_proc_name_str);
        return -2;
    }

    return 0;
}


/*
************************************************************************
**  write_hbmlm_ckpointfile                                           **
**      Parameters: (None)                                            **
**      Returns:    (Nothing)                                         **
**                                                                    **
**      This routine writes the checkpoint file based on the data     **
**      in the hbmlm_lm_info structure and the hbmlm_cl_info table.   **
**      The fields are separated by semi-colons (';'),                **
**      and entries are separated by new-lines ('\n').                **
************************************************************************
*/
void    write_hbmlm_ckpointfile()
{
    FILE                   *ckpt_fd;

    struct in_addr          ipaddr_wk;

    struct hbmlm_cl_fields *cl_entry_curr_ptr;

    struct tm               tm_struct_wk;
    struct tm              *tm_struct_wk_ptr;

    char                    date_time_str[HBM_DATE_TIME_LEN];
    char                   *date_time_fmt_out_str = "%Y/%m/%d %T GMT";

    int                     i;
    int                     retcd;

    unsigned int            secs_left;

    tm_struct_wk_ptr = &tm_struct_wk;

/*  First get the current time (and convert to UTC)     */

    retcd = gettimeofday(&time_curr, NULL);

/*
**  Now write the table file.                                         **
*/
    if  ((ckpt_fd = fopen(hbmlm_lm_info.LM_ckptFileNameWk_str, "w")) == NULL)
    {
        fprintf(stderr,
                "\n%s: Error opening hbmlm checkpoint file",
                hbmlm_proc_name_str);
        fprintf(stderr,
                "\n%s: Continuing without writing file\n",
                hbmlm_proc_name_str);
        return;
    }

/*
**  First output the header literal and the current date and time     **
**  for identification and "version control".                         **
*/
#ifdef HAVE_GMTIME_R
    gmtime_r((time_t *) &(time_curr.tv_sec), tm_struct_wk_ptr);
#else
    tm_struct_wk_ptr = gmtime((time_t *) &(time_curr.tv_sec));
#endif
    strftime(date_time_str,
             HBM_DATE_TIME_LEN,
             date_time_fmt_out_str,
             tm_struct_wk_ptr);
#ifndef HAVE_GMTIME_R
    free(tm_struct_wk_ptr);
#endif
    fprintf(ckpt_fd, "HBMLM CHECKPOINT:%s\n",
            date_time_str);
/*
**  Now write the local monitor data to the file.                     **
*/
    fprintf(ckpt_fd, "LM Data:%s;%s;%hu;%hu;%u;%lu\n",
            hbmlm_lm_info.LM_hostIPNum_str,
            hbmlm_lm_info.LM_hostName_str,
            hbmlm_lm_info.LM_portNumReg,
            hbmlm_lm_info.LM_portNumRpt,
            hbmlm_lm_info.LM_defaultReportInterval,
            hbmlm_lm_info.LM_ClientsCt);
/*
**  Now write the client data (for each client process).
*/
    for (cl_entry_curr_ptr = hbmlm_cl_list.head;
         cl_entry_curr_ptr;
         cl_entry_curr_ptr = cl_entry_curr_ptr->next)
    {
        fprintf(ckpt_fd, "Client Data:%u;%s;%s;%d;",
                    cl_entry_curr_ptr->CL_procPID,
                    cl_entry_curr_ptr->CL_procName_str,
                    cl_entry_curr_ptr->CL_procNameRpt_str,
                    cl_entry_curr_ptr->CL_procStatus);
                    
#ifdef HAVE_GMTIME_R
        gmtime_r(&(cl_entry_curr_ptr->CL_regTime),
                 tm_struct_wk_ptr);
#else
        tm_struct_wk_ptr = gmtime(&(cl_entry_curr_ptr->CL_regTime));
#endif
        strftime(date_time_str,
                 HBM_DATE_TIME_LEN,
                 date_time_fmt_out_str,
                 tm_struct_wk_ptr);
#ifndef HAVE_GMTIME_R
        free(tm_struct_wk_ptr);
#endif
        fprintf(ckpt_fd, "%s;%s;%hu;",
                date_time_str,
                cl_entry_curr_ptr->CL_DCHostIPNum_str,
                cl_entry_curr_ptr->CL_DCPortNum);

#ifdef HAVE_GMTIME_R
        gmtime_r(&(cl_entry_curr_ptr->CL_blockedTime),
                 tm_struct_wk_ptr);
#else
        tm_struct_wk_ptr = gmtime(&(cl_entry_curr_ptr->CL_blockedTime));
#endif
        strftime(date_time_str,
                 HBM_DATE_TIME_LEN,
                 date_time_fmt_out_str,
                 tm_struct_wk_ptr);
#ifndef HAVE_GMTIME_R
        free(tm_struct_wk_ptr);
#endif
        fprintf(ckpt_fd, "%s;",
                date_time_str);
        secs_left = cl_entry_curr_ptr->CL_cpuSecs;
        if  (secs_left >= HBM_TIME_SECS_PER_DAY)
        {
            fprintf(ckpt_fd, "%u-",
                    secs_left / HBM_TIME_SECS_PER_DAY);
            secs_left %= HBM_TIME_SECS_PER_DAY;
            fprintf(ckpt_fd, "%2.2u:",
                    secs_left / HBM_TIME_SECS_PER_HOUR);
            secs_left %= HBM_TIME_SECS_PER_HOUR;
        }
        else if  (secs_left >= HBM_TIME_SECS_PER_HOUR)
        {
            fprintf(ckpt_fd, "%2.2u:",
                    secs_left / HBM_TIME_SECS_PER_HOUR);
            secs_left %= HBM_TIME_SECS_PER_HOUR;
        }
        fprintf(ckpt_fd, "%2.2u:%2.2u;",
                (secs_left / HBM_TIME_SECS_PER_MIN),
                (secs_left % HBM_TIME_SECS_PER_MIN));

#ifdef HAVE_GMTIME_R
        gmtime_r(&(cl_entry_curr_ptr->CL_shutdownTime),
                 tm_struct_wk_ptr);
#else
        tm_struct_wk_ptr = gmtime(&(cl_entry_curr_ptr->CL_shutdownTime));
#endif
        strftime(date_time_str,
                 HBM_DATE_TIME_LEN,
                 date_time_fmt_out_str,
                 tm_struct_wk_ptr);
#ifndef HAVE_GMTIME_R
        free(tm_struct_wk_ptr);
#endif
        fprintf(ckpt_fd, "%s;%u;%u;",
                date_time_str,
                cl_entry_curr_ptr->CL_shutdownMsgCt,
                cl_entry_curr_ptr->CL_rptNumber);

#ifdef HAVE_GMTIME_R
        gmtime_r(&(cl_entry_curr_ptr->CL_rptTimeLast),
                 tm_struct_wk_ptr);
#else
        tm_struct_wk_ptr = gmtime(&(cl_entry_curr_ptr->CL_rptTimeLast));
#endif
        strftime(date_time_str,
                 HBM_DATE_TIME_LEN,
                 date_time_fmt_out_str,
                 tm_struct_wk_ptr);
#ifndef HAVE_GMTIME_R
        free(tm_struct_wk_ptr);
#endif
        fprintf(ckpt_fd, "%s;%u;",
                date_time_str,
                cl_entry_curr_ptr->CL_rptInterval);

#ifdef HAVE_GMTIME_R
        gmtime_r(&(cl_entry_curr_ptr->CL_rptTimeNext),
                 tm_struct_wk_ptr);
#else
        tm_struct_wk_ptr = gmtime(&(cl_entry_curr_ptr->CL_rptTimeNext));
#endif
        strftime(date_time_str,
                 HBM_DATE_TIME_LEN,
                 date_time_fmt_out_str,
                 tm_struct_wk_ptr);
#ifndef HAVE_GMTIME_R
        free(tm_struct_wk_ptr);
#endif
        fprintf(ckpt_fd, "%s\n",
                date_time_str);
    }
/*
**  Close the file and return.
*/
    fclose(ckpt_fd);
    unlink(hbmlm_lm_info.LM_ckptFileName_str);
    link(hbmlm_lm_info.LM_ckptFileNameWk_str,
         hbmlm_lm_info.LM_ckptFileName_str);
    unlink(hbmlm_lm_info.LM_ckptFileNameWk_str);

    return;
}


/*
************************************************************************
**  format_Rptmsg_LM_fields                                           **
**      Parameters: (None)                                            **
**      Returns:    (None)                                            **
**                                <0 - error setting up the port      **
**                                                                    **
**      This routine sets up the LM information fields in the buffer  **
**      used for Report messages to Data Collectors.                  **
************************************************************************
*/
void    format_Rptmsg_LM_fields()
{
    bzero(hbm_msg_Rptbuff, HBM_BUFF_SIZE_MSG);

    strcpy(hbm_msg_Rptbuff,
           "HBMLM");

    hbm_msg_RptmsgLength_ptr = hbm_msg_Rptbuff + 5;

    hbm_msg_RptLMhostName_ptr = hbm_msg_RptmsgLength_ptr + 4;
    strcpy(hbm_msg_RptLMhostName_ptr, hbmlm_lm_info.LM_hostName_str);

    hbm_msg_RptCLprocessName_ptr = hbm_msg_RptLMhostName_ptr
                                 + hbmlm_lm_info.LM_hostNameLen + 1;
}


u_short calc_cksum( register u_short   *msg_ptr,
                    register int        msg_len)
{
    register long       sum;
    u_short             oddbyte;
    register u_short    cksum;

    sum = 0;
    while (msg_len > 1)
    {
        sum += *msg_ptr++;
        msg_len -= 2;
    }
    if  (msg_len > 1)
    {
        oddbyte = 0;
        *((u_char *) &oddbyte) = *(u_char *)msg_ptr;
        sum += oddbyte;
    }
    sum = (sum >> 16) + (sum & 0xffff);
    sum += (sum >> 16);
    cksum = (u_short) (~sum);

    return cksum;
}


/*
************************************************************************
**  format_send_rptmsg                                                **
**      Parameters: cl_entry_ptr        pointer to the client info    **
**                                      table entry for the client    **
**                                      to be reported on.            **
**      Returns:    (Nothing)                                         **
**                                                                    **
**      This routine formats a report message for the client and      **
**      sends it to the appropriate data collector.                   **
************************************************************************
*/
void    format_send_rptmsg( struct hbmlm_cl_fields *cl_entry_ptr)
{
    int     retcd;
    char   *msg_buff_ptr;
    int     nleft;
    int     nsent;

    if  (   (cl_entry_ptr->CL_procStatus ==
                                HBM_PROCSTATUS_SHUTDOWN_NORMAL)
         || (cl_entry_ptr->CL_procStatus ==
                                HBM_PROCSTATUS_SHUTDOWN_ABNORMAL)
         || (cl_entry_ptr->CL_procStatus ==
                                HBM_PROCSTATUS_SHUTDOWN_DIED))
    {
        cl_entry_ptr->CL_shutdownMsgCt++;
        if  (cl_entry_ptr->CL_shutdownMsgCt > 5)
        {
            return;
        }
    }

    strcpy(hbm_msg_RptCLprocessName_ptr,
           cl_entry_ptr->CL_procNameRpt_str);

    hbm_msg_RptCLprocessPID_ptr = hbm_msg_RptCLprocessName_ptr;
    while (*hbm_msg_RptCLprocessPID_ptr != '\0')
    {
        hbm_msg_RptCLprocessPID_ptr++;
    }
    hbm_msg_RptCLprocessPID_ptr++;

    PackUInt32(hbm_msg_RptCLprocessPID_ptr,
               cl_entry_ptr->CL_procPID);

    hbm_msg_RptCLstatus_ptr = hbm_msg_RptCLprocessPID_ptr + 4;
    PackUInt32(hbm_msg_RptCLstatus_ptr,
               cl_entry_ptr->CL_procStatus);
    
    (cl_entry_ptr->CL_rptNumber)++;
    hbm_msg_RptCLrptNumber_ptr = hbm_msg_RptCLstatus_ptr + 4;
    PackUInt32(hbm_msg_RptCLrptNumber_ptr,
               cl_entry_ptr->CL_rptNumber);
    
    hbm_msg_RptCLrptTime_ptr = hbm_msg_RptCLrptNumber_ptr + 4;
    PackUInt32(hbm_msg_RptCLrptTime_ptr,
               time_curr.tv_sec);
    cl_entry_ptr->CL_rptTimeLast = time_curr.tv_sec;
    while (cl_entry_ptr->CL_rptTimeNext <= time_curr.tv_sec)
    {
        cl_entry_ptr->CL_rptTimeNext += cl_entry_ptr->CL_rptInterval;
    }

    hbm_msg_RptCLrptInterval_ptr = hbm_msg_RptCLrptTime_ptr + 4;
    PackUInt32(hbm_msg_RptCLrptInterval_ptr,
               cl_entry_ptr->CL_rptInterval);

    hbm_msg_RptCLregTime_ptr = hbm_msg_RptCLrptInterval_ptr + 4;
    PackUInt32(hbm_msg_RptCLregTime_ptr,
               cl_entry_ptr->CL_regTime);

    hbm_msg_RptCLblockedTime_ptr = hbm_msg_RptCLregTime_ptr + 4;
    PackUInt32(hbm_msg_RptCLblockedTime_ptr,
               cl_entry_ptr->CL_blockedTime);

    hbm_msg_RptCLcpuSecs_ptr = hbm_msg_RptCLblockedTime_ptr + 4;
    PackUInt32(hbm_msg_RptCLcpuSecs_ptr,
               cl_entry_ptr->CL_cpuSecs);

    hbm_msg_RptCLshutdownTime_ptr = hbm_msg_RptCLcpuSecs_ptr + 4;
    PackUInt32(hbm_msg_RptCLshutdownTime_ptr,
               cl_entry_ptr->CL_shutdownTime);

    hbm_msg_RptCLshutdownMsgCt_ptr = hbm_msg_RptCLshutdownTime_ptr + 4;
    PackUInt32(hbm_msg_RptCLshutdownMsgCt_ptr,
               cl_entry_ptr->CL_shutdownMsgCt);

    hbm_msg_RptmsgLength = hbm_msg_RptCLshutdownMsgCt_ptr
                         - hbm_msg_Rptbuff + 5;
    PackUInt32(hbm_msg_RptmsgLength_ptr, hbm_msg_RptmsgLength);

    msg_buff_ptr = hbm_msg_Rptbuff;
    nleft = hbm_msg_RptmsgLength;

    while (nleft)
    {
        nsent = sendto(sock_fd_rpt,
                       hbm_msg_Rptbuff,
                       hbm_msg_RptmsgLength,
                       0,
                       (struct sockaddr *)&(cl_entry_ptr->CL_DCaddrRpt),
                       sizeof(struct sockaddr));
        if  (nsent <= 0)
        {
            fprintf(stderr,
                    "\n%s: HBM: Error sending report\n",
                    hbmlm_proc_name_str);
            perror(hbmlm_proc_name_str);
            nleft = 0;
        }
        else
        {
            msg_buff_ptr += nsent;
            nleft -= nsent;
        }
    }

    return;
}


void    delete_cl_entries_shutdown_procs()
{
    struct hbmlm_cl_fields *cl_entry_ptr,
                           *cl_entry_next_ptr;

    for (cl_entry_ptr = hbmlm_cl_list.head;
         cl_entry_ptr;
         cl_entry_ptr = cl_entry_next_ptr)
    {
        cl_entry_next_ptr = cl_entry_ptr->next;

        if  (cl_entry_ptr->CL_shutdownMsgCt >= 5)
        {
            ListRemove(&hbmlm_cl_list, cl_entry_ptr);

            if  (cl_entry_ptr->CL_procName_str ==
                                    cl_entry_ptr->CL_procNameRpt_str)
            {
                free(cl_entry_ptr->CL_procName_str);
            }
            else
            {
                free(cl_entry_ptr->CL_procName_str);
                free(cl_entry_ptr->CL_procNameRpt_str);
            }
            cl_entry_ptr->CL_procName_str = NULL;
            cl_entry_ptr->CL_procNameRpt_str = NULL;
            free(cl_entry_ptr->CL_DCHostIPNum_str);
            cl_entry_ptr->CL_DCHostIPNum_str = NULL;
            free(cl_entry_ptr);

            hbmlm_lm_info.LM_ClientsCt--;
        }
    }
}


void    send_rpt_msgs()
{
    struct hbmlm_cl_fields *cl_entry_ptr;

    for (cl_entry_ptr = hbmlm_cl_list.head;
         cl_entry_ptr;
         cl_entry_ptr = cl_entry_ptr->next)
    {
        if  (   (cl_entry_ptr->CL_rptTimeNext <= time_curr.tv_sec)
             || (quit_now))
        {
            format_send_rptmsg(cl_entry_ptr);
        }
    }
    delete_cl_entries_shutdown_procs();
}

        
/*
************************************************************************
**  process_reg_unreg_msg                                             **
**      Parameters: sock_fd_reg             the socket fd to get the  **
**                                          message from              **
**      Returns:    (Nothing)                                         **
**                                                                    **
**      This routine reads in a message from the specified socket,    **
**      verifies that it is a register/unregister message, and        **
**      processes the message.                                        **
**                                                                    **
**      A report is sent to the appropriate DC if the message is      **
**      valid.                                                        **
************************************************************************
*/
void    process_reg_unreg_msg(  int     sock_fd_reg)
{
    struct hbmlm_cl_fields *cl_entry_curr_ptr,
                           *cl_entry_new_ptr;

    struct sockaddr_in      from_addr;

#if  (   (defined TARGET_ARCH_CRAYT3E) \
      || (defined TARGET_ARCH_HPUX) \
      || (defined TARGET_ARCH_IRIX) \
      || (defined TARGET_ARCH_SOLARIS))
    int                     from_addr_len;
#else
    size_t                  from_addr_len;
#endif

    int                     cl_proc_PID;

    char                   *msg_buff_ptr,
                           *msg_buff_ptr_wk1,
                           *msg_buff_ptr_wk2,
                           *msg_buff_ptr_msgend;

    int                     i,
                            str1_len,
                            str2_len,
                            str3_len;

    int                     msg_len,
                            nleft,
                            nread,
                            nsent;

    int                     retcd;

/*
**  Command port was selected,                                        **
**  so get the message and process the request.                       **
*/
    cl_entry_curr_ptr = cl_entry_new_ptr = NULL;

    bzero((char *)&from_addr, sizeof(from_addr));
    from_addr.sin_family = AF_INET;
    from_addr_len = sizeof(from_addr);
    msg_len = recvfrom(sock_fd_reg,
                       hbm_msg_array,
                       HBM_BUFF_SIZE_MSG,
                       0,
                       (struct sockaddr *) &from_addr,
                       &from_addr_len);

    if  (errno == EAGAIN)
    {
        return;
    }
    if  (strncmp(hbm_msg_array, "HBMCL", 5))
    {
        return;
    }

    hbm_msg_RegCmsgLength_ptr = hbm_msg_array + 5;
    UnpackUInt32(hbm_msg_RegCmsgLength_ptr, hbm_msg_RegCmsgLength);

    if  (msg_len != hbm_msg_RegCmsgLength)
    {
        fprintf(stderr,
                "\n%s: Registration/Unregistration message wrong length",
                hbmlm_proc_name_str);
        fprintf(stderr,
                "\n%s:   Rcvd length: %d  Expected length: %d\n",
                hbmlm_proc_name_str,
                msg_len,
                hbm_msg_RegCmsgLength);
        return;
    }

/*
**  Got the message, so let's process it.           **
*/
    retcd = gettimeofday(&time_curr, NULL);

    hbm_msg_RegCmsgType_ptr = hbm_msg_RegCmsgLength_ptr + 4;

    if  (*hbm_msg_RegCmsgType_ptr == HBM_MSGTYPE_REGISTER)
    {
        DEBUGF2("Processing register msg from port %d\n", sock_fd_reg );

        hbm_msg_RegCmsg = hbm_msg_array;

        hbm_msg_RegCprocessPID_ptr = hbm_msg_RegCmsgType_ptr + 1;

        hbm_msg_RegCprocessName_str = hbm_msg_RegCprocessPID_ptr + 4;

        hbm_msg_RegCprocessNameRpt_str = hbm_msg_RegCprocessName_str;
        while (*hbm_msg_RegCprocessNameRpt_str != '\0')
        {
            hbm_msg_RegCprocessNameRpt_str++;
        }
        hbm_msg_RegCprocessNameRpt_str++;
        str1_len = hbm_msg_RegCprocessNameRpt_str
                                - hbm_msg_RegCprocessName_str;

        hbm_msg_RegCDChostIPNum_str = hbm_msg_RegCprocessNameRpt_str;
        while (*hbm_msg_RegCDChostIPNum_str != '\0')
        {
            hbm_msg_RegCDChostIPNum_str++;
        }
        hbm_msg_RegCDChostIPNum_str++;

        str2_len = hbm_msg_RegCDChostIPNum_str
                                - hbm_msg_RegCprocessNameRpt_str;

        hbm_msg_RegCDCportNum_ptr = hbm_msg_RegCDChostIPNum_str;
        while (*hbm_msg_RegCDCportNum_ptr != '\0')
        {
            hbm_msg_RegCDCportNum_ptr++;
        }
        hbm_msg_RegCDCportNum_ptr++;

        str3_len = hbm_msg_RegCDCportNum_ptr
                                - hbm_msg_RegCDChostIPNum_str;

        if  (   (str1_len <= 1)
             || (str2_len <= 1)
             || (str3_len <= 1))
        {
            return;
        }

        cl_entry_new_ptr = (struct hbmlm_cl_fields *)
                           malloc(sizeof(struct hbmlm_cl_fields));
        bzero((char *) cl_entry_new_ptr, sizeof(*cl_entry_new_ptr));

        UnpackUInt32(hbm_msg_RegCprocessPID_ptr,
                     cl_entry_new_ptr->CL_procPID);
        cl_entry_new_ptr->CL_procName_str =
                        (char *) malloc(str1_len);
        strcpy(cl_entry_new_ptr->CL_procName_str,
               hbm_msg_RegCprocessName_str);
        if  (!strcmp(hbm_msg_RegCprocessName_str,
                     hbm_msg_RegCprocessNameRpt_str))
        {
            cl_entry_new_ptr->CL_procNameRpt_str =
                                    cl_entry_new_ptr->CL_procName_str;
        }
        else
        {
            cl_entry_new_ptr->CL_procNameRpt_str =
                                    (char *) malloc(str2_len);
            strcpy(cl_entry_new_ptr->CL_procNameRpt_str,
                   hbm_msg_RegCprocessNameRpt_str);
        }
        cl_entry_new_ptr->CL_procStatus = HBM_PROCSTATUS_ACTIVE;
        cl_entry_new_ptr->CL_regTime = time_curr.tv_sec;
        cl_entry_new_ptr->CL_DCHostIPNum_str =
                        (char *) malloc(str3_len);
        strcpy(cl_entry_new_ptr->CL_DCHostIPNum_str,
               hbm_msg_RegCDChostIPNum_str);
        UnpackUInt32(hbm_msg_RegCDCportNum_ptr,
                     cl_entry_new_ptr->CL_DCPortNum);
        cl_entry_new_ptr->CL_DCaddrRpt.sin_family = AF_INET;
        cl_entry_new_ptr->CL_DCaddrRpt.sin_addr.s_addr =
                        inet_addr(hbm_msg_RegCDChostIPNum_str);
        cl_entry_new_ptr->CL_DCaddrRpt.sin_port =
                        htons(cl_entry_new_ptr->CL_DCPortNum);
        if  (   (cl_entry_new_ptr->CL_DCaddrRpt.sin_addr.s_addr <= 0)
             || (cl_entry_new_ptr->CL_DCaddrRpt.sin_port == 0))
        {
            free(cl_entry_new_ptr->CL_procName_str);
            if  (cl_entry_new_ptr->CL_procNameRpt_str !=
                                    cl_entry_new_ptr->CL_procName_str)
            {
                free(cl_entry_new_ptr->CL_procNameRpt_str);
            }
            free(cl_entry_new_ptr->CL_DCHostIPNum_str);
            free(cl_entry_new_ptr);
            cl_entry_new_ptr = NULL;
            return;
        }
        cl_entry_new_ptr->CL_blockedTime = 0;
        cl_entry_new_ptr->CL_cpuSecs = 0;
        cl_entry_new_ptr->CL_shutdownTime = 0;
        cl_entry_new_ptr->CL_shutdownMsgCt = 0;
        cl_entry_new_ptr->CL_rptNumber = 0;
        cl_entry_new_ptr->CL_rptTimeLast = 0;
        cl_entry_new_ptr->CL_rptInterval =
                        hbmlm_lm_info.LM_defaultReportInterval;

        for (cl_entry_curr_ptr = hbmlm_cl_list.head;
             cl_entry_curr_ptr;
             cl_entry_curr_ptr = cl_entry_curr_ptr->next)
        {
            if  (   (!strcmp(cl_entry_new_ptr->CL_DCHostIPNum_str,
                             cl_entry_curr_ptr->CL_DCHostIPNum_str))
                 && (cl_entry_new_ptr->CL_DCPortNum ==
                             cl_entry_curr_ptr->CL_DCPortNum)
                 && (   (cl_entry_new_ptr->CL_rptTimeNext == 0)
                     || (cl_entry_new_ptr->CL_rptTimeNext >
                             cl_entry_curr_ptr->CL_rptTimeNext)))
            {
                cl_entry_new_ptr->CL_rptTimeNext =
                             cl_entry_curr_ptr->CL_rptTimeNext;
            }
        }
        if  (cl_entry_new_ptr->CL_rptTimeNext == 0)
        {
            for (cl_entry_curr_ptr = hbmlm_cl_list.head;
                 cl_entry_curr_ptr;
                 cl_entry_curr_ptr = cl_entry_curr_ptr->next)
            {
                if  (   (cl_entry_new_ptr->CL_rptTimeNext == 0)
                     || (cl_entry_new_ptr->CL_rptTimeNext >
                             cl_entry_curr_ptr->CL_rptTimeNext))
                {
                    cl_entry_new_ptr->CL_rptTimeNext =
                             cl_entry_curr_ptr->CL_rptTimeNext;
                }
            }
        }
        if  (cl_entry_new_ptr->CL_rptTimeNext == 0)
        {
            cl_entry_new_ptr->CL_rptTimeNext = time_curr.tv_sec;
        }

        if  ((cl_entry_curr_ptr = add_cl_entry(cl_entry_new_ptr)) == NULL)
        {
            fprintf(stderr,
                    "\n%s: Error adding process for register",
                    hbmlm_proc_name_str);
            fprintf(stderr,
                    "\n%s:   Process Name: %s  PID: %hu\n",
                    hbmlm_proc_name_str,
                    hbm_msg_RegCprocessName_str,
                    cl_entry_new_ptr->CL_procPID);
            free(cl_entry_new_ptr->CL_procName_str);
            if  (cl_entry_new_ptr->CL_procNameRpt_str !=
                                    cl_entry_new_ptr->CL_procName_str)
            {
                free(cl_entry_new_ptr->CL_procNameRpt_str);
            }
            free(cl_entry_new_ptr->CL_DCHostIPNum_str);
            free(cl_entry_new_ptr);
        }
        cl_entry_new_ptr = NULL;
    }
    else if  (   (*hbm_msg_RegCmsgType_ptr == HBM_MSGTYPE_SHUTDOWN_NORMAL)
              || (*hbm_msg_RegCmsgType_ptr == HBM_MSGTYPE_SHUTDOWN_ABNORMAL))
    {
        DEBUGF2("Processing shutdown msg from port %d\n", sock_fd_reg );
        hbm_msg_UnRegCmsg = hbm_msg_array;
        hbm_msg_UnRegCmsgLength_ptr = hbm_msg_UnRegCmsg + 5;
        hbm_msg_UnRegCmsgLength = hbm_msg_RegCmsgLength;
        hbm_msg_UnRegCmsgType_ptr = hbm_msg_UnRegCmsgLength_ptr + 4;
        hbm_msg_UnRegCprocessPID_ptr = hbm_msg_UnRegCmsgType_ptr +1;
        hbm_msg_UnRegCprocessName_str = hbm_msg_UnRegCprocessPID_ptr + 4;

        UnpackUInt32(hbm_msg_UnRegCprocessPID_ptr, cl_proc_PID);

        cl_entry_curr_ptr = find_cl_entry(cl_proc_PID);

        if  (cl_entry_curr_ptr == NULL)
        {
            fprintf(stderr,
                    "\n%s: Unregistration message received from unknown process",
                    hbmlm_proc_name_str);
            fprintf(stderr,
                    "\n%s:   Process Name: %s  PID: %hu\n",
                    hbmlm_proc_name_str,
                    hbm_msg_UnRegCprocessName_str,
                    cl_proc_PID);
        }
        else if  (strcmp(cl_entry_curr_ptr->CL_procName_str,
                         hbm_msg_UnRegCprocessName_str))
        {
            fprintf(stderr,
                    "\n%s: Unregistration message received wrong process name",
                    hbmlm_proc_name_str);
            fprintf(stderr,
                    "\n%s: PID: %hu  Process Name: %s    Rcvd Process Name: %s\n",
                    hbmlm_proc_name_str,
                    cl_proc_PID,
                    cl_entry_curr_ptr->CL_procName_str,
                    hbm_msg_UnRegCprocessName_str);
        }
        else
        {
            if  (*hbm_msg_RegCmsgType_ptr == HBM_MSGTYPE_SHUTDOWN_NORMAL)
            {
                cl_entry_curr_ptr->CL_procStatus =
                        HBM_PROCSTATUS_SHUTDOWN_NORMAL;
            }
            else
            /* => (*hbm_msg_RegCmsgType_ptr == HBM_MSGTYPE_SHUTDOWN_ABNORMAL) */
            {
                cl_entry_curr_ptr->CL_procStatus =
                        HBM_PROCSTATUS_SHUTDOWN_ABNORMAL;
            }
            cl_entry_curr_ptr->CL_blockedTime = 0;
            cl_entry_curr_ptr->CL_shutdownTime = time_curr.tv_sec;
        }
    }

    if  (cl_entry_curr_ptr != NULL)
    {
        write_hbmlm_ckpointfile();

        /*  Send a reply message.  */

        strcpy(hbm_msg_RegCmsgResp, "HBMLM");
        if  (*hbm_msg_RegCmsgType_ptr == HBM_MSGTYPE_REGISTER)
        {
            hbm_msg_RegCmsgResp[5] = HBM_MSGTYPE_RESPONSE_REGISTER;
        }
        else if  (   (*hbm_msg_RegCmsgType_ptr ==
                                        HBM_MSGTYPE_SHUTDOWN_NORMAL)
                  || (*hbm_msg_RegCmsgType_ptr ==
                                        HBM_MSGTYPE_SHUTDOWN_ABNORMAL))
        {
            hbm_msg_RegCmsgResp[5] = HBM_MSGTYPE_RESPONSE_SHUTDOWN;
        }
        retcd = gettimeofday(&time_curr, NULL);
        PackUInt32(hbm_msg_RegCmsgResp+6, time_curr.tv_sec);

        from_addr.sin_family = AF_INET;

        nleft = hbm_msg_RegCmsgResp_len;
        msg_buff_ptr = hbm_msg_RegCmsgResp;
        while (nleft > 0)
        {
            nsent = sendto(sock_fd_reg,
                           msg_buff_ptr,
                           nleft,
                           0,
                           (struct sockaddr *) &from_addr,
                           sizeof(from_addr));
            if  (nsent <= 0)
            {
                fprintf(stderr,
                        "\n%s: HBM: Error sending report\n",
                        hbmlm_proc_name_str);
                perror(hbmlm_proc_name_str);
                nleft = 0;
            }
            else
            {
                msg_buff_ptr += nsent;
                nleft -= nsent;
            }
        }

        format_send_rptmsg(cl_entry_curr_ptr);
        write_hbmlm_ckpointfile();
    }

    DEBUGF2("Processing of msg from port %d successful\n", sock_fd_reg);
    return;
}


/*
************************************************************************
**  process_rpt_msg                                                   **
**      Parameters: sock_fd_rpt             the socket to get the     **
**                                          message from              **
**      Returns:    (Nothing)                                         **
**                                                                    **
**      This placeholder routine reads and discards a message         **
**      received on the UDP port used for sending reports to DCs.     **
************************************************************************
*/
void    process_rpt_msg(    int     sock_fd_rpt)
{
    int             msg_len;

    struct sockaddr from_addr;

#if  (   (defined TARGET_ARCH_CRAYT3E) \
      || (defined TARGET_ARCH_HPUX) \
      || (defined TARGET_ARCH_IRIX) \
      || (defined TARGET_ARCH_SOLARIS))
    int             from_addr_len;
#else
    size_t          from_addr_len;
#endif

    bzero((void *)&from_addr, sizeof(from_addr));
    from_addr.sa_family = AF_INET;
    from_addr_len = sizeof(from_addr);
    msg_len = recvfrom(sock_fd_rpt,
                       hbm_msg_array,
                       HBM_BUFF_SIZE_MSG,
                       0,
                       (struct sockaddr *) &from_addr,
                       &from_addr_len);

    if  (errno == EAGAIN)
    {
        return;
    }
    if  (strncmp(hbm_msg_array, "HBMDC", 5))
    {
        return;
    }
}


void    call_ps_update_client_data()
{
    int                     pipe_fd[2];

    int                     wk_int;

    int                     got_val,
                            retcd;

    int                     pid_fork,
                            pid_wait,
                            status;

    int                     ps_read;
    char                    ps_buff[HBM_BUFF_SIZE_PS],
                           *ps_buff_ptr_wk1,
                           *ps_buff_ptr_wk2,
                           *ps_buff_ptr_last_read,
                           *ps_buff_ptr_end = ps_buff + HBM_BUFF_SIZE_PS - 1;

    int                     ps_output_pid;
    char                    ps_output_status;
    int                     ps_output_cpu_days,
                            ps_output_cpu_hours,
                            ps_output_cpu_secs,
                            ps_output_cpu_mins;
    char                    ps_output_procname_array[256];

    struct hbmlm_cl_fields *cl_entry_ptr;

    for (cl_entry_ptr = hbmlm_cl_list.head;
         cl_entry_ptr;
         cl_entry_ptr = cl_entry_ptr->next)
    {
        cl_entry_ptr->CL_updated = 0;
    }

    if  (retcd = pipe(pipe_fd))
    {
        fprintf(stderr,
                "\n%s: Error creating pipe for ps to monitor clients",
                hbmlm_proc_name_str);
        fprintf(stderr,
                "\n%s: Client status info not updated\n",
                hbmlm_proc_name_str);
        perror(hbmlm_proc_name_str);
        return;
    }
    if  ((pid_fork = fork()) < 0)
    {
        fprintf(stderr,
                "\n%s: Error forking child for ps to monitor clients",
                hbmlm_proc_name_str);
        fprintf(stderr,
                "\n%s: Client status info not updated\n",
                hbmlm_proc_name_str);
        perror(hbmlm_proc_name_str);
        close(pipe_fd[0]);
        close(pipe_fd[1]);
        return;
    }
    if  (pid_fork == 0)
    {
/*
**      Child Code
*/
/*
**      Close stdout and the read pipe fd,
**      then dup the write pipe fd to 1 (stdout) and close it.
**      Finally, exec ps with the appropriate parms.
*/
        close(1);
        close(pipe_fd[0]);
        if  (dup2(pipe_fd[1],1) < 0)
        {
            fprintf(stderr,
                    "\n%s: Error duping write pipe for ps to monitor clients",
                    hbmlm_proc_name_str);
            fprintf(stderr,
                    "\n%s: Client status info not updated\n",
                    hbmlm_proc_name_str);
            perror(hbmlm_proc_name_str);
            close(pipe_fd[1]);
            exit(0);
        }
        close(pipe_fd[1]);

#if  (defined TARGET_ARCH_AIX)
        execl( "/bin/ps", "ps", "-e", "-o", "pid status time comm", (char *)0 );
#elif  (defined TARGET_ARCH_HPUX)
        execl( "/bin/ps", "ps", "-e", "-l", "-w", (char *)0 );
#elif  (defined TARGET_ARCH_IRIX)
        execl( "/bin/ps", "ps", "-e", "-o", "pid state time comm", (char *)0 );
#elif  (defined TARGET_ARCH_SUNOS)
        execl( "/bin/ps", "ps", "-agx", (char *)0 );
#elif  (defined TARGET_ARCH_SOLARIS)
        execl( "/bin/ps", "ps", "-e", "-o", "pid s time comm", (char *)0 );
#else
        execl( "/bin/ps", "ps", "-e", "-o", "pid state time comm", (char *)0 );
#endif

        fprintf( stderr,"\nexecl of \"ps\" failed\n");
        exit(0);
/*
**      End of Child Code.
*/
    }
/*
**  Parent Code from here...
*/
    close(pipe_fd[1]);

    if  (hbm_read_line(pipe_fd[0],
                       ps_buff,
                       HBM_BUFF_SIZE_PS) > 0)
    {
        while ((ps_read = hbm_read_line(pipe_fd[0],
                                        ps_buff,
                                        HBM_BUFF_SIZE_PS)) > 0)
        {
            ps_output_pid = 0;

            ps_buff_ptr_last_read = ps_buff + ps_read - 1;

/*
**          Parse the status line:
**              PID, ps_oputput_status_ptr, cpu seconds, minutes, ps_output_procname
*/
#if  (defined TARGET_ARCH_HPUX)
            sscanf(ps_buff,
                   "%d %c %d %d",
                   &wk_int,
                   &ps_output_status,
                   &wk_int,
                   &ps_output_pid);
            ps_buff_ptr_wk1 = index(ps_buff, ':');
            ps_buff_ptr_wk2 = ps_buff_ptr_wk1;
            while (!isspace(*ps_buff_ptr_wk2))
            {
                ps_buff_ptr_wk2--;
            }
            ps_buff_ptr_wk2++;
            if  (*(ps_buff_ptr_wk1 - 3) == '-')
            {
                sscanf(ps_buff,
                       "%d-%d:%d:%d %[ -z]\n",
                       &ps_output_cpu_days,
                       &ps_output_cpu_hours,
                       &ps_output_cpu_mins,
                       &ps_output_cpu_secs,
                       ps_output_procname_array);
            }
            else
            {
                ps_output_cpu_days = 0;
                if  (*(ps_buff_ptr_wk1 + 3) == ':')
                {
                    sscanf(ps_buff,
                           "%d:%d:%d %[ -z]\n",
                           &ps_output_cpu_hours,
                           &ps_output_cpu_mins,
                           &ps_output_cpu_secs,
                           ps_output_procname_array);
                }
                else
                {
                    ps_output_cpu_hours = 0;
                    sscanf(ps_buff,
                           "%d:%d %[ -z]\n",
                           &ps_output_cpu_mins,
                           &ps_output_cpu_secs,
                           ps_output_procname_array);
                }
            }

/* #if  (   (defined TARGET_ARCH_AIX) \
         || (defined TARGET_ARCH_IRIX) \
         || (defined TARGET_ARCH_SOLARIS)) */
#elif  (!(defined TARGET_ARCH_SUNOS))

            ps_buff_ptr_wk1 = index(ps_buff, ':');
            if  (*(ps_buff_ptr_wk1 - 3) == '-')
            {
                sscanf(ps_buff,
                       "%d %c %d-%d:%d:%d %[ -z]\n",
                       &ps_output_pid,
                       &ps_output_status,
                       &ps_output_cpu_days,
                       &ps_output_cpu_hours,
                       &ps_output_cpu_mins,
                       &ps_output_cpu_secs,
                       ps_output_procname_array);
            }
            else
            {
                ps_output_cpu_days = 0;
                if  (*(ps_buff_ptr_wk1 + 3) == ':')
                {
                    sscanf(ps_buff,
                           "%d %c %d:%d:%d %[ -z]\n",
                           &ps_output_pid,
                           &ps_output_status,
                           &ps_output_cpu_hours,
                           &ps_output_cpu_mins,
                           &ps_output_cpu_secs,
                           ps_output_procname_array);
                }
                else
                {
                    ps_output_cpu_hours = 0;
                    sscanf(ps_buff,
                           "%d %c %d:%d %[ -z]\n",
                           &ps_output_pid,
                           &ps_output_status,
                           &ps_output_cpu_mins,
                           &ps_output_cpu_secs,
                           ps_output_procname_array);
                }
            }

#else /*  TARGET_ARCH_SUNOS defined  */
/*
**          Get the PID, and skip the pointer past it.
*/
            while (   (ps_buff_ptr_wk1 <= ps_buff_ptr_last_read)
                   && (isspace(*ps_buff_ptr_wk1)))
            {
                ps_buff_ptr_wk1++;
            }
            got_val = sscanf(ps_buff_ptr_wk1, "%d", &ps_output_pid);
            while (   (ps_buff_ptr_wk1 <= ps_buff_ptr_last_read)
                   && (!isspace(*ps_buff_ptr_wk1)))
            {
                ps_buff_ptr_wk1++;
            }
/*
**          Skip the pointer past the terminal.
**          (The pointed to character is a space,
**           terminal id field begins with next (may be blank),
**           then we have spaces until first status character.)
*/
            ps_buff_ptr_wk1++;
            while (   (ps_buff_ptr_wk1 <= ps_buff_ptr_last_read)
                   && (!isspace(*ps_buff_ptr_wk1)))
            {
                ps_buff_ptr_wk1++;
            }
            while (   (ps_buff_ptr_wk1 <= ps_buff_ptr_last_read)
                   && (isspace(*ps_buff_ptr_wk1)))
            {
                ps_buff_ptr_wk1++;
            }
/*
**          The pointer is on the first status character
**          (the only one we're interested in).
*/
            if  (ps_buff_ptr_wk1 <= ps_buff_ptr_last_read)
            {
                ps_output_status = *ps_buff_ptr_wk1;
            }
/*
**          Skip the pointer to the CPU time.
*/
            ps_buff_ptr_wk1++;
            while (   (ps_buff_ptr_wk1 <= ps_buff_ptr_last_read)
                   && (!isdigit(*ps_buff_ptr_wk1)))
            {
                ps_buff_ptr_wk1++;
            }
            ps_buff_ptr_wk2 = index(ps_buff_ptr_wk1, ':');
            if  (ps_buff_ptr_wk2)
            {
                if  (*(ps_buff_ptr_wk2 - 3) == '-')
                {
                    sscanf(ps_buff_ptr_wk1,
                           "%d-%d:%d:%d %[ -z]\n",
                           &ps_output_cpu_days,
                           &ps_output_cpu_hours,
                           &ps_output_cpu_mins,
                           &ps_output_cpu_secs,
                           ps_output_procname_array);
                }
                else
                {
                    ps_output_cpu_days = 0;
                    if  (*(ps_buff_ptr_wk2 + 3) == ':')
                    {
                        sscanf(ps_buff_ptr_wk1,
                               "%d:%d:%d %[ -z]\n",
                               &ps_output_cpu_hours,
                               &ps_output_cpu_mins,
                               &ps_output_cpu_secs,
                               ps_output_procname_array);
                    }
                    else
                    {
                        ps_output_cpu_hours = 0;
                        sscanf(ps_buff_ptr_wk1,
                               "%d:%d %[ -z]\n",
                               &ps_output_cpu_mins,
                               &ps_output_cpu_secs,
                               ps_output_procname_array);
                    }
                }
            }
#endif  /* TARGET_ARCH_xxxx defined */
            ps_output_cpu_secs += (ps_output_cpu_days
                                   * HBM_TIME_SECS_PER_DAY);
            ps_output_cpu_secs += (ps_output_cpu_hours
                                   * HBM_TIME_SECS_PER_HOUR);
            ps_output_cpu_secs += (ps_output_cpu_mins
                                   * HBM_TIME_SECS_PER_MIN);

/*
**          Done parsing the status line
*/
/*
**          Now have all the fields,
**          get the client entry (if any) and update it.
*/
            if  (cl_entry_ptr = find_cl_entry(ps_output_pid))
            {
                if  (strcmp(ps_output_procname_array,
                            cl_entry_ptr->CL_procName_str))
                {
                    fprintf(stderr,
                            "\n%s: Warning -- process name from ps doesn't "
                            "match client info for pid %d",
                            hbmlm_proc_name_str,
                            ps_output_pid);
                    fprintf(stderr,
                            "\n%s:            Reg name: %s  PS name: %s",
                            hbmlm_proc_name_str,
                            cl_entry_ptr->CL_procName_str,
                            ps_output_procname_array);
                    fprintf(stderr,
                            "\n%s:            Replacing Reg name with PS name\n",
                            hbmlm_proc_name_str);
                    free(cl_entry_ptr->CL_procName_str);
                    cl_entry_ptr->CL_procName_str =
                                (char *)malloc(strlen(ps_output_procname_array) + 1);
                    strcpy(cl_entry_ptr->CL_procName_str,
                           ps_output_procname_array);
                }
/*
**              if  (the client is active)
*/
#if  (defined TARGET_ARCH_AIX)
                if  (   (ps_output_status == 'A')
                     || (ps_output_status == 'R'))
#elif  (defined TARGET_ARCH_HPUX)
                if  (ps_output_status == 'R')
#elif  (   (defined TARGET_ARCH_IRIX) \
        || (defined TARGET_ARCH_SOLARIS) \
        || (defined TARGET_ARCH_SUNOS))
                if  (   (ps_output_status == '0')
                     || (ps_output_status == 'R'))
#else
                if  (   (ps_output_status == 'O')
                     || (ps_output_status == 'R'))
#endif /* defined TARGET_ARCH_xxxx */
                {
                    cl_entry_ptr->CL_updated = 1;
                    if  (   (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_NORMAL)
                         && (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_ABNORMAL)
                         && (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_DIED))
                    {
                        cl_entry_ptr->CL_procStatus = HBM_PROCSTATUS_ACTIVE;
                        cl_entry_ptr->CL_shutdownTime = 0;
                        cl_entry_ptr->CL_shutdownMsgCt = 0;
                    }
                    cl_entry_ptr->CL_blockedTime = 0;
                    cl_entry_ptr->CL_cpuSecs = ps_output_cpu_secs;
                }
/*
**              else if  (the client is blocked)
*/
#if  (   (defined TARGET_ARCH_AIX) \
      || (defined TARGET_ARCH_HPUX))
                else if  (   (ps_output_status == 'I')
                          || (ps_output_status == 'S')
                          || (ps_output_status == 'T')
                          || (ps_output_status == 'W')
                          || (ps_output_status == 'X'))
#elif  (defined TARGET_ARCH_IRIX)
                else if  (   (ps_output_status == 'I')
                          || (ps_output_status == 'S')
                          || (ps_output_status == 'T')
                          || (ps_output_status == 'X'))
#elif  (   (defined TARGET_ARCH_SOLARIS) \
        || (defined TARGET_ARCH_SUNOS))
                else if  (   (ps_output_status == 'S')
                          || (ps_output_status == 'T'))
#else
                else if  (   (ps_output_status == 'S')
                          || (ps_output_status == 'T'))
#endif /* defined TARGET_ARCH_xxxx */
                {
                    if  (ps_output_cpu_secs > cl_entry_ptr->CL_cpuSecs)
                    {
                        cl_entry_ptr->CL_updated = 1;
                        if  (   (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_NORMAL)
                             && (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_ABNORMAL)
                             && (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_DIED))
                        {
                            cl_entry_ptr->CL_procStatus = HBM_PROCSTATUS_ACTIVE;
                            cl_entry_ptr->CL_shutdownTime = 0;
                            cl_entry_ptr->CL_shutdownMsgCt = 0;
                        }
                        cl_entry_ptr->CL_blockedTime = 0;
                        cl_entry_ptr->CL_cpuSecs = ps_output_cpu_secs;
                    }
                    else
                    {
                        cl_entry_ptr->CL_updated = 1;
                        if  (   (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_NORMAL)
                             && (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_ABNORMAL)
                             && (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_DIED))
                        {
                            cl_entry_ptr->CL_procStatus = HBM_PROCSTATUS_BLOCKED;
                            cl_entry_ptr->CL_blockedTime = time_curr.tv_sec;
                            cl_entry_ptr->CL_shutdownTime = 0;
                            cl_entry_ptr->CL_shutdownMsgCt = 0;
                        }
                        else
                        {
                            cl_entry_ptr->CL_blockedTime = 0;
                        }
                        cl_entry_ptr->CL_cpuSecs = ps_output_cpu_secs;
                    }
                }
/*
**              else if  (the client is dead/zombie)
*/
#if  (defined TARGET_ARCH_AIX)
                else if  (   (ps_output_status == '0')
                          || (ps_output_status == 'O')
                          || (ps_output_status == 'Z'))
#elif  (defined TARGET_ARCH_HPUX)
                else if  (   (ps_output_status == '0')
                          || (ps_output_status == 'Z'))
#elif  (   (defined TARGET_ARCH_IRIX) \
        || (defined TARGET_ARCH_SOLARIS) \
        || (defined TARGET_ARCH_SUNOS))
                else if  (ps_output_status == 'Z')
#else
                else if  (ps_output_status == 'Z')
#endif /* defined TARGET_ARCH_xxxx */
                {
                    cl_entry_ptr->CL_updated = 1;
                    if  (   (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_NORMAL)
                         && (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_ABNORMAL)
                         && (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_DIED))
                    {
                        cl_entry_ptr->CL_procStatus =
                                        HBM_PROCSTATUS_SHUTDOWN_DIED;
                        cl_entry_ptr->CL_blockedTime = 0;
                        cl_entry_ptr->CL_shutdownTime = time_curr.tv_sec;
                        cl_entry_ptr->CL_shutdownMsgCt = 0;
                    }
                    cl_entry_ptr->CL_cpuSecs = ps_output_cpu_secs;
                }
            }
        }  /* end while */

        for (cl_entry_ptr = hbmlm_cl_list.head;
             cl_entry_ptr;
             cl_entry_ptr = cl_entry_ptr->next)
        {
            if  (   (!(cl_entry_ptr->CL_updated))
                 && (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_NORMAL)
                 && (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_ABNORMAL)
                 && (cl_entry_ptr->CL_procStatus !=
                                        HBM_PROCSTATUS_SHUTDOWN_DIED))
            {
                cl_entry_ptr->CL_procStatus =
                                        HBM_PROCSTATUS_SHUTDOWN_DIED;
                cl_entry_ptr->CL_blockedTime = 0;
                cl_entry_ptr->CL_shutdownTime = time_curr.tv_sec;
                cl_entry_ptr->CL_shutdownMsgCt = 0;
            }
        }
    }  /* end  if (ps_read > 0)  */
    close(pipe_fd[0]);
    pid_wait = wait(&status);
}


#ifdef  TARGET_ARCH_HPUX
void    shutdown_quit()
#else
void    shutdown_quit(  int     dummy)
#endif
{
    quit_now = 1;
}


main(   int     argc,
        char   *argv[])
{
    char                   *parm_fname_str = NULL;

#if  (defined TARGET_ARCH_HPUX)
    int                     fd_readset;
    size_t                  max_fdp1;
#else
    fd_set                  fd_readset;
    int                     max_fdp1;
#endif  /*  TARGET_ARCH_HPUX  */

    struct timeval          time_left;

    struct hbmlm_cl_fields *cl_entry_ptr;

    int                     i,
                            error,
                            retcd,
                            retcd_time;

/*
************************************************************************
**  get the parameter file, then get the parameters:                  **
**                                                                    **
**      dflt_rpt_interval --  default report (to HBMDCs) interval     **
**      port_num          --  Port (socket) number for recieving      **
**                            registration/unregistration messages    **
**      ckpt_fname_str    --  Name of checkpoint file                 **
************************************************************************
*/
    hbmlm_proc_name_str = *argv;

    bzero((char *) &hbmlm_lm_info, sizeof(struct hbmlm_lm_fields));

/*
**  Get the HBMLM host IP Number (same as for any clients).
*/
    if  (error = get_LM_machine_info())
    {
        fprintf(stderr,
                "\n%s: Error getting machine information",
                hbmlm_proc_name_str);
        fprintf(stderr,
                "\n%s: Continuing without heartbeat processing\n",
                hbmlm_proc_name_str);
        return (-100 + error);
    }

/*
**  Initialize the Internet address fields.
*/
    hbmlm_lm_info.LM_addr_reg.sin_family = AF_INET;
    hbmlm_lm_info.LM_addr_reg.sin_addr.s_addr = htonl(INADDR_ANY);
    hbmlm_lm_info.LM_addr_rpt.sin_family = AF_INET;
    hbmlm_lm_info.LM_addr_rpt.sin_addr.s_addr = htonl(INADDR_ANY);

/*
**  Get the hbmlm parameter file name,
**  then get the parameter values.
*/
    parm_fname_str = get_hbmlm_parmfile_name(argc, argv);

    if  (error = get_hbmlm_parms(parm_fname_str))
    {
        return (-200 + error);
    }

    DEBUGF1("Parameter file parsed\n");
    DEBUGF2("Local Monitor hostname: %s\n", hbmlm_lm_info.LM_hostName_str );
    DEBUGF2("Local Monitor IP addr:  %s\n", hbmlm_lm_info.LM_hostIPNum_str );

/*
**  Check for and read the checkpoint file for previous run
**  information, including information about client processes
**  requiring continuing (resumed) monitoring.
*/
    hbmlm_lm_info.LM_ClientsCt = 0;
    read_hbmlm_ckpointfile();
    format_Rptmsg_LM_fields();

/*
************************************************************************
**  Now we have all the parameters, so we set up the sockets.         **
************************************************************************
*/
    if  (retcd = set_up_udp_port(&sock_fd_reg,
                                 &(hbmlm_lm_info.LM_addr_reg)))
    {
        fprintf(stderr,
                "\n%s: Error setting up UDP (reg/unreg) port\n",
                hbmlm_proc_name_str);
        close(sock_fd_reg);
        return retcd;
    }
#ifdef  TARGET_ARCH_HPUX
    if  (sock_fd_reg >= (8 * sizeof(int)))
    {
        fprintf(stderr,
                "\n%s: HBM: Funky HP-UX error giving fd number "
                "too large to process.\n",
                hbmlm_proc_name_str);
        return -300;
    }
#endif  /*  TARGET_ARCH_HPUX defined  */

    DEBUGF2("Listening for regs/unregs on port %d\n", hbmlm_lm_info.LM_addr_reg.sin_port);

    if  (retcd = set_up_udp_port(&sock_fd_rpt,
                                 &(hbmlm_lm_info.LM_addr_rpt)))
    {
        fprintf(stderr,
                "\n%s: Error setting up UDP (rpt) port\n",
                hbmlm_proc_name_str);
        close(sock_fd_rpt);
        close(sock_fd_reg);
        return retcd;
    }
#ifdef  TARGET_ARCH_HPUX
    if  (sock_fd_rpt >= (8 * sizeof(int)))
    {
        fprintf(stderr,
                "\n%s: HBM: Funky HP-UX error giving fd number "
                "too large to process.\n",
                hbmlm_proc_name_str);
        return -400;
    }
#endif  /*  TARGET_ARCH_HPUX defined  */

    DEBUGF2("Sending reports on port %d\n", hbmlm_lm_info.LM_addr_rpt.sin_port);

    max_fdp1 = sock_fd_reg;
    if  (sock_fd_rpt > max_fdp1)
    {
        max_fdp1 = sock_fd_rpt;
    }

    max_fdp1++;

    signal(SIGINT, shutdown_quit);
    signal(SIGQUIT, shutdown_quit);

    retcd_time = gettimeofday(&time_curr, NULL);
    if  (hbmlm_lm_info.LM_ClientsCt == 0)
    {
        time_to_monitor_write.tv_sec = 0;
        time_to_monitor_write.tv_usec = 0;
    }
    else
    {
/*      Go through all client records, setting rptTimeNext to now.    */
        for (cl_entry_ptr = hbmlm_cl_list.head;
             cl_entry_ptr;
             cl_entry_ptr = cl_entry_ptr->next)
        {
            cl_entry_ptr->CL_rptTimeNext = time_curr.tv_sec;
        }
        time_to_monitor_write.tv_sec = time_curr.tv_sec;
        time_to_monitor_write.tv_usec = 0;
    }
    write_hbmlm_ckpointfile();

    /***********************/
    /*   Main Select Loop  */
    /***********************/

    DEBUGF1("Entering main select loop\n");

    while (!quit_now)
    {
#ifdef  TARGET_ARCH_HPUX
        bzero((char *) &fd_readset, sizeof(int));
        fd_readset |= (1 << sock_fd_reg);
        fd_readset |= (1 << sock_fd_rpt);
#else
        FD_ZERO(&fd_readset);
        FD_SET(sock_fd_reg, &fd_readset);
        FD_SET(sock_fd_rpt, &fd_readset);
#endif

        if  (!hbmlm_cl_list.head)
        {
            if  (!quit_now)
            {
                retcd = select(max_fdp1, &fd_readset, NULL, NULL, NULL);
                DEBUGF1("select completed\n");
            }
            else
            {
                retcd = 0;
            }
        }
        else
        {
            cl_entry_ptr = hbmlm_cl_list.head;
            time_to_monitor_write.tv_sec =
                                cl_entry_ptr->CL_rptTimeNext;
            time_to_monitor_write.tv_usec = 0;

            for (cl_entry_ptr = cl_entry_ptr->next;
                 cl_entry_ptr;
                 cl_entry_ptr = cl_entry_ptr->next)
            {
                if  (time_to_monitor_write.tv_sec > cl_entry_ptr->CL_rptTimeNext)
                {
                    time_to_monitor_write.tv_sec = cl_entry_ptr->CL_rptTimeNext;
                }
            }
            retcd_time = gettimeofday(&time_curr, NULL);
            if  (time_to_monitor_write.tv_sec <= time_curr.tv_sec)
            {
                time_left.tv_sec = time_left.tv_usec = 0;
            }
            else
            {
                time_left.tv_sec = time_to_monitor_write.tv_sec - time_curr.tv_sec;
/*              (time_to_monitor_write.tv_usec is 0, so always <= time_curr.tv_usec)  */
                if  (time_curr.tv_usec)
                {
                    time_left.tv_sec--;
                    time_left.tv_usec = 1000000 - time_curr.tv_usec;
                }
            }

            if  (!quit_now)
            {
                retcd = select(max_fdp1, &fd_readset, NULL, NULL, &time_left);
                DEBUGF1("timed select completed\n");
            }
            else
            {
                retcd = 0;
            }
        }

        if  (   (retcd < 0)
             && (errno != EINTR))
        {
            if  (errno == EBADF)
            {
                fprintf(stderr,
                        "\n%s: Error in select(), "
                        "invalid descriptor specified\n",
                        hbmlm_proc_name_str);
                perror(hbmlm_proc_name_str);
            }
            if  (errno == EFAULT)
            {
                fprintf(stderr,
                        "\n%s: Error in select(), "
                        "invalid pointer parameter\n",
                        hbmlm_proc_name_str);
                perror(hbmlm_proc_name_str);
            }
            return 1;
        }

        if  (retcd > 0)
        {
#ifdef  TARGET_ARCH_HPUX
            if  (fd_readset & (1 << sock_fd_reg))
#else
            if  (FD_ISSET(sock_fd_reg, &fd_readset))
#endif  /*  TARGET_ARCH_HPUX defined  */
            {
                process_reg_unreg_msg(sock_fd_reg);
            }
#ifdef  TARGET_ARCH_HPUX
            if  (fd_readset & (1 << sock_fd_rpt))
#else
            if  (FD_ISSET(sock_fd_rpt, &fd_readset))
#endif  /*  TARGET_ARCH_HPUX defined  */
            {
                process_rpt_msg(sock_fd_rpt);
            }
        }
        
        retcd_time = gettimeofday(&time_curr, NULL);
        if  (   (time_curr.tv_sec >= time_to_monitor_write.tv_sec)
             || (quit_now))
        {
            DEBUGF2("start performing timed updates, curr time is %d\n", time_curr.tv_sec);

            if  (hbmlm_lm_info.LM_ClientsCt)
            {
                call_ps_update_client_data();
            }
            write_hbmlm_ckpointfile();
            retcd_time = gettimeofday(&time_curr, NULL);
            send_rpt_msgs();
            write_hbmlm_ckpointfile();
        }
    }
    
    close(sock_fd_reg);
    close(sock_fd_rpt);

    exit(0);
}
