/*****************************************************************/
/*      server.h                                                 */
/*      Version 0.0                      Henri Casanova          */
/*****************************************************************/

/* Included file */

#include "global.h"
#include "netsolve_arch.h"
#include "object.h"

/* UGLY */

/* Macros */

#define ARGUMENT 0
#define CONSTANT 1
#define FORMULA  2

#define CONNECT_TIMEOUT  100  /* In seconds                          */

#define REST_DELAY_INCR	 45  /* In seconds                          */

#define HASH_PB          10  /* Number of entries in the hash table */
#define HASH_SV          10  /* Number of entries in the hash table */
#define SERVER_MODULE    3   /* Size of a server module             */

#define MAX_FAILURE      0  /* Removed after 1  failures          */

#define MAX_LINE_LENGTH  150  /* MAx line length in input file       */

/* Servers */

typedef struct ProblemItem problem_item;
typedef struct PbList pb_list; /* List of problem codes */
struct PbList
{
  problem_item *problem; /* Problem descriptor   */
  pb_list *next;         /* Next problem         */
  pb_list *previous;     /* Previous problem     */
};

typedef struct ServerDesc server_desc; /* ServerDescriptor */
struct ServerDesc
{
  char hostname[MAX_HOST_NAME];   /* host name          */
  unsigned int IPaddr;            /* IP addr            */
  int data_format;                /* Data Format        */
  int speed;                      /* Speed              */
  int host_status;                /* Up/Down            */
  int server_status;              /* Running/Stopped    */
  int number_proc;                /* Number of procs    */
  int mpp;                        /* 0 or 1             */
  int connection;                 /* usec for connect   */
  int transfer;                   /* bytes/sec          */
  int workload;                   /* Workload           */
};

typedef struct ServerItem server_item; /* list of descriptor */
struct ServerItem
{
  server_desc server;    /* Server descriptor */
  server_item *previous;
  server_item *next;
  int number_failed;     /* Number of failures           */
  time_t first_stopped;  /* First stopped                */
  time_t last_pong;      /* Last measure                 */
  int time_ponged;       /* Number of pong measures      */
  int dirty;             /* 1: Has been sent to a client */
  time_t creation;       /* Date of creation             */
  int number_problems;   /* number of problems           */
  pb_list *first_pb;     /* First Problem                */
};

/************************/
/* Global Variables     */
/************************/

 server_item *sv_hash_table[HASH_SV]; /* Hash Table of resources */
 server_item *ag_hash_table[HASH_SV]; /* Hash Table of agents    */

#define MAX_COMP 100

 int computation_pids[MAX_COMP];   /* Array of computation processes*/
 int computation_rest[MAX_COMP];   /* Array of computation processes*/

 int CONDOR;                       /* 0: NO    1: YES               */
 char *CONDOR_path;

 char *scratch_path;

 int work_pid;                     /* pid of the process in charge  */
                                  /* of the workload broadcast     */
 int lcl_sock;                     /* UNIX socket descriptor        */
 int rmt_sock;                     /* Listning socket               */
 server_item *my_self;             /* Server item of this process   */
 int client_desc;                  /* Service socket descriptor     */
 char my_host[MAX_HOST_NAME];      /* Name of my host               */
 unsigned int my_address;          /* my IP address                 */
 struct hostent *my_hp;            /* my host descriptor            */
 int my_data_format;               /* Data format                   */
 char my_arch[10];                 /* My architecture               */
 int mpp;                          /* I'm a MPP                     */
 int number_processors;            /* For servers                   */

 int number_computations;          /* Number of ongoing computation */

 char *root_path;                  /* $(NETSOLVE_ROOT_SERVER)       */
 char tmp_path[30];                    
 char log_path[30];

#define ONLY_COMPUTING 0
#define ALL_SERVERS    1

/* Statistics */

 int nb_pb_solved;     /* Number of solved problems */
 int nb_lcl_request;   /* Number of client request  */
 time_t up_date;       /* Date of start             */
 char char_up_date[26];/* String version            */

/************/
/* Problems */
/************/

struct server_item;

typedef struct SvList sv_list; /* List of server numbers */
struct SvList
{
  server_item *server;       /* Server descriptor                */
  sv_list *next;             /* Next server                      */
  sv_list *previous;         /* Previous server                  */
};

struct ProblemItem
{
  problem_desc problem;      /* Problem descriptor               */
  problem_item *previous;    /* Previous Problem                 */
  problem_item *next;        /* Next problem                     */
  int number_servers;        /* Number of servers in the list    */
  sv_list *first_server;     /* First server knowing the problem */
};

 problem_item *pb_hash_table[HASH_PB]; /* Hash table */

/* Miscelleanous */

#define PONG_PERIOD   3600      /* Distance request every hour                */
#define RECONTACT_PERIOD 450000 /* Distance request every five days           */
#define PONG_TRIALS   24        /* 24 trials (one day)                        */
#define WORKLOAD_TIME 120       /* the workload is broadcasted every 120 sec  */
#define FORCE_BROADCAST 10      /* The broadcast occurs every 20 minutes      */
#define WORKLOAD_CONFIDENCE 20  /* [-20;+20]                                  */
#define TO_BE_A -1              /* To be a arranged : default value           */
                                /* before receiving the report                */

#define DISTANCE_LENGTH 40960
#define DISTANCE_TRY    10

/*
 * Sorting structure
 */

typedef struct ServerScore server_score; /* List of server numbers */
struct ServerScore
{
  int index;
  double score;
};

/*
 * the information needed by the parallel server.  This will simply
 * add things to server.h
 * for the pvm version specify -DPVM to the compiler
*/

/* need to do away with this, don't know how yet */

/* for the matrix descriptors for dense matrices which are type 1*/

#define DTYPE 0
#define CTXT_ 1
#define M_ 2
#define N_ 3
#define MB_ 4
#define NB_ 5
#define RSRC_ 6
#define CSRC_ 7
#define LLD_ 8

/* some simple macros for the LLD computation */
#ifndef MIN
#define MIN(a,b) (a)<(b) ? (a):(b)
#endif
#ifndef MAX
#define MAX(a,b) (a)>(b) ? (a):(b)
#endif
#define ICEIL(a,b) ((a)+(b)-1)/(b)

/*
 * values to correspond to the type of parallel machine in use
 */

#define PVM 1
#define MPI 2
#define MPL 3
#define NX 4

int MPInodes;
char MPImachinefile[256];

/*
 * these globals are to ensure that the necessary information is available
 * in all the various fork() and subroutine operations in the parallel version.
 */
 int NUM_PROCS;
 int MYROW,MYCOL;
 int PROC_ROWS, PROC_COLS;
 int MYCONTXT;
 int ROW_BLOCK, COL_BLOCK;
 void *SIZE_PTR;

/*
 * User restiction data_structure.
 */

typedef struct RestrictionItem restriction_item; 
struct RestrictionItem
{
  char domain[128];
  int max_pending;
  int pending;
};

#define NB_RESTRICTIONS 20

 int nb_restrictions;
 restriction_item restrictions[NB_RESTRICTIONS];

FILE* globalfile;
