/*****************************************************************/
/*      agent.c                                                  */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  main()                                                       */
/*****************************************************************/

/*
 *  This is the main funtion of the NetSolve agent
 */

#include "server.h"
#include "agent_prot.h"
#include "shared_prot.h"

#ifdef __STDC__
main(int argc,char **argv)
#else
main(argc,argv)
int argc;
char **argv;
#endif
{
  int ag_port = AG_PORT;
  int pid;

  fd_set readfd;                      /* For select()        */
 
  struct sockaddr_in rmt_client;      /* INET socket address */
  int len_client_add;                 /* address length      */
  
  struct hostent *hp;                 /* Host descriptor     */
  int cc;
  int i;
 
  char command_line[256];
  

/***************************
 * Initialize structure    *
 ***************************/

for (i=0;i<HASH_SV;i++)
  sv_hash_table[i] = NULL;
for (i=0;i<HASH_SV;i++)
  ag_hash_table[i] = NULL;
for (i=0;i<HASH_PB;i++)
  pb_hash_table[i] = NULL;

/**********************
 * Getting the path   *
 **********************/

  if ((root_path=(char *)getenv("NETSOLVE_SERVER_ROOT")) == NULL)
  {
    fprintf(stderr,"The NETSOLVE_SERVER_ROOT variable should be set\n");
    exit(0);
  }
  
/**********************
 * Geting my hostname *
 **********************/

  if (gethostname(my_host,MAX_HOST_NAME) == -1)
  {
    perror("main() gethostname()");
    term();
  }
  if ((my_hp = gethostbyname(my_host)) == NULL)
  {
    perror("main() gethostbyname()");
    term();
  }
  BCOPY(my_hp->h_addr_list[0],&my_address,sizeof(unsigned int));
  strcpy(my_host,my_hp->h_name);

  
/**********************
 * Arguments analysis *
 **********************/

  if (argc >= 3)
  {
    fprintf(stderr,"Usage : %s [agent name ]\n",argv[0]);
    term();
  }


/******************************
 * Use the configuration file *
 *******************************/

#ifdef VIEW
  fprintf(stderr,"initializing......\n");
#endif

  if (agent_init() == -1)
    term();
  if (argc == 2)
    if (contact_master(argv[1]) == -1)
      term();

/*******************
 * Set the signals *
 *******************/

  set_signals();

#ifdef HBM

/*******************
 * Running the HBM *
 *******************/

/***********************************
 * Clean_up the HBM files          *
 ***********************************/

  sprintf(command_line,"rm -rf /tmp/hbmdc*");
  system(command_line); 

/***********************************
 * Start up the HBM data collector *
 ***********************************/

#ifdef VIEW
  fprintf(stderr,"starting up the Heart Beat Monitor Data Collector.....\n");
#endif
  pid = fork();
  if (pid == -1)
    perror("fork");

  if (pid == 0) /* The HBMDC */
  {
    char path[256];

    sprintf(path,"%s/Globus/build/%s/",root_path,my_arch);
    if (chdir(path)== -1)
    {
      fprintf(stderr,"Warning: Unable to start the HBM data collector.\n");
      perror("chdir");
      exit(0);
    }
    if (execl("./hbmdc","hbdmc",(char *)0) == -1)
    {
      fprintf(stderr,"Warning: Unable to start the HBM data collector.\n");
      perror("execl");
      exit(0);
    }
  }

/***************************************
 * Start up the HBM NetSolve interface *
 ***************************************/

#ifdef VIEW
  fprintf(stderr,
     "starting up the Heart Beat Monitor NetSolve Interface.....\n");
#endif
  pid = fork();
  if (pid == -1)
    perror("fork");

  if (pid == 0) /* The HBM NetSolve interface */
  {
    char path[256];

    sprintf(path,"%s/bin/%s/",root_path,my_arch);
    if (chdir(path)== -1)
    {
      fprintf(stderr,"Warning: Unable to start the HBM NetSolve Interface.\n");
      perror("chdir");
      exit(0);
    }
    if (execl("./hbcproxy","hbcproxy",(char *)0) == -1)
    {
      fprintf(stderr,"Warning: Unable to start the HBM NetSolve Interface.\n");
      perror("execl");
      exit(0);
    }
  }

#endif

/*********************
 * Create a log file *
 *********************/

#ifdef VIEW
  fprintf(stderr,"creating a log file......\n");
#endif
  new_agent_log();

/**********************
 * Startup of socket  *
 **********************/

#ifdef VIEW
  fprintf(stderr,"enabling network connections......\n");
#endif
  /* remote socket */
  if ((rmt_sock = establish_socket(&ag_port)) == -1)
  {
    perror("main() create_socket_INET()");
    term();
  }

/*************
 * MAIN LOOP *
 *************/
 
#ifdef VIEW
  fprintf(stderr,"Agent running.\n");
#endif

  while(1)
  {
    listen(rmt_sock,MAXCONN);
    len_client_add = sizeof(rmt_client);
    if ((client_desc = accept(rmt_sock,
              (struct sockaddr *)&rmt_client,&len_client_add)) == -1)
    {
      if ((errno == EINTR)||(errno ==ECHILD))
        continue;
#ifdef VIEW
        perror("main() accept()");
#endif
        continue;
    }
    /* I generate the reply */
    if (proc_mes() == -1)
    {
      if (gen_sv_failure(client_desc) == -1)
      {
#ifdef VIEW
        fprintf(stderr,"WARNING : Impossible to notify my failure\n");
#endif
      }
    }
/****************/
  } /* END LOOP */
/****************/
 
}
