/*****************************************************************/
/*      proc_mes.c                                               */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  proc_mes()                                                   */
/*  gen_all_struct()                                             */
/*  gen_pb_failed()                                              */
/*  gen_pb_refused()                                             */
/*  asses_score()                                                */
/*  are_you_alive_request()                                      */
/*---------------------------------------------------------------*/
/* INFO_REQUEST                                                  */
/* PROBLEM_REQUEST                                               */
/* CLIENT_F_REPORT                                               */
/* FAILURE_REPORT                                                */
/* DISTANCE_REPORT                                               */
/* ARE_YOU_ALIVE                                                 */
/* IS_ALIVE                                                      */
/* I_AM_NEW                                                      */
/* I_EXIST                                                       */
/* WORKLOAD                                                      */
/* I_DIE                                                         */
/* POLL_CONTACT                                                  */
/* POLL_LIST_PB                                                  */
/* POLL_LIST_SV                                                  */
/* POLL_GET_PB                                                   */
/* POLL_GET_SV                                                   */
/* POLL_STAT                                                     */
/* POLL_KILL                                                     */
/* JOB_COMPLETED                                                 */
/*****************************************************************/

#include "server.h"
#include "agent_prot.h"
#include "shared_prot.h"

/*
 * proc_mes : the server/agent received a message on my remote socket
 *   the first 4 bytes of the message are an integer that gives the
 *   message ID. According to this ID, the message is processed.
 */

#ifdef __STDC__
int proc_mes()
#else
int proc_mes()
#endif
{

  int message_type;
  int encoding;
  FILE *stream;
  char logbuf[256];
  time_t current;

  /* Get message ID */

  encoding = init_recv(client_desc);
  read_int(&message_type);

  switch(message_type){

/***********************/
  case INFO_REQUEST:   
/***********************/
{
  /* A client wants the description of a problem */

  char code[CODE_LENGTH]; /* FORMAT + CODE */
  problem_item *problem;
  int status,i;
  int format;
  format_spec *format_descriptor;
  constant *current_constant;
  formula *current_form;
  matlab_merge *current_matlab;

  read_string(code);
  end_message();
  
#ifdef VIEW
  fprintf(stderr,"Information request for problem '%s'\n",code);
#endif
  
  problem = (problem_item *)find_problem(code);
  if (problem == 0)
  {
#ifdef VIEW
    fprintf(stderr,"\tUnknown problem\n");
#endif
    status = NO_SUCH_PB;
    init_send(encoding,client_desc);
    write_int(&status);
    end_message();
    end_com();
    return 1;
  }

  status = 1;
  init_send(encoding,client_desc);
  write_int(&status);
  write_string(problem->problem.path);
  write_string(problem->problem.desc);
  write_int(&(problem->problem.nb_obj_input));
  for (i=0;i<problem->problem.nb_obj_input;i++)
  {
    write_int(&(problem->problem.input_obj[i].data_type));
    write_int(&(problem->problem.input_obj[i].object_type));
    write_string(problem->problem.input_obj[i].object_desc);
  }
  write_int(&(problem->problem.nb_obj_output));
  for (i=0;i<problem->problem.nb_obj_output;i++)
  {
    write_int(&(problem->problem.output_obj[i].data_type));
    write_int(&(problem->problem.output_obj[i].object_type));
    write_string(problem->problem.output_obj[i].object_desc);
  }
  /* Send the desired format */
  format_descriptor = &(problem->problem.formats[0]);
  write_int(&(format_descriptor->nb_args));
  for (i=0;i<format_descriptor->nb_args;i++)
    write_string(format_descriptor->args_desc[i]);
  /* Sends the constants */
  write_int(&(format_descriptor->nb_constants));
  current_constant = format_descriptor->consts;
  for (i=0;i<format_descriptor->nb_constants;i++)
  {
    write_string(current_constant->mnemonic); 
    write_int(&(current_constant->value)); 
    current_constant = current_constant->next;
  }
  /* Sends the formulae */
  write_int(&(format_descriptor->nb_formulae));
  current_form = format_descriptor->forms;
  for (i=0;i<format_descriptor->nb_formulae;i++)
  {
    write_string(current_form->mnemonic);
    write_string(current_form->expression);
    current_form = current_form->next;
  }
  /* Send the merge */
  current_matlab = problem->problem.matlab;
  write_int(&(problem->problem.nb_matlab));
  for (i=0;i<problem->problem.nb_matlab;i++)
  { 
    write_int(&(current_matlab->output1));
    write_int(&(current_matlab->output2));
    current_matlab = current_matlab->next;
  }
  end_message();
  end_com();
#ifdef VIEW
    fprintf(stderr,"\tInformation sent back\n");
#endif
  return 1;
}

/***********************/
  case PROBLEM_REQUEST:   
/***********************/
{
  /* This is a request for solving a problem */
  int client_type;
  char code[CODE_LENGTH];
  int problem_size;
  int data_size,result_size;
  problem_item *problem;
  server_item **list_of_servers;
  int len_list;
  problem_desc pb_desc;
  int status,i;

  read_int(&client_type);
  read_string(code);
#ifdef VIEW
  fprintf(stderr,"Client wants to solve '%s'\n",code);
#endif
  read_int(&problem_size);
  read_int(&data_size);
  read_int(&result_size);
  read_int(&(pb_desc.nb_obj_input));
  pb_desc.input_obj = (object *)calloc(pb_desc.nb_obj_input,sizeof(object));
  for (i=0;i<pb_desc.nb_obj_input;i++)
  {
    read_int(&(pb_desc.input_obj[i].data_type));
    read_int(&(pb_desc.input_obj[i].object_type));
  }
  read_int(&(pb_desc.nb_obj_output));
  pb_desc.output_obj = (object *)calloc(pb_desc.nb_obj_output,sizeof(object));
  for (i=0;i<pb_desc.nb_obj_output;i++)
  {
    read_int(&(pb_desc.output_obj[i].data_type));
    read_int(&(pb_desc.output_obj[i].object_type));
  }
  /* END OF A RECEIVE PHASE */
  end_message();

  /*
   * Does this problem exist ?
   */
  problem = (problem_item *)find_problem(code);
  if (problem == 0)
  {  /* There is no such problem */
    free(pb_desc.output_obj);
    free(pb_desc.input_obj);
#ifdef VIEW
    fprintf(stderr,"\tUnknown problem\n");
#endif
    if (gen_rep_pb_req(encoding,NO_SUCH_PB,(server_item **)0,
                       0,0,0,0,(problem_item *)0) == -1)
    {
      perror("proc_mes() gen_rep_pb_req()");
      return -1;
    }
    end_com();
    return 1;
  }
  /*
   * Is the problem correctly specified ?
   */
  pb_desc.complexity = NULL;
  status = problem_cmp(pb_desc,problem->problem);
  free(pb_desc.output_obj);
  free(pb_desc.input_obj);

  if ((status == -1)||(status == -3))
  {
  /* The input parameters are not well specified */
#ifdef VIEW
    fprintf(stderr,"\tBad number of input objects\n");
#endif
    if (gen_rep_pb_req(encoding,BAD_INPUT,(server_item **)0,
                       0,0,0,0,(problem_item *)0) == -1)
    {
      perror("proc_mes() gen_rep_pb_req()");
      return -1;
    }
    end_com();
    return 1;
  }
  if ((status == -2) || (status == -4))
  {
#ifdef VIEW
    fprintf(stderr,"\tBad number of input objects\n");
#endif
    if (gen_rep_pb_req(encoding,BAD_OUTPUT,(server_item **)0,
                       0,0,0,0,(problem_item *)0) == -1)
    {
      perror("proc_mes() gen_rep_pb_req()");
      return -1;
    }
    end_com();
    return 1;
  }
  if (status == -5)
  {
  /* Nothing here */
  }

 /*
  * From now, the problem is well specified
  */
  nb_lcl_request++;
  sprintf(logbuf,"Request #%d : problem='%s', size=%d",nb_lcl_request,code,problem_size);
  append_to_log(logbuf);
  add_request_to_log(client_desc);
  list_of_servers = (server_item **)give_server_list(problem,&len_list);
  if (gen_rep_pb_req(encoding,CORRECT_PB,list_of_servers,len_list,
      problem_size,data_size,result_size,problem) == -1)
  {
    perror("proc_mes() gen_rep_pb_rq()");
    if (list_of_servers != NULL)
      free(list_of_servers);
    return -1;
  }
  end_com();
#ifdef VIEW
    fprintf(stderr,"\tList of computational servers returned\n");
#endif
  if (list_of_servers != NULL)
      free(list_of_servers);
  return 1;
}
  /*******************/
  case CLIENT_F_REPORT:
  /*******************/
{
  /* Failure report from a client */

  unsigned int IPaddr;
  server_item *server;
  int type;

  read_int(&type);
  read_IPaddr(&IPaddr);
  end_message();
  end_com();
  if ((server = (server_item *)find_server(IPaddr)) == 0)
    return 1;
#ifdef VIEW
  fprintf(stderr,"Client reported failure for server %s (",
          server->server.hostname);
  if (type == FAILED)
    fprintf(stderr,"FAILED)\n");
  else
    fprintf(stderr,"STOPPED)\n");
#endif
  time(&current);
  sprintf(logbuf,"Report of failure for %s (%s); %s",
          server->server.hostname,(type==FAILED ? "FAILED":"STOPPED"),
          ctime(&current));
  append_to_log(logbuf);
  if (server->dirty == 1)
  {
    server->server.server_status = type;
    if (type == FAILED)
    {
      server->number_failed++;
      if (server->number_failed > MAX_FAILURE)
      {
#ifdef VIEW
        fprintf(stderr,"\tToo many failure for %s (%d): Marked for removal \n",
                server->server.hostname,server->number_failed);
#endif
        remove_server(server);
      }
    }
    if (type == STOPPED)
    {
      time_t date;

      if (server->first_stopped == 0)
      {
        time(&date);
        server->first_stopped = date;
      }
    }
  }
  return 1;
}
  
  /*******************/
  case FAILURE_REPORT:
  /*******************/
{
  /* Failure report form a utility process */
  time_t date;
  unsigned int IPaddr;
  server_item *server;

  read_long(&date);
#ifdef VIEW
  fprintf(stderr,"Failure report from a utility process\n");
#endif
  read_IPaddr(&IPaddr);
  end_message();
  end_com();
  /* END OF A RECEIVE PHASE */
  if ((server = (server_item *)find_server(IPaddr)) == 0)
  {
    if ((server = (server_item *)find_agent(IPaddr)) == 0)
      return 1;
  }
  if (date > (server->creation))
  {
    (server->server).host_status = 0;
#ifdef VIEW
    fprintf(stderr,"\tMarked %s down and stopped\n",(server->server).hostname);
#endif
    time(&current);
    sprintf(logbuf,"***Marked %s down and stopped, %s",
        (server->server).hostname,ctime(&current));
    append_to_log(logbuf);
    server->server.host_status = DOWN;
    server->server.server_status = STOPPED;
    if (server->first_stopped == 0)
      server->first_stopped = date;
      
  }
return 1;
}
  /*******************/
  case DISTANCE_REPORT:
  /*******************/
{
  /* Bandwidth/Latency report */
  unsigned int IPaddr;
  int delay1,delay2;
  server_item *server;

  read_IPaddr(&IPaddr);
  read_int(&delay1);
  read_int(&delay2);
  end_message();
  end_com();
  /* END OF A RECEIVE PHASE */
  if ((server = (server_item *)find_server(IPaddr)) == 0)
    return 1;
  time(&(server->last_pong));
  if ((delay1 == -1)||(delay2 == -1))
  {
    if ((server->server).host_status == UP)
      server->time_ponged = 1; /* First failure */
    else
      server->time_ponged++; /* Subsequent failure */

#ifdef VIEW
      fprintf(stderr,"%d pong failure for %s\n",
             server->time_ponged,
             server->server.hostname);
#endif
  /* Who knows  ???? */
/*
    server->server.host_status = DOWN;
*/
  }
  else
  {
    if ((server->server).host_status == DOWN)
    {
#ifdef VIEW
      fprintf(stderr,"%s seems to be UP\n",server->server.hostname);
      fprintf(stderr,"Is the server on %s running ?\n",
              server->server.hostname);
      are_you_alive_request(server->server.IPaddr);
#endif
      server->server.host_status = UP;
    }
    else
    {
      server->time_ponged++;
#ifdef VIEW
      fprintf(stderr,"'%s', %dth pong trial : %d %d\n",
              server->server.hostname,server->time_ponged,
              delay1,delay2); 
#endif
      if (server->time_ponged == 1)
      {
        server->server.connection = delay1;
        server->server.transfer   = delay2;
      }
      else
      {
        (server->server).connection = ((server->time_ponged -1)*
          (int)(server->server.connection)+(int)delay1) /
          (server->time_ponged);
        (server->server).transfer =   ((server->time_ponged -1)*
          (int)(server->server.transfer)  +(int)delay2) /
          (server->time_ponged);
      }
    }
  }
  return 1;
}

/***********************/
  case ARE_YOU_ALIVE:      
/***********************/
{
  /* Of course I am ALIVE !! */
  int answer = ALIVE;

  end_message();
  init_send(encoding,client_desc);
  write_int(&answer);
  end_message();
  end_com();
  return 1;
}

/***********************/
  case IS_ALIVE:     
/***********************/
{
  /* The computational server is alive */
  unsigned int IPaddr;
  server_item *server;

  read_IPaddr(&IPaddr);
  end_message();
  end_com();
  if ((server = (server_item *)find_server(IPaddr)) == 0)
    return 1;
#ifdef VIEW
   fprintf(stderr,"--> %s remote is alive !!\n",server->server.hostname);
#endif
  server->server.host_status = UP;
  server->server.server_status = RUNNING;
  server->number_failed = 0;
  return 1;
}

/***********************/
  case I_AM_NEW:        
/***********************/
{
  /* A new server wants to check in */
  server_desc sv_desc;
  server_item *server;
  int number_of_pb;
  problem_desc *list_failed_pb;
  int j;
  problem_desc pb_desc;
  problem_item *problem;
  int nb_failed_pb=0;
  int restOK = 1;
 /* 
  * First, we check the problems of the newcomer.
  */

  read_string(sv_desc.hostname);
  read_IPaddr(&(sv_desc.IPaddr));
#ifdef VIEW
  fprintf(stderr,"Computational server on '%s' checking in\n",sv_desc.hostname);
#endif
  read_int(&(sv_desc.data_format));
  read_int(&(sv_desc.speed));
  read_int(&(sv_desc.workload));
  read_int(&(sv_desc.number_proc));
  read_int(&(sv_desc.mpp));
  sv_desc.host_status = UP;
  sv_desc.server_status = RUNNING;
  sv_desc.connection = TO_BE_A; /* Waiting a value */
  sv_desc.transfer   = TO_BE_A; /* Waiting a value */
  read_int(&number_of_pb);
  
  /* Remove the entry for the corresponding agent/server */
  if (number_of_pb != 0)
  {
    if ((server = (server_item *)find_server(sv_desc.IPaddr)) != 0)
    {
#ifdef DEBUG
      fprintf(stderr,"I have to remove %s\n",sv_desc.hostname);
#endif
      if (remove_server(server) == -1)
      {
        perror("proc_mes() remove_server()");
        end_message(); /* END OF A RECEIVE PHASE */
        return -1;
      }
    }
  }
  else
  {
    if ((server = (server_item *)find_agent(sv_desc.IPaddr)) != 0)
    {
#ifdef DEBUG
      fprintf(stderr,"I have to remove agent %s\n",sv_desc.hostname);
#endif
      if (remove_agent(server) == -1)
      {
        perror("proc_mes() remove_agent()");
        end_message(); /* END OF A RECEIVE PHASE */
        return -1;
      }
    }
  }

  /* Create a new entry for the corresponding agent/server */
  if (number_of_pb != 0)
  {
    if ((server =(server_item *)new_server(sv_desc)) == 0)
    {
      perror("proc_mes() new_server()");
      end_message(); /* END OF A RECEIVE PHASE */
      return -1;
    }
  }
  else
  {
    if ((server =(server_item *)new_agent(sv_desc)) == 0)
    {
      perror("proc_mes() new_agent()");
      end_message(); /* END OF A RECEIVE PHASE */
      return -1;
    }
  }

  /* Checking the problems one by one */
  list_failed_pb = (problem_desc *)calloc(number_of_pb,sizeof(problem_desc));
  for (j=0;j<number_of_pb;j++)
  {
    read_problem(&pb_desc);
    problem = (problem_item *)find_problem(pb_desc.code);
    if (problem != 0)  /* THE PROBLEM ALREADY EXISTS */
    {
#ifdef DEBUG
      fprintf(stderr,
        "received %s %d %d \n local  %d %d \n",
        pb_desc.code,
        pb_desc.nb_obj_input,pb_desc.nb_obj_output,
        (problem->problem).nb_obj_input,(problem->problem).nb_obj_output);
#endif
      if ((problem_cmp(pb_desc,problem->problem) != 1)||
          (problem_cmp_format(pb_desc,problem->problem) != 1))
      { /* The problem is wrong */ 
#ifdef VIEW
        fprintf(stderr,"\tThe problem %s is not correctly specified\n",pb_desc.code);
#endif
        BCOPY(&(problem->problem),&(list_failed_pb[nb_failed_pb]),
              sizeof(problem_desc));
        nb_failed_pb++;
      }
      add_server_to_pb(problem,server);
      add_pb_to_server(problem,server);
    }
    else  /* THE PROBLEM DOESN'T EXIST */
    {
      /* We create it */
#ifdef DEBUG
      fprintf(stderr,"The problem %s doesn't exist\n",pb_desc.code);
#endif
      /* We have to create the problem */
      problem = (problem_item *)new_problem(pb_desc);
      if (problem == 0)
      {
        perror("proc_mes() new_problem()");
        return -1;
      }
      add_pb_to_server(problem,server);
      add_server_to_pb(problem,server);
    }
  }
  if (list_failed_pb != NULL)
    free(list_failed_pb);

  end_message();
  
 /*
  * Then we construct the reply
  */
  if (nb_failed_pb == 0) 
  {
    /* Send all the structure */
    if (gen_all_struct() == -1)
    {
      perror("proc_mes() gen_all_struct()");
      return -1;
    }

#ifdef VIEW
    fprintf(stderr,"\tNew server has successfully checked in.\n");
#endif
    time(&current);
    sprintf(logbuf,"Server '%s' check_in, %s\n",
        sv_desc.hostname,ctime(&current));
    append_to_log(logbuf);
    if (number_of_pb != 0)
      distance_request(1,server,0);
  }
  else
  {
#ifdef VIEW
    fprintf(stderr,"New server has been refused.\n");
#endif

    time(&current);
    sprintf(logbuf,"Server '%s' refused, %s\n",
        sv_desc.hostname,ctime(&current));
    append_to_log(logbuf);
    remove_server(server);
    if (gen_pb_failed(list_failed_pb,nb_failed_pb) == -1)
    {
      perror("proc_mes() gen_pb_failed()");
      return -1;
    }
  }
  end_com();
  return 1; 
}
/***********************/
  case I_EXIST:       
/***********************/
{
  /* A new server has checked in */
  server_desc sv_desc;
  server_item *server;
  int number_of_pb;
  int j;
  problem_desc pb_desc;
  problem_item *problem;
  int delay;

  read_int(&delay); /* delay for the pong */
  read_string(sv_desc.hostname);
  read_IPaddr(&(sv_desc.IPaddr));
#ifdef VIEW
  fprintf(stderr,"A server on '%s' has cheked in to an agent\n",sv_desc.hostname);
#endif
  read_int(&(sv_desc.data_format));
  read_int(&(sv_desc.speed));
  read_int(&(sv_desc.workload));
  read_int(&(sv_desc.number_proc));
  read_int(&(sv_desc.mpp));
 /*
  * From here we deal with the message received
  */
  sv_desc.host_status = UP;
  sv_desc.server_status = RUNNING;
  sv_desc.connection = TO_BE_A; /* Waiting a value */
  sv_desc.transfer   = TO_BE_A; /* Waiting a value */
  read_int(&number_of_pb);
  /* We remove if it already exists */
  if (number_of_pb != 0)
  {
    if ((server = (server_item *)find_server(sv_desc.IPaddr)) != 0)
    {
      if (remove_server(server) == -1)
      {
        perror("proc_mes() remove_server()");
        end_message(); /* END of a receive Phase */
        return -1;
      }
    }
  }
  else
  {
    if ((server = (server_item *)find_agent(sv_desc.IPaddr)) != 0)
    {
      if (remove_agent(server) == -1)
      {
        perror("proc_mes() remove_agent()");
        end_message(); /* END of a receive Phase */
        return -1;
      }
    }
  }

  if (number_of_pb != 0)
  {
    if ((server =(server_item *)new_server(sv_desc)) == 0)
    {
      perror("proc_mes() new_server()");
      end_message(); /* END OF A RECEIVE PHASE */
      return -1;
    }
  }
  else
  {
    if ((server =(server_item *)new_agent(sv_desc)) == 0)
    {
      perror("proc_mes() new_agent()");
      end_message(); /* END OF A RECEIVE PHASE */
      return -1;
    }
  }
  if ((number_of_pb != 0)&&(my_self->number_problems == 0))
    distance_request(1,server,delay);
  for (j=0;j<number_of_pb;j++)
  {
    read_problem(&pb_desc);
    problem = (problem_item *)find_problem(pb_desc.code);
    if (problem == 0) /* THE PROBLEM DOESN'T EXIST YET */
    { 
      problem = (problem_item *)new_problem(pb_desc);
      if (problem == 0)
      {
        perror("proc_mes() new_problem()");
        end_message(); /* END OF A RECEIVE PHASE */
        return -1;
      }
      add_pb_to_server(problem,server); 
      add_server_to_pb(problem,server); 
    }
    else /* THE PROBLEM ALREADY EXISTS */
    {
      add_pb_to_server(problem,server);
      add_server_to_pb(problem,server);
      /* Skip the problem */
    }
  }
  end_message(); /* END OF A READ PHASE */
  end_com();
#ifdef VIEW
  fprintf(stderr,"\tServer on '%s added\n",sv_desc.hostname);
#endif
  time(&current);
  sprintf(logbuf,"\n+++Server on '%s' has checked in at %s\n",sv_desc.hostname,
      ctime(&current));
  append_to_log(logbuf);
  return 1; 
}
  /*******************/
  case WORKLOAD:
  /*******************/
{
  unsigned int IPaddr;
  server_item *server;
  int workload;

  read_IPaddr(&IPaddr);
  server = (server_item *)find_server(IPaddr);
  if (server == 0)
  { /* The server is unknown */
    struct hostent *hp;
    struct sockaddr_in server_INET_add;
    int sock,connection,message;

#ifdef VIEW
    fprintf(stderr,"Some unknown server contacted me !\n");
    fprintf(stderr,"\tTrying to find out more\n");
#endif
    end_message();
    end_com();
    /* Get info on the unknown server */
    sock = connect_to_server("unknown",IPaddr);
    if (sock < 0)
      return 1;
     
    init_send(DATA_XDR,sock);
    message = WHO_ARE_YOU;
    write_int(&message);
    end_message();
    end_com(); 
    return 1;
  }
  /* The server is known */

  /*
   * Here, modify the status !!!
   */

  if (server->server.host_status == DOWN)
    server->server.host_status = UP;
  if (server->server.server_status == STOPPED)
    server->server.server_status = RUNNING;

 /*
  * What to do for a FAILED server ????
  */

  read_int(&workload);
#ifdef VIEW
  fprintf(stderr,"Workload update from '%s' : %d\n",server->server.hostname,workload);
#endif

  end_message();
  server->server.workload = workload;
  end_com();
  return 1;
}
  /*******************/
  case I_DIE:
  /*******************/
{
  unsigned int IPaddr;
  server_item *server;

  read_IPaddr(&IPaddr);
  end_message();
  end_com();
  server = (server_item *)find_server(IPaddr);
  if (server == 0)
  {
    server = (server_item *)find_agent(IPaddr);
    if (server == 0) /* The server is unknown */
      break;
    else
    {
#ifdef DEBUG
      fprintf(stderr,"I remove %s\n",server->server.hostname);
#endif
      remove_agent(server);
      return 1;
    }
  }
  /* The server is known */
#ifdef VIEW
  fprintf(stderr,"Server on '%s' has checked out\n",server->server.hostname);
#endif
  time(&current);
  sprintf(logbuf,"Server on '%s' has checked out at %s",
        server->server.hostname,ctime(&current));
  append_to_log(logbuf);
  remove_server(server);
  return 1;
}
  /*******************/
  case POLL_CONTACT:
  /*******************/
{
  end_message();
  init_send(DATA_XDR,client_desc);
  write_string(my_arch);
  end_message();
  end_com();
  return 1;
} 
  /*******************/
  case POLL_LIST_PB:
  /*******************/
{
  int detail;

  read_int(&detail);
  end_message();
  fprintf(stderr,"LIST DETAIL = %d\n",detail);
  poll_gen_pb_list(encoding,detail);
  fprintf(stderr,"GERENATED\n");
  end_com();
  return 1;
}
  /*******************/
  case POLL_LIST_SV:
  /*******************/
{
  int detail;

  read_int(&detail);
  end_message();
  poll_gen_sv_list(encoding,detail);
  end_com();
  return 1;
}
  /*******************/
  case POLL_GET_PB:
  /*******************/
{
  poll_gen_get_pb();
  end_com();
  return 1;
}
  /*******************/
  case POLL_GET_SV:
  /*******************/
{
  poll_gen_get_sv();
  end_com();
  return 1;
}
  /*******************/
  case POLL_STAT:
  /*******************/
{
  end_message();
  poll_gen_stat();
  end_com();
  return 1;
}
  /*******************/
  case POLL_KILL:
  /*******************/
{
  end_message();
  end_com();
  bailout();
  term();
  return 1;
}
  /*******************/
  case JOB_COMPLETED:
  /*******************/
{
  unsigned int IPaddr;
  server_item *server;
  int type;

  read_IPaddr(&IPaddr);
  end_message();
  end_com();

  if ((server = (server_item *)find_server(IPaddr)) == 0)
    return 1;
#ifdef DEBUG
  fprintf(stderr,"I received a job completion for %s\n",
          server->server.hostname);
#endif
  time(&current);
  sprintf(logbuf,"**Job completion from %s at %s",
          server->server.hostname,ctime(&current));
  append_to_log(logbuf);

  type = server->server.server_status;
  if (type == RUNNING)
  {
#ifdef DEBUG
    fprintf(stderr,"I decrease the workload of %s by 100 if possible\n",
      server->server.hostname);
#endif
    if ((server->server.workload) > 100)
      (server->server.workload) -= 100;
  }
  return 1;
}

/*****************/
  }  /* END CASE */
/*****************/
  return -1;
}

/*
 * gen_all_struct() : Pack all my structure 
 */

#ifdef __STDC__
int gen_all_struct()
#else
int gen_all_struct()
#endif
{
  int number_of_pb=0;
  int number_of_sv=0;
  int type=YOU_ARE_NEW;
  problem_item *problem;
  server_item *server;
  pb_list *current;
  int j;

  number_of_pb = 0;
  for (j=0;j<HASH_PB;j++)
  {
    problem = pb_hash_table[j];
    while(problem != 0)
    {
      number_of_pb++;
      problem=problem->next;
    }
  }
  /* Then the servers */
  number_of_sv = 0;
  for (j=0;j<HASH_SV;j++)
  {
    server = sv_hash_table[j];
    while(server != 0)
    {
        number_of_sv++;
        server=server->next;
    }
  }
  for (j=0;j<HASH_SV;j++)
  {
    server = ag_hash_table[j];
    while(server != 0)
    {
        number_of_sv++;
        server=server->next;
    }
  }
  /* I actually send the data */
  /* BEGINING OF A SEND PHASE */
  /* XDR by default */
  init_send(DATA_XDR,client_desc);
  write_int(&type);
  write_int(&number_of_pb);
  for (j=0;j<HASH_PB;j++)
  {
    problem = pb_hash_table[j];
    while(problem != 0)
    {
      write_problem(problem->problem);
      problem=problem->next;
    }
  }
  write_int(&number_of_sv);
  for (j=0;j<HASH_SV;j++)
  {
    server = sv_hash_table[j];
    while(server != 0)
    {
      write_string((server->server).hostname);
      write_IPaddr(&((server->server).IPaddr));
      write_int(&((server->server).data_format));
      write_int(&((server->server).speed));
      write_int(&((server->server).host_status));
      write_int(&((server->server).server_status));
      write_int(&((server->server).workload));
      write_int(&((server->server).number_proc));
      write_int(&((server->server).mpp));
      write_int(&(server->number_problems));

      current=server->first_pb;
      while(current != 0)
      {
        write_string(((current->problem)->problem).code);
        current=current->next;
      }
      server=server->next;
    }
  }
  for (j=0;j<HASH_SV;j++)
  {
    server = ag_hash_table[j];
    while(server != 0)
    {
      write_string((server->server).hostname);
      write_IPaddr(&((server->server).IPaddr));
      write_int(&((server->server).data_format));
      write_int(&((server->server).speed));
      write_int(&((server->server).host_status));
      write_int(&((server->server).server_status));
      write_int(&((server->server).workload));
      write_int(&((server->server).number_proc));
      write_int(&((server->server).mpp));
      write_int(&(server->number_problems));

      current=server->first_pb;
      while(current != 0)
      {
        write_string(((current->problem)->problem).code);
        current=current->next;
      }
      server=server->next;
    }
  }
  /* END OF A SEND PHASE */
  end_message();
  return 1;
}

/*
 * gen_pb_failure() : The problem of the newcomer is WRONG
 */

#ifdef __STDC__
int gen_pb_failed(problem_desc *list_failed_pb,int nb_failed_pb)
#else
int gen_pb_failed(list_failed_pb,nb_failed_pb)
problem_desc *list_failed_pb;
int nb_failed_pb;
#endif
{
  int type=WRONG_PROBLEM;
  int j;

  /* BEGINNING OF A SEND PHASE */
  /* XDR by default */
  init_send(DATA_XDR,client_desc);
  write_int(&type);
  write_int(&nb_failed_pb);
  for (j=0;j<nb_failed_pb;j++)
    write_problem(list_failed_pb[j]);
  /* END OF A SEND PHASE */
  end_message();
  return 1;
}
  

/*
 * gen_pb_refused() : The problem doesn't correspond to the specification
 *                    I can't solve it...
 */

#ifdef __STDC_
int gen_pb_refused(int encoding)
#else
int gen_pb_refused(encoding)
int encoding;
#endif
{
  int message_type;

  message_type = PB_REFUSED;
  init_send(encoding,client_desc);
  write_int(&message_type);
  /* END OF A SEND PHASE */
  end_message();
  return 1;
}

/*
 * server_compare() : For the qsort
 */
#ifndef PLAT_HPPA
#ifdef __STDC__
int server_compare(const void *x,const void *y)
#else
int server_compare(x,y)
const void *x;
const void *y;
#endif
#else
#ifdef __STDC__
int server_compare(void *x,void *y)
#else
int server_compare(x,y)
void *x;
void *y;
#endif
#endif
{
  if (((server_score *)x)->score > ((server_score *)y)->score)
    return 1;
  if (((server_score *)x)->score < ((server_score *)y)->score)
    return -1;
  return 0;
}

/*
 * gen_rep_pb_req() : Generates the reply message
 */

#ifdef __STDC__
int gen_rep_pb_req(int encoding, int status,server_item **server_list,
      int len_list,int problem_size,int data_size,int result_size,problem_item *problem)
#else
int gen_rep_pb_req(encoding,status,server_list,len_list,
                   problem_size,data_size,result_size,problem)
int encoding;
int status;
server_item **server_list;
int len_list;
int problem_size;
int data_size;
int result_size;
problem_item *problem;
#endif
{
  int i,j;
  int list_length; 	      /* The real length processed */
  server_item *server=NULL;
  double asses_score();
  int server_compare();
  server_score *to_sort;
  int len_list_cleaned;
  int first;

 /*
  * It's a error reply message
  */
  if (status<0)
  {
    init_send(encoding,client_desc);
    write_int(&status);
    end_message();
    return 1;
  }
 /*
  * It a succesfull reply message
  */
  
  /* 
   * I take only the 'good' servers
   */

  len_list_cleaned = len_list;
  for (j=0;j<len_list;j++)
  {
    server_item *server = (server_item *)server_list[j];
    if (((server->server).host_status != UP) ||
        (((server->server).server_status != RUNNING) && 
        ((server->server).server_status != FAILED)))
     len_list_cleaned--; 
  }

  to_sort = (server_score *)calloc(len_list_cleaned,sizeof(server_score));

  /* I compute the score */
  i = 0;
  for (j=0;j<len_list;j++)
  {
    server_item *server = (server_item *)server_list[j];
    if (((server->server).host_status != UP) ||
        (((server->server).server_status != RUNNING) &&
        ((server->server).server_status != FAILED)))
    {
    }
    else
    {
    to_sort[i].score=(double)asses_score(server_list[j],
                      problem_size,data_size,result_size,problem);
    to_sort[i].index = j;
    i++;
    }
  }
  /* I sort */
  qsort(to_sort,len_list_cleaned,sizeof(server_score),server_compare);

  /* Now the len_list servers are sorted */

  if (len_list_cleaned > LEN_RETURN_LIST)
    list_length = LEN_RETURN_LIST;
  else
    list_length = len_list_cleaned;

#ifdef VIEW
  fprintf(stderr,"\tI send back %d servers to try\n",list_length);
#endif

  init_send(encoding,client_desc);
  write_int(&status);
  write_int(&list_length);
  first = 0;
  for (i=0;i<list_length;i++)
  {
    int tmp;

    server = (server_item *)server_list[to_sort[i].index];
    if (((server->server).host_status == UP)&&
        (((server->server).server_status == RUNNING) ||
        ((server->server).server_status == FAILED)))
    {
      write_string((server->server).hostname);
      write_IPaddr(&((server->server).IPaddr));
      write_int(&((server->server).data_format));
      tmp = (int)(to_sort[i].score);
      write_int(&tmp);
      server->dirty = 1;
      if (i==first)
      {
        ((server->server).workload)+= 100;
      }
    }
    else /* Skip the down/stopped servers */
      first++;
  }
  /* END OF A SEND PHASE */
  end_message();
  if (to_sort != NULL)
    free(to_sort);
  return 1;
}

/*
 *  asses_score()
 *
 *  Compute the estimated time to perform the solve.
 */

#ifdef __STDC__
double asses_score(server_item *server,int problem_size,int data_size,
   int result_size,problem_item *problem)
#else
double asses_score(server,problem_size,data_size,result_size,problem)
server_item *server;
int problem_size;
int data_size;
int result_size;
problem_item *problem;
#endif
{
/* This procedure should be smart */
  double send;
  double recv;
  double result;

  double nb_flops;
  double speed_eff;
  double communication;
  double computation;
  double eval_flops();

  if ((server->server.host_status == DOWN)||
      (server->server.server_status == STOPPED))
    return -1;

/* Compute an estimate of the number of flops */
  nb_flops = (double)eval_flops(problem->problem,problem_size);
/* Compute an estimate of the actual Kflops of the machine */
  if (server->server.workload< (100*(server->server.number_proc -1)))
    speed_eff = (double)server->server.speed;
  else
    speed_eff = ((double)server->server.speed*100.*
                  (double)(server->server.number_proc))/
                (100.*(double)(server->server.number_proc)
                +(double)server->server.workload - 100.*
                 (double)((server->server.number_proc)-1));
/* compute an estimate of the time required to send the data*/
  send = ((double)server->server.connection/1000000.)+
         (double)data_size/(double)server->server.transfer;
/* Compute an estimate of the time required to recv the result */
  if (result_size == -1)
    recv = 0.;
  else
    recv = (double)result_size/(double)server->server.transfer;

  communication = send+recv;
  computation   = (double)nb_flops/((double)speed_eff*1000.);
  result = communication+computation;

  if (server->server.transfer == TO_BE_A)
  {
#ifdef DEBUG
    fprintf(stderr,"Network data on %s not available yet\n",server->server.hostname);
#endif
    result = 1000000.00;
  }
#ifdef VIEW
  fprintf(stderr,"\t  '%s' : %.5f+%.5f=%.5f)\n",
         server->server.hostname,communication,computation,result);
#endif
  return(result);
}

/*
 * Are_you_alive ?
 * 
 */
#ifdef __STDC__
int are_you_alive_request(unsigned int IPaddr)
#else
int are_you_alive_request(IPaddr)
unsigned int IPaddr;
#endif
{
  int pid;
  int answer;
  struct sockaddr_in server_add;
  int type;
  int sock;
  struct sockaddr_in my_server_add; /* Server INET address */

  /* Son */
  sock = connect_to_server("unknown",IPaddr);
  if (sock == -3)
  {
    perror("are_you_alive() socket()\n");
    exit(0);
  }
  if (sock < 0) 
    exit(0);
  
  type = ARE_YOU_ALIVE;

  init_send(DATA_XDR,sock);
  write_int(&type);
  end_message(); 
  init_recv(sock);
  read_int(&answer);
  end_message();
  end_com();
  if (answer != ALIVE) 
    exit(0);
  
  /* I report the ALIVE to the local agent */

   sock = connect_to_agent(my_host,my_address);
   if (sock == -3)
     exit(0);
   if (sock == -2)
   {
     fprintf(stderr,"Impossible to contact host %s\n",my_host);
     exit(0);
   }
   if (sock == -1)
   {
     fprintf(stderr,"No NetSolve agent running ??\n");
     exit(0); 
   } 
  type = IS_ALIVE;

  init_send(DATA_XDR,sock);
  write_int(&type);
  write_IPaddr(&IPaddr);
  end_message();
  end_com();
  exit(0);
}
