/*****************************************************************/
/*      server_util.c                                            */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  agent_init()                                                 */
/*  contact_master()                                             */
/*  print_cache()                                                */
/*  eval_flops()                                                 */
/*****************************************************************/

#include "server.h"
#include "agent_prot.h"
#include "shared_prot.h"

#ifdef __STDC__
int agent_init()
#else
int agent_init()
#endif
{
  server_desc descriptor;
  FILE *stream;
  int i;

  number_computations = 0;
  for (i=0;i<MAX_COMP;i++)
    computation_pids[i] = 0;

  strcpy(descriptor.hostname,my_host);

  descriptor.IPaddr = my_address;
  my_data_format = data_format();
  descriptor.data_format = my_data_format;

  descriptor.speed = 0;
  descriptor.workload = 0;
  descriptor.number_proc = number_processors;
  descriptor.mpp = mpp;
  descriptor.connection = 0;
  descriptor.transfer = 1000000;
  descriptor.host_status   = UP;
  descriptor.server_status = RUNNING;
#ifdef DEBUG
  fprintf(stderr,"my address is %u.%u.%u.%u\n",BYTE(&my_address,0),
                                       BYTE(&my_address,1),
                                       BYTE(&my_address,2),
                                       BYTE(&my_address,3));
#endif
  if ((my_self = (server_item *)new_agent(descriptor)) == 0)
  {
    perror("init_server() new_agent()");
    return -1;
  }
  nb_pb_solved = 0;
  nb_lcl_request = 0;
  time(&up_date);
  strcpy(char_up_date,ctime(&up_date));
  return 1;
}

/*
 *  contact_master()
 *
 *   contact the master server
 */

#ifdef __STDC__
int contact_master(char *name)
#else
int contact_master(name)
char *name;
#endif
{
int sock;
struct sockaddr_in server_INET_add; /* Server INET socket address  */
struct hostent *hp;                 /* host descriptor             */
int connection;
pb_list *list;                /* List of problems                */
problem_desc *problem;        /* pointer to a problem descriptor */
int type;                     /* Message type                    */
int nb_problems=0;           
int status;                   /* Reply status                    */
int nb_failed_pb;             /* Nb of failed problems           */
problem_desc pb_desc;         /* Problem descriptor              */
server_item *server;
int i,j;
char master_host[256];
unsigned int master_address;

  if ((hp = gethostbyname(name)) == NULL)
  {
    fprintf(stderr,"Cannot resolve hostname : '%s'\n",name);
    return -1;
  }
  strcpy(master_host,hp->h_name);
  BCOPY(hp->h_addr_list[0],&master_address,sizeof(unsigned int));


#ifdef VIEW
  fprintf(stderr,"contacting '%s' ...\n",master_host);
#endif
  sock = connect_to_agent(master_host,master_address);
  if (sock == -3)
    return -1;
  if (sock == -2)
  {
      fprintf(stderr,"Impossible to locate host %s\n",master_host);
     return -1;
  }
  if (sock == -1)
  {
      fprintf(stderr,"Impossible to contact agent\n");
      return -1;
  }

  nb_problems = my_self->number_problems;
  list = my_self->first_pb;
  type = I_AM_NEW;
  init_send(DATA_XDR,sock);
  write_int(&type);
  write_string(my_host);
  write_IPaddr(&((my_self->server).IPaddr));
  write_int(&((my_self->server).data_format));
  write_int(&((my_self->server).speed));
  write_int(&((my_self->server).workload));
  write_int(&((my_self->server).number_proc));
  write_int(&((my_self->server).mpp));
  write_int(&nb_problems);
  list = my_self->first_pb;
  for (j=0;j<nb_problems;j++)
  {
    problem = &((list->problem)->problem);
    write_problem(*problem);
    list = list->next;
  }
  end_message();

  /* Get the answer */
  /* BEGINNING OF A RECEIVE PHASE */
  init_recv(sock);  
  read_int(&status);
  if (status == WRONG_PROBLEM) /* Report errors and exit */
  {
    read_int(&nb_failed_pb);
    fprintf(stderr,"----------------------------\n");
    for (j=0;j<nb_failed_pb;j++)
    {
      int i;

      read_problem(&pb_desc);
      fprintf(stderr,"Problem %s badly specified - Should be :\n",pb_desc.code);
      fprintf(stderr,"Nb of objects in input : %d\n",pb_desc.nb_obj_input);
      fprintf(stderr,"Nb of objects in output : %d\n",pb_desc.nb_obj_output);
      fprintf(stderr,"Complexity : %s\n",pb_desc.complexity);
      fprintf(stderr,"Input :\n");
      for (i=0;i< pb_desc.nb_obj_input;i++)
      {
        switch (pb_desc.input_obj[i].data_type)
        {
     
          case NETSOLVE_ASCII :
          {
             fprintf(stderr," ASCII ");
             break;
          }
          case NETSOLVE_CHAR :
          {
             fprintf(stderr," Character ");
             break;
          }
          case NETSOLVE_I :
          {
             fprintf(stderr," Integer ");
             break;
          }
          case NETSOLVE_B :
          {
            fprintf(stderr," Byte ");
            break;
          }
          case NETSOLVE_S :
          {
             fprintf(stderr," Single precision real ");
             break;
          }
          case NETSOLVE_D :
          {
             fprintf(stderr," Double precision real ");
             break;
          }
          case NETSOLVE_C :
          {
             fprintf(stderr," Single precision complex ");
             break;
          }
          case NETSOLVE_Z :
          {
             fprintf(stderr," Double precision complex ");
             break;
          }
          case NETSOLVE_EXTERNAL :
          {
             fprintf(stderr," ");
             break;
          }
          default :
          {
            fprintf(stderr,"Unknown data type ???");
            return -1;
          }
        }
        switch (pb_desc.input_obj[i].object_type)
        {

          case NETSOLVE_FILE :
          {
             fprintf(stderr," FILE \n");
             break;
          }
          case NETSOLVE_MATRIX :
          {
             fprintf(stderr," Matrix \n");
             break;
          }
          case NETSOLVE_VECTOR :
          {
             fprintf(stderr," Vector \n");
             break;
          }
          case NETSOLVE_UPF :
          {
             fprintf(stderr," User Provided Function \n");
             break;
          }
          case NETSOLVE_SCALAR :
          {
             fprintf(stderr," Scalar \n");
             break;
          }
          default :
          {
            fprintf(stderr,"Unknown object type ???");
            return -1;
          }
        }
      }
      fprintf(stderr,"Output : \n");
      for (i=0;i< pb_desc.nb_obj_output;i++)
      {
        switch (pb_desc.output_obj[i].data_type)
        {

          case NETSOLVE_ASCII :
          {
             fprintf(stderr," ASCII ");
             break;
          }
          case NETSOLVE_CHAR :
          {
             fprintf(stderr," Character ");
             break;
          }
          case NETSOLVE_B :
          {
            fprintf(stderr," Byte ");
            break;
          }
          case NETSOLVE_I :
          {
             fprintf(stderr," Integer ");
             break;
          }
          case NETSOLVE_S :
          {
             fprintf(stderr," Single precision real ");
             break;
          }
          case NETSOLVE_D :
          {
             fprintf(stderr," Double precision real ");
             break;
          }
          case NETSOLVE_C :
          {
             fprintf(stderr," Single precision complex ");
             break;
          }
          case NETSOLVE_Z :
          {
             fprintf(stderr," Double precision complex ");
             break;
          }
          default :
          {
            fprintf(stderr,"Unknown data type ???");
            return -1;
          }
        }
        switch (pb_desc.output_obj[i].object_type)
        {

          case NETSOLVE_FILE :
          {
             fprintf(stderr," File \n");
             break;
          }
          case NETSOLVE_MATRIX :
          {
             fprintf(stderr," Matrix \n");
             break;
          }
          case NETSOLVE_VECTOR :
          {
             fprintf(stderr," Vector \n");
             break;
          }
          case NETSOLVE_SCALAR :
          {
             fprintf(stderr," Scalar \n");
             break;
          }
          default :
          {
            fprintf(stderr,"Unknown object type ???");
            return -1;
          }
        }
      }
      fprintf(stderr,"Formats : %d\n",pb_desc.nb_formats);
      for (i=0;i<pb_desc.nb_formats;i++)
      {
        int j;
        
        fprintf(stderr,"format #%d - %d argument\n",i,pb_desc.formats[i].nb_args);
        for (j=0;j<pb_desc.formats[i].nb_args;j++)
        {
          fprintf(stderr,"--> %s\n",pb_desc.formats[i].args_desc[j]);
        }
      }
     
      fprintf(stderr,"----------------------------\n");
    } 
    end_message();
    end_com();
    return -1;
  }
  if (status == SV_FAILURE) /* The master has failed !!! */
  {
    fprintf(stderr,"The agent %s failed\n",master_host);
    fprintf(stderr,"Sorry, impossible to check in\n");
    end_message();
    end_com();
    return -1;
  }
  /* From now, status = YOU_ARE_NEW */
#ifdef VIEW
  fprintf(stderr,"%s accepted my application\n",master_host);
#endif
  if (get_all_struct() == -1)
  {
    perror("contact_master() get_all_struct()");
    end_com();
    return -1;
  }
  /* I care about the distance only if I am an agent */
  if (my_self->number_problems == 0)
    distance_request(0,(server_item *)0,0);

  end_message();
  end_com();

/*****************************************/
/*      I broadcast my existence !!! WOW */
/*****************************************/
  if (1)
  { 
    time_t date;
    int rest_delay=5;

    type = I_EXIST;
    for (i=0;i<HASH_SV;i++)
    {
      server = sv_hash_table[i];
      while(server != 0)
      {
        if (1)
        {
#ifdef VIEW
          fprintf(stderr,"I report my existence to %s\n",server->server.hostname);
#endif
          time(&date);
          sock = connect_to_agent(server->server.hostname,
                           server->server.IPaddr);
          if (sock == -3)
            return -1;
          if (sock < 0)
          {
#ifdef VIEW
            fprintf(stderr,"WARNING :Server %s unreachable\n",server->server.hostname);
#endif
            server->server.server_status = STOPPED;
            server->server.host_status = DOWN;
            server=server->next;
            continue;
          }
          /* XDR or RAW */
          if (my_data_format == (server->server.data_format))
            init_send(DATA_RAW,sock);
          else
            init_send(DATA_XDR,sock);
          write_int(&type);
	  write_int(&rest_delay);
          rest_delay+= REST_DELAY_INCR;
          write_string(my_host);
          write_IPaddr(&((my_self->server).IPaddr));
          write_int(&((my_self->server).data_format));
          write_int(&((my_self->server).speed));
          write_int(&((my_self->server).workload));
          write_int(&((my_self->server).number_proc));
          write_int(&((my_self->server).mpp));
          write_int(&nb_problems);
          list = my_self->first_pb;
          for (j=0;j<nb_problems;j++)
          {
            problem = &((list->problem)->problem);
            write_problem(*problem);
            list = list->next;
          }
          end_message();
          end_com();
        }
        server = server->next;
      }
    }
    /* Now the agents */
    for (i=0;i<HASH_SV;i++)
    {

      server = ag_hash_table[i];
      while(server != 0)
      {
        if ((server->server.IPaddr != my_address)&&
            (server->server.IPaddr != master_address))
        {
#ifdef VIEW
          fprintf(stderr,"I report my existence to %s\n",server->server.hostname);
#endif
          time(&date);
          sock = connect_to_agent(server->server.hostname,
                                   server->server.IPaddr);
          if (sock == -3)
            return -1;

          if (sock < 0)
          {
#ifdef VIEW
            fprintf(stderr,"WARNING :Server %s unreachable\n",server->server.hostname);
#endif
            server->server.server_status = STOPPED;
            server->server.host_status = DOWN;
            server=server->next;
            continue;
          }
          /* XDR or RAW */
          if (my_data_format == (server->server.data_format))
            init_send(DATA_RAW,sock);
          else
            init_send(DATA_XDR,sock);
          write_int(&type);
          write_int(&rest_delay);
          rest_delay+= REST_DELAY_INCR;
          write_string(my_host);
          write_IPaddr(&((my_self->server).IPaddr));
          write_int(&((my_self->server).data_format));
          write_int(&((my_self->server).speed));
          write_int(&((my_self->server).workload));
          write_int(&((my_self->server).number_proc));
          write_int(&((my_self->server).mpp));
          write_int(&nb_problems);
          list = my_self->first_pb;
          for (j=0;j<nb_problems;j++)
          {
            problem = &((list->problem)->problem);
            write_problem(*problem);
            list = list->next;
          }
          end_message();
          end_com();
        }
        server = server->next;
      }
    }
  }
  return 1;
}

/*
 * get_all_struct() : get all the data structure from the master
 */

#ifdef __STDC__
int get_all_struct()
#else
int get_all_struct()
#endif
{
  int number_of_pb;
  int number_of_sv;

  problem_desc *pb_desc;
  server_desc sv_desc;
  server_item *server;
  problem_item *problem;

  int nb_of_problems;
  char code[CODE_LENGTH];

  int i,j;

  pb_desc = (problem_desc *)calloc(1,sizeof(problem_desc));
  read_int(&number_of_pb);
  for (j=0;j<number_of_pb;j++)
  {
    read_problem(pb_desc);
    if ((problem_item *)find_problem(pb_desc->code) == NULL)
    {
      if ((problem_item *)new_problem(*pb_desc) == NULL)
      {
        perror("get_all_struct() new_problem()");
        return -1;
      }
    }
    else
    {
      free_pb_desc(pb_desc);
    }
  }
  read_int(&number_of_sv);
  for (j=0;j<number_of_sv;j++)
  {
    read_string(sv_desc.hostname);
    read_IPaddr(&(sv_desc.IPaddr));
    read_int(&(sv_desc.data_format));
    read_int(&(sv_desc.speed));
    read_int(&(sv_desc.host_status));
    read_int(&(sv_desc.server_status));
    read_int(&(sv_desc.workload));
    read_int(&(sv_desc.number_proc));
    read_int(&(sv_desc.mpp));
    sv_desc.connection = TO_BE_A; /* Waiting a value */
    sv_desc.transfer   = TO_BE_A; /* Waiting a value */
    read_int(&nb_of_problems);
    if ((strcmp(sv_desc.hostname,my_host) == 0)&&
        (nb_of_problems == my_self->number_problems))
    {
        /* It is myself !!! */
    }
    else
    {
      if (nb_of_problems != 0)
      {
        if ((server =(server_item *)new_server(sv_desc)) == NULL)
        {
          perror("get_all_struct() new_server()");
          return -1;
        }
      }
      else
      {
        if ((server =(server_item *)new_agent(sv_desc)) == NULL)
        {
          perror("get_all_struct() new_agent()");
          return -1;
        }
      }
    }
    for (i=0;i<nb_of_problems;i++)
    {
      read_string(code);
      if ((strcmp(sv_desc.hostname,my_host) == 0)&&
         (nb_of_problems == my_self->number_problems))
      {
         /* It's myself */
      }
      else
      { 
        problem =(problem_item *)find_problem(code);
        add_pb_to_server(problem,server);
        add_server_to_pb(problem,server);
      }
    }
  }
  free(pb_desc);
  return 1;
}

/*
 * print_cache() : For debugging purposes
 */ 

#ifdef __SDTC__
print_cache(int detail)
#else
print_cache(detail)
int detail;
#endif
{
  int i;
  server_item *server;
  problem_item *problem;
  pb_list *current;
  sv_list *cur;

  fprintf(stderr,"--------------------------------------------\n");
  fprintf(stderr,"SERVERS\n");
  for (i=0;i<HASH_SV;i++)
  {
    server = sv_hash_table[i];
    while(server!=0)
    {
      fprintf(stderr,"Server : %s\n",server->server.hostname);
      fprintf(stderr,"\t IPaddr =%u.%u.%u.%u d_format = %d, spd = %d, sts = %d ser_sts = %d conn = %d transf = %d wload = %d number_proc : %d MPP : %d\n",
             BYTE(&(server->server.IPaddr),0),
             BYTE(&(server->server.IPaddr),1),
             BYTE(&(server->server.IPaddr),2),
             BYTE(&(server->server.IPaddr),3),
             server->server.data_format,
             server->server.speed,
             server->server.host_status,
             server->server.server_status,
             server->server.connection,
             server->server.transfer,
             server->server.workload,
             server->server.number_proc,
             server->server.mpp);
      fprintf(stderr,"\t NUMBER OF PROBLEMS : %d\n",server->number_problems);
      fprintf(stderr,"\t problems :");
      current = server->first_pb;
      while(current!=0)
      {
        fprintf(stderr,"|%s|",current->problem->problem.code);
        current=current->next;
      }
      fprintf(stderr,"\n");
      server=server->next;
    }
  }
  fprintf(stderr,"AGENTS\n");
  for (i=0;i<HASH_SV;i++)
  {
    server = ag_hash_table[i];
    while(server!=0)
    {
      fprintf(stderr,"Server : %s\n",server->server.hostname);
      fprintf(stderr,"\t IPaddr =%u.%u.%u.%u d_format = %d, spd = %d, sts = %d ser_sts = %d conn = %d transf = %d wload = %d number_proc : %d MPP : %d\n",
             BYTE(&(server->server.IPaddr),0),
             BYTE(&(server->server.IPaddr),1),
             BYTE(&(server->server.IPaddr),2),
             BYTE(&(server->server.IPaddr),3),
             server->server.data_format,
             server->server.speed,
             server->server.host_status,
             server->server.server_status,
             server->server.connection,
             server->server.transfer,
             server->server.workload,
             server->server.number_proc,
             server->server.mpp);
      fprintf(stderr,"\t NUMBER OF PROBLEMS : %d\n",server->number_problems);
      fprintf(stderr,"\t problems :");
      current = server->first_pb;
      while(current!=0)
      {
        fprintf(stderr,"|%s|",current->problem->problem.code);
        current=current->next;
      }
      fprintf(stderr,"\n");
      server=server->next;
    }
  }


  if(detail == 0)
    return 1;
#if 0
  for (i=0;i<HASH_PB;i++)
  {
    problem = pb_hash_table[i];
    while(problem!=0)
    {
      int j;

      fprintf(stderr,"Problem : %s\n",problem->problem.code);
      fprintf(stderr,"\t input  %d %d compl : %s\n",
             problem->problem.nb_obj_input,
             problem->problem.nb_obj_output,
             problem->problem.complexity);
      fprintf(stderr,"Input :\n");
      for (j=0;j< problem->problem.nb_obj_input;j++)
      {
        switch (problem->problem.input_obj[j].data_type)
        {
          case NETSOLVE_ASCII :
          {
             fprintf(stderr," ASCII ");
             break;
          }
          case NETSOLVE_CHAR :
          {
             fprintf(stderr," Character ");
             break;
          }
          case NETSOLVE_I :
          {
             fprintf(stderr," Integer ");
             break;
          }
          case NETSOLVE_S :
          {
             fprintf(stderr," Single precision real ");
             break;
          }
          case NETSOLVE_D :
          {
             fprintf(stderr," Double precision real ");
             break;
          }
          case NETSOLVE_C :
          {
             fprintf(stderr," Single precision complex ");
             break;
          }
          case NETSOLVE_Z :
          {
             fprintf(stderr," Double precision complex ");
             break;
          }
          default :
          {
            fprintf(stderr,"Unknown data type ???");
            return -1;
          }
        }
        switch (problem->problem.input_obj[j].object_type)
        {

          case NETSOLVE_FILE :
          {
             fprintf(stderr," File \n");
             break;
          }
          case NETSOLVE_MATRIX :
          {
             fprintf(stderr," Matrix \n");
             break;
          }
          case NETSOLVE_VECTOR :
          {
             fprintf(stderr," Vector \n");
             break;
          }
          case NETSOLVE_SCALAR :
          {
             fprintf(stderr," Scalar \n");
             break;
          }
          default :
          {
            fprintf(stderr,"Unknown object type ???");
            return -1;
          }
        }
      }
      fprintf(stderr,"Output : \n");
      for (j=0;j< problem->problem.nb_obj_output;j++)
      {
        switch (problem->problem.output_obj[j].data_type)
        {

          case NETSOLVE_ASCII :
          {
             fprintf(stderr," ASCII ");
             break;
          }
          case NETSOLVE_CHAR :
          {
             fprintf(stderr," Character ");
             break;
          }
          case NETSOLVE_I :
          {
             fprintf(stderr," Integer ");
             break;
          }
          case NETSOLVE_S :
          {
             fprintf(stderr," Single precision real ");
             break;
          }
          case NETSOLVE_D :
          {
             fprintf(stderr," Double precision real ");
             break;
          }
          case NETSOLVE_C :
          {
             fprintf(stderr," Single precision complex ");
             break;
          }
          case NETSOLVE_Z :
          {
             fprintf(stderr," Double precision complex ");
             break;
          }
          default :
          {
            fprintf(stderr,"Unknown data type ???");
            return -1;
          }
        }
        switch (problem->problem.output_obj[j].object_type)
        {

          case NETSOLVE_FILE :
          {
             fprintf(stderr," File \n");
             break;
          }
          case NETSOLVE_MATRIX :
          {
             fprintf(stderr," Matrix \n");
             break;
          }
          case NETSOLVE_VECTOR :
          {
             fprintf(stderr," Vector \n");
             break;
          }
          case NETSOLVE_SCALAR :
          {
             fprintf(stderr," Scalar \n");
             break;
          }
          default :
          {
            fprintf(stderr,"Unknown object type ???");
            return -1;
          }
        }
      }
      fprintf(stderr,"\t NUMBER OF SERVERS : %d\n",problem->number_servers);
      cur = problem->first_server;
      while(cur!=0)
      {
        fprintf(stderr,"\t \t -->  %s \n",cur->server->server.hostname);
        cur=cur->next;
      }
      problem=problem->next;
    }
  }
#endif
  fprintf(stderr,"--------------------------------------------\n");
}

/*
 * eval_flops()
 */

#ifdef __STDC__
double eval_flops(problem_desc pb_desc,int size)
#else
double eval_flops(pb_desc,size)
problem_desc pb_desc;
int size;
#endif
{
  char *complexity=(char *)strdup(pb_desc.complexity);
  char *token1,*token2;
  double result;
  
  token1 = (char *)strtok(complexity," ,\t");
  token2 = (char *)strtok(NULL," ,\t");

  result = (double)(atoi(token1)*power(size,atoi(token2)));
  free(complexity);
  return result;
}
