/*****************************************************************/
/*      pong.c                                                   */
/*      Version 1.0                      Albert Cohen            */
/*---------------------------------------------------------------*/
/*  latency()                                                    */
/*  bandwidth()                                                  */
/*  main()                                                       */
/*---------------------------------------------------------------*/
/* Revision 1 : Henri Casanova  Jul 13th 1994                    */
/*****************************************************************/

/*
 * pong : get network measures
 */

#define TIMEOUT 400

#include <sys/time.h>
#include <time.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include "bfunc.h"
#include "global.h"

#define UNREACHABLE_HOST -1

#define PORT 7
#define MAXLENGTH 65536
#define LENGTH 1024

int local; /* Local socket */
int size;

#ifdef __STDC__
void panic()
#else
void panic() 
#endif
{
  fprintf(stdout,"-1\n-1\n");
  exit(0);
}

/*-----------------------------------------------------------------------------
  conection time in micro-seconds
  */
#ifdef __STDC__
int latency(char *host_name,unsigned int host_address)
#else
int latency(host_name,host_address)
char *host_name;
unsigned int host_address;
#endif
{
  struct sockaddr_in host;	/*Destination IP address */
  int s, i;
  int host_length; 		/*Length of this address*/
  char *inbuf;			/*incoming messages buffer*/
  char *outbuf;			/*outgoing messages buffer*/
  struct hostent *hp;

  /*for timing*/
  struct timeval tv[2];
  int dt = 0, maxdt = 0;           

  inbuf=(char *)calloc(MAXLENGTH,sizeof(char)); /*buffer Allocation*/
  outbuf=(char *)calloc(MAXLENGTH,sizeof(char));

  /* MAIN LOOP */
  for (i=0;i<LATENCY_TIMES;i++)
  {
    local = socket(AF_INET,SOCK_STREAM,0); 
    BZERO((char *)&host,sizeof(host));
    host.sin_family = AF_INET;
    host.sin_port = htons(PORT);
    hp = gethostbyaddr((char *)&host_address,sizeof(unsigned int),AF_INET);
    if (hp == NULL)
    {
      hp = gethostbyname((char *)host_name);
      if (hp == NULL)
      {
        close(local);
        return UNREACHABLE_HOST;
      }
    }
    BCOPY(hp->h_addr,&host.sin_addr,hp->h_length);
    host_length = sizeof(host);
    gettimeofday(&tv[0], (struct timezone*)0);
    /* Connection */
    if(connect(local,(struct sockaddr *)&host,sizeof(host))==-1)
      return UNREACHABLE_HOST;
    gettimeofday(&tv[1], (struct timezone*)0);
    dt += ((tv[1].tv_sec - tv[0].tv_sec) * 1000000
         + tv[1].tv_usec - tv[0].tv_usec);
    if(send(local,outbuf,1,0)==-1)
      return UNREACHABLE_HOST;
    if(recv(local,inbuf,1,0)==-1)
      return UNREACHABLE_HOST; 
    shutdown(local,2);
    close(local);
    sleep(5);
  }
  return (dt/LATENCY_TIMES);
}

/*-----------------------------------------------------------------------------
  bandwidth in bytes/s
  */
#ifdef __STDC__
int bandwidth(char *host_name,unsigned int host_address)
#else
int bandwidth(host_name,host_address)
char *host_name;
unsigned int host_address;
#endif
{
  struct sockaddr_in host;	/*Destination IP address */
  int i;
  int host_length; 		/*Length of this address*/
  char *inbuf;			/*incoming messages buffer*/
  char *outbuf;			/*outgoing messages buffer*/
  struct hostent *hp;

  /*for timing*/
  struct timeval tv1, tv2;
  int dt=0;            

  inbuf=(char *)calloc(MAXLENGTH,sizeof(char)); /*buffer Allocation*/
  outbuf=(char *)calloc(MAXLENGTH,sizeof(char));

  local=socket(AF_INET,SOCK_STREAM,0); 
  BZERO((char *)&host,sizeof(host));
  host.sin_family = AF_INET;
  host.sin_port = htons(PORT);
  hp = gethostbyaddr((char *)&host_address,sizeof(unsigned int),AF_INET);
  if (hp == NULL)
  {
    hp = gethostbyname((char *)host_name);
    if (hp == NULL)
      return UNREACHABLE_HOST;
  }
  BCOPY(hp->h_addr,&host.sin_addr,hp->h_length);
  host_length = sizeof(host);
  /* Connection */
  if(connect(local,(struct sockaddr *)&host,host_length)==-1)
    return UNREACHABLE_HOST;
  /* MAIN LOOP */
  i = 0;
  if(send(local,outbuf,size,0) == -1)
    return UNREACHABLE_HOST;
  if(recv(local,inbuf,size,0) == -1)
    return UNREACHABLE_HOST;

  do {
  gettimeofday(&tv1, (struct timezone*)0);
    if(send(local,outbuf,size,0) == -1)
      return UNREACHABLE_HOST;
    if(recv(local,inbuf,size,0) == -1)
      return UNREACHABLE_HOST;
    gettimeofday(&tv2, (struct timezone*)0);
    dt += ((tv2.tv_sec - tv1.tv_sec) * 1000000+tv2.tv_usec - tv1.tv_usec) / 2;
    i++;
  } while(dt < BANDWIDTH_TIME);

  /*Shutdown*/
  shutdown(local,2);
  close(local);
  return (int)((double)size * (double)i * 1000000. / (double)dt);
}

/*
 * Usage : pong <hostname> <IP@>
 *
 */

#ifdef __STDC__
void main(int argc, char **argv)
#else
void main(argc, argv)
     int argc;
     char *argv[];
#endif
{
  void handler();
  int connection;
  int band;

  size = LENGTH;

  (void)signal(SIGALRM,handler);
  alarm(TIMEOUT);
  if(argc != 3) panic();
  connection = my_latency(argv[1],(unsigned int)atoi(argv[2]));
  if (connection == UNREACHABLE_HOST)
  {
#ifdef VIEW
    fprintf(stderr,"PONG(%s) : Retrying\n",argv[1]);
#endif
    connection = my_latency(argv[1],(unsigned int)atoi(argv[2]));
  }
  band = my_bandwidth(argv[1],(unsigned int)atoi(argv[2]));
  if (band == UNREACHABLE_HOST)
  {
#ifdef VIEW
    fprintf(stderr,"PONG(%s) : Retrying\n",argv[1]);
#endif
    band = my_bandwidth(argv[1],(unsigned int)atoi(argv[2]));
  }

  if ((band == UNREACHABLE_HOST) &&(connection == UNREACHABLE_HOST))
  {
    /* Falling back to Homemade protocol */
#ifdef VIEW
    fprintf(stderr,"PONG(%s) : Falling back to homemade protocol\n",argv[1]);
#endif
    connection = my_latency(argv[1],(unsigned int)atoi(argv[2]));
    if (connection == UNREACHABLE_HOST)
    {
#ifdef VIEW
      fprintf(stderr,"PONG(%s) : Retrying\n",argv[1]);
#endif
      connection = my_latency(argv[1],(unsigned int)atoi(argv[2]));
    }
    band = my_bandwidth(argv[1],(unsigned int)atoi(argv[2]));
    if (band == UNREACHABLE_HOST)
    {
#ifdef VIEW
      fprintf(stderr,"PONG(%s) : Retrying\n",argv[1]);
#endif
      band = my_bandwidth(argv[1],(unsigned int)atoi(argv[2]));
    }
  }
  fprintf(stdout,"%d\n%d\n",connection,band);
#ifdef DEBUG    
  fprintf(stderr,"PONG(%s) : %d %d\n",argv[1],connection,band);
#endif
}

#ifdef __STDC__
void handler()
#else
void handler()
#endif
{
  shutdown(local,2);
  close(local);
  fprintf(stdout,"-1\n-1\n");
  exit(0);
}
  
/*-----------------------------------------------------------------------------
   homemade !! conection time in micro-seconds
  */

#ifdef __STDC__
int my_latency(char *host_name,unsigned int host_address)
#else
int my_latency(host_name,host_address)
char *host_name;
unsigned int host_address;
#endif
{
  struct sockaddr_in host;      /*Destination IP address */
  int s, i;
  int host_length;              /*Length of this address*/
  char *inbuf;                  /*incoming messages buffer*/
  char *outbuf;                 /*outgoing messages buffer*/
  struct hostent *hp;

  /*for timing*/
  struct timeval tv[2];
  int dt = 0, maxdt = 0;

  inbuf=(char *)calloc(MAXLENGTH,sizeof(char)); /*buffer Allocation*/
  outbuf=(char *)calloc(MAXLENGTH,sizeof(char));

  /* MAIN LOOP */
  for (i=0;i<LATENCY_TIMES;i++)
  {
    int type;
    int size = 1;
    int end = -1;

    local = socket(AF_INET,SOCK_STREAM,0);
    BZERO((char *)&host,sizeof(host));
    host.sin_family = AF_INET;
    host.sin_port = htons(SV_PORT);
    hp = gethostbyaddr((char *)&host_address,sizeof(unsigned int),AF_INET);
    if (hp == NULL)
    {
      hp = gethostbyname((char *)host_name);
      if (hp == NULL)
      {
        close(local);
        return UNREACHABLE_HOST;
      }
    }
    BCOPY(hp->h_addr,&host.sin_addr,hp->h_length);
    host_length = sizeof(host);
    gettimeofday(&tv[0], (struct timezone*)0);
    /* Connection */
    if(connect(local,(struct sockaddr *)&host,host_length)==-1)
      return UNREACHABLE_HOST;
    gettimeofday(&tv[1], (struct timezone*)0);
    dt += ((tv[1].tv_sec - tv[0].tv_sec) * 1000000
         + tv[1].tv_usec - tv[0].tv_usec);
    init_send(DATA_XDR,local);
    type = PONG_LATENCY;
    if (write_int(&type) == -1)
      return UNREACHABLE_HOST;
    end_message();
    end_com();
    sleep(5);
  }
  return (dt/LATENCY_TIMES);
}


/*-----------------------------------------------------------------------------
  Homemade !! bandwidth in bytes/s
  */
#ifdef __STDC__
int my_bandwidth(char *host_name,unsigned int host_address)
#else
int my_bandwidth(host_name,host_address)
char *host_name;
unsigned int host_address;
#endif
{
  int type;
  int size = LENGTH;
  int end = -1;

  struct sockaddr_in host;      /*Destination IP address */
  int i;
  int host_length;              /*Length of this address*/
  int *inbuf;                  /*incoming messages buffer*/
  int *outbuf;                 /*outgoing messages buffer*/
  struct hostent *hp;

  /*for timing*/
  struct timeval tv1, tv2;
  int dt=0;

  inbuf=(int *)calloc(MAXLENGTH,sizeof(int)); /*buffer Allocation*/
  outbuf=(int *)calloc(MAXLENGTH,sizeof(int));


  local=socket(AF_INET,SOCK_STREAM,0);
  BZERO((char *)&host,sizeof(host));
  host.sin_family = AF_INET;
  host.sin_port = htons(SV_PORT);
  hp = gethostbyaddr((char *)&host_address,sizeof(unsigned int),AF_INET);
  if (hp == NULL)
  {
    hp = gethostbyname((char *)host_name);
    if (hp == NULL)
      return UNREACHABLE_HOST;
  }
  BCOPY(hp->h_addr,&host.sin_addr,hp->h_length);
  host_length = sizeof(host);
  /* Connection */
  if(connect(local,(struct sockaddr *)&host,host_length)==-1)
    return UNREACHABLE_HOST;
  /* MAIN LOOP */
  i = 0;
  init_send(DATA_RAW,local);
  type = PONG_BANDWIDTH;
  if (write_int(&type) == -1)
    return UNREACHABLE_HOST;

  do {
  gettimeofday(&tv1, (struct timezone*)0);
    if (write_int(&size) == -1)
      return UNREACHABLE_HOST;
    if(write_int_array(outbuf,size) == -1)
      return UNREACHABLE_HOST;
    end_message();
    init_recv(local);
    if(read_int_array(inbuf,size) == -1)
      return UNREACHABLE_HOST;
    end_message();
    init_send(DATA_XDR,local);

    gettimeofday(&tv2, (struct timezone*)0);
    dt += ((tv2.tv_sec - tv1.tv_sec) * 1000000+tv2.tv_usec - tv1.tv_usec) / 2;
    i++;
  } while(dt < BANDWIDTH_TIME);

  if (write_int(&end) == -1)
      return UNREACHABLE_HOST;
  end_message();
  /*Shutdown*/
  end_com();
  return (int)((double)(4*size) * (double)i * 1000000. / (double)dt);
}

