/*****************************************************************/
/*      code_generator.c                                         */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/* main()                                                        */
/* process_code()                                                */
/* process_line()                                                */
/* process_mnemonic()                                            */
/* lookup()							 */
/* main()                                                        */
/*****************************************************************/

#include "code_generator.h"

/*
 *  MAIN PROCEDURE
 */

#ifdef __STDC__
main(int argc,char **argv)
#else
main(argc,argv)
int argc;
char **argv;
#endif
{

   char *token1,*token2,*token3;
   char name[256];
   char *buf;
   int i,j;
   int count=0;
   char *tmp;
   char *upper_code;
   FILE *configfile;
   FILE *initcodefile;
   FILE *tmp_include;
   char initcodefilename[128];
   char configfilename[128];
   char command[256];
   int filesnumber = 0;
   char *files[128];
   int loop;
   int nb_includes=0;
   char *includes[50];
   int nb_libs=0;
   char *libs[512];
   int file_lib[512];

  condorserver = 0;

  /* Check he arguments */

  if ((argc != 1)&&(argc != 2))
  {
    fprintf(stderr,"Usage : %s [configuration file]\n",argv[0]);
    exit(0);
  }

  /* Check the environment */

  if ((root = (char *)getenv("NETSOLVE_SERVER_ROOT")) == 0)
  {
    fprintf(stderr,"The NETSOLVE_SERVER_ROOT environment variable should be set\n");
    exit(0);
  }

  /* Open the def_file */
  sprintf(name,"%s/include/%s\0",root,DEF_FILE);
  target_def = fopen(name,"w");
  if (target_def == NULL)
  {
    fprintf(stderr,"Impossible to open %s\n",name);
    exit(0);
  }
  if (put_header_def()<0)
  {
    perror("put_header_def()");
    exit(0);
  }
  
  /* Opening the configuration file */
  
  if (argc == 1)
  {
    fprintf(stderr,"Using the default configuration file\n");
    sprintf(configfilename,"%s/server_config",root);
  }
  else
    sprintf(configfilename,argv[1]);

  configfile = fopen(configfilename,"r");
  if (configfile == NULL)
  {
    fprintf(stderr,"Cannot open file '%s'\n",configfilename);
    exit(0);
  }

  /* Opening the init source file */

  sprintf(initcodefilename,"%s/src/Server/%s",root,TARGETINITCODE);
  initcodefile = fopen(initcodefilename,"w");
  if (initcodefile == NULL)
  {
    fprintf(stderr,"Cannot open file '%s'\n",initcodefilename);
    fclose(configfile);
    exit(0);
  }

  /* Putting the header in the init source file */
  
  fprintf(initcodefile,"/******************************************/\n");
  fprintf(initcodefile,"/* Generated file from the code_generator */\n");
  fprintf(initcodefile,"/******************************************/\n\n");
  fprintf(initcodefile,"#include \"server.h\"\n\n");
  fprintf(initcodefile,"int problem_init()\n");
  fprintf(initcodefile,"{\n");
  fprintf(initcodefile,"problem_desc *pb_desc;\n");
  fprintf(initcodefile,"matlab_merge *tmp;\n");
  fprintf(initcodefile,"constant *consttmp;\n");
  fprintf(initcodefile,"formula *formtmp;\n");
  fprintf(initcodefile,"problem_item *pb_item;\n");
  fprintf(initcodefile,"char *buf;\n\n");

  /* Getting to work */

  fprintf(stderr,"Parsing the configuration file ...\n");
  
  while(1)
  {
    tmp = fgets(line,256,configfile);
    if (tmp == NULL)
    {
      fclose(configfile);
      break;
    }
    if (strncmp(line,"@CONDOR:",8) == 0)
      condorserver = 1;
    if (strncmp(line,"@PROBLEMS:\n",11) == 0)
      break;
  }
  if (tmp == NULL)
  {
    fprintf(stderr,"The configuration file '%s' does not contain problem files!!\n",
        configfilename);
    exit(0);
  }
  while(1)
  {
    char *tmp;
    tmp = fgets(line,256,configfile);
    if (tmp == NULL)
    {
      fclose(configfile);
      break;
    }
    if (*tmp == '@')
    {
      fclose(configfile);
      break;
    }
    if (line[0] == '#')
      continue; 
    line[strlen(line)-1] = '\0';
    strcpy(files[count++]=(char *)malloc(sizeof(char)*(1+strlen(line))),line);
  }

  if (count == 0)
  {
     fprintf(stderr,"The configuration files does not contain problem files\n");
     exit(0);
  }

  /*************/
  /* MAIN LOOP */
  /*************/

  filesnumber = count;

  PARALLEL=(int *)calloc(filesnumber,sizeof(int));
  if(PARALLEL==NULL) 
  {
    perror("Allocating PARALLEL array");
    exit(1);
  }

  fprintf(stderr,"Generating the code ...\n");
  for (count = 0;count < filesnumber;count++)
  {
    char *last_slash;
    char file_name[100];
    char fullname[256];
  
    /* Remove a potential /tmp/NetSolve_includes file */
    system("rm -f /tmp/NetSolve_includes");

    /* Open the source file */
    sprintf(fullname,"%s/%s",root,files[count]);
    sourcefile = fopen(fullname,"r");
    if (sourcefile == NULL)
    {
      fprintf(stderr,"Impossible to open the source file : %s\n",fullname);
      exit(0);
    }
    fprintf(stderr,"Processing file '%s'\n",files[count]);
    /* Find out the last word of files[count] */
    last_slash = (char *)strrchr(files[count],'/');
    if (last_slash == NULL)
    {
      strcpy(file_name,files[count]);
    }
    else
    {
      strcpy(file_name,&(files[count][(int)(last_slash-files[count]+1)]));
    }
    /* Open the target file */
    sprintf(name,"%s/src/Server/%s-%s.c\0",root,TARGET,file_name);
    targetfile = fopen(name,"w");
    if (targetfile == NULL)
    {
      fprintf(stderr,"Impossible to open the target file : %s\n",name);
      exit(0);
    }
    if (put_header() <0)
    {
      perror("put_header()");
      exit(0);
    }
    linecount = 0;
    problemcount = 0;
  
    while((char *)read_line() != NULL)
    {
      int status;
      char **const_list;
      int function_index = 0;

      /* Allocate space for the pb_desc */
  
      token1 = (char *)strtok(line," \t\n");
      token2 = (char *)strtok(NULL," \t\n");
      if ((token1 == NULL)||(token2 == NULL))
      {
        fprintf(stderr,"'@PROBLEM <problem name>'  expected at line %d\n",
                linecount);
        exit(0);
      }
      if (strcmp(token1,"@PROBLEM") != 0)
      {
        fprintf(stderr,"'@PROBLEM <problem name>'  expected at line %d\n",
                linecount);
        exit(0);
      }
      else
      {
        pb_desc = (problem_desc *)calloc(1,sizeof(problem_desc));
      }
      if (strlen(token2) > CODE_LENGTH)
      {
        fprintf(stderr,"'%s' : Problem name too long at line %d\n",token2,linecount);
        exit(0);
      }
      strcpy(pb_desc->code,token2);
      if ((char *)read_line() == NULL)
      {
        fprintf(stderr,"Interrupted problem section at line %d\n",
                linecount);
        exit(0);
      }

      /* Update initcodefile */

      fprintf(initcodefile,"/* Problem */\n");
      fprintf(initcodefile,"  pb_desc=(problem_desc *)calloc(1,sizeof(problem_desc));\n\n");
      fprintf(initcodefile,"  strcpy(pb_desc->code,\"%s\");\n",pb_desc->code);
      fprintf(initcodefile,"  strcpy(pb_desc->file,\"%s\");\n",file_name);

      /* FUNCTION : @FUNCTION <function name> - Not used here, just checking */
  
      token1 = (char *)strtok(line," \t\n");
      token2 = (char *)strtok(NULL," \t\n");
      while (token1 != NULL)
      {
        if (strcmp(token1,"@FUNCTION") == 0)
        {
          if (token2 == NULL)
          {
            fprintf(stderr,"'@FUNCTION <function name>'  expected at line %d\n",
                           linecount);
            exit(0);
          }
          function_name[function_index++] = (char *)strdup(token2);
          if ((char *)read_line() == NULL)
          {
            fprintf(stderr,"Interrupted problem section at line %d\n",
                           linecount);
            exit(0);
          }
          token1 = (char *)strtok(line," \t\n");
          token2 = (char *)strtok(NULL," \t\n");
        }
        else
          break;
      }

      /* Update initcodefile */

         /* NO CODE TO GENERATE HERE */

      /* INCLUDE : @INCLUDE <path/file.h> - Used here */
      while (token1 != NULL)
      {
        if (strcmp(token1,"@INCLUDE") == 0)
        {
          if (token2 == NULL)
          {
            fprintf(stderr,"'@INCLUDE <header file>'  expected at line %d\n",
                           linecount);
            exit(0);
          }
          includes[nb_includes++] = (char *)strdup(token2);
          if ((char *)read_line() == NULL)
          {
            fprintf(stderr,"Interrupted problem section at line %d\n",
                           linecount);
            exit(0);
          }
          token1 = (char *)strtok(line," \t\n");
          token2 = (char *)strtok(NULL," \t\n");
        }
        else
          break;
      }

      /* LIB : @LIB <path/file.[a|o]> - Used here */
      while (token1 != NULL)
      {
        if (strcmp(token1,"@LIB") == 0)
        {
          if (token2 == NULL)
          {
            fprintf(stderr,"'@LIB <library|object file>'  expected at line %d\n",
                           linecount);
            exit(0);
          }
          libs[nb_libs] = (char *)strdup(token2);
          file_lib[nb_libs] = count;
          nb_libs++;
          if ((char *)read_line() == NULL)
          {
            fprintf(stderr,"Interrupted problem section at line %d\n",
                           linecount);
            exit(0);
          }
          token1 = (char *)strtok(line," \t\n");
          token2 = (char *)strtok(NULL," \t\n");
        }
        else
          break;
      }


      /* LANGUAGE : @LANGUAGE <language name> - Not used here, just checking */
  
      if ((token1 == NULL)||(token2 == NULL))
      {
        fprintf(stderr,"'@LANGUAGE <language name>'  expected at line %d\n",
                linecount);
        exit(0);
      }
      if (strcmp(token1,"@LANGUAGE") != 0)
      {
        fprintf(stderr,"'@LANGUAGE <language name>'  expected at line %d\n",
                linecount);
        exit(0);
      }
  
      if (strcmp(token2,"C") == 0)
        current_language = C;
      else if (strcmp(token2,"FORTRAN") == 0)
      {
        current_language = F;
      }
      else
      {
        fprintf(stderr,"'%s ' unknown language at line %d\n",token2,linecount);
        exit(0);
      }
  
      if ((char *)read_line() == NULL)
      {
        fprintf(stderr,"Interrupted problem section at line %d\n",
                linecount);
        exit(0);
      }

      /* Update initcodefile */

         /* NO CODE TO GENERATE HERE */

  
      /* MAJOR : @MAJOR [COL,ROW] */
      
      token1 = (char *)strtok(line," \t\n");
      token2 = (char *)strtok(NULL," \t\n");

      if ((token1 == NULL)||(token2 == NULL))
      {
        fprintf(stderr,"'@MAJOR [COL|ROW]' expected at line %d\n",linecount);
        exit(0);
      }
      if (strcmp(token1,"@MAJOR") != 0)
      {
        fprintf(stderr,"token1 = %s\n",token1);
        fprintf(stderr,"'@MAJOR [COL|ROW]' expected at line %d\n",linecount);
        exit(0);
      }
      if ((strcmp(token2,"ROW")&&strcmp(token2,"COL")) != 0)
      {
        fprintf(stderr,"token2 = %s\n",token2);
        fprintf(stderr,"'@MAJOR [COL|ROW]' expected at line %d\n",linecount);
        exit(0);
      }

      /* Update initcodefile */

      fprintf(initcodefile,"  pb_desc->maj=%s_MAJOR;\n",token2);

      /* PATH :  @PATH <path>  - Not really used here - just checking */
  
      if ((char *)read_line() == NULL)
      {
        fprintf(stderr,"Interrupted problem section at line %d\n",
                linecount);
        exit(0);
      }
      token1 = (char *)strtok(line," \t\n");
      token2 = (char *)strtok(NULL," \t\n");
      if ((token1 == NULL)||(token2 == NULL))
      {
        fprintf(stderr,"'@PROBLEM <problem name>'  expected at line %d\n",
                linecount);
        exit(0);  
      }
      if (strcmp(token1,"@PATH") != 0)
      {
        fprintf(stderr,"@PATH expected at line %d\n",linecount);
        exit(0);  
      }
      if (strlen(token2) > PATH_LENGTH)
      {
        fprintf(stderr,"'%s' : Path too long at line %d\n",linecount);
        exit(0);  
      }
      if ((token2[0] != '/')||(token2[strlen(token2)-1] != '/'))
      {
        fprintf(stderr,"line %d : A path should begin and end with '/'\n",
                linecount);
        exit(0);  
      }
      strcpy(pb_desc->path,token2);

      /* Update initcodefile */

      fprintf(initcodefile,"  strcpy(pb_desc->path,\"%s\");\n",token2);
  
      /* DESCRIPTION - Not really used here - just checking */
  
      if ((char *)read_line() == NULL)
      {
        fprintf(stderr,"Interrupted problem section at line %d\n",
                linecount);
        exit(0);  
      }
      if (strncmp(line,"@DESCRIPTION",12) != 0)
      {
        fprintf(stderr,"@DESCRIPTION expected at line %d\n",linecount);
        exit(0);  
      }
      if ((char *)read_line() == NULL)
      {
        fprintf(stderr,"Interrupted problem section at line %d\n",
                linecount);
        exit(0);  
      }
      buf = pb_desc->desc;
      fprintf(initcodefile,"  buf = pb_desc->desc;\n");
      while ((strncmp(line,"@INPUT",6) != 0) &&
            (strncmp(line,"@PARALLEL",9) != 0))
      {
        line[strlen(line)-1] = '\n';
        strcpy(buf,line);
        buf += strlen(line);
        line[strlen(line)-1] = '\0';
        fprintf(initcodefile,"  strcpy(buf,\"");
        for (loop = 0;loop < strlen(line);loop++)
        {
          if (line[loop]=='"')
            fprintf(initcodefile,"\\\"");
          else
            fprintf(initcodefile,"%c",line[loop]);
        }
        fprintf(initcodefile,"\\n\");\n");
        fprintf(initcodefile,"  buf += %d;\n",strlen(line)+1);
        if (buf > (pb_desc->desc + DESC_LENGTH))
        {
          fprintf(stderr,"Description to long at line %d\n",
                          linecount);
          exit(0);  
        }
        if ((char *)read_line() == NULL)
        {
          fprintf(stderr,"Interrupted problem section at line %d\n",
                  linecount);
          exit(0);  
        }
      }
      if (buf == pb_desc->desc)
      {
        (pb_desc->desc)[0] = '?'; 
        (pb_desc->desc)[1] = '\0'; 
        fprintf(initcodefile,"  strcpy(pb_desc->desc,\"?\");");
      }
      else
      {
        (pb_desc->desc)[buf - (pb_desc->desc)-1] = '\0';
        fprintf(initcodefile,"  (pb_desc->desc)[buf - (pb_desc->desc)-1] = '\\0';\n");
      }
      
      /* PARALLEL: @PARALLEL <PVM or MPI or MPL or NX >*/
      token1 = (char *)strtok(line," \t\n");
      token2 = (char *)strtok(NULL," \t\n");
      if ((token1 == NULL)||(token2 == NULL))
      {
        fprintf(stderr,
                "'@INPUT <# of input objects>'  or @PARALLEL <type> expected at line %d\n",
                linecount);
        exit(0);
      }
      if (strcmp(token1,"@PARALLEL") == 0)
      {
        if(strcmp(token2, "PVM") == 0) {
          fprintf(initcodefile,"pb_desc->parallel = PVM;\n");
          PARALLEL[count]=PVM;
          }
        else if(strcmp(token2, "MPI") == 0) {
          fprintf(initcodefile,"pb_desc->parallel = MPI;\n");
          PARALLEL[count]=MPI;
          }
        else if(strcmp(token2, "MPL") != 0) {
          fprintf(initcodefile,"pb_desc->parallel = MPL;\n");
          PARALLEL[count]=MPL;
          }
        else if(strcmp(token2, "NX") != 0) {
          fprintf(initcodefile,"pb_desc->parallel = NX;\n");
          PARALLEL[count]=NX;
          }
        else {
          fprintf(stderr,"One of PVM, MPI, MPL, NX expected at line %d\n",
              linecount);
          exit(0);
          }
        if ((char *)read_line() == NULL)
        {
          fprintf(stderr,"Interrupted problem section at line %d\n",
                  linecount);
          exit(0);
        }
        token1 = (char *)strtok(line," \t\n");
        token2 = (char *)strtok(NULL," \t\n");
      }
      else
      {
        fprintf(initcodefile,"pb_desc->parallel = 0;\n");
      }

      /* INPUT :  @INPUT <number of input objects> */
  
      if ((token1 == NULL)||(token2 == NULL))
      {
        fprintf(stderr,"'@INPUT <# of input objects>' expected at line %d\n",
                linecount);
        exit(0);  
      }
      if (strcmp(token1,"@INPUT") != 0)
      {
        fprintf(stderr,"@INPUT expected at line %d\n",linecount);
        exit(0);  
      }
  
      if (is_number(token2) != 0)
      {
        fprintf(stderr,"'%s' : Should be a number at line %d\n",
                token2,linecount);
        exit(0);  
      }
      pb_desc->nb_obj_input = atoi(token2);
      pb_desc->input_obj = (object *)malloc(sizeof(object)*(pb_desc->nb_obj_input));
  
      fprintf(initcodefile,"  pb_desc->nb_obj_input = %d;\n",atoi(token2));
      fprintf(initcodefile,"  pb_desc->input_obj = ");
      fprintf(initcodefile,"(object *)malloc(sizeof(object)*(pb_desc->nb_obj_input));\n");

      if ((char *)read_line() == NULL)
      {
        fprintf(stderr,"Interrupted problem section at line %d\n",
              linecount);
      exit(0);  
      }
      for (i=0;i<pb_desc->nb_obj_input;i++)
      {
        token1 = (char *)strtok(line," \t\n");
        token2 = (char *)strtok(NULL," \t\n");
        token3 = (char *)strtok(NULL," \t\n");
        if ((token1 == NULL)||(token2 == NULL)||(token3 == NULL))
        {
          fprintf(stderr,
                  "@OBJECT <object type> <data type>' expected at line %d\n",
                  linecount);
          exit(0);  
        }
        if (strcmp(token1,"@OBJECT") != 0)
        {
          fprintf(stderr,
                  "@OBJECT <object type> <data type> expected at line %d\n",
                  linecount);
          exit(0);  
        }
        if (strcmp(token2,"MATRIX") == 0)
          pb_desc->input_obj[i].object_type = NETSOLVE_MATRIX;
        else if (strcmp(token2,"FILE") == 0)
          pb_desc->input_obj[i].object_type = NETSOLVE_FILE;
        else if (strcmp(token2,"VECTOR") == 0)
          pb_desc->input_obj[i].object_type = NETSOLVE_VECTOR;
        else if (strcmp(token2,"UPF") == 0)
          pb_desc->input_obj[i].object_type = NETSOLVE_UPF;
        else if (strcmp(token2,"SCALAR") == 0)
          pb_desc->input_obj[i].object_type = NETSOLVE_SCALAR;
        else
        {
          fprintf(stderr,
            "Object type at line %d should be one of the following:\n",
            linecount);
          fprintf(stderr,"        'MATRIX'\n");
          fprintf(stderr,"        'VECTOR'\n");
          fprintf(stderr,"        'SCALAR'\n");
          fprintf(stderr,"        'FILE'\n");
          fprintf(stderr,"        'UPF'\n");
          exit(0);  
        }
        if (strcmp(token3,"I") == 0)
          pb_desc->input_obj[i].data_type = NETSOLVE_I;
        else if (strcmp(token3,"ASCII") == 0)
          pb_desc->input_obj[i].data_type = NETSOLVE_ASCII;
        else if (strcmp(token3,"B") == 0)
          pb_desc->input_obj[i].data_type = NETSOLVE_B;
        else if (strcmp(token3,"CHAR") == 0)
          pb_desc->input_obj[i].data_type = NETSOLVE_CHAR;
        else if (strcmp(token3,"S") == 0)
          pb_desc->input_obj[i].data_type = NETSOLVE_S;
        else if (strcmp(token3,"EXTERNAL") == 0)
          pb_desc->input_obj[i].data_type = NETSOLVE_EXTERNAL;
        else if (strcmp(token3,"D") == 0)
          pb_desc->input_obj[i].data_type = NETSOLVE_D;
        else if (strcmp(token3,"C") == 0)
          pb_desc->input_obj[i].data_type = NETSOLVE_C;
        else if (strcmp(token3,"Z") == 0)
          pb_desc->input_obj[i].data_type = NETSOLVE_Z;
        else
        {
          fprintf(stderr,
            "Data type at line %d should be one of the following:\n",
            linecount);
          fprintf(stderr,"        'I'        : integer\n");
          fprintf(stderr,"        'S'        : single precision real\n");
          fprintf(stderr,"        'D'        : double precision real\n");
          fprintf(stderr,"        'C'        : single precision complex\n");
          fprintf(stderr,"        'Z'        : double precision complex\n");
          fprintf(stderr,"        'ASCII'    : ASCII\n");
          fprintf(stderr,"        'EXTERNAL' : User Provided Function\n");
          exit(0);  
        }
       
        fprintf(initcodefile,"  pb_desc->input_obj[%d].object_type = %d;\n",
                  i,pb_desc->input_obj[i].object_type);
        fprintf(initcodefile,"  pb_desc->input_obj[%d].data_type = %d;\n",
                  i,pb_desc->input_obj[i].data_type);

        /* OBJECT DESCRIPTION - Not really used here - just checking */
  
        if ((char *)read_line() == NULL)
        {
          fprintf(stderr,"Interrupted problem section at line %d\n",
                  linecount);
          exit(0);  
        }
        buf = pb_desc->input_obj[i].object_desc;

        fprintf(initcodefile,"  buf = pb_desc->input_obj[%d].object_desc;\n",i);

        while ((strncmp(line,"@OBJECT",7) != 0)&&
               (strncmp(line,"@OUTPUT",7) != 0))
        {
          line[strlen(line)-1] = '\n';
          strcpy(buf,line);
          buf += strlen(line);

          line[strlen(line)-1] = '\0';
          fprintf(initcodefile,"  strcpy(buf,\"");
          for (loop = 0;loop < strlen(line);loop++)
          {
            if (line[loop]=='"')
              fprintf(initcodefile,"\\\"");
            else
              fprintf(initcodefile,"%c",line[loop]);
          }
          fprintf(initcodefile,"\\n\");\n");
          fprintf(initcodefile,"  buf += %d;\n",strlen(line)+1);

          if (buf > (pb_desc->input_obj[i].object_desc + OBJ_LENGTH))
          {
            fprintf(stderr,"Description to long at line %d\n",
                            linecount);
            exit(0);  
          }
          if ((char *)read_line() == NULL)
          {
            fprintf(stderr,"Interrupted problem section at line %d\n",
                    linecount);
            exit(0);  
          }
        }
        if (buf == pb_desc->input_obj[i].object_desc)
        {
          pb_desc->input_obj[i].object_desc[0] = ' ';
          pb_desc->input_obj[i].object_desc[0] = '\0';

          fprintf(initcodefile,"  strcpy(pb_desc->input_obj[%d].object_desc,\"?\");",i);
        }
        else
        {
          pb_desc->input_obj[i].object_desc[buf
             - pb_desc->input_obj[i].object_desc-1] = '\0';
          fprintf(initcodefile,"  pb_desc->input_obj[%d].object_desc[buf\n",i);
          fprintf(initcodefile,"    - pb_desc->input_obj[%d].object_desc-1] = '\\0';\n",i);
        }
      }
  
      /* OUTPUT :  @OUTPUT <number of output objects> */
  
      token1 = (char *)strtok(line," \t\n");
      token2 = (char *)strtok(NULL," \t\n");
      if ((token1 == NULL)||(token2 == NULL))
      {
        fprintf(stderr,"'@OUTPUT <# of output objects>' expected at line %d\n",
                linecount);
        exit(0);  
      }
      if (strcmp(token1,"@OUTPUT") != 0)
      {
        fprintf(stderr,"@OUTPUT expected at line %d\n",linecount);
        exit(0);  
      }
  
      if (is_number(token2) != 0)
      {
        fprintf(stderr,"'%s' : Should be a number at line %d\n",
                token2,linecount);
        exit(0);  
      }
      pb_desc->nb_obj_output = atoi(token2);
      pb_desc->output_obj = (object *)malloc(sizeof(object)*pb_desc->nb_obj_output);
  
      fprintf(initcodefile,"  pb_desc->nb_obj_output = %d;\n",atoi(token2));
      fprintf(initcodefile,"  pb_desc->output_obj = ");
      fprintf(initcodefile,"(object *)malloc(sizeof(object)*(pb_desc->nb_obj_output));\n");

      if ((char *)read_line() == NULL)
      {
        fprintf(stderr,"Interrupted problem section at line %d\n",
              linecount);
      exit(0);  
      }
      for (i=0;i<pb_desc->nb_obj_output;i++)
      {
        token1 = (char *)strtok(line," \t\n");
        token2 = (char *)strtok(NULL," \t\n");
        token3 = (char *)strtok(NULL," \t\n");
        if ((token1 == NULL)||(token2 == NULL)||(token3 == NULL))
        {
          fprintf(stderr,
                  "@OBJECT <object type> <data type>' expected at line %d\n",
                  linecount);
          exit(0);  
        }
        if (strcmp(token1,"@OBJECT") != 0)
        {
          fprintf(stderr,
                  "@OBJECT <object type> <data type> expected at line %d\n",
                  linecount);
          exit(0);  
        }
        if (strcmp(token2,"MATRIX") == 0)
          pb_desc->output_obj[i].object_type = NETSOLVE_MATRIX;
        else if (strcmp(token2,"FILE") == 0)
          pb_desc->output_obj[i].object_type = NETSOLVE_FILE;
        else if (strcmp(token2,"VECTOR") == 0)
          pb_desc->output_obj[i].object_type = NETSOLVE_VECTOR;
        else if (strcmp(token2,"SCALAR") == 0)
          pb_desc->output_obj[i].object_type = NETSOLVE_SCALAR;
        else if (strcmp(token2,"UPF") == 0)
          pb_desc->output_obj[i].object_type = NETSOLVE_UPF;
        else
        {
          fprintf(stderr,
            "Object type at line %d should be one of the following:\n",
            linecount);
          fprintf(stderr,"        'MATRIX'\n");
          fprintf(stderr,"        'VECTOR'\n");
          fprintf(stderr,"        'SCALAR'\n");
          fprintf(stderr,"        'FILE'\n");
          fprintf(stderr,"        'UPF'\n");
          exit(0);  
        }
        if (strcmp(token3,"I") == 0)
          pb_desc->output_obj[i].data_type = NETSOLVE_I;
        else if (strcmp(token3,"ASCII") == 0)
          pb_desc->output_obj[i].data_type = NETSOLVE_ASCII;
        else if (strcmp(token3,"B") == 0)
          pb_desc->output_obj[i].data_type = NETSOLVE_B;
        else if (strcmp(token3,"CHAR") == 0)
          pb_desc->output_obj[i].data_type = NETSOLVE_CHAR;
        else if (strcmp(token3,"S") == 0)
          pb_desc->output_obj[i].data_type = NETSOLVE_S;
        else if (strcmp(token3,"S") == 0)
          pb_desc->output_obj[i].data_type = NETSOLVE_S;
        else if (strcmp(token3,"D") == 0)
          pb_desc->output_obj[i].data_type = NETSOLVE_D;
        else if (strcmp(token3,"EXTERNAL") == 0)
          pb_desc->output_obj[i].data_type = NETSOLVE_EXTERNAL;
        else if (strcmp(token3,"C") == 0)
          pb_desc->output_obj[i].data_type = NETSOLVE_C;
        else if (strcmp(token3,"Z") == 0)
          pb_desc->output_obj[i].data_type = NETSOLVE_Z;
        else
        {
          fprintf(stderr,
            "Data type at line %d should be one of the following:\n",
            linecount);
          fprintf(stderr,"        'I' : integer\n");
          fprintf(stderr,"        'S' : single precision real\n");
          fprintf(stderr,"        'D' : double precision real\n");
          fprintf(stderr,"        'C' : single precision complex\n");
          fprintf(stderr,"        'Z' : double precision complex\n");
          fprintf(stderr,"        'ASCII' : double precision complex\n");
          fprintf(stderr,"        'EXTERNAL' : UPF\n");
          exit(0);  
        }

        fprintf(initcodefile,"  pb_desc->output_obj[%d].object_type = %d;\n",
                  i,pb_desc->output_obj[i].object_type);
        fprintf(initcodefile,"  pb_desc->output_obj[%d].data_type = %d;\n",
                  i,pb_desc->output_obj[i].data_type);

        /* OBJECT DESCRIPTION - Not really used here - just checking */
  
        if ((char *)read_line() == NULL)
        {
          fprintf(stderr,"Interrupted problem section at line %d\n",
                  linecount);
          exit(0);  
        }
        buf = pb_desc->output_obj[i].object_desc;

        fprintf(initcodefile,"  buf = pb_desc->output_obj[%d].object_desc;\n",i);

        while ((strncmp(line,"@OBJECT",7) != 0)&&
               (strncmp(line,"@COMPLEXITY",10) != 0)&&
               (strncmp(line,"@MATLAB_MERGE",13) != 0))
        {
          line[strlen(line)-1] = '\n';
          strcpy(buf,line);
          buf += strlen(line);

          line[strlen(line)-1] = '\0';
          fprintf(initcodefile,"  strcpy(buf,\"");
          for (loop = 0;loop < strlen(line);loop++)
          {
            if (line[loop]=='"')
              fprintf(initcodefile,"\\\"");
            else
              fprintf(initcodefile,"%c",line[loop]);
          }
          fprintf(initcodefile,"\\n\");\n");
          fprintf(initcodefile,"  buf += %d;\n",strlen(line)+1);

          if (buf > (pb_desc->output_obj[i].object_desc + OBJ_LENGTH))
          {
            fprintf(stderr,"Description to long at line %d\n",
                            linecount);
            exit(0);  
          }
          if ((char *)read_line() == NULL)
          {
            fprintf(stderr,"Interrupted problem section at line %d\n",
                    linecount);
            exit(0);  
          }
        }
        if (buf == pb_desc->output_obj[i].object_desc)
        {
          pb_desc->output_obj[i].object_desc[0] = '?';
          pb_desc->output_obj[i].object_desc[0] = '\0';
          
          fprintf(initcodefile,"  strcpy(pb_desc->output_obj[%d].object_desc,\"?\");",i);
        }
        else
        {
          pb_desc->output_obj[i].object_desc[buf -
            pb_desc->output_obj[i].object_desc-1] = '\0';

          fprintf(initcodefile,"  pb_desc->output_obj[%d].object_desc[buf\n",i);
          fprintf(initcodefile,"    - pb_desc->output_obj[%d].object_desc-1] = '\\0';\n",i);
        }
      }
  
      /* MATLAB_MERGE */

      if (strncmp(line,"@MATLAB_MERGE",13) == 0)
      {
        while(strncmp(line,"@MATLAB_MERGE",13) == 0)
        {
          matlab_merge *tmp;

          token1 = (char *)strtok(line," \t\n");
          token2 = (char *)strtok(NULL," \n\t");
          token3 = (char *)strtok(NULL," \t\n");
          if ((token1 == NULL)||
              (token2 == NULL)||
              (token3 == NULL))
          {
            fprintf(stderr,"'@MATLAB_MERGE <output #> <output #>' expected at line %d\n",
                    linecount);
            exit(0);
          }
          if (strcmp(token1,"@MATLAB_MERGE") != 0)
          {
            fprintf(stderr,"'@MATLAB_MERGE <output #> <output #>' expected at line %d\n",
                    linecount);
            exit(0);
          }
          if (is_number(token2) != 0)
          {
            fprintf(stderr,"'%s' should be a number at line %d\n",
                    token2,linecount);
            exit(0);
          }
          if (is_number(token3) != 0)
          {
            fprintf(stderr,"'%s' should be a number at line %d\n",
                    token3,linecount);
            exit(0);
          }
          if ((atoi(token2) >= pb_desc->nb_obj_output)||
             (atoi(token3) >= pb_desc->nb_obj_output))
          {
            fprintf(stderr,"Only %d objects in output at line %d\n",
                    pb_desc->nb_obj_output,linecount);
            exit(0);
          }
          fprintf(initcodefile,"tmp = (matlab_merge *)calloc(1,sizeof(matlab_merge));\n");
          fprintf(initcodefile,"tmp->output1 = %d;\n",atoi(token2));
          fprintf(initcodefile,"tmp->output2 = %d;\n",atoi(token3));
          if ((atoi(token2) != (atoi(token3)-1)))
          {
            fprintf(stderr,"Merges can occur only with consecutive objects at line %d\n",
                    linecount);
            exit(0);
          }
          fprintf(initcodefile,"tmp->next = pb_desc->matlab;\n");
          fprintf(initcodefile,"pb_desc->matlab = tmp;\n");
          fprintf(initcodefile,"(pb_desc->nb_matlab)++;\n");
          
          if ((char *)read_line() == NULL)
          {
            fprintf(stderr,"Interrupted problem section at line %d\n",
                    linecount);
            exit(0);
          }
        }
      }

      /* COMPLEXITY :  @COMPLEXITY <complexity> - Not really used here */
  
      token1 = (char *)strtok(line," \t\n");
      token2 = (char *)strtok(NULL," \t\n");
      if ((token1 == NULL)||(token2 == NULL))
      {
        fprintf(stderr,"'@COMPLEXITY <complexity>'  expected at line %d\n",
                linecount);
        exit(0);  
      }
      if (strcmp(token1,"@COMPLEXITY") != 0)
      {
        fprintf(stderr,"'@COMPLEXITY <complexity>'  expected at line %d\n",
                linecount);
        exit(0);  
      }
      pb_desc->complexity = (char *)strdup(token2);
   
      fprintf(initcodefile,"  pb_desc->complexity = (char *)strdup(\"%s\");\n",token2);
  
      /* FORMATS   : @FORMATS <number of formats>     */
  
      if ((char *)read_line() == NULL)
      {
        fprintf(stderr,"'@FORMATS <number of formats>' expected at line %d\n",
                linecount);
        exit(0);
      }
      token1 = (char *)strtok(line," \t\n");
      token2 = (char *)strtok(NULL," \t\n");
      if ((token1 == NULL)||(token2 == NULL))
      {
        fprintf(stderr,"'@FORMATS <number of formats>'  expected at line %d\n",
                linecount);
        exit(0);
      }
      if (strcmp(token1,"@FORMATS") != 0)
      {
        fprintf(stderr,"'@FORMATS <number of formats>'  expected at line %d\n",
                linecount);
        exit(0);
      }
      if (is_number(token2) != 0)
      {
        fprintf(stderr,"'%s' : Should be a number at line %d\n",
                token2,linecount);
        exit(0);
      }
      pb_desc->nb_formats = atoi(token2);
      pb_desc->formats = (format_spec *)malloc(sizeof(format_spec)*pb_desc->nb_formats);

      fprintf(initcodefile,"  pb_desc->nb_formats = %d;\n",atoi(token2));
      fprintf(initcodefile,"  pb_desc->formats = ");
      fprintf(initcodefile,
        " (format_spec *)calloc(1,sizeof(format_spec)*pb_desc->nb_formats);\n");
  
      if ((char *)read_line() == NULL)
      {
        fprintf(stderr,"Format section interrupted at line %d\n",linecount);
        exit(0);
      }
  
      for (i=0;i<pb_desc->nb_formats;i++)
      {
        char **mnemonic_list;
        char **formula_list;
        int *const_list;
        int mnemonic_index = 0;
        int nb_arg = 0;
        int nb_const = 0;
        int nb_formula = 0;
  
        /* Maximum case */
        mnemonic_list =
              (char **)calloc(5*(pb_desc->nb_obj_input+pb_desc->nb_obj_output),
              sizeof(char *));
        formula_list =
              (char **)calloc(5*(pb_desc->nb_obj_input+pb_desc->nb_obj_output),
              sizeof(char *));
        const_list =
              (int *)calloc(5*(pb_desc->nb_obj_input+pb_desc->nb_obj_output),
              sizeof(int));
  
       
        token1 = (char *)strtok(line," \t\n");
        if ((token1 == NULL)||(token2 == NULL))
        {
          fprintf(stderr,"'@FORMAT' expected at line %d\n",linecount);
          exit(0);
        }
        if (strcmp(token1,"@FORMAT") != 0)
        {
          fprintf(stderr,"'@FORMAT' expected at line %d\n",linecount);
          exit(0);
        }
        if ((char *)read_line() == NULL)
        {
          fprintf(stderr,"Interrupted format section at line %d\n",linecount);
          exit(0);
        }
        while(strncmp(line,"@ARG",4) == 0)
        {
          int type_mnemonic;
          char *token_const;
  
          token1 = (char *)strtok(line," \t\n");
          token2 = (char *)strtok(NULL,"=\n");
          token_const = (char *)strtok(NULL," \t\n");
          if (token2 == NULL)
          {
            fprintf(stderr,"Argument list expected at line %d\n",
                  linecount);
            exit(0);
          }
          /* Check the mnemonics */
          mnemonic_list[mnemonic_index] = (char *)strdup(token2);
          if (strncmp(token1,"@ARG",4) != 0)
          {
            fprintf(stderr,"Invalid line at line %d\n",linecount);
            exit(0);
          }
          mnemonic_index++;
          nb_arg++;
          token3 = (char *)strtok(token2," \t,");
          if (token3 == NULL)
          {
            fprintf(stderr,"Invalid mnemonic at line %d\n",linecount);
            exit(0);
          }
          else
          {
            type_mnemonic = check_mnemonic(token3);
          }
          while(token3 != NULL)
          {
            int type;
  
            type= check_mnemonic(token3);
            if (type<0)
              exit(0);
            if (type != type_mnemonic)
            {
              fprintf(stderr,"Incoherent integer/pointer mnemonic specification\n");
              fprintf(stderr,"for the %d th format of '%s'\n",i,pb_desc->code);
              exit(0);
            }
  
            token3 = (char *)strtok(NULL," \t,");
          }
          if ((char *)read_line() == NULL)
          {
            fprintf(stderr,"Format section interrupted at line %d\n",linecount);
            exit(0);
          }
        }
  
        /* Constants */
        while(strncmp(line,"@CONST",4) == 0)
        {
          int type_mnemonic;
          char *token_const;
          int const_value;
  
          token1 = (char *)strtok(line," \t\n");
          token2 = (char *)strtok(NULL,"=\n");
          token_const = (char *)strtok(NULL," \t\n");
          if (token2 == NULL)
          {
            fprintf(stderr,"Argument expected at line %d\n",
                  linecount);
            exit(0);
          }
          /* Check the mnemonics */
          mnemonic_list[mnemonic_index] = (char *)strdup(token2);
  
          if (token_const == NULL)
          {
            fprintf(stderr,"A constant value must be specified at line %d\n",linecount);
            exit(0);
          }
          if (is_number(token_const) != 0)
          {
            fprintf(stderr,"Invalid constant value must be specified at line %d\n",linecount);
            exit(0);
          }
          const_value = atoi(token_const);
          const_list[nb_const] = const_value;
  
          mnemonic_index++;
          nb_const++;
  
          type_mnemonic = check_mnemonic(token2);
          if (type_mnemonic<0)
              exit(0);
          if (type_mnemonic != 1)
          {
            fprintf(stderr,"Only an integer can be a constant\n");
            exit(0);
          }
          if ((char *)read_line() == NULL)
          {
            fprintf(stderr,"Format section interrupted at line %d\n",linecount);
            exit(0);
          }
        }
  
        /* Formulae */
  
        while(strncmp(line,"@COMP",5) == 0)
        {
          int  type_mnemonic;
          char *token_form_line[10];
          int i,j;
          int length = 0;
          char *token_form = NULL;
          char *there;
          char *tmp;

          token1 = (char *)strtok(line," \t\n");
          token1 = (char *)strdup(token1);
          token2 = (char *)strtok(NULL,"=\n");
          token2 = (char *)strdup(token2);
          /* Get the different lines */
          tmp = (char *)strtok(NULL,"\n");
          for (i=0;i<10;i++)
          {
  
            if (tmp == NULL)
            {
              i = 0;
              length = 0;
              break;
            }
            if (tmp[strlen(tmp)-1] == '\\')
            {
              tmp[strlen(tmp)-1] = '\0';
              token_form_line[i] = (char *)strdup(tmp);  
              length+=strlen(tmp);
            } 
            else
            {
              token_form_line[i] = (char *)strdup(tmp);
              length+=strlen(tmp);
              break;
            }
            if ((char *)read_line() == NULL)
            {
              fprintf(stderr,"Format section interrupted at line %d\n",linecount);
              exit(0);
            }
            tmp = (char *)strtok(line,"\n");; 
          }
          i++;

          /* concatenate into a big stuff */
          if (length != 0)
            token_form = (char *)calloc(1+length,sizeof(char));
          else
            i = 0;
          there = token_form;
          for(j=0;j<i;j++)
          {
            strncpy(there,token_form_line[j],strlen(token_form_line[j]));
            there+=strlen(token_form_line[j]);
            free(token_form_line[j]);
          } 
          token_form[length] = '\0';
          if (token2 == NULL)
          {
            fprintf(stderr,"'@COMP <arg list> <formula>' expected at line %d\n",
                    linecount);
            exit(0);
          }
          if (token_form == NULL)
          {
            fprintf(stderr,"A expression must be specified at line %d\n",
                    linecount);
            exit(0);
          }
          if (check_expression(token_form,mnemonic_list,mnemonic_index) < 0)
          {
            exit(0);
          }
          mnemonic_list[mnemonic_index] = (char *)strdup(token2);
    	  formula_list[nb_formula] = (char *)strdup(token_form);
          nb_formula++;
          mnemonic_index++;
          type_mnemonic = check_mnemonic(token2);
          if (type_mnemonic <0)
            exit(0);
          if (type_mnemonic != 1)
          {
            fprintf(stderr,"Only an integer can be computed by an ");
            fprintf(stderr,"expression at line %d\n",linecount);
            exit(0);
          }
          if ((char *)read_line() == NULL)
          {
            fprintf(stderr,"Format section interrupted at line %d\n",linecount);
            exit(0);
          }
        }
        /* Now the list of mnemonics is constructed */
        if (check_mnemonic_list(mnemonic_list,i) < 0)
        {
          exit(0);
        }
        else
        {    /* Update the problem descriptor */
          int j;
 
  
          pb_desc->formats[i].nb_args = nb_arg;
          pb_desc->formats[i].nb_constants = nb_const;
          pb_desc->formats[i].nb_formulae = nb_formula;
          pb_desc->formats[i].args_desc = 
                   (char **)malloc(sizeof(char *)*nb_arg);

          fprintf(initcodefile,"  pb_desc->formats[%d].nb_args = %d;\n",i,nb_arg);
          fprintf(initcodefile,"  pb_desc->formats[%d].nb_constants = %d;\n",i,nb_const);
          fprintf(initcodefile,"  pb_desc->formats[%d].nb_formulae = %d;\n",i,nb_formula);
          fprintf(initcodefile,"  pb_desc->formats[%d].args_desc =\n",i);
          fprintf(initcodefile,"    (char **)malloc(sizeof(char *)*%d);\n",nb_arg);
  
          for (j=0;j<nb_arg;j++)
          {
            pb_desc->formats[i].args_desc[j] = (char *)strdup(mnemonic_list[j]);

            fprintf(initcodefile,"  pb_desc->formats[%d].args_desc[%d] =",i,j);
            fprintf(initcodefile,"    (char *)strdup(\"%s\");\n",mnemonic_list[j]);

            free(mnemonic_list[j]);
          }
          for (j=0;j<nb_const;j++)
          {
            constant *tmp;
  
            tmp = (constant *)malloc(sizeof(constant));
            tmp->next = pb_desc->formats[i].consts;
            tmp->value = const_list[j];
            tmp->mnemonic = (char *)strdup(mnemonic_list[nb_arg+j]);

            fprintf(initcodefile,"  consttmp = (constant *)malloc(sizeof(constant));\n");
            fprintf(initcodefile,"  consttmp->next = pb_desc->formats[%d].consts;\n",i);
            fprintf(initcodefile,"  consttmp->value = %d;\n",
                     const_list[j]);
            fprintf(initcodefile,"  consttmp->mnemonic = (char *)strdup(\"%s\");\n",
                     mnemonic_list[nb_arg+j]);

            if (((tmp->mnemonic)[0] == 'O')||
               (((tmp->mnemonic)[1]) == 'O')||
               (((tmp->mnemonic)[0]) == 'I'))
            {
              fprintf(stderr,
                "'%s' cannot be a constant (at line %d)\n",tmp->mnemonic,
                  linecount);
              exit(0);
            }
            pb_desc->formats[i].consts = tmp;

            fprintf(initcodefile,"  pb_desc->formats[%d].consts = consttmp;\n",i);

          }
          for (j=0;j<nb_formula;j++)
          {
            formula *tmp;
      
            tmp = (formula *)malloc(sizeof(formula));
            tmp->next = pb_desc->formats[i].forms;
            tmp->expression = (char *)strdup(formula_list[j]);
            tmp->mnemonic = (char *)strdup(mnemonic_list[j+nb_arg+nb_const]);

            fprintf(initcodefile,"  formtmp = (formula *)malloc(sizeof(formula));\n");
            fprintf(initcodefile,"  formtmp->next = pb_desc->formats[%d].forms;\n",i);
            fprintf(initcodefile,"  formtmp->expression = (char *)strdup(\"%s\");\n",
                   formula_list[j]);
            fprintf(initcodefile,"  formtmp->mnemonic = (char *)strdup(\"%s\");\n",
                   mnemonic_list[j+nb_arg+nb_const]);

            if (((tmp->mnemonic)[0] == 'O')||
               (((tmp->mnemonic)[1]) == 'O')||
               (((tmp->mnemonic)[0]) == 'I'))
            {
              fprintf(stderr,
                "'%s' cannot be an expression (at line %d)\n",tmp->mnemonic,
                  linecount);
              exit(0);
            }
            pb_desc->formats[i].forms = tmp;

            fprintf(initcodefile,"  pb_desc->formats[%d].forms = formtmp;\n",i); 

          }
        }
      }
      /* CODE : main thing to analyze here */
  
      token1 = (char *)strtok(line," \t\n");
      if (token1 == NULL)
      {
        fprintf(stderr,"'@CODE' expected at line %d\n",
                linecount);
        exit(0);
      }
      if (strcmp(token1,"@CODE") != 0)
      {  
        fprintf(stderr,"'@CODE' expected at line %d\n",
                linecount);
        exit(0);
      }
      /* Insert the pseudo-header */
      sprintf(buf,"#if (defined(flag_%s)||(!defined(UPF)))\n",pb_desc->code);
      write_line(buf);
      write_line("/****************************/\n");
      sprintf(buf,"/* Problem : %s */\n",pb_desc->code);
      write_line(buf);
      write_line("/****************************/\n");
      upper_code = (char *)strdup(pb_desc->code);
      uppercase(upper_code);
      sprintf(buf,"if (strcmp(buf1,\"%s\") == 0)\n{\n\n",upper_code);
      write_line(buf);
  

      /* Process the code */

      if (process_code() < 0)
      {
        fprintf(stderr,"Impossible to generate the code for : %s\n",
             pb_desc->code);
        exit(0);
      }
  
      /* Insert the end of problem section */
  
      write_line("\n}\n\n");
      write_line("#endif\n");
  
  
      /* UPDATE THE DEFINITION FILE IF NECESSARY */
  
      if (current_language == F)
      {
        char string[256];
        char caps_name[256];
    
        for (i=0;i<function_index;i++)
        {
          fwrite("#ifdef ADD_\n",1,strlen("#ifdef ADD_\n"),target_def);
          sprintf(string,"#define %s %s_\n",function_name[i],function_name[i]);
          fwrite(string,1,strlen(string),target_def);
          fwrite("#endif\n",1,strlen("#endif\n"),target_def);

          fwrite("#ifdef F2C\n",1,strlen("#ifdef F2C\n"),target_def);
          sprintf(string,"#define %s %s__\n",function_name[i],function_name[i]);
          fwrite(string,1,strlen(string),target_def);
          fwrite("#endif\n",1,strlen("#endif\n"),target_def);
    
          fwrite("#ifdef UPCASE\n",1,strlen("#ifdef UPCASE\n"),target_def);
          caps(caps_name,function_name[i]);
          sprintf(string,"#define %s %s\n",function_name[i],caps_name);
          fwrite(string,1,strlen(string),target_def);
          fwrite("#endif\n\n",1,strlen("#endif\n\n"),target_def);
        }
      }

      /* FINISH TO UPDATE INITCODEFILE */
      fprintf(initcodefile,"  if ((pb_item = (problem_item *)new_problem(*pb_desc)) == 0)\n");
      fprintf(initcodefile,"  {\n    perror(\"problem_init() new_problem()\");\n");
      fprintf(initcodefile,"       return -1;\n  }\n");
      fprintf(initcodefile,"  add_server_to_pb(pb_item,my_self);\n");
      fprintf(initcodefile,"  add_pb_to_server(pb_item,my_self);\n");
      fprintf(initcodefile,"  free(pb_desc);\n\n\n");

      /* FREE SOME STUFF */
      free(pb_desc->input_obj);
      free(pb_desc->output_obj);
      free(pb_desc);
    }
  
  /* TERMINATE THE GENERATION */
  
    write_line("\n\n\nreturn 1;\n}\n");
    fclose(sourcefile);
    fclose(targetfile);
    fprintf(stderr,"Writing file %s-%s.c\n",TARGET,file_name);

    /* Put the include statements   */
    /* into a separate file for now */
    tmp_include = fopen("/tmp/NetSolve_includes","a");
 
    /* Remove the redundant includes */
    for (i=0;i<nb_includes;i++)
    {
      int redundant = 0;
      for (j=0;j<i;j++)
        if (!strcmp(includes[i],includes[j]))
          redundant = 1;
      if (redundant)
        continue;
      fprintf(tmp_include,"#include %s\n",includes[i]);
    }
    fclose(tmp_include);

    /* Put the includes before the target */
    sprintf(command,"cat /tmp/NetSolve_includes/ %s > /tmp/NetSolve_final",name);
    system(command);
    sprintf(command,"mv /tmp/NetSolve_final %s",name);
    system(command);
    system("rm -f  /tmp/NetSolve_includes");
  }
  fclose(target_def);
  fprintf(stderr,"Code generated\n");
  fprintf(stderr,"Generating the Makefile...\n"); 
  generate_makefile(filesnumber,files,libs,nb_libs,file_lib);
  fprintf(stderr,"Makefile generated\n");
  /* Putting the tail of the problem init source file */
  fprintf(initcodefile,"  return 1;\n");
  fprintf(initcodefile,"}\n");
  fclose(initcodefile);
  fprintf(stderr,"You may now compile the server\n");
  exit(0);
}
  
/*
 * generate_makefile()
 */
#ifdef __STDC__
int generate_makefile(int filesnumber,char **files,char **libs,int nb_libs,int *file_lib)
#else
int generate_makefile(filesnumber,files,libs,nb_libs,file_lib)
int filesnumber;
char **files;
char **libs;
int nb_libs;
int *file_lib;
#endif
{
  char makefile[256];
  char *last_slash;
  char file_name[256];
  int i = 0;
  int j,k;
  

  sprintf(makefile,"%s/src/Server/%s\0",root,TARGETMAKEFILE);
  target_makefile = fopen(makefile,"w");
  if (target_makefile == NULL)
  {
    fprintf(stderr,"Impossible to open %s\n",makefile);
    exit(0);
  }
  
  fprintf(target_makefile,"#\n");
  fprintf(target_makefile,"#  generated ALL :\n");
  fprintf(target_makefile,"#\n");
  fprintf(target_makefile,"ALL = $(XDIR)/server \\\n");
  for (i = 0;i <filesnumber-1;i++)
  {
    last_slash = (char *)strrchr(files[i],'/');
    if (last_slash == NULL)
      strcpy(file_name,files[i]);
    else
      strcpy(file_name,&(files[i][(int)(last_slash-files[i]+1)]));
    fprintf(target_makefile,"         $(XDIR)/service-%s \\\n",file_name);
  }
  last_slash = (char *)strrchr(files[i],'/');
  if (last_slash == NULL)
    strcpy(file_name,files[i]);
  else
    strcpy(file_name,&(files[i][(int)(last_slash-files[i]+1)]));
  fprintf(target_makefile,"         $(XDIR)/service-%s\n\n",file_name);

  fprintf(target_makefile,"\ndefault : $(ALL)\n\n");
  fprintf(target_makefile,"#\n");
  fprintf(target_makefile,"# Makefile.numerical : generated section\n");
  fprintf(target_makefile,"#\n\n");
  

  for (i=0;i<filesnumber;i++)
  {
    last_slash = (char *)strrchr(files[i],'/');
    if (last_slash == NULL)
      strcpy(file_name,files[i]);
    else
      strcpy(file_name,&(files[i][(int)(last_slash-files[i]+1)]));

    fprintf(target_makefile,"###########################################\n");
    fprintf(target_makefile,"# SECTION : numerical-%s \n\n",file_name);
    if(PARALLEL[i]>0) 
    {
      fprintf(target_makefile,
        "$(XDIR)/service-%s := CFLAGS += -DPARALLEL=%d\n",
         file_name,PARALLEL[i]);
      if(PARALLEL[i]==PVM)
      {
         fprintf(target_makefile,
           "$(XDIR)/service-%s := CFLAGS += -I$(PVM_INCDIR)\n",
            file_name);
      }
      if(PARALLEL[i]==MPI)
      {
         fprintf(target_makefile,
           "$(XDIR)/service-%s := CFLAGS += -I$(MPI_INCDIR)\n",
            file_name);
      }
    }


    if (!condorserver)
    {
      fprintf(target_makefile,"$(XDIR)/service-%s : \\\n",file_name);
      if (PARALLEL[i]==PVM)
      {
        fprintf(target_makefile,
          "\t\t$(PVM_XDIR) \\\n");
      }
      fprintf(target_makefile,"\t\t$(XDIR) \\\n"); 
      fprintf(target_makefile,"\t\t$(OBJDIR) \\\n"); 
      if(PARALLEL[i]>0) 
      {
        fprintf(target_makefile,"\t\t$(OBJDIR)/pservice.o \\\n");
        fprintf(target_makefile,"\t\t$(OBJDIR)/pdlawrite.o \\\n");
        fprintf(target_makefile,"\t\t$(OBJDIR)/pdlaread.o \\\n");
        fprintf(target_makefile,"\t\t$(OBJDIR)/fortran_C_file_access.o \\\n");
        fprintf(target_makefile,"\t\t$(OBJDIR)/blacs_routines.o \\\n");
        fprintf(target_makefile,"\t\t$(OBJDIR)/pcompute.o \\\n");
      }
      else 
      {
        fprintf(target_makefile,"\t\t$(OBJDIR)/service.o \\\n");
        fprintf(target_makefile,"\t\t$(OBJDIR)/compute.o \\\n");
      }
      fprintf(target_makefile,"\t\t$(UPF) \\\n"); 
      fprintf(target_makefile,"\t\t$(OBJDIR)/pack_util.o \\\n"); 
      fprintf(target_makefile,"\t\t$(OBJDIR)/numerical-%s.o \\\n",file_name); 
      fprintf(target_makefile,"\t\t$(OBJDIR)/socket_util.o \\\n");
      fprintf(target_makefile,"\t\t$(OBJDIR)/log_file.o \\\n");
      fprintf(target_makefile,"\t\t$(OBJDIR)/gen_mes_all.o \\\n");
      fprintf(target_makefile,"\t\t$(OBJDIR)/list_processing.o \n");
  
      if(PARALLEL[i]>0)
      {
        fprintf(target_makefile,"\t$(LINKER) $(UPF) $(OBJDIR)/pservice.o \\\n");
        fprintf(target_makefile,"\t    $(OBJDIR)/pdlawrite.o \\\n");
        fprintf(target_makefile,"\t    $(OBJDIR)/pdlaread.o \\\n");
        fprintf(target_makefile,"\t    $(OBJDIR)/fortran_C_file_access.o \\\n");
        fprintf(target_makefile,"\t    $(OBJDIR)/blacs_routines.o \\\n");
        fprintf(target_makefile,"\t    $(OBJDIR)/pcompute.o \\\n");
      }
      else 
      {
        fprintf(target_makefile,"\t$(LINKER) $(UPF) $(OBJDIR)/service.o \\\n");
        fprintf(target_makefile,"\t    $(OBJDIR)/compute.o \\\n");
      }
  
      fprintf(target_makefile,"\t    $(OBJDIR)/pack_util.o \\\n");
      fprintf(target_makefile,"\t    $(OBJDIR)/numerical-%s.o \\\n",file_name);
      fprintf(target_makefile,"\t    $(OBJDIR)/socket_util.o \\\n");
      fprintf(target_makefile,"\t    $(OBJDIR)/log_file.o \\\n");
      fprintf(target_makefile,"\t    $(OBJDIR)/gen_mes_all.o \\\n");
      fprintf(target_makefile,"\t    $(OBJDIR)/list_processing.o \\\n");
/*
      if(PARALLEL[i]>0) 
      {
        fprintf(target_makefile,"\t    $(SCALA_LIB) \\\n");
        if(PARALLEL[i]==PVM) 
        {
          fprintf(target_makefile,"\t    $(PVM_BLACS_LIB) \\\n");
          fprintf(target_makefile,"\t    $(PVM_LIB) \\\n");
        }
        else 
          if(PARALLEL[i]==MPI) 
          {
            fprintf(target_makefile,"\t    $(MPI_BLACS_LIB) \\\n");
            fprintf(target_makefile,"\t    $(MPI_LIB) \\\n");
          }
      }
*/
      /* Putting the libs */
      for (j=0;j<nb_libs;j++)
      {
        int already_seen;
  
        already_seen = 0;

        /* Is it for this file ? */
        if (file_lib[j] != i)
          continue;

        /* Have we seen it before ? */
        for (k=0;k<j;k++)
        {
          if (file_lib[k] != i)
            continue;
          if (!strcmp(libs[j],libs[k]))
            already_seen = 1;
        }
        if (already_seen)
          continue;

        /* Add it to the makefile */
        fprintf(target_makefile,"\t    %s \\\n",libs[j]);
      }
  
      fprintf(target_makefile,"\t    $(F_FROM_C_LIB) \\\n");
      fprintf(target_makefile,"\t    $(LIBS) $(HBMLIB) $(ARCHDLIB) -o $@ \n");
  
      /*
       * this removes  the object files so that multiple parallel versions
       * can be made and the objects will use the correct libraries.
       */
      if(PARALLEL[i]>0) 
      {
        fprintf(target_makefile,"\trm -f $(PARALLEL)\n");
        if(PARALLEL[i]==PVM) 
        {
          fprintf(target_makefile,
          "\tcp $(XDIR)/service-%s $(PVM_XDIR)\n",file_name);
        }
      }
    }
    else  /* CONDOR */
    {
      fprintf(target_makefile,"$(XDIR)/service-%s : \\\n",file_name);
      fprintf(target_makefile,"\t\t$(XDIR) \\\n");
      fprintf(target_makefile,"\t\t$(OBJDIR) \\\n");
      fprintf(target_makefile,"\t\t$(OBJDIR)/service.o \\\n");
      fprintf(target_makefile,"\t\t$(OBJDIR)/compute.o \\\n");

      fprintf(target_makefile,"\t\t$(UPF) \\\n");
      fprintf(target_makefile,"\t\t$(OBJDIR)/numerical-%s.o \\n",file_name);

      fprintf(target_makefile,"\t$(CONDOR1) $(UPF) $(OBJDIR)/service.o \\\n");
      fprintf(target_makefile,"\t    $(OBJDIR)/compute.o \\\n");

      fprintf(target_makefile,"\t    $(OBJDIR)/numerical-%s.o \\\n",file_name);

      fprintf(target_makefile,"\t    $(NUMLIB) $(F_FROM_C_LIB) \\\n");
      fprintf(target_makefile,"\t    (HBMLIB) $(CONDOR2) \\\n");
      fprintf(target_makefile,"\t    $(LIBS) $(ARCHDLIB) -o $@ \n");

    }

    fprintf(target_makefile,"\n");

    fprintf(target_makefile,"$(OBJDIR)/numerical-%s.o : \\\n",file_name);
    fprintf(target_makefile,"\t\t$(OBJDIR) \\\n");
    fprintf(target_makefile,"\t\t$(SDIR)/numerical-%s.c \n",file_name);
    fprintf(target_makefile,"\t$(CC) $(CFLAGS) -c $(SDIR)/numerical-%s.c -o $@ \n\n",file_name);
  
    fprintf(target_makefile,"# end of section : numerical-%s \n",file_name);
    fprintf(target_makefile,"###########################################\n\n");

  }
  if (!condorserver)
  {
    fprintf(target_makefile,"$(OBJDIR)/service.o : $(OBJDIR) $(SDIR)/service.c $(HDRS)\n");
    fprintf(target_makefile,"\t$(CC) $(CFLAGS) -c $(SDIR)/service.c -o $@\n");
    fprintf(target_makefile,"\n");
    fprintf(target_makefile,"$(OBJDIR)/compute.o : $(OBJDIR) $(SDIR)/compute.c $(HDRS)\n");
    fprintf(target_makefile,"\t$(CC) $(CFLAGS) -c $(SDIR)/compute.c -o $@\n");
  }
  else
  {
    fprintf(target_makefile,"$(OBJDIR)/service.o : $(OBJDIR) $(SDIR)/service.c $(HDRS)\n");
    fprintf(target_makefile,"\t$(CC) $(CFLAGS) -DISCONDOR -c $(SDIR)/service.c -o $@\n");
    fprintf(target_makefile,"\n");
    fprintf(target_makefile,"$(OBJDIR)/compute.o : $(OBJDIR) $(SDIR)/compute.c $(HDRS)\n");
    fprintf(target_makefile,"\t$(CC) $(CFLAGS) -DISCONDOR -c $(SDIR)/compute.c -o $@\n");
  }

  fclose(target_makefile);
  return;
}


/*
 * process_code()
 *
 * Actually processes the code itself, with replacement of the macros
 *
 */

#ifdef __STDC__
int process_code()
#else
int process_code()
#endif
{
  int status;

  while(1)
  {
    if ((char *)read_line() == 0)
    {
      fprintf(stderr,"Interrupted @CODE section at line %d\n",linecount);
      return -1;
    }
    if (strncmp(line,"@END_CODE",9) == 0) /* End of the code */
      return 1;
    if ((status = process_line()) <0)
      return status;
  }
}

/*
 * process_line()
 *
 *  Actually processes a line
 *
 */

#ifdef __STDC__
int process_line()
#else
int process_line()
#endif
{
  char outputline[MAX_LINE_LENGTH];
  int output_offset = 0;
  int input_offset = 0;

  int waiting_quote = 0;
  char *replacement;
  char *process_mnemonic();
 
  static char prev_last_previous = ' ';
  static char last_previous = ';';

  int i;
  int first = 1;


  for (i=0;i<MAX_LINE_LENGTH;i++)
    outputline[i] = '\0';
  
  while(line[input_offset] != '\n')
  {
    char mnemonic[MNEMONIC_LENGTH+1];

    /** Carriage return after each comma **/
    if ((line[input_offset] == '"')&&(waiting_quote == 0))
    {
      waiting_quote = 1;
      outputline[output_offset++] = line[input_offset++];
      continue;
    }
    if ((line[input_offset] == '"')&&(waiting_quote == 1))
    {
      waiting_quote = 0;
      outputline[output_offset++] = line[input_offset++];
      continue;
    }
    if (line[input_offset] == ',')
    {
      first = 0;
      outputline[output_offset++] = line[input_offset++];
      if (waiting_quote == 0)
      {
        outputline[output_offset] = '\n';
        write_line(outputline);
        for (i=0;i<MAX_LINE_LENGTH;i++)
            outputline[i] = '\0';
        output_offset = 0;
      }
      if (line[input_offset] == '\n')
      {
        if (input_offset-2 >= 0)
          prev_last_previous = line[input_offset-2];
        last_previous = line[input_offset-1];
        write_line(outputline);
        return 1;
      }
      continue;
    }
    /** Carriage return after each = **/
    if ((line[input_offset] == '=')&&
        (line[input_offset-1] != '=')&&
        (line[input_offset+1] != '='))
    {
      first = 0;
      outputline[output_offset++] = line[input_offset++];
      if (waiting_quote == 0)
      {
        outputline[output_offset] = '\n';
        write_line(outputline);
        for (i=0;i<MAX_LINE_LENGTH;i++)
          outputline[i] = '\0';
        output_offset = 0;
      }
      if (line[input_offset] == '\n')
      {
        if (input_offset-2 >= 0)
          prev_last_previous = line[input_offset-2];
        last_previous = line[input_offset-1];
        write_line(outputline);
        return 1;
      }
      continue;
    }

    /** Not a special character **/
    if ((line[input_offset] != '@')&&
        (line[input_offset] != '\\'))
    {
      if (line[input_offset] != ' ' && line[input_offset] != '\t')
        first = 0;
      outputline[output_offset++] = line[input_offset++];
      if (line[input_offset] == '\n')
      {
        if (input_offset-2 >= 0)
          prev_last_previous = line[input_offset-2];
        last_previous = line[input_offset-1];
        outputline[output_offset] = '\n';
        write_line(outputline);
        return 1;
      }
      continue;
    }
    
    /** De-specified character **/
    if (line[input_offset] == '\\')
    {
      first = 0;
      input_offset++;
      outputline[output_offset++] = line[input_offset++];
      if (line[input_offset] == '\n')
      {
        if (input_offset-2 >= 0)
          prev_last_previous = line[input_offset-2];
        last_previous = line[input_offset-1];
        outputline[output_offset] = '\n';
        write_line(outputline);
        return 1;
      }
      continue;
    }

    /** FROM NOW : SPECIAL CHARACTER **/

    input_offset++;
    for (i=0;i<MNEMONIC_LENGTH+1;i++)
      mnemonic[i] = '\0';
    for (i=0;i<MNEMONIC_LENGTH;i++)
    {
      if (line[input_offset] == '@')
        break;
      mnemonic[i] = line[input_offset++]; 
    }
    if (i==MNEMONIC_LENGTH)
    {
      fprintf(stderr,"Unterminated or too long mnemonic at line %d\n",linecount);
      return -1;
    }

    /* Check the validity of the mnemonic */
    i = 0;
    while(mnemonic[i] != '\0')
    {
      if ((mnemonic[i] != 'I') && (mnemonic[i] != 'O') &&
          (mnemonic[i] != 'l') && (mnemonic[i] != 'd') &&
          (mnemonic[i] != 'm') && (mnemonic[i] != 'n') &&
          (mnemonic[i] != '0') && (mnemonic[i] != '1') &&
          (mnemonic[i] != '2') && (mnemonic[i] != '3') &&
          (mnemonic[i] != '4') && (mnemonic[i] != '5') &&
          (mnemonic[i] != '6') && (mnemonic[i] != '7') &&
          (mnemonic[i] != '8') && (mnemonic[i] != '9'))
      {
        fprintf(stderr,"Illegal mnemonic character '%c ' in '%s ' at line %d\n",
                mnemonic[i],mnemonic,linecount);
        return -1;
      }
      i++;
    }

    input_offset++;
    if (first == 1)
    {
      if (last_previous != ';' && last_previous != '}' 
          && last_previous != '{')
        first = 0;
      if (last_previous == '/' && prev_last_previous == '*')
        first = 1;
    }
    i = input_offset;
    while((line[i] == ' ')||(line[i] == ')')) i++;
    if (line[i] == '[')
      first = 0;
    replacement = (char *)process_mnemonic(first,mnemonic,&output_offset);
    if (replacement == NULL)
    {
      return -1;
    }

    i = 0;
    while(line[i] != '\n')
    {
      if (line[i] != ' ' && line[i] != '\t')
      {
        if (i-1 >= 0)
          last_previous = line[i-1];
        last_previous = line[i];
      }
      i++;
    }
    for (i=0;replacement[i]!='\0';i++)
      outputline[output_offset++] = replacement[i];
    
    free(replacement);
  }
}

/*
 * process_mnemonic()
 *
 *  Do the substitution in the target file
 */

#ifdef __STDC__
char *process_mnemonic(int first, char *mnemonic,int *output_offset)
#else
char *process_mnemonic(first,mnemonic,output_offset)
int first;
char *mnemonic;
int *output_offset;
#endif
{
  char buf[MAX_LINE_LENGTH];
  int i;
  char data_type;

  char *replacement=NULL;

  for (i=0;i<MAX_LINE_LENGTH;i++)
    buf[i] = '\0';

  switch (mnemonic[0])
  {
    case 'I' :         /* INPUT PARAMETER */
    {
      int index;
      if (is_number((char *)mnemonic+1) == -1)
      {
        fprintf(stderr,"Invalid mnemonic '%s' at line %d\n",
                       mnemonic,linecount);
        return NULL;
      }
      index = atoi((char *)mnemonic+1);

      if (index >= pb_desc->nb_obj_input)
      {
        fprintf(stderr,
           "Line %d :Only %d objects in input to %s in mnemonic @%s@\n",
           linecount,pb_desc->nb_obj_input,pb_desc->code,mnemonic);
        return NULL;
      }
      data_type = pb_desc->input_obj[index].data_type;

      if (first == 0)
      {
        switch (data_type)
        {
          case NETSOLVE_ASCII :
          {
            sprintf(buf,"((char *)(data_input[%d]))",index);
            break;
          }
          case NETSOLVE_B :
          {
            sprintf(buf,"((char *)(data_input[%d]))",index);
            break;
          }
          case NETSOLVE_I :
          {
            sprintf(buf,"((int *)(data_input[%d]))",index);
            break;
          }
          case NETSOLVE_CHAR :
          {
            sprintf(buf,"((char *)(data_input[%d]))",index);
            break;
          }
          case NETSOLVE_S :
          {
            sprintf(buf,"((float *)(data_input[%d]))",index);
            break;
          }          
          case NETSOLVE_D :
          {
            sprintf(buf,"((double *)(data_input[%d]))",index);
            break;
          }
          case NETSOLVE_C :
          {
            sprintf(buf,"((scomplex *)(data_input[%d]))",index);
            break;
          }
          case NETSOLVE_Z :
          {
            sprintf(buf,"((dcomplex *)(data_input[%d]))",index);
            break;
          }
          case NETSOLVE_EXTERNAL :
          {
            /* Nothing here */
            break;
          }
          default :
          {
            fprintf(stderr,"Unknown data type ??\n");
            return NULL;
          }
        }
      }
      else
        sprintf(buf,"(data_input[%d])",index);

      return ((char *)strdup(buf));
    }
    case 'O' :             /* OUTPUT PARAMETER */
    {
      int index;
      if (is_number((char *)mnemonic+1) == -1)
      {
        fprintf(stderr,"Invalid mnemonic '%s' at line %d\n",
                       mnemonic,linecount); 
        return NULL;
      }
      index = atoi((char *)mnemonic+1);

      if (index >= pb_desc->nb_obj_output)
      {
        fprintf(stderr,
           "Line %d :Only %d objects in output to %s in mnemonic @%s@\n",
           linecount,pb_desc->nb_obj_output,pb_desc->code,mnemonic);
        return NULL;
      }
      data_type = pb_desc->output_obj[index].data_type;

      if (first == 0)
      {
        switch (data_type)
        {
          case NETSOLVE_ASCII :
          {
            sprintf(buf,"((char *)(data_output[%d]))",index);
            break;
          }
          case NETSOLVE_B :
          {
            sprintf(buf,"((char *)(data_output[%d]))",index);
            break;
          }
          case NETSOLVE_I :
          {
            sprintf(buf,"((int *)(data_output[%d]))",index);
            break;
          }
          case NETSOLVE_CHAR :
          {
            sprintf(buf,"((char *)(data_output[%d]))",index);
            break;
          }
          case NETSOLVE_S :
          {
            sprintf(buf,"((float *)(data_output[%d]))",index);
            break;
          }
          case NETSOLVE_D :
          {
            sprintf(buf,"((double *)(data_output[%d]))",index);
            break;
          }
          case NETSOLVE_C :
          {
            sprintf(buf,"((scomplex *)(data_output[%d]))",index);
            break;
          }
          case NETSOLVE_Z :
          {
            sprintf(buf,"((dcomplex *)(data_output[%d]))",index);
            break;
          }
          default :
          {
            fprintf(stderr,"Unknown data type ??\n");
            return NULL;
            }
        }
      }
      else
        sprintf(buf,"(data_output[%d])",index);

      return ((char *)strdup(buf));
    }
    case 'm' :  /* M DIMENSION */
    {
      switch (mnemonic[1])
      {
        case 'I' :            /* In input */
        {
          int index;
          int object_type;

          if (is_number((char *)mnemonic+2) == -1)
          {
            fprintf(stderr,"Invalid mnemonic '%s' at line %d\n",
                           mnemonic,linecount); 
            return NULL;
          }
          index = atoi((char *)mnemonic+2);
          if (index >= pb_desc->nb_obj_input)
          {
            fprintf(stderr,
              "Line %d :Only %d objects in input to %s in mnemonic @%s@\n",
              linecount,pb_desc->nb_obj_input,pb_desc->code,mnemonic);
            return NULL;
          }
          object_type = pb_desc->input_obj[index].object_type;
          if ((object_type != NETSOLVE_MATRIX) &&
              (object_type != NETSOLVE_VECTOR))
          {
            fprintf(stderr,
              "Line %d :Object #%d in input to %s does not have a 'm-dimension'\n",
              linecount,index,pb_desc->code);
            return NULL;
          }
          sprintf(buf,"((int *)(&(sizes_input[%d].m)))",index);
          return ((char *)strdup(buf));
        }
        case 'O' :             /* IN OUTPUT */
        {
          int index;
          int object_type;

          if (is_number((char *)mnemonic+2) == -1)
          {
            fprintf(stderr,"Invalid mnemonic '%s' at line %d\n",mnemonic,linecount);
            return NULL;
          }
          index = atoi((char *)mnemonic+2);
          if (index >= pb_desc->nb_obj_output)
          {
            fprintf(stderr,
              "Line %d :Only %d objects in output to %s in mnemonic @%s@\n",
              linecount,pb_desc->nb_obj_output,pb_desc->code,mnemonic);
            return NULL;
          }
          object_type = pb_desc->output_obj[index].object_type;
          if ((object_type != NETSOLVE_MATRIX) &&
              (object_type != NETSOLVE_VECTOR))
          {
            fprintf(stderr,
              "Line %d :Object #%d in output to %s does not have a 'm-dimension'\n",
              linecount,index,pb_desc->code);
            return NULL;
          }
          sprintf(buf,"((int *)(&(sizes_output[%d].m)))",index);
          return ((char *)strdup(buf));
        }
        default :     /* ERROR */
        {
          fprintf(stderr,"Line %d : Invalid mnemonic @%s @\n",
                  linecount,mnemonic);
          return NULL;
        }
      }
    }
    case 'n' : /* N Dimension */
    {
      switch (mnemonic[1])
      {
        case 'I' :            /* In input */
        {
          int index;
          int object_type;

          if (is_number((char *)mnemonic+2) == -1)
          {
            fprintf(stderr,"Invalid mnemonic '%s' at line %d\n",
                    mnemonic,linecount);
            return NULL;
          }
          index = atoi((char *)mnemonic+2);
          if (index >= pb_desc->nb_obj_input)
          {
            fprintf(stderr,
              "Line %d :Only %d objects in input to %s in mnemonic @%s@\n",
              linecount,pb_desc->nb_obj_input,pb_desc->code,mnemonic);
            return NULL;
          }
          object_type = pb_desc->input_obj[index].object_type;
          if (object_type != NETSOLVE_MATRIX)
          {
            fprintf(stderr,
              "Line %d :Object #%d in input to %s does not have a 'n-dimension'\n",
              linecount,index,pb_desc->code);
            return NULL;
          }
          sprintf(buf,"((int *)(&(sizes_input[%d].n)))",index);

          return ((char *)strdup(buf));
        }
        case 'O' :             /* IN OUTPUT */
        {
          int index;
          int object_type;

          if (is_number((char *)mnemonic+2) == -1)
          {
            fprintf(stderr,"Invalid mnemonic '%s' at line %d\n",mnemonic,linecount);
            return NULL;
          }
          index = atoi((char *)mnemonic+2);
          if (index >= pb_desc->nb_obj_output)
          {
            fprintf(stderr,
              "Line %d :Only %d objects in output to %s in mnemonic @%s@\n",
              linecount,pb_desc->nb_obj_input,pb_desc->code,mnemonic);
            return NULL;
          }
          object_type = pb_desc->output_obj[index].object_type;
          if (object_type != NETSOLVE_MATRIX)
          {
            fprintf(stderr,
              "Line %d :Object #%d in output to %s does not have a 'n-dimension'\n",
              linecount,index,pb_desc->code);
            return NULL;
          }
          sprintf(buf,"((int *)(&(sizes_output[%d].n)))",index);
          return ((char *)strdup(buf));
        }
        default :     /* ERROR */
        {
          fprintf(stderr,"Line %d : Invalid mnemonic @%s @\n",
                  linecount,mnemonic);
          return NULL;
        }
      }
    }
    case 'l' : /* Leading Dimension */
    {
      switch (mnemonic[1])
      {
        case 'I' :            /* In input */
        {
          int index;
          int object_type;

          if (is_number((char *)mnemonic+2) == -1)
          {
            fprintf(stderr,"Invalid mnemonic '%s' at line %d\n",mnemonic,linecount);
            return NULL;
          }
          index = atoi((char *)mnemonic+2);
          if (index >= pb_desc->nb_obj_input)
          {
            fprintf(stderr,
              "Line %d :Only %d objects in input to %s in mnemonic @%s@\n",
              linecount,pb_desc->nb_obj_input,pb_desc->code,mnemonic);
            return NULL;
          }
          object_type = pb_desc->input_obj[index].object_type;
          if (object_type != NETSOLVE_MATRIX)
          {
            fprintf(stderr,
              "Line %d :Object #%d in input to %s does not have a 'leading dimension'\n",
              linecount,index,pb_desc->code);
            return NULL;
          }
          sprintf(buf,"((int *)(&(sizes_input[%d].l)))",index);
          return ((char *)strdup(buf));
        }
        case 'O' :             /* IN OUTPUT */
        {
          int index;
          int object_type;

          if (is_number((char *)mnemonic+2) == -1)
          {
            fprintf(stderr,"Invalid mnemonic '%s' at line %d\n",mnemonic,linecount);
            return NULL;
          }
          index = atoi((char *)mnemonic+2);
          if (index >= pb_desc->nb_obj_output)
          {
            fprintf(stderr,
              "Line %d :Only %d objects in output to %s in mnemonic @%s@\n",
              linecount,pb_desc->nb_obj_input,pb_desc->code,mnemonic);
            return NULL;
          }
          object_type = pb_desc->output_obj[index].object_type;
          if (object_type != NETSOLVE_MATRIX)
          {
            fprintf(stderr,
              "Line %d :Object #%d in output to %s does not have a 'leading dimension'\n",
              linecount,index,pb_desc->code);
            return NULL;
          }
          sprintf(buf,"((int *)(&(sizes_output[%d].l)))",index);
          return ((char *)strdup(buf));
        }
        default :     /* ERROR */
        {
          fprintf(stderr,"Line %d : Invalid mnemonic @%s @\n",
                  linecount,mnemonic);
          return NULL;
        }
      }
    }
    case 'd' : /* ScaLAPACK Description */
    {
      switch (mnemonic[1])
      {
        case 'I' :            /* In input */
        {
          int index;
          int object_type;

          if (is_number((char *)mnemonic+2) == -1)
          {
            fprintf(stderr,"Invalid mnemonic '%s' at line %d\n",mnemonic,linecount);
            return NULL;
          }
          index = atoi((char *)mnemonic+2);
          if (index >= pb_desc->nb_obj_input)
          {
            fprintf(stderr,
              "Line %d :Only %d objects in input to %s in mnemonic @%s@\n",
              linecount,pb_desc->nb_obj_input,pb_desc->code,mnemonic);
            return NULL;
          }
          object_type = pb_desc->input_obj[index].object_type;
          if (object_type != NETSOLVE_MATRIX)
          {
            fprintf(stderr,
              "Line %d :Object #%d in input to %s does not have a 'descriptor'\n",
              linecount,index,pb_desc->code);
            return NULL;
          }
          if (first == 0)
            sprintf(buf,"((int *)(sizes_input[%d].d))",index);
          else
            sprintf(buf,"(sizes_input[%d].d)",index);
          return ((char *)strdup(buf));
        }
        case 'O' :             /* IN OUTPUT */
        {
          int index;
          int object_type;

          if (is_number((char *)mnemonic+2) == -1)
          {
            fprintf(stderr,"Invalid mnemonic '%s' at line %d\n",mnemonic,linecount);
            return NULL;
          }
          index = atoi((char *)mnemonic+2);
          if (index >= pb_desc->nb_obj_output)
          {
            fprintf(stderr,
              "Line %d :Only %d objects in output to %s in mnemonic @%s@\n",
              linecount,pb_desc->nb_obj_input,pb_desc->code,mnemonic);
            return NULL;
          }
          object_type = pb_desc->output_obj[index].object_type;
          if (object_type != NETSOLVE_MATRIX)
          {
            fprintf(stderr,
              "Line %d :Object #%d in output to %s does not have a 'descriptor'\n",
              linecount,index,pb_desc->code);
            return NULL;
          }
          if (first == 0)
            sprintf(buf,"((int *)(sizes_output[%d].d))",index);
          else
            sprintf(buf,"(sizes_output[%d].d)",index);
          return ((char *)strdup(buf));
        }
        default :     /* ERROR */
        {
          fprintf(stderr,"Line %d : Invalid mnemonic @%s @\n",
                  linecount,mnemonic);
          return NULL;
        }
      }
    }
    default : /* Should never happen */
    {
      fprintf(stderr,"Invalid mnemonic '%s' at line %d\n",mnemonic,linecount);
      return NULL;
    }
  }

}


/*
 * check_mnemonic_list()
 *
 *   given a problem descriptor and a list of mnemonic, checks if
 *   all problem features are represented and if all mnemonics are
 *   correctly written
 */

#ifdef __STDC__
int check_mnemonic_list(char **mnemonic_list,int format_number)
#else
int check_mnemonic_list(mnemonic_list,format_number)
char **mnemonic_list;
int format_number;
#endif
{
  char *token;
  int i;

  for (i=0;i<pb_desc->nb_obj_input;i++)
  {
    int object_type;
    int status;
    char buf[10];

    object_type = pb_desc->input_obj[i].object_type;

    sprintf(buf,"I%d",i);
    status = lookup(mnemonic_list,buf);
    if (status == -1)
    {
      fprintf(stderr,"Problem : %s - Format %d : One argument should contain 'I%d'\n",
              pb_desc->code,format_number,i);
      return -1;
    }
    if (status == -2)
    {
      fprintf(stderr,"Problem : %s - Format %d : One SINGLE argument should contain 'I%d'\n",
              pb_desc->code,format_number,i);
      return -1;
    }
    switch (object_type)
    {
      case NETSOLVE_MATRIX :
      {
        sprintf(buf,"mI%d",i);
        status = lookup(mnemonic_list,buf);
        if (status == -1)
        {
          fprintf(stderr,"Problem : %s - Format %d : One argument should contain 'mI%d'\n",
                  pb_desc->code,format_number,i);
          return -1;
        }
        if (status == -2)
        {
          fprintf(stderr,"Problem : %s - Format %d : One SINGLE argument should contain 'mI%d'\n",
                  pb_desc->code,format_number,i);
          return -1;
        }
        sprintf(buf,"nI%d",i);
        status = lookup(mnemonic_list,buf);
        if (status == -1)
        {
          fprintf(stderr,"Problem : %s - Format %d : One argument should contain 'nI%d'\n",
                  pb_desc->code,format_number,i);
          return -1;
        }
        if (status == -2)
        {
          fprintf(stderr,"Problem : %s - Format %d : One SINGLE argument should contain 'nI%d'\n",
                  pb_desc->code,format_number,i);
          return -1;
        }
        sprintf(buf,"lI%d",i);
        status = lookup(mnemonic_list,buf);
        if (status == -1)
        {
          fprintf(stderr,"Problem : %s - Format %d : One argument should contain 'lI%d'\n",
                  pb_desc->code,format_number,i);
          return -1;
        }
        if (status == -2)
        {
          fprintf(stderr,"Problem : %s - Format %d : One SINGLE argument should contain 'lI%d'\n",
                  pb_desc->code,format_number,i);
          return -1;
        }
      }
      case NETSOLVE_VECTOR :
      {
        sprintf(buf,"mI%d",i);
        status = lookup(mnemonic_list,buf);
        if (status == -1)
        {
          fprintf(stderr,"Problem : %s - Format %d : One argument should contain 'mI%d'\n",
                  pb_desc->code,format_number,i);
          return -1;
        }
        if (status == -2)
        {
          fprintf(stderr,"Problem : %s - Format %d : One SINGLE argument should contain 'mI%d'\n",
                  pb_desc->code,format_number,i);
          return -1;
        }
      }
      case NETSOLVE_SCALAR :
      {
        break; /* Nothing special here */
      }
      case NETSOLVE_EXTERNAL :
      {
        break; /* Nothing special here */
      }
      case NETSOLVE_FILE :
      {
        break; /* Nothing special here */
      }
    }
  }

  for (i=0;i<pb_desc->nb_obj_output;i++)
  {
    int object_type;
    int status;
    char buf[10];

    object_type = pb_desc->output_obj[i].object_type;

    sprintf(buf,"O%d",i);
    status = lookup(mnemonic_list,buf);
    if (status == -1)
    {
      fprintf(stderr,"Problem : %s - Format %d : One argument should contain 'O%d'\n",
              pb_desc->code,format_number,i);
      return -1;
    }
    if (status == -2)
    {
      fprintf(stderr,"Problem : %s - Format %d : One SINGLE argument should contain 'O%d'\n",
              pb_desc->code,format_number,i);
      return -1;
    }
    switch (object_type)
    {
      case NETSOLVE_MATRIX :
      {
        sprintf(buf,"mO%d",i);
        status = lookup(mnemonic_list,buf);
        if (status == -2)
        {
          fprintf(stderr,"Problem : %s - Format %d : One SINGLE argument should contain 'mO%d'\n",
                  pb_desc->code,format_number,i);
          return -1;
        }
        sprintf(buf,"nO%d",i);
        status = lookup(mnemonic_list,buf);
        if (status == -2)
        {
          fprintf(stderr,"Problem : %s - Format %d : One SINGLE argument should contain 'nO%d'\n",
                  pb_desc->code,format_number,i);
          return -1;
        }
        sprintf(buf,"lO%d",i);
        status = lookup(mnemonic_list,buf);
        if (status == -2)
        {
          fprintf(stderr,"Problem : %s - Format %d : One SINGLE argument should contain 'lO%d'\n",
                  pb_desc->code,format_number,i);
          return -1;
        }
      }
      case NETSOLVE_VECTOR :
      {
        sprintf(buf,"mO%d",i);
        status = lookup(mnemonic_list,buf);
        if (status == -2)
        {
          fprintf(stderr,"Problem : %s - Format %d : One SINGLE argument should contain 'mO%d'\n",
                  pb_desc->code,format_number,i);
          return -1;
        }
      }
      case NETSOLVE_SCALAR :
      {
        break; /* Nothing special here */
      }
      case NETSOLVE_EXTERNAL :
      {
        break; /* Nothing special here */
      }
      case NETSOLVE_FILE :
      {
        break; /* Nothing special here */
      }
    }
  }
        return 1;
}

/*
 * lookup()
 *
 * Returns -1 if the mnemonic is not in the list
 * Returns -2 if the mnemonic is more than once in the list
 */
#ifdef __STDC__
int lookup(char** list, char *string)
#else
int lookup(list,string)
char **list;
char *string;
#endif
{
  char *token;
  int i=0;
  int result=-1;
  char *buf;

  while(list[i] != NULL)
  {

    buf = (char *)strdup(list[i]);

    token = (char *)strtok(buf," \t,");
    while(token != NULL)
    {
      if (strcmp(string,token) == 0)
      {
        if (result == -1)
          result = 1;
        else
          result = -2;
      }
      token = (char *)strtok(NULL," \t,");
    }
    free(buf);
    i++;
  }
  return result;
}

/*
 * check_mnemonic()
 *
 *   return -1 if a mnemonic is incorrectly specified
 *   return 1 if the mnemonic specifie an integer
 *   return 2 if the mnemonic specifies a pointer
 *
 */
#ifdef __STDC__
int check_mnemonic(char* string)
#else
int check_mnemonic(string)
char *string;
#endif
{

  switch(string[0])
  {
    case '?' :  /* Ignored mnemonic */
    {
      return 1;
    }
    case 'I' :
    {
      int index;

      if (is_number((char *)string+1) == -1)
      {
        fprintf(stderr,"'%s' : invalid mnemonic at line %d\n",string,linecount);
        return -1;
      }
      index = atoi((char *)string+1);

      if (index >= pb_desc->nb_obj_input)
      {
        fprintf(stderr,
           "Line %d :Only %d objects in input to %s in mnemonic '%s'\n",
           linecount,pb_desc->nb_obj_input,pb_desc->code,string);
        return -1;
      }
      return 2;
    }
    case 'O' :
    {
      int index;

      if (is_number((char *)string+1) == -1)
      {
        fprintf(stderr,"'%s' : invalid mnemonic at line %d\n",string,linecount);
        return -1;
      }
      index = atoi((char *)string+1);

      if (index >= pb_desc->nb_obj_output)
      {
        fprintf(stderr,
           "Line %d :Only %d objects in output to %s in mnemonic '%s'\n",
           linecount,pb_desc->nb_obj_output,pb_desc->code,string);
        return -1;
      }
      return 2;
    }
    case 'm' :
    {
      switch (string[1])
      {
        case 'I' :            /* In input */
        {
          int index;
          int object_type;

          if (is_number((char *)string+2) == -1)
          {
            fprintf(stderr,"'%s' : invalid mnemonic at line %d\n",string,linecount);
            return -1;
          }
          index = atoi((char *)string+2);
          if (index >= pb_desc->nb_obj_input)
          {
            fprintf(stderr,
              "Line %d :Only %d objects in input to %s in mnemonic '%s'\n",
              linecount,pb_desc->nb_obj_input,pb_desc->code,string);
            return -1;
          }
          object_type = pb_desc->input_obj[index].object_type;
          if ((object_type != NETSOLVE_MATRIX) &&
              (object_type != NETSOLVE_VECTOR))
          {
            fprintf(stderr,
              "Line %d :Object #%d in input to %s does not have a 'm-dimension'\n",
              linecount,index,pb_desc->code);
            return -1;
          }
          return 1;
        }
        case 'O' :             /* IN OUTPUT */
        {
          int index;
          int object_type;

          if (is_number((char *)string+2) == -1)
          {
            fprintf(stderr,"'%s' : invalid mnemonic at line %d\n",string,linecount);
            return -1;
          }
          index = atoi((char *)string+2);
          if (index >= pb_desc->nb_obj_output)
          {
            fprintf(stderr,
              "Line %d :Only %d objects in output to %s in mnemonic '%s'\n",
              linecount,pb_desc->nb_obj_output,pb_desc->code,string);
            return -1;
          }
          object_type = pb_desc->output_obj[index].object_type;
          if ((object_type != NETSOLVE_MATRIX) &&
              (object_type != NETSOLVE_VECTOR))
          {
            fprintf(stderr,
              "Line %d :Object #%d in output to %s does not have a 'm-dimension'\n",
              linecount,index,pb_desc->code);
            return -1;
          }
          return 1;
        }
        default :     /* ERROR */
        {
          fprintf(stderr,"Line %d : Invalid mnemonic '%s'\n",
                  linecount,string);
          return -1;
        }
      }
    }
    case 'n' :
    {
      switch (string[1])
      {
        case 'I' :            /* In input */
        {
          int index;
          int object_type;

          if (is_number((char *)string+2) == -1)
          {
            fprintf(stderr,"'%s' : invalid mnemonic at line %d\n",string,linecount);
            return -1;
          }
          index = atoi((char *)string+2);
          if (index >= pb_desc->nb_obj_input)
          {
            fprintf(stderr,
              "Line %d :Only %d objects in input to %s in mnemonic '%s'\n",
              linecount,pb_desc->nb_obj_input,pb_desc->code,string);
            return -1;
          }
          object_type = pb_desc->input_obj[index].object_type;
          if (object_type != NETSOLVE_MATRIX)
          {
            fprintf(stderr,
              "Line %d :Object #%d in input to %s does not have a 'n-dimension'\n",
              linecount,index,pb_desc->code);
            return -1;
          }
          return 1;
        }
        case 'O' :             /* IN OUTPUT */
        {
          int index;
          int object_type;

          if (is_number((char *)string+2) == -1)
          {
            fprintf(stderr,"'%s' : invalid mnemonic at line %d\n",string,linecount);
            return -1;
          }
          index = atoi((char *)string+2);
          if (index >= pb_desc->nb_obj_output)
          {
            fprintf(stderr,
              "Line %d :Only %d objects in output to %s in mnemonic '%s'\n",
              linecount,pb_desc->nb_obj_output,pb_desc->code,string);
            return -1;
          }
          object_type = pb_desc->output_obj[index].object_type;
          if (object_type != NETSOLVE_MATRIX)
          {
            fprintf(stderr,
              "Line %d :Object #%d in output to %s does not have a 'n-dimension'\n",
              linecount,index,pb_desc->code);
            return -1;
          }
          return 1;
        }
        default :     /* ERROR */
        {
          fprintf(stderr,"Line %d : Invalid mnemonic '%s'\n",
                  linecount,string);
          return -1;
        }
      }
    }
    case 'l' :
    {
      switch (string[1])
      {
        case 'I' :            /* In input */
        {
          int index;
          int object_type;

          if (is_number((char *)string+2) == -1)
          {
            fprintf(stderr,"'%s' : invalid mnemonic at line %d\n",string,linecount);
            return -1;
          }
          index = atoi((char *)string+2);
          if (index >= pb_desc->nb_obj_input)
          {
            fprintf(stderr,
              "Line %d :Only %d objects in input to %s in mnemonic '%s'\n",
              linecount,pb_desc->nb_obj_input,pb_desc->code,string);
            return -1;
          }
          object_type = pb_desc->input_obj[index].object_type;
          if (object_type != NETSOLVE_MATRIX)
          {
            fprintf(stderr,
              "Line %d :Object #%d in input to %s does not have a 'leading dimension'\n",
              linecount,index,pb_desc->code);
            return -1;
          }
          return 1;
        }
        case 'O' :             /* IN OUTPUT */
        {
          int index;
          int object_type;

          if (is_number((char *)string+2) == -1)
          {
            fprintf(stderr,"'%s' : invalid mnemonic at line %d\n",string,linecount);
            return -1;
          }
          index = atoi((char *)string+2);
          if (index >= pb_desc->nb_obj_output)
          {
            fprintf(stderr,
              "Line %d :Only %d objects in output to %s in mnemonic '%s'\n",
              linecount,pb_desc->nb_obj_output,pb_desc->code,string);
            return -1;
          }
          object_type = pb_desc->output_obj[index].object_type;
          if (object_type != NETSOLVE_MATRIX)
          {
            fprintf(stderr,
              "Line %d :Object #%d in output to %s does not have a 'leading dimension'\n",
              linecount,index,pb_desc->code);
            return -1;
          }
          return 1;
        }
        default :     /* ERROR */
        {
          fprintf(stderr,"Line %d : Invalid mnemonic '%s'\n",
                  linecount,string);
          return -1;
        }
      }
    }
    default :
    {
      fprintf(stderr,"'%s' : invalid mnemonic at line %d\n",string,linecount);
      return -1;
    }
  }
}

/*
 * check_expression()
 *
 *  Checks that :
 *    1. The expression is syntacticly correct
 *    2. The mnemonics used have been seen BEFORE
 *
 */

#ifdef __STDC__
int check_expression(char *string,char **mnemonic_list,int mnemonic_index)
#else
int check_expression(string,mnemonic_list,mnemonic_index)
char *string;
char **mnemonic_list;
int mnemonic_index;
#endif
{
  int result;
  char *buf;    /* Sub-expression                         */

  if (strncmp(string,"op(",3) == 0) /* OPERATOR */
  {
    char *op;
    char *arg1,*arg2;

    get_op_arg1_arg2(string,&op,&arg1,&arg2);
    if (op == NULL)
    {
      fprintf(stderr,"Invalid operator in 'op' expression at %d\n",linecount);
      return -1;
    }
    if (arg1 == NULL)
    {
      fprintf(stderr,"Invalid first argument in 'op' expression at %d\n",
              linecount);
      return -1;
    }
    if (arg2 == NULL)
    {
      fprintf(stderr,"Invalid second argument in 'op' expression at %d\n",
              linecount);
      return -1;
    }
    /* Examine the operator */
    if ((strcmp(op,"+") != 0) && 
        (strcmp(op,"-") != 0) &&
        (strcmp(op,"*") != 0))
    {
      fprintf(stderr,"Invalid operator in 'op' expression at %d\n",linecount);
      return -1;
    }
    if (check_expression(arg1,mnemonic_list,mnemonic_index) != 1)
    {
      fprintf(stderr,"Invalid first argument in 'op' expression at %d\n",
              linecount);
      return -1;
    }
    if (check_expression(arg2,mnemonic_list,mnemonic_index) != 1)
    {
      fprintf(stderr,"Invalid second argument in 'op' expression at %d\n",
              linecount);
      return -1;
    }
    free(op);free(arg1);free(arg2);
    return 1;

  }
  else if (strncmp(string,"array(",6) == 0) /* ARRAY ACCESS */
  {
    char *array;
    char *index;
    int number;
    
    get_array_index(string,&array,&index);
    if (array == NULL)
    {
      fprintf(stderr,"Invalid array in 'array' expression at %d\n",linecount);
      return -1;
    }
    if (index == NULL)
    {
      fprintf(stderr,"Invalid index in 'array' expression at %d\n",linecount);
      return -1;
    }
    /* Check the array */
    if (check_mnemonic(array) < 0)
    {
      fprintf(stderr,"Invalid mnemonic '%s' at line %d\n",
                     array,linecount);
      return -1;
    }
    if (check_mnemonic(array) != 2)
    {
      fprintf(stderr,"'%s' should be an array at line %d\n",
                     array,linecount);
      return -1;
    }
    if (array[0] == 'O')
    {
      fprintf(stderr,
              "Only input parameters can be used in expressions at %d\n",
              linecount);
      return -1;
    }
    number = atoi(array+1);
    if ((pb_desc->input_obj[number].data_type != NETSOLVE_I)&&
        (pb_desc->input_obj[number].data_type != NETSOLVE_CHAR))
    {
      fprintf(stderr,"The array accessed at %d should be of",linecount);
      fprintf(stderr," integers or chars\n");
      return -1;
    }
    if (lookup(mnemonic_list,array) ==  -1)
    {
      fprintf(stderr,"First time mnemonic '%s' seen at line %d\n",
                     array,linecount);
      return -1;
    }
    /* -2 is impossible as a returned value of lookup() */
    /* Therefore the array is okay here                 */

    /* Check the index */
    if (check_expression(index,mnemonic_list,mnemonic_index) != 1)
    {
      fprintf(stderr,"'%s' is an invalid index expression at line %d\n",
                     index,linecount);
      return -1;
    }
    /* Everything os OK */
    free(array);free(index);
    return 1;
  }
  else if (strncmp(string,"'",1) == 0) /* CHARACTER CONSTANT */
  {
    if (strlen(string) != 3)
    {
      fprintf(stderr,"Invalid character constant %s at line %d\n",
              string,linecount);
      return -1;
    }
    if (string[2] != '\'')
    {
      fprintf(stderr,"Invalid character constant %s at line %d\n",
              string,linecount);
      return -1;
    }
    return 1;
  }
  else if (strncmp(string,"if(",3) == 0) /* CONDITIONAL */
  {
    char *lhs;
    char *rhs;
    char *ifstmt;
    char *elsestmt;
    char *start = string + 3;
    int lhv,rhv;


    if (get_members(start,&lhs,&rhs,&ifstmt,&elsestmt) == -1)
      return -1;

    if (check_expression(lhs,mnemonic_list,mnemonic_index) != 1)
    {
      fprintf(stderr," The Left Hand side of the expression is invalid on line %d\n",linecount);
      return -1;
    }
    if (check_expression(rhs,mnemonic_list,mnemonic_index) != 1)
    {
      fprintf(stderr," The Right hand side of the  expression is invalid on line %d\n",linecount);
      return -1;
    }
    if (check_expression(ifstmt,mnemonic_list,mnemonic_index) != 1)
    {
      fprintf(stderr," The If clause is invalid on line %d\n",linecount);
      return -1;
    }
    if (check_expression(elsestmt,mnemonic_list,mnemonic_index) != 1)
    {
      fprintf(stderr," The Else clause is invalid on line %d\n",linecount);
      return -1;
    }
    free(lhs);free(rhs);free(ifstmt);free(elsestmt);
    return 1;
  }
  else if (is_number(string) == 0)
  {
    return 1;
  }  
  else /* MNEMONIC : must be integer */
  {
    if (check_mnemonic(string) < 0)
    {
      fprintf(stderr,"Invalid mnemonic '%s' at line %d\n",string,linecount);
      return -1;
    }
    if (check_mnemonic(string) != 1)
    {
      fprintf(stderr,"mnemonic '%s' at line %d is not an integer\n",
              string,linecount);
      return -1;
    }
    if ((string[0] == 'O')||(string[1] == 'O'))
    {
      fprintf(stderr,
              "Only input parameters can be used in expressions at %d\n",
              linecount);
      return -1;
    }
    if (lookup(mnemonic_list,string) == -1)
    {
      fprintf(stderr,"First time mnemonic '%s' seen at line %d\n",
                     string,linecount);
      return -1;
    }
    return 1;
  }
   
}

/*
 * get_op_arg1_arg2()
 *
 * extract 'op','arg1','arg2' from an op(x,y,z) expression.
 *
 */
#ifdef __STDC__
int get_op_arg1_arg2(char *string,char **op,char **arg1,char **arg2)
#else
int get_op_arg1_arg2(string,op,arg1,arg2)
char *string;
char **op;
char **arg1;
char **arg2;
#endif
{
  int nb_parent_seen = 0;
  char *buf;
  char *tmp;
  char *tmp_string;
  
  tmp_string = (char *)strdup(string);

  if (tmp_string[strlen(tmp_string)-1] != ')')
  {
    fprintf(stderr,"Unbalanced parenthesis at %d\n",linecount);
    *op = NULL;
    *arg1 = NULL;
    *arg2 = NULL;
    free(tmp_string);
    return 1;
  }
  tmp_string[strlen(tmp_string)-1] = '\0';

  buf = (char *)strdup(tmp_string+3);
  /* Get the op */
  tmp = (char *)strchr(buf,',');
  if (tmp == NULL)
  {
    *op = NULL;
    *arg1 = NULL;
    *arg2 = NULL;
    free(tmp_string);
    return 1;
  }
  *tmp = '\0';
  *op = (char *)strdup(buf);
  buf = tmp+1;
  tmp = buf;
  while(1)
  {
    if (*tmp == '\0')
    {
      fprintf(stderr,"Unbalanced parenthesis at %d\n",linecount);
      *arg1 = NULL;
      *arg2 = NULL;
      free(tmp_string);
      return 1;
    }
    if (*tmp == '(')
      nb_parent_seen++;
    if (*tmp == ')')
      nb_parent_seen--;
    if (nb_parent_seen <0)
    {
      fprintf(stderr,"Unbalanced parenthesis at %d\n",linecount);
      *arg1 = NULL;
      *arg2 = NULL;
      free(tmp_string);
      return 1;
    }
    if ((*tmp == ',')&&(nb_parent_seen == 0))
      break; 
    tmp++; 
  } 
  *tmp = '\0';
  *arg1 = (char *)strdup(buf);
  buf = tmp+1;
  tmp = buf;
  nb_parent_seen = 0;
  while(1)
  {
    if (*tmp == '\0')
    {
      if (nb_parent_seen != 0)
      {
        fprintf(stderr,"Unbalanced parenthesis at %d\n",linecount);
        *arg2 = NULL;
        free(tmp_string);
        return 1;
      }
      break;
    }
    if (*tmp == '(')
      nb_parent_seen++;
    if (*tmp == ')')
      nb_parent_seen--;
    if (nb_parent_seen <0)
    {
      fprintf(stderr,"Unbalanced parenthesis at %d\n",linecount);
      *arg2 = NULL;
      free(tmp_string);
      return 1;
    }
    if ((*tmp == ',')&&(nb_parent_seen == 0))
      break;
    tmp++;
  }
  *tmp = '\0';
  *arg2 = (char *)strdup(buf);
  buf = tmp+1;
  free(tmp_string);
  return 1;  
}


/*
 * get_array_index()
 *
 * extract 'array','index' from an array(x,y) expression.
 *
 */
#ifdef __STDC__
int get_array_index(char *string,char** array,char** index)
#else
int get_array_index(string,array,index)
char *string;
char **array;
char **index;
#endif
{
  int nb_parent_seen = 0;
  char *buf = NULL;
  char *tmp;
  char *tmp_string;

  tmp_string = (char *)strdup(string);

  if (tmp_string[strlen(tmp_string)-1] != ')')
  {
    fprintf(stderr,"Unbalanced parenthesis at %d\n",linecount);
    *array = NULL;
    *index = NULL;
    return 1;
  }
  tmp_string[strlen(tmp_string)-1] = '\0';

  buf = (char *)strdup(tmp_string+6);
  tmp = buf;
  while(1)
  {
    if (*tmp == '\0')
    {
      fprintf(stderr,"Unbalanced parenthesis at %d\n",linecount);
      *array = NULL;
      *index = NULL;
      return 1;
    }
    if (*tmp == '(')
      nb_parent_seen++;
    if (*tmp == ')')
      nb_parent_seen--;
    if (nb_parent_seen <0)
    {
      fprintf(stderr,"Unbalanced parenthesis at %d\n",linecount);
      *array = NULL;
      *index = NULL;
      return 1;
    }
    if ((*tmp == ',')&&(nb_parent_seen == 0))
      break;
    tmp++;
  }
  *tmp = '\0';
  *array = (char *)strdup(buf);
  buf = tmp+1;
  tmp = buf;
  nb_parent_seen = 0;
  while(1)
  {
    if (*tmp == '\0')
    {
      if (nb_parent_seen != 0)
      {
        fprintf(stderr,"Unbalanced parenthesis at %d\n",linecount);
        *index = NULL;
        return 1;
      }
      break;
    }
    if (*tmp == '(')
      nb_parent_seen++;
    if (*tmp == ')')
      nb_parent_seen--;
    if (nb_parent_seen <0)
    {
      fprintf(stderr,"Unbalanced parenthesis at %d\n",linecount);
      *index = NULL;
      return 1;
    }
    if ((*tmp == ',')&&(nb_parent_seen == 0))
      break;
    tmp++;
  }
  *tmp = '\0';
  *index = (char *)strdup(buf);
  return 1;
}

/*
 * get_members()
 */
#ifdef __STDC__
int get_members(char *string,char **lhs,char **rhs,
              char **ifstmt,char **elsestmt)
#else
int get_members(string,lhs,rhs,ifstmt,elsestmt)
char *string;
char **lhs;
char **rhs;
char **ifstmt;
char **elsestmt;
#endif
{
  int nbpar = 0;
  int length;
  char *start;
  int comma[2];
  int nb=0;
  int i;
  int egal = 0;

  for (i=0;i<2;i++)
    comma[i] = 0;
  
  start = (char *)strdup(string);
  length = strlen(start);
  start[--length] = '\0';
 
  /* Looking for ','s */

  for (i=0;i<length;i++)
  {
    if (start[i] == ',')
      if (nbpar == 0)
      {
        if (nb == 2) 
        {
          fprintf(stderr,"An if expression must be of the form :\n");
          fprintf(stderr," if(<expr1>,<expr2>,<expr3>) at line %d\n",linecount);
          return -1;
        }
        else
          comma[nb++] = i; 
      }
    if (start[i] == '(')
      nbpar++;
    if (start[i] == ')')
      nbpar--;
  } 
  if ((comma[0] == 0)||
      (comma[1] == 0)||
      (comma[0] == comma[1])||
      (nbpar != 0))
  {
    fprintf(stderr,"An if expression must be of the form :\n");
    fprintf(stderr," if(<expr1>,<expr2>,<expr3>) at line %d\n",linecount);
    return -1;
  }

  /* now, we have the 'if' and 'else' */
   
  *(start+comma[0]) = '\0';
  *(start+comma[1]) = '\0';
  *ifstmt   = (char *)strdup(start+comma[0]+1);
  *elsestmt = (char *)strdup(start+comma[1]+1);

  /* Deal with the expression */

  for (i=0;i<comma[0];i++)
  {
    if (start[i] == '=')
      if (nbpar == 0)
        egal = i;
    if (start[i] == '(')
      nbpar++;
    if (start[i] == ')')
      nbpar--;
  }
  if ((nbpar != 0)||(egal == 0))
  {
    fprintf(stderr,"The expression must be of the form :\n");
    fprintf(stderr,"   <expr1> = <expr2> at line %d\n",linecount);
    return -1;
  }
  *(start+egal) = '\0';
  *lhs = (char *)strdup(start);
  *rhs = (char *)strdup(start+egal+1);
  free(start);
  return 1;
}
