#include <stdio.h>
#include <string.h>
#include "global.h" 
/* global.h is from netsolve */

#define WORKINGSIZE 4096
#define DEBUGGING 1
#define HBM_WAIT 5
#define debug(a) if (DEBUGGING) (a)

#ifdef __STDC__
main () 
#else
main ()
#endif
{
   int hbm_sock;
   int pid;
   int client_fd;
   int hbm_port = HBM_INTERFACE_PORT;
   void reap();   

/* Set up the signals to avoid Zombies */
#ifdef  SYSVSIGNAL
   (void)signal(SIGCLD, reap);
#else
   (void)signal(SIGCHLD, reap);
#endif
   

/* Set up socket here; hbm_port is a global constant to be defined elsewhere */

   if ((hbm_sock = establish_socket(&hbm_port)) == -1) {
      perror("HBM collector monitor:  Error on socket open");
      exit(-1);
   }

   listen(hbm_sock, MAXCONN);     /* MAXCONN should be spec'd in global.h */
   while (1) {
   /* Block until contacted.  Params NULL, so we don't know calling machine */
      if ((client_fd = accept(hbm_sock, NULL, NULL)) == -1) /* error */
         perror("HBM collector monitor:  Error accepting connection");
      pid = fork();
      if (pid == -1) {
         perror("HBM collector monitor:  ");
      } else if (pid != 0) {
         close(client_fd);
         /* Parent */
         /* Do nothing for now */
      } else {
         /* Child */
         int problem_pid;
         int problem_status;

      /* Find out what process we're interested in */
         init_recv(client_fd);
         read_int(&problem_pid);
         end_message();

      /* Loop until client_pid process is over */
         while (1) {
            sleep(HBM_WAIT);	/* Wait HBM_WAIT seconds before each checkup */
            problem_status = hbmapi_report(NULL, problem_pid);
            init_send(DATA_XDR, client_fd);
            write_int(&problem_status);
            end_message();
         /* Check problem status.  If it is inactive, don't follow it  */
         /* anymore and exit.  The client side will figure this out by */
         /* itself. */
         /* Currently we are allowing the status to be one of:         */
         /* active, blocked, or overdue. */
            if (problem_status != 1 &&
                problem_status != 0 &&
                problem_status != 2 &&
                problem_status != 3) { 
               exit(1);		
            }
         }
      } /* End of child */
   } /* End main while loop */
}
 

/* Given hostname (not IP address) and pid, returns HBM status of the process. */
/* hostname can be NULL, but this could (unlikely) be problematic if two */
/* monitored processes on different machines had the same PID. */

#ifdef __STDC__
int hbmapi_report(char *hostname, int pid) 
#else
int hbmapi_report(hostname,pid) 
char *hostname;
int pid;
#endif
{
   char *pointer;
   unsigned int client_pid, client_status;
   char client_machine_name[WORKINGSIZE];
   char client_pid_try[WORKINGSIZE];
   char client_status_try[WORKINGSIZE];
   char working[WORKINGSIZE];
   FILE *checkpointfile;
   
   /* Keep trying to open the checkpointfile until successful */
   while ((checkpointfile = fopen("/tmp/hbmdc_checkpoint", "r")) == NULL) {
      sleep(1);
   }
      
   while (fgets(working, WORKINGSIZE, checkpointfile)) {
   /* Check first 12 characters to see if this is a client entry */
      if (strncmp(working, "Client Data:", 12) == 0) {
      /* It is, so get text fields of interest (Otherwise, get next line) */
         pointer = strchr(working, ';'); /* Start of name */
         pointer++;
         strcpy(client_machine_name, pointer);
         pointer = strchr(pointer, ';'); /* Start of something */
         pointer++;
         pointer = strchr(pointer, ';'); /* Start of next thing */
         pointer++;
         pointer = strchr(pointer, ';'); /* Start of PID */
         pointer++;
         strcpy(client_pid_try, pointer);
         pointer = strchr(pointer, ';'); /* Start of status */
         pointer++;
         strcpy(client_status_try, pointer);

      /* Terminate and atoi text fields as necessary */
         *(strchr(client_machine_name, ';')) = '\0';
         *(strchr(client_pid_try, ';')) = '\0';
         *(strchr(client_status_try, ';')) = '\0';
         client_pid = atoi(client_pid_try);
         client_status = atoi(client_status_try);

      /* Check fields.  If they match with input, break loop and return */
         if (hostname) {		/* Make sure hostname exists */
            if (strcmp(hostname, client_machine_name) == 0) 
               if (pid == client_pid) 
                  break;
         } else if (pid == client_pid)
                  break;

      } /* End client entry work if */
   }  /* End loop through file */
   return (client_status);
}

#ifdef __STDC__
void reap(int sig)
#else
void
reap(sig)
        int sig;
#endif
{
        int es = 0;
        int pid;
#ifndef NOWAIT3
#if defined(RUSAGE_SELF)
        struct rusage rus;
#else
        int rus;
#endif
#endif

#ifdef  NOWAIT3
#ifdef  NOWAITPID
        pid = wait(&es);
#else
        pid = waitpid(-1, &es, WNOHANG);
        while (waitpid(-1, &es, WNOHANG) > 0);
#endif
#else   /*NOWAIT3*/
        pid = wait3(&es, WNOHANG,&rus);
        while (wait3(&es, WNOHANG,&rus) > 0);
#endif  /*NOWAIT3*/

#ifdef  SYSVSIGNAL
        (void)signal(SIGCLD, reap);
#endif

}

