/*****************************************************************************
 * file: blacs_routines.c           for Netsolve parallel operation
 *                                  to incorporate ScaLAPACK
 * 31 July 1997      Kim Buckner
 * 15 February 1998   Henri Casanova
 *
 *****************************************************************************
 *  int *desc_int()
 *  int Cnumroc()
 *  void Cpdtran()
 *  int *make_ipiv()
 ****************************************************************************/
#include "server.h"
#include <math.h>

/*
 * this initializes a matrix descriptor based upon the type of 
 * of function is being callled.  currently only supports the 
 * basic type
 */

#ifdef __STDC__
int *desc_init(int rows,int cols,int type)
#else
int *desc_init(rows,cols,type)
int rows;
int cols;
int type;
#endif
{
  int zero=0;
  int *desc;
  char logbuf[80];

  switch(type){
    case 1:
      desc=(int *)calloc(9,sizeof(int));
      if(desc==NULL) {
	if((MYROW==0)&&(MYCOL==0)) {
	  sprintf(logbuf,"Could not allocate a matrix descriptor\n");
	  append_to_log(logbuf);
	  }
#ifdef VIEW
      fprintf(stderr,"Could not allocate a matrix descriptor\n");
#endif
      para_term();
      exit(1);
      }
#ifdef ADD_
      desc[LLD_]=numroc_(&rows,&ROW_BLOCK,&MYROW,&zero,&PROC_ROWS);
#else 
      desc[LLD_]=numroc(&rows,&ROW_BLOCK,&MYROW,&zero,&PROC_ROWS);
#endif

      if(desc[LLD_]<1)
	desc[LLD_]=1;

      desc[M_]=rows;
      desc[N_]=cols;
      desc[DTYPE]=1;
      desc[CTXT_]=MYCONTXT;
      desc[RSRC_]=0;
      desc[CSRC_]=0;
      desc[MB_]=ROW_BLOCK;
      desc[NB_]=COL_BLOCK;

      return desc;
      break;

    default:
#ifdef VIEW
      fprintf(stderr,"Unknown descriptor type\n");
#endif
      return NULL;
  }/* end of switch */
}

/*
 * Cnumroc
 */
#ifdef __STDC__
int Cnumroc(int rows,int row_block, int myrow,int isrcproc, int nprocs)
#else
int Cnumroc(rows,row_block,myrow,isrcproc,nprocs)
int rows;
int row_block; 
int myrow;
int isrcproc; 
int nprocs;
#endif
{
#ifdef ADD_
  return numroc_(&rows,&row_block,&myrow,&isrcproc,&nprocs);
#endif
#ifdef UPCASE
  return NUMROC(&rows,&row_block,&myrow,&isrcproc,&nprocs);
#endif
#ifdef NOCHANGE
  return numeroc(&rows,&row_block,&myrow,&isrcproc,&nprocs);
#endif
}

/*
 * Cpdtran
 */
#ifdef __STDC__
void Cpdtran(int m,int n,double alpha,double *a,int ia,int ja,
              int *desca,double beta,double *c,int ic,int jc, int *descc)
#else
void Cpdtran(m,n,alpha,a,ia,ja,desca,beta,c,ic,jc,descc)
int m;
int n;
double alpha;
double *a;
int ia;
int ja;
int *desca;
double beta;
double *c;
int ic;
int jc;
int *descc;
#endif
{
#ifdef ADD_
  pdtran_(&m,&n,&alpha,a,&ia,&ja,desca,&beta,c,&ic,&jc,descc);
#endif
#ifdef UPCASE
  PDTRAN(&m,&n,&alpha,a,&ia,&ja,desca,&beta,c,&ic,&jc,descc);
#endif
#ifdef NOCHANGE
  pdtran(&m,&n,&alpha,a,&ia,&ja,desca,&beta,c,&ic,&jc,descc);
#endif
}

/*
 * clean up routine for the parallel machine
 */
#ifdef __STDC__
int para_term()
#else
int para_term()
#endif
{
   Cblacs_gridexit(MYCONTXT);
   Cblacs_exit(0);
   return 0;
}

#ifdef __STDC__
int *make_ipiv(int size)
#else
int *make_ipiv(size)
int size;
#endif
{
  int rtn;
  int *tmp;

  rtn=Cnumroc(size,ROW_BLOCK,MYROW,0,PROC_ROWS);

  rtn+=ROW_BLOCK;
  tmp=(int *)calloc(rtn,sizeof(int));
  return tmp;
}
