/*****************************************************************/
/*      compute.c                                                */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  compute()                                                    */
/*****************************************************************/

#include "server.h"
#include "server_prot.h"
#include "shared_prot.h"

/*
 *  Compute is called within the service process.
 *  It calls linear_algebra() to actually solve the problem.
 * 
 *  The end_com() is done by the calling routine : main() (service.c)
 */

#ifdef __STDC__
int compute(problem_desc *pb_desc,input_sizes *sizes_input,
            input_sizes *sizes_output, void **data_input,
            void **data_output, int sock,
            int mode, char *condordir,int client_major)
#else
int compute(pb_desc,sizes_input,sizes_output,
            data_input,data_output,sock,mode,condordir,client_major)
problem_desc *pb_desc;
input_sizes *sizes_input;
input_sizes *sizes_output;
void **data_input;
void **data_output;
int sock;
int mode;
char *condordir;
int client_major;
#endif
{
  int status,i,type;
  int total_size;

  char command_line[128];
  char stdoutfilename[128];
  FILE *file;
  char filename[128];
  int  stdoutlength;
  int  stdoutfile=-1;
  char *stdoutbuf;

  /*
   * Call the underlying Linear Algebra Software
   */

  /*
   * Redirect the stdout
   */
  
  /* File names */
#ifdef ISCONDOR
  sprintf(stdoutfilename,"%s/NetSolve_stdout",condordir);
#else
  sprintf(stdoutfilename,"/tmp/NetSolve_stdout.%d",getpid());
#endif

  /* Opening the files */
  if ((stdoutfile = open(stdoutfilename,O_RDWR|O_CREAT,0600)) < 0)
  {
#ifdef VIEW
    fprintf(stderr,"Warning : Could not open file '%s'\n",stdoutfilename);
#endif
  }
   
  /* Redirecting the files */
  close(1);
  dup(stdoutfile);  

  status = linear_algebra(pb_desc,sizes_input,sizes_output,
                          data_input,data_output);


  fflush(stdout);

  /* Closing the files */
  close(stdoutfile);

  /* Getting the sizes */
  if (stdoutfile >= 0)
  {
    struct stat st;

    if (stat(stdoutfilename,&st) != -1)
    {
      int cc; 
      stdoutfile = open(stdoutfilename,O_RDONLY);
  
      stdoutlength = (int)(st.st_size)+1;
      stdoutbuf = (char *)malloc(sizeof(char)*(stdoutlength));
      cc = read(stdoutfile,stdoutbuf,stdoutlength-1);
      stdoutbuf[stdoutlength-1] = '\0';
      close(stdoutfile);
    }
    else
    {
      stdoutbuf = (char *)strdup("Server Error : No stdout available");
      stdoutlength = strlen(stdoutbuf)+1;
    }
  }
  else
  {
    stdoutbuf = (char *)strdup("Server Error : No stdout available");
    stdoutlength = strlen(stdoutbuf)+1;
  }

  /* Deleting the files */

  unlink(stdoutfilename);

  switch(status)
  {
    case BAD_DIMENSION :
      type = DM_MISMATCH;
      break;
    case BAD_VALUES :
      type = BAD_VALUES;
      break;
    case FAILURE :
      type = SV_FAILURE;
      break;
  }

  if (status < 0)
  {
#ifdef ISCONDOR
    sprintf(filename,"%s/type",condordir);
    file = fopen(filename,"w");
    if (file == NULL)
    {
#ifdef DEBUG
      fprintf(stderr,"Cannot create file '%s'\n",filename);
#endif
      create_done_file(condordir,-1);
      exit(0);
    }
    if (fwrite(&type,sizeof(int),1,file) != 1)
    {
      create_done_file(condordir,-1);
      exit(0);
    }
    fclose(file);
    sprintf(filename,"%s/stdout",condordir);
    file = fopen(filename,"w");
    if (file == NULL)
    {
#ifdef DEBUG
      fprintf(stderr,"Cannot create file '%s'\n",filename);
#endif
      create_done_file(condordir,-1);
      exit(0);
    }
    if (fwrite(&stdoutlength,sizeof(int),1,file) != 1)
    {
      create_done_file(condordir,-1);
      exit(0);
    }
    if (fwrite(stdoutbuf,sizeof(char),stdoutlength,file) != stdoutlength)
    {
      create_done_file(condordir,-1);
      exit(0);
    }
    fclose(file);
#else
/*
    init_send(mode,sock);
*/
    write_int(&type);
    write_int(&stdoutlength);
    write_char_array(stdoutbuf,stdoutlength);
    end_message();
    end_com();
#endif
    free(stdoutbuf);
    return -1;
  }

  /* Send back the solution */

  status = SOLUTION;

#ifdef ISCONDOR
  sprintf(filename,"%s/type",condordir);
  file = fopen(filename,"w");
  if (file == NULL)
  {
#ifdef DEBUG
    fprintf(stderr,"Cannot create file '%s'\n",filename);
#endif
    create_done_file(condordir,-1);
    exit(0);
  }
  if (fwrite(&status,sizeof(int),1,file) != 1)
  {
    create_done_file(condordir,-1);
    exit(0);
  }
  fclose(file);
  sprintf(filename,"%s/stdout",condordir);
  file = fopen(filename,"w");
  if (file == NULL)
  {
#ifdef DEBUG
    fprintf(stderr,"Cannot create file '%s'\n",filename);
#endif
    create_done_file(condordir,-1);
    exit(0);
  }
  if (fwrite(&stdoutlength,sizeof(int),1,file) != 1)
  {
    create_done_file(condordir,-1);
    exit(0);
  }

  if (fwrite(stdoutbuf,sizeof(char),stdoutlength,file) != stdoutlength)
  {
    create_done_file(condordir,-1);
    exit(0);
  }
  fclose(file);
#else
  init_send(mode,sock);
  write_int(&status);
  write_int(&stdoutlength);
  write_char_array(stdoutbuf,stdoutlength);
#endif

#ifdef DEBUG
  fprintf(stderr,"Sending back the result of the computation\n");
#endif


  /* Re-transposing if necessary */
  if (client_major != pb_desc->maj)
  {
    for (i=0;i<pb_desc->nb_obj_output;i++)
    {
      if (pb_desc->output_obj[i].object_type == NETSOLVE_MATRIX)
      {
        /* Need to transpose it */
        transpose_matrix(pb_desc->maj,
                         pb_desc->output_obj[i].data_type,&(data_output[i]),
                         sizes_output[i].m,sizes_output[i].n);
      }
    }
  }

  for (i=0;i<pb_desc->nb_obj_output;i++)
  {
    if (condordir)
    {
      sprintf(filename,"%s/output%d",condordir,i);
      file = fopen(filename,"w");
      if (file == NULL)
      {  
#ifdef DEBUG
        fprintf(stderr,"Unable to create file '%s'\n",filename);
#endif
        create_done_file(condordir,-1);
        exit(0);
      }
    }
    switch(pb_desc->output_obj[i].object_type)
    {
      case NETSOLVE_MATRIX :
      {
#ifdef DEBUG
        fprintf(stderr,"MATRIX %dx%d\n",sizes_output[i].m,sizes_output[i].n);
#endif

#ifdef ISCONDOR
        if (fwrite(&(client_major),sizeof(int),1,file) != 1)
        {
          create_done_file(condordir,-1);
          exit(0);
        }
        if (fwrite(&(sizes_output[i].m),sizeof(int),1,file) != 1)
        {
          create_done_file(condordir,-1);
          exit(0);
        }
        if (fwrite(&(sizes_output[i].n),sizeof(int),1,file) != 1)
        {
          create_done_file(condordir,-1);
          exit(0);
        }
#else
        write_int(&(pb_desc->output_obj[i].object_type));
        write_int(&(pb_desc->output_obj[i].data_type));
        write_int(&(client_major));
        write_int(&(sizes_output[i].m));
        write_int(&(sizes_output[i].n));
#endif
        total_size = (sizes_output[i].m)*(sizes_output[i].n);
        break;
      }
      case NETSOLVE_VECTOR :
      {
#ifdef DEBUG
        fprintf(stderr,"VECTOR %d\n",sizes_output[i].m);
#endif
#ifdef ISCONDOR
        if (fwrite(&(sizes_output[i].m),sizeof(int),1,file) != 1)
        {
          create_done_file(condordir,-1);
          exit(0);
        }
#else
        write_int(&(pb_desc->output_obj[i].object_type));
        write_int(&(pb_desc->output_obj[i].data_type));
        write_int(&(sizes_output[i].m));
#endif
        total_size = (sizes_output[i].m);
        sizes_output[i].n = 1;
        break;
      }
      case NETSOLVE_SCALAR :
      {
#ifdef DEBUG
        fprintf(stderr,"SCALAR\n");
#endif
#ifdef ISCONDOR
#else
        write_int(&(pb_desc->output_obj[i].object_type));
        write_int(&(pb_desc->output_obj[i].data_type));
#endif
        total_size = 1;
        sizes_output[i].m = 1;
        sizes_output[i].n = 1;
        break;
      }
      case NETSOLVE_FILE:
      {
        struct stat st;
#ifdef DEBUG
        fprintf(stderr,"FILE\n");
#endif
#ifdef ISCONDOR
#else
        write_int(&(pb_desc->output_obj[i].object_type));
        write_int(&(pb_desc->output_obj[i].data_type));
#endif
        if (stat(data_output[i],&st) == -1)
        {
          fprintf(stderr,"Impossible to asccess file '%s'\n",data_output[i]);
          break;
        }
        sizes_output[i].m = (int)(st.st_size)-1;
        sizes_output[i].n = 1;
        total_size = sizes_output[i].m;
        write_int(&(sizes_output[i].m));
        break;
      }
      default :
      {
#ifdef DEBUG
        fprintf(stderr,"Invalid Output object structure ??\n");
#endif
        return -1;
      }
    }
    switch(pb_desc->output_obj[i].data_type)
    {
      case NETSOLVE_CHAR :
      {
#ifdef ISCONDOR
        fwrite(data_output[i],sizeof(char),total_size,file);
#else
        write_char_array(data_output[i],total_size);
#endif
        break;
      }
      case NETSOLVE_ASCII :
      {
#ifdef ISCONDOR
        int cc;
        char buffer[256];
        FILE *f;

        f = fopen(data_output[i],"r");
        if (f == NULL)
        {
          fprintf(stderr,"Impossible to open file '%s'\n",data_output[i]);
          break;
        }
        while ((cc = fread(buffer,1,256,f)) != 0)
          fwrite(buffer,sizeof(char),cc,file);
#else
        int fd;
        fd = open(data_output[i],O_RDONLY);
        if (fd < 0)
        {
          fprintf(stderr,"Impossible to open file '%s'\n",data_output[i]);
          break;
        }
        fprintf(stderr,"Calling write_char_array_file\n");
        write_char_array_file(fd,total_size);
        close(fd);
#endif
        break;
      }
      case NETSOLVE_B :
      {
#ifdef ISCONDOR
        fwrite(data_output[i],sizeof(char),total_size,file);
#else
        int kk, num_vec, vec_len;

        if((pb_desc->output_obj[i].object_type == NETSOLVE_MATRIX) &&
           (client_major == ROW_MAJOR))
        {
          num_vec = sizes_output[i].m;
          vec_len = sizes_output[i].n;
        }
        else
        {
          num_vec = sizes_output[i].n;
          vec_len = sizes_output[i].m;
        }

        for (kk=0;kk<num_vec;kk++)
        {
          write_byte_array((char *)((char *)(data_output[i])+
                   kk*(vec_len)),vec_len);
        }
#endif
        break;
      }
      case NETSOLVE_I :
      {
#ifdef ISCONDOR
        fwrite(data_output[i],sizeof(int),total_size,file);
#else
        write_int_array(data_output[i],total_size);
#endif
        break;
      }
      case NETSOLVE_S :
      {
#ifdef ISCONDOR
        fwrite(data_output[i],sizeof(float),total_size,file);
#else
        write_single_array(data_output[i],total_size);
#endif
        break;
      }
      case NETSOLVE_D :
      {
#ifdef ISCONDOR
        fwrite(data_output[i],sizeof(double),total_size,file);
#else
        write_double_array((double *)(data_output[i]),total_size);
#endif
        break;
      }
      case NETSOLVE_C :
      {
#ifdef ISCONDOR
        fwrite(data_output[i],sizeof(scomplex),total_size,file);
#else
        write_scomplex_array(data_output[i],total_size);
#endif
        break;
      }
      case NETSOLVE_Z :
      {
#ifdef ISCONDOR
        fwrite(data_output[i],sizeof(dcomplex),total_size,file);
#else
        write_dcomplex_array(data_output[i],total_size);
#endif
        break;
      }
      default :
      {
#ifdef DEBUG
        fprintf(stderr,"Invalid Output object type ??\n");
#endif
        return -1;
      }
    }
    if (condordir)
      fclose(file);
  }
  free(stdoutbuf);
#ifndef ISCONDOR
  end_message();
#endif
  return 1;
}
