/*****************************************************************/
/*      pcompute.c                                               */
/*      Version 1.0                      Kim Buckner             */
/*      Version 1.1                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  pcompute()                                                   */
/*****************************************************************/

#include "server.h"
#include "server_prot.h"
#include "shared_prot.h"

/*
 *  Compute is called within the service process.
 *  It calls linear_algebra() to actually solve the problem.
 * 
 *  The end_com() is done by the calling routine : main() (service.c)
 */

#ifdef __STDC__
int pcompute(problem_desc *pb_desc,input_sizes *sizes_input,
            input_sizes *sizes_output, void **data_input,
            void **data_output, int sock,
            int mode, char *condordir,int client_major)
#else
int pcompute(pb_desc,sizes_input,sizes_output,
            data_input,data_output,sock,mode,condordir,client_major)
problem_desc *pb_desc;
input_sizes *sizes_input;
input_sizes *sizes_output;
void **data_input;
void **data_output;
int sock;
int mode;
char *condordir;
int client_major;
#endif
{
  int status,i,type;
  int total_size;

  int zero = 0;
  int one = 1;
  double *work;

  char command_line[128];
  char stdoutfilename[128];
  FILE *file;
  char filename[128];
  int  stdoutlength;
  int  stdoutfile=-1;
  char *stdoutbuf;
  time_t date_start,date_end;

  /*
   * Redirect the stdout
   */
  
  /* File names */
  if ((MYROW == 0)&&(MYCOL == 0))
  {
    sprintf(stdoutfilename,"%s/NetSolve_stdout",condordir);

    /* Opening the files */
    if ((stdoutfile = open(stdoutfilename,O_RDWR|O_CREAT,0600)) < 0)
    {
#ifdef VIEW
      fprintf(stderr,"Warning : Could not open file '%s'\n",stdoutfilename);
#endif
    }
     
    /* Redirecting the files */
    close(1);
    dup(stdoutfile);  
  }

  status = linear_algebra(pb_desc,sizes_input,sizes_output,
                          data_input,data_output);

  if ((MYROW==0)&&(MYCOL==0))
  {
    fflush(stdout);
  
    /* Closing the files */
    close(stdoutfile);
  
    /* Getting the sizes */
    if (stdoutfile >= 0)
    {
      struct stat st;
  
      if (stat(stdoutfilename,&st) != -1)
      {
        int cc; 
        stdoutfile = open(stdoutfilename,O_RDONLY);
    
        stdoutlength = (int)(st.st_size)+1;
        stdoutbuf = (char *)malloc(sizeof(char)*(stdoutlength));
        cc = read(stdoutfile,stdoutbuf,stdoutlength-1);
        stdoutbuf[stdoutlength-1] = '\0';
        close(stdoutfile);
      }
      else
      {
        stdoutbuf = (char *)strdup("Server Error : No stdout available");
        stdoutlength = strlen(stdoutbuf)+1;
      }
    }
    else
    {
      stdoutbuf = (char *)strdup("Server Error : No stdout available");
      stdoutlength = strlen(stdoutbuf)+1;
    }
  
    /* Deleting the files */
  
    unlink(stdoutfilename);
  
  }
  if (status < 0)
  {
    if ((MYROW != 0)||(MYCOL != 0))
      return -1;
  }
 
  if ((MYROW==0)&&(MYCOL==0))
  {
    switch(status)
    {
    case BAD_DIMENSION :
      type = DM_MISMATCH;
      break;
    case BAD_VALUES :
      type = BAD_VALUES;
      break;
    case FAILURE :
      type = SV_FAILURE;
      break;
    }

    if (status < 0)
    {
      sprintf(filename,"%s/type",condordir);
      file = fopen(filename,"w");
      if (file == NULL)
      {
        fprintf(stderr,"Cannot create file '%s'\n",filename);
        create_done_file(condordir,-1);
        exit(0);
      }
      if (fwrite(&type,sizeof(int),1,file) != 1)
      {
        create_done_file(condordir,-1);
        exit(0);
      }
      fclose(file);
      sprintf(filename,"%s/stdout",condordir);
      file = fopen(filename,"w");
      if (file == NULL)
      {
        fprintf(stderr,"Cannot create file '%s'\n",filename);
        create_done_file(condordir,-1);
        exit(0);
      }
      if (fwrite(&stdoutlength,sizeof(int),1,file) != 1)
      {
        create_done_file(condordir,-1);
        exit(0);
      }
      if (fwrite(stdoutbuf,sizeof(char),stdoutlength,file) != stdoutlength)
      {
        create_done_file(condordir,-1);
        exit(0);
      }
      fclose(file);
      free(stdoutbuf);
      return -1;
    }
    status = SOLUTION;

    sprintf(filename,"%s/type",condordir);
    file = fopen(filename,"w");
    if (file == NULL)
    {
#ifdef DEBUG
      fprintf(stderr,"Cannot create file '%s'\n",filename);
#endif
      create_done_file(condordir,-1);
      exit(0);
    }
    if (fwrite(&status,sizeof(int),1,file) != 1)
    {
      create_done_file(condordir,-1);
      exit(0);
    }
    fclose(file);
    sprintf(filename,"%s/stdout",condordir);
    file = fopen(filename,"w");
    if (file == NULL)
    {
#ifdef DEBUG
      fprintf(stderr,"Cannot create file '%s'\n",filename);
#endif
      create_done_file(condordir,-1);
      exit(0);
    }
    if (fwrite(&stdoutlength,sizeof(int),1,file) != 1)
    {
      create_done_file(condordir,-1);
      exit(0);
    }
  
    if (fwrite(stdoutbuf,sizeof(char),stdoutlength,file) != stdoutlength)
    {
      create_done_file(condordir,-1);
      exit(0);
    }
    fclose(file);
  }

  /* Re-transposing if necessary */
  if (client_major != pb_desc->maj)
  {
#ifdef VIEW
    if ((MYROW==0)&&(MYCOL==0))
      fprintf(stderr,"Re=majoring the matrices\n");
#endif
    for (i=0;i<pb_desc->nb_obj_output;i++)
    {
      if (pb_desc->output_obj[i].object_type == NETSOLVE_MATRIX)
      {
        /* Need to transpose it */
        Cpdtran(sizes_output[i].m,sizes_output[i].n,
           1.0,(double *)(data_output[i]),1,1,
           sizes_output[i].d,
           0.0,(double *)(data_output[i]),1,1,
           sizes_output[i].d);
      }
    }
  }


#ifdef DEBUG
  if ((MYROW==0)&&(MYCOL==0))
    fprintf(stderr,"Sending back the result of the computation\n");
#endif

  for (i=0;i<pb_desc->nb_obj_output;i++)
  {
    sprintf(filename,"%s/output%d",condordir,i);
    file = fopen(filename,"w");
    if (file == NULL)
    {  
#ifdef DEBUG
      fprintf(stderr,"Unable to create file '%s'\n",filename);
#endif
      create_done_file(condordir,-1);
      exit(0);
    }
    switch(pb_desc->output_obj[i].object_type)
    {
      case NETSOLVE_MATRIX :
      {
        if ((MYROW==0)&&(MYCOL==0))
        {
#ifdef DEBUG
          fprintf(stderr,"MATRIX %dx%d\n",sizes_output[i].m,sizes_output[i].n);
#endif

          if (fwrite(&(client_major),sizeof(int),1,file) != 1)
          {
            create_done_file(condordir,-1);
            exit(0);
          }
          if (fwrite(&(sizes_output[i].m),sizeof(int),1,file) != 1)
          {
            create_done_file(condordir,-1);
            exit(0);
          }
          if (fwrite(&(sizes_output[i].n),sizeof(int),1,file) != 1)
          {
            create_done_file(condordir,-1);
            exit(0);
          }
        } 
        total_size = (sizes_output[i].m)*(sizes_output[i].n);
        break;
      }
      case NETSOLVE_VECTOR :
      {
        if ((MYROW==0)&&(MYCOL==0))
        {
#ifdef DEBUG
          fprintf(stderr,"VECTOR %d\n",sizes_output[i].m);
#endif
          if (fwrite(&(sizes_output[i].m),sizeof(int),1,file) != 1)
          {
            create_done_file(condordir,-1);
            exit(0);
          }
        }
        total_size = (sizes_output[i].m);
        sizes_output[i].n = 1;
        break;
      }
      case NETSOLVE_SCALAR :
      {
        if ((MYROW==0)&&(MYCOL==0))
        {
#ifdef DEBUG
          fprintf(stderr,"SCALAR\n");
#endif
        }
        total_size = 1;
        sizes_output[i].m = 1;
        sizes_output[i].n = 1;
        break;
      }
      default :
      {
        if ((MYROW==0)&&(MYCOL==0))
        {
#ifdef DEBUG
          fprintf(stderr,"Invalid Output object structure ??\n");
#endif
        }
        return -1;
      }
    }
    switch(pb_desc->output_obj[i].data_type)
    {
      case NETSOLVE_D : 
      {
        if (pb_desc->output_obj[i].object_type == NETSOLVE_MATRIX)
             /* 2-D BLOCK CYCLIC DISTRIBUTION !!! */
        {
#ifdef VIEW
          if ((MYROW == 0)&&(MYCOL==0))
            fprintf(stderr,"opening file '%s'\n",filename);
#endif
          globalfile = file;
          work = (double *)malloc(sizeof(double)*(sizes_output[i].d)[MB_]);
#ifdef ADD_
          pdlawrite_(&(sizes_output[i].m),
                     &(sizes_output[i].n),
                     data_output[i],
                     &one,&one,sizes_output[i].d,
                     &zero,&zero,work);
#endif
#ifdef UPCASE
          PDLAWRITE(&(sizes_output[i].m),
                     &(sizes_output[i].n),
                     data_output[i],
                     &one,&one,sizes_output[i].d,
                     &zero,&zero,work);
#endif
#ifdef NOCHANGE
          pdlawrite(&(sizes_output[i].m),
                     &(sizes_output[i].n),
                     data_output[i],
                     &one,&one,sizes_output[i].d,
                     &zero,&zero,work);
#endif
          fclose(globalfile);
        }
        else
        {  /* NOT 2-D BLOCK CYCLIC */
          if ((MYROW==0)&&(MYCOL==0))
            fwrite(data_output[i],sizeof(double),total_size,file);
        }
        break;
      }
      case NETSOLVE_I :
      {
        if ((MYROW == 0)&&(MYCOL==0))
          fwrite(data_output[i],sizeof(int),total_size,file);
        fclose(file);
      }
      case NETSOLVE_CHAR :
      {
        if ((MYROW == 0)&&(MYCOL==0))
          fwrite(data_output[i],sizeof(char),total_size,file);
        fclose(file);
      }
      default : 
        break;
    }
  }
  free(stdoutbuf);
  return 1;
}
