#include "server.h"

int pdlaread( FILE *file, double *a, int *desca, int irread, 
               int icread, double *work )
{
/*
 *  -- ScaLAPACK example --
 *     University of Tennessee, Knoxville, Oak Ridge National Laboratory,
 *     and University of California, Berkeley.
 * 
 *     written by Antoine Petitet, August 1995 (petitet@cs.utk.edu)
 *     painfully converted to C by Henri Casanova, February 1998
 *
 *  Purpose
 *  =======
 *
 *  PDLAREAD reads from a file a matrix and distribute
 *  it to the process grid.
 *
 *  Only the process of coordinates {irread, icread} reads the file.
 *
 *  work must be of size >= mb_ = desca( mb_ ).
 *
 *  =====================================================================
 */

 /*     .. Parameters .. */
  int nin = 11;
  int block_cyclic_2d = 1;

 /*     .. Local Scalars .. */
  int h, i, ib, ictxt, icurcol, icurrow, ii, j, jb;
  int jj, k, lda, m, mycol, myrow, n, npcol, nprow;

 /*     .. Local Arrays .. */
  int iwork[2];

 /*     .. External Subroutines ..*/
  extern Cblacs_gridinfo();
  extern infog2l();
  extern Cdgerv2d();
  extern Cdgesd2d();
  extern Cigebs2d();
  extern Cigebr2d();

 /*     Get grid parameters */

  ictxt = desca[CTXT_];
  Cblacs_gridinfo(ictxt, &nprow, &npcol, &myrow, &mycol);

  if ((myrow==irread)&&(mycol==icread))
  {
    int dummy;

    fread(&dummy,sizeof(int),1,file);
    fread(&(iwork[0]),sizeof(int),1,file);
    fread(&(iwork[1]),sizeof(int),1,file);
    Cigebs2d(ictxt,"All"," ",2,1,iwork,2);
  }
  else
  {
    Cigebr2d(ictxt,"All"," ",2,1,iwork,2,irread,icread);
  }
 
  m = iwork[0];
  n = iwork[1];

  if ((m <=0) || (n <= 0))
    return -1;

  if ((m>desca[M_]) || (n > desca[N_]))
  {
    if ((myrow==0)&&(mycol==0))
    {
      fprintf(stderr,"PDLAREAD: Matrix too big to fit in\n");
      fprintf(stderr,"Abort....\n");
    }
    return -1;
  }

  ii = 1; jj = 1;
  icurrow = desca[RSRC_];
  icurcol = desca[CSRC_];
  lda = desca[LLD_];

 /* Loop over column blocks */

  for (j=0;j<n;j+=desca[NB_])
  {
    jb = MIN(desca[NB_],n-j+1);
    for (h=0;h<=jb-1;h++)    /* ???? C/FORTRAN */
    {
      /* Loop over block of rows */
      for (i=0;i<m;i+=desca[MB_])
      {
        ib = MIN(desca[MB_],m-i+1);
        if ((icurrow == irread)&&(icurcol == icread))
        {
          if ((myrow == irread)&&(mycol == icread))
          {
            for (k=0;k<=ib-1;k++)  /* ???? C/FORTRAN */
            {
              fread(&(a[ii+k+(jj+h-1)*lda]),sizeof(double),1,file);
            }
          }
        }
        else
        {
          if ((myrow==icurrow)&&(mycol==icurcol))
          {
            Cdgerv2d(ictxt,ib,1,&(a[ii+(jj+h-1)*lda]),lda,irread,icread);
          }
          else if ((myrow == irread)&&(mycol == icread))
          {
            for (k=0;k<ib;k++)
              fread(&(work[k]),sizeof(double),1,file);
            Cdgesd2d(ictxt,ib,1,work,desca[MB_],icurrow,icurcol);
          }
        }
        if (myrow == icurrow)
          ii = ii + ib;
        icurrow = (icurrow+1) % nprow;
      }
      ii = 1;
      icurrow = desca[RSRC_];
    }
    if (mycol == icurcol)
      jj = jj + jb;
    icurcol = (icurcol+1)%npcol;
  }

  if ((myrow == irread)&&(mycol == icread))
    fclose(file);
  
  return 1;
}
