/*****************************************************************/
/*      pservice.c                                               */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  main()                                                       */
/*  create_done_file()                                           */
/*****************************************************************/

#include "server.h"
#include "server_prot.h"
#include "shared_prot.h"

/*
 * LINUX :
 *
 *  Faking a Fortran-like main routine for Linux !!!
 */
#ifdef PLAT_LINUX
MAIN__(){}
#endif

/*
 * service 
 *
 *  Process in charge with a problem solve
 */

#ifdef __STDC__
int main(int argc,char **argv)
#else
int main(argc,argv)
int argc;
char **argv;
#endif
{
  char hbmhost[256];
  int zero = 0;
  char HOSTNAME[256];
  double *work;
  int mypnum,rtn,j;
  int i,dummy;
  int local_matrix_size;
  int total_size;
  int mode;
  int sock;
  problem_desc pb_desc;
  input_sizes *sizes_input;
  input_sizes *sizes_output;
  void **data_input;
  void **data_output;
  FILE *file;
  char filename[128];
  char command_line[256];
  char logbuf[256];
  time_t current;
  char *dirname;
  int *majors;
  int client_major;
  int index_param=1;

#if (PARALLEL==PVM)

  temp=(char *)getenv("PVM_ROOT");
  if(temp==NULL){
    fprintf(stderr,"PVM_ROOT not set\n");
    para_term();
    exit(0);
    }
  sprintf(wd,"%s/bin/",temp);
  temp=(char *)getenv("PVM_ARCH");
  if(temp==NULL){
    fprintf(stderr,"PVM_ARCH not set\n");
    para_term();
    exit(0);
    }
  strcat(wd,temp);
  if (chdir(wd) == -1)
  {
    perror("chdir");
    para_term();
    exit(0);
  }
#endif

#if (PARALLEL==MPI)
  /* Calling MPI_Init() */
  MPI_Init(&argc,&argv);
  argv = &(argv[1]);
  argc--;
#endif

  gethostname(HOSTNAME,256);
  Cblacs_pinfo(&mypnum,&rtn);

#if (PARALLEL==PVM)
  {
    int narch, nhosts;
    struct pvmhostinfo **hostp=NULL;
    int rtnpvm_config;

    rtnpvm_config=pvm_config(&nhosts,&narch,hostp);
    if(rtnpvm_config < 0) {
      fprintf(stderr,"pvmd not responding\n");
      para_term();
      exit(1);
      }
    NUM_PROCS=nhosts;
    /* create  the blacs_setup.dat file and write my name to it */

    if (mypnum == 0)
    {
      temp=(char *)getenv("PVM_ROOT");
      if(temp==NULL){
        fprintf(stderr,"PVM_ROOT not set\n");
        para_term();
        exit(1);
        }
      sprintf(blacs_file,"%s/bin/",temp);
      temp=(char *)getenv("PVM_ARCH");
      if(temp==NULL){
        fprintf(stderr,"PVM_ARCH not set\n");
        para_term();
        exit(1);
        }
      strcat(blacs_file,temp);
      strcat(blacs_file,"/blacs_setup.dat");
      tmp_file=fopen(blacs_file,"w");
      if(tmp_file==NULL) {
        fprintf(stderr,"Could not create the blacs_setup.dat file\n");
        para_term();
        exit(1);
        }
      fprintf(stderr,"argv[argc-1 (%d)] = %s\n",
            argc-1,argv[argc-1]);
      fprintf(tmp_file,argv[argc-1]);
      fclose(tmp_file);
    }

    Cblacs_setup(&mypnum,&NUM_PROCS);
  }
#endif

  ROW_BLOCK=COL_BLOCK=40;
  PROC_ROWS=2;
  PROC_COLS=4;
  NUM_PROCS=PROC_ROWS*PROC_COLS;


  Cblacs_get(-1,0,&MYCONTXT);
  Cblacs_gridinit(&MYCONTXT, "R", PROC_ROWS, PROC_COLS);
  Cblacs_gridinfo(MYCONTXT,&i,&j,&MYROW,&MYCOL);

  if (MYROW == -1)
  {
    para_term();
    exit(0);
  }

#ifdef VIEW
  if ((MYROW==0)&&(MYCOL==0))
    fprintf(stderr,"BLACS grid initialized: %d x %d\n",PROC_ROWS,PROC_COLS);
#endif

  strcpy(hbmhost,argv[index_param++]);
  strcpy(pb_desc.code,argv[index_param++]);
  pb_desc.nb_obj_input = atoi(argv[index_param++]);
  if ((pb_desc.input_obj = 
           (object *)calloc(pb_desc.nb_obj_input,sizeof(object))) == NULL)
  {
    create_done_file(dirname,-1);
    para_term();
    exit(0);
  }
  for (i = 0;i< pb_desc.nb_obj_input; i++)
  {
    pb_desc.input_obj[i].data_type = atoi(argv[index_param++]);
    pb_desc.input_obj[i].object_type = atoi(argv[index_param++]);
  }
  pb_desc.nb_obj_output = atoi(argv[index_param++]);

  if ((pb_desc.output_obj =
          (object *)calloc(pb_desc.nb_obj_output,sizeof(object))) == NULL)
  {
    create_done_file(dirname,-1);
    para_term();
    exit(0);
  }
  for (i = 0;i< pb_desc.nb_obj_output; i++)
  {
    pb_desc.output_obj[i].data_type = 
                    atoi(argv[index_param++]);
    pb_desc.output_obj[i].object_type = 
                    atoi(argv[index_param++]);
  }
  sock = atoi(argv[index_param++]);
  mode = atoi(argv[index_param++]);
  
#if (PARALLEL==MPI)
  index_param++;
#else
  /* do something here with the name of the executable */
  /* argv[index_param++]                               */
#endif

  dirname = argv[index_param++];

  pb_desc.maj = atoi(argv[index_param++]);

  /* Allocate the sizes arrays */

  if ((sizes_input = 
     (input_sizes *)calloc(pb_desc.nb_obj_input,
           sizeof(input_sizes))) == NULL)
  {
    create_done_file(dirname,-1);
    para_term();
    exit(0);
  }
 
  if ((majors = (int *)calloc(pb_desc.nb_obj_input,sizeof(int))) == NULL)
  {
    create_done_file(dirname,-1);
    para_term();
    exit(0);
  }
  for (i=0;i<pb_desc.nb_obj_input;i++)
    majors[i] = -1;
  
  if ((sizes_output = 
    (input_sizes *)calloc(pb_desc.nb_obj_output,
           sizeof(input_sizes))) == NULL)
  {
    create_done_file(dirname,-1);
    para_term();
    exit(0);
  }
  /* Allocate the spaces */

  if ((data_input = 
         (void **)calloc(pb_desc.nb_obj_input,sizeof(void *))) == NULL)
  {
    create_done_file(dirname,-1);
    para_term();
    exit(0);
  }

  if ((data_output = 
         (void **)calloc(pb_desc.nb_obj_output,sizeof(void *)) ) == NULL)
  {
    create_done_file(dirname,-1);
    para_term();
    exit(0);
  }


  /* Reading the matrix sizes ONLY  and allocating the SPACE */

  for (i=0;i<pb_desc.nb_obj_input;i++)
  {
#ifdef VIEW
    if ((MYROW==0)&&(MYCOL==0))
      fprintf(stderr,"Object #%d \n",i);
#endif
    sprintf(filename,"%s/input%d",dirname,i);
    file = fopen(filename,"r");
    if (file == NULL)
    {
      create_done_file(dirname,-1);
      para_term();
      exit(0);
    }
    switch(pb_desc.input_obj[i].object_type)
    {
      case NETSOLVE_MATRIX :
        /* SO FAR< WE ASSUME ONLY 2-D DISTRIBUTE DOUBLE MATRICES */
        /* Getting the sizes */
        if (fread(&(majors[i]),sizeof(int),1,file) != 1)
        {
          fclose(file);
          create_done_file(dirname,-1);
          para_term();
          exit(0);
        }
        if (fread(&(sizes_input[i].m),sizeof(int),1,file) != 1)
        {
          fclose(file);
          create_done_file(dirname,-1);
          para_term();
          exit(0);
        }
        if (fread(&(sizes_input[i].n),sizeof(int),1,file) != 1)
        {
          fclose(file);
          create_done_file(dirname,-1);
          para_term();
          exit(0);
        }
        fclose(file);

        /* Allocate the space */
        local_matrix_size = 
          Cnumroc(sizes_input[i].m,ROW_BLOCK,MYROW,0,PROC_ROWS) *
          Cnumroc(sizes_input[i].n,COL_BLOCK,MYCOL,0,PROC_COLS);

        data_input[i] = (double *)calloc(local_matrix_size,DOUBL);  

        /* Construct a descriptor */
        sizes_input[i].d = (int *)desc_init(sizes_input[i].m,
                  sizes_input[i].n,1);
        work = (double *)malloc(sizeof(double)*(sizes_input[i].d)[MB_]);
        globalfile = fopen(filename,"r");
#ifdef ADD_
        pdlaread_(data_input[i],sizes_input[i].d,&zero,&zero,work);
#endif
#ifdef UPCASE
        PDLAREAD(data_input[i],sizes_input[i].d,&zero,&zero,work);
#endif
#ifdef NOCHANGE
        pdlaread(data_input[i],sizes_input[i].d,&zero,&zero,work);
#endif

        fclose(globalfile);
        free(work);
        break;

      case NETSOLVE_VECTOR :
        /* NON-DISTRIBUTED SO FAR */
      {
        if (fread(&(sizes_input[i].m),sizeof(int),1,file) != 1)
        {
          fclose(file);
          create_done_file(dirname,-1);
          para_term();
          exit(0);
        }
        sizes_input[i].n = 1;
        data_input[i] = (double *)calloc(sizes_input[i].m,DOUBL);  
        fclose(file);
        break;
      }
      case NETSOLVE_SCALAR :
      {
        sizes_input[i].m = 1;
        sizes_input[i].n = 1;
        /* Allocate the space */
        data_input[i] = (double *)calloc(1,DOUBL);  
        fclose(file);
        break;
      }
      case NETSOLVE_UPF :
      {
        /* Do nothing */
        fclose(file);
        break;
      }
      default :
      {
        fprintf(stderr,"Unknown object type ??\n");
        create_done_file(dirname,-1);
        fclose(file);
        para_term();
        exit(0);
      }
    }         
  }

#ifdef VIEW
  if ((MYROW==0)&&(MYCOL==0))
     fprintf(stderr,"Data obtained...\n");
#endif

  /* Remajoring the matrices ?? */
  client_major = -1;
  for (i=0;i<pb_desc.nb_obj_input;i++)
  {
    if (majors[i] == -1)
      continue;
    if (client_major == -1)
    {
      client_major = majors[i];
    }
    else
    {
      if (client_major != majors[i])
      {
#ifdef VIEW
        if ((MYCOL==0)&&(MYROW==0))
          fprintf(stderr,"Incompatible majors in input !!\n");
#endif
        create_done_file(dirname,-1);
        para_term();
        exit(0);
      }
    }
  }
  /* At this point, the majors are consistent   */
  /* I can do the real re-majoring if necessary */
  if (pb_desc.maj != client_major)
  {
#ifdef VIEW
    if ((MYROW==0)&&(MYCOL==0))
      fprintf(stderr,"Re-majoring the matrices\n");
#endif
    /* Transpose everything */
    for (i=0;i<pb_desc.nb_obj_input;i++)
    {
      if (pb_desc.input_obj[i].object_type == NETSOLVE_MATRIX)
      {
        /* Need to transpose it */
        Cpdtran(sizes_input[i].m,sizes_input[i].n,
           1.0,(double *)(data_input[i]),1,1,
           sizes_input[i].d,
           0.0,(double *)(data_input[i]),1,1,
           sizes_input[i].d);
      }
    }
  }


#ifdef VIEW
  if ((MYROW==0)&&(MYCOL==0))
    fprintf(stderr,"Solving '%s'....\n",pb_desc.code);
#endif

/*
 * Perform the computations
 */

  if  (pcompute(&pb_desc,sizes_input,sizes_output,
                   data_input,data_output,sock,
                   mode,dirname,client_major) == -1)
  {
    if ((MYROW == 0)&&(MYCOL == 0))
    {
      char *stdoutbuf;
      int stdoutlength;
      int type = SV_FAILURE;
  
      stdoutbuf = (char *)strdup("Server Error : No stdout available");
      stdoutlength = strlen(stdoutbuf)+1;
  
      sprintf(filename,"%s/type",dirname);
      file = fopen(filename,"w");
      if (file == NULL)
      {
#ifdef DEBUG
        fprintf(stderr,"Cannot create file '%s'\n",filename);
#endif
        create_done_file(dirname,-1);
        para_term();
        exit(0);
      } 
      if (fwrite(&type,sizeof(int),1,file) != 1)
      {
        create_done_file(dirname,-1);
        para_term();
        exit(0);
      }
      fclose(file);
      sprintf(filename,"%s/stdout",dirname);
      file = fopen(filename,"w");
      if (file == NULL)
      {
#ifdef DEBUG
        fprintf(stderr,"Cannot create file '%s'\n",filename);
#endif
        create_done_file(dirname,-1);
        para_term();
        exit(0);
      }
      if (fwrite(&stdoutlength,sizeof(int),1,file) != 1)
      {
        create_done_file(dirname,-1);
        para_term();
        exit(0);
      }
      if (fwrite(&stdoutbuf,sizeof(char),stdoutlength,file) != stdoutlength)
      {
        create_done_file(dirname,-1);
        para_term();
        exit(0);
      }
      fclose(file);
      free(stdoutbuf);
    }
  }
  
#ifdef VIEW
  if ((MYROW==0)&&(MYCOL==0))
    fprintf(stderr,"'%s' Completed\n",pb_desc.code);
#endif

  if ((MYROW == 0)&&(MYCOL == 0))
    create_done_file(dirname,1);
  para_term();
  exit(0);
} 

#ifdef __STDC__
int create_done_file(char *dirname,int code)
#else
int create_done_file(dirname,code)
char *dirname;
int code;
#endif
{
  FILE *file;
  char name[128];

  sprintf(name,"%s/done",dirname);
  file = fopen(name,"w");
  if (file == NULL)
  {
#ifdef DEBUG
     fprintf(stderr,"Cannot create file '%s'\n",name);
#endif
     return -1; 
  } 
  if (code == -1)
    fprintf(file,"ERROR");
  else
    fprintf(file,"DONE");

  fclose(file);
  return 1;
}
