/*****************************************************************/
/*      restriction_util.c                                       */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  restriction_included()                                       */
/*****************************************************************/

#include "server.h"
#include "server_prot.h"
#include "shared_prot.h"

/*
 * restriction_included()
 *  does r1 include r2
 */

#ifdef __STDC__
int restriction_included(restriction_item *r1,int nr1,
               restriction_item *r2,int nr2)
#else
int restriction_included(r1,nr1,r2,nr2)
restriction_item *r1;
int nr1;
restriction_item *r2;
int nr2;
#endif
{
  int i;

  for (i=0;i<nr2;i++)
    if (!(restriction_contains(r1,nr1,r2[i])))
      return 0;
  return 1;
}

/*
 * restriction_contains()
 *   r contains item
 */
#ifdef __STDC__
int restriction_contains(restriction_item *r,int nr,
                  restriction_item item)
#else
int restriction_contains(r,nr,item)
restriction_item *r;
int nr;
restriction_item item;
#endif
{
  int i;
  for (i=0;i<nr;i++)
  {
    if(item_contains(r[i].domain,item.domain))
    {
      if (r[i].max_pending >= item.max_pending)
        return 1;
    }
  }
  return 0;
}

/*
 * item_contains()
 *    item1 contains item2
 */
#ifdef __STDC__
int item_contains(char *s1,char *s2)
#else
int item_contains(s1,s2)
char *s1,*s2;
#endif
{
  char *lastdot1;
  char *lastdot2;
  char *domain1,*domain2;

  domain1 = (char *)strdup(s1);
  domain2 = (char *)strdup(s2);

  if (strcmp(domain1,"*") == 0)
    return 1;
  while(1)
  {
    lastdot1 = (char *)strrchr(domain1,'.');
    lastdot2 = (char *)strrchr(domain2,'.');

    if (lastdot1 == NULL)
    {
      if ((strcmp(domain1,"*") == 0)||
          (strcmp(domain1,domain2) == 0))
      {
        free(domain1); free(domain2);
        return 1;        
      }
      else
      {
        free(domain1); free(domain2);        
        return 0;
      }
    }
    if (lastdot2 == NULL)
    {
      free(domain1); free(domain2);
      return 0;
    }
    
    if((strcmp(lastdot1,".*") == 0)||
       (strcmp(lastdot1,lastdot2) == 0))
    {
      lastdot1[0] = '\0';
      lastdot2[0] = '\0';
    } 
    else
    {
      free(domain1); free(domain2);
      return 0;
    }
  }
}


/*
 * get_restriction_group()
 *   Check for incoming request
 */
#ifdef __STDC__
int get_restriction_group(int s)
#else
int get_restriction_group(s)
int s;
#endif
{
  struct sockaddr_in name;
  int namelen;
  unsigned int IPaddr=0;
  int i,slot;

  namelen = sizeof(name);
  if (getpeername(s,(struct sockaddr *)&name,&namelen) != 0)
  {
#ifdef VIEW
    fprintf(stderr,"Unknown peer!!\n");
#endif
    return -1;
  }
  else
  {
    struct hostent *hp=NULL;
    char hostname[100];

    bcopy(&(name.sin_addr),&IPaddr,sizeof(IPaddr));
    hp = gethostbyaddr((char *)&IPaddr,sizeof(IPaddr),AF_INET);
    if (hp != NULL)
       strcpy(hostname,hp->h_name);
    else
    {
#ifdef VIEW
      fprintf(stderr,"No name in DNS for peer!!\n");
#endif
      return -1;
    }
    lowercase(hostname);

    slot = -1;
    for (i=0;i<nb_restrictions;i++)
    {
      if (item_contains(restrictions[i].domain,hostname) ==0)
        continue;
      if (slot == -1)
        slot = i;
      else
      {
        if (strlen(restrictions[i].domain) > strlen(restrictions[i].domain))
          slot = i;
      }
    }
    return slot;
  }
}

