/*****************************************************************/
/*      server.c                                                 */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  main()                                                       */
/*****************************************************************/

/*
 *  This is the code of the NetSolve server
 *  The server should never stop running and sleeps
 *  during its call to select
 */

#include "server.h"
#include "server_prot.h"
#include "shared_prot.h"

#ifdef __STDC__
main(int argc,char **argv)
#else
main(argc,argv)
int argc;
char **argv;
#endif
{
  int pid;

  int sv_port = SV_PORT;

  fd_set readfd;                      /* For select()        */
 
  struct sockaddr_in rmt_client;      /* INET socket address */
  int len_client_add;                 /* address length      */
  
  struct hostent *hp;                 /* Host descriptor     */
  int cc;
  int i;

  char configfile[256];
  FILE *stream;
  char command_line[256];


/***************************
 * Initialize structure    *
 ***************************/

for (i=0;i<HASH_SV;i++)
  sv_hash_table[i] = NULL;
for (i=0;i<HASH_SV;i++)
  ag_hash_table[i] = NULL;
for (i=0;i<HASH_PB;i++)
  pb_hash_table[i] = NULL;


/**********************
 * Getting the path   *
 **********************/

  if ((root_path=(char *)getenv("NETSOLVE_SERVER_ROOT")) == NULL)
  {
    fprintf(stderr,"The NETSOLVE_SERVER_ROOT variable should be set\n");
    exit(0);
  }

/***********************
 * Getting my hostname *
 ***********************/

  if (gethostname(my_host,MAX_HOST_NAME) == -1)
  {
    perror("main() gethostname()");
    term();
  }
  if ((my_hp = gethostbyname(my_host)) == NULL)
  {
    perror("main() gethostbyname()");
    term();
  }
  BCOPY(my_hp->h_addr_list[0],&my_address,sizeof(unsigned int));
  strcpy(my_host,my_hp->h_name);

/**********************
 * Arguments analysis *
 **********************/
  
  switch(argc)
  {
  case 1: /* We define a default configuration file name */
    sprintf(configfile,"%s/server_config",root_path);
    break;
  case 2: /* There is a configuration file specified */
    sprintf(configfile,"%s",argv[1]);
    break;
  default:
    fprintf(stderr,"Usage : %s [configuration file]\n",argv[0]);
    exit(0); 
    break;
  }

  /******************************
   * Use the configuration file *
   *******************************/

#ifdef VIEW
  fprintf(stderr,"initializing......\n");
#endif

  if ((CONDOR = set_CONDOR(configfile)) == -1)
    term();
  if (setMPI(configfile) == -1)
    term();
  if (set_number_processors(configfile) == -1)
    term();
  if (server_init() == -1)
    term(); 
  if (problem_init() == -1)
    term();
  if (set_restrictions(configfile) == -1)
    term();
  if (set_master(configfile) == -1)
    term();
  if (set_scratchspace(configfile) == -1)
    term();


/*******************
 * Set the signals *
 *******************/

  set_signals();

/********************
 * Create a log file
 ********************/

#ifdef VIEW
  fprintf(stderr,"creating a log file......\n");
#endif
  new_server_log();

/**********************
 * Startup of sockets *
 **********************/

#ifdef VIEW
  fprintf(stderr,"enabling network connections......\n");
#endif

  /* remote socket */
  if ((rmt_sock = establish_socket(&sv_port)) == -1)
  {
    perror("main() create_socket_INET()");
    term();
  }

/********************************
 * Running the workload manager *
 ********************************/

  if ((work_pid = run_workload_process()) == -1)
    fprintf(stderr,"Warning : Impossible to fork the workload process\n");

#ifdef HBM

/*******************
 * Running the HBM *
 *******************/

/***********************************
 * Clean_up the HBM files          *
 ***********************************/
  
  sprintf(command_line,"rm -rf /tmp/hbmlm*");
  system(command_line); 
 

/***********************************
 * Start up the HBM data collector *
 ***********************************/

#ifdef VIEW
  fprintf(stderr,"starting up the Heart Beat Monitor Local Monitor.....\n");
#endif
  pid = fork();
  if (pid == -1)
    perror("fork");

  if (pid == 0) /* The HBMLM */
  {
    char path[256];

    sprintf(path,"%s/Globus/build/%s/",root_path,my_arch);
    if (chdir(path)== -1)
    {
      fprintf(stderr,"Warning: Unable to start the HBM local monitor.\n");
      perror("chdir");
      exit(0);
    }
    if (execl("./hbmlm","hbdlm",(char *)0) == -1)
    {
      fprintf(stderr,"Warning: Unable to start the HBM local monitor.\n");
      perror("execl");
      exit(0);
    }
  }

#endif


/*************
 * MAIN LOOP *
 *************/
 

#ifdef VIEW
  fprintf(stderr,"Server running.\n");
#endif

  while(1)
  {

    /* I get the message */
    listen(rmt_sock,MAXCONN);
    len_client_add = sizeof(rmt_client);
    if ((client_desc = accept(rmt_sock,
              (struct sockaddr *)&rmt_client,&len_client_add)) == -1)
    {
      if ((errno == EINTR)||(errno ==ECHILD))
        continue;
#ifdef VIEW
      perror("main() accept()");
#endif
      continue;
    }
    /* I generate the reply */
    pid = fork();
    if (pid != 0) /* Father */
    {
      close(client_desc);
      continue;
    }
    else /* Child */
    {
      close(rmt_sock);
      if (proc_mes() == -1)
      {
        if (gen_sv_failure(client_desc) == -1)
        {
#ifdef VIEW
          fprintf(stderr,"WARNING : Impossible to notify my failure\n");
#endif
        }
      }
      exit(0);
    }
/****************/
  } /* END LOOP */
/****************/
 
}
