/*****************************************************************/
/*      server_config_file.c                                     */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/* set_CONDOR()                                                  */
/* set_number_processors()                                       */
/* set_master()                                                  */
/* set_restrictions()                                            */
/*****************************************************************/

#include "server.h"
#include "server_prot.h"
#include "shared_prot.h"

/*
 * set_CONDOR()
 */
#ifdef __STDC__
int set_CONDOR(char *filename)
#else
int set_CONDOR(filename)
char *filename;
#endif
{
  FILE *file;
  char line[MAX_LINE_LENGTH];
  int linecount = 0;
  char *token1,*token2;

  /* Opening the file */
  file = fopen(filename,"r");

  if (file == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  while(read_line(file,line,&linecount) != NULL)
  {
    if (strncmp(line,"@CONDOR:",8))
      continue;
    fclose(file);
    token1 = (char *)strtok(line,":\t\n");
    token2 = (char *)strtok(NULL," \t\n");
    if (token2 == NULL)
      return 0;

    CONDOR_path = (char *)strdup(token2);
    CONDOR_path[strlen(CONDOR_path)] = '\0';
    return 1;
  }
  fclose(file);
  return 0;
}

/* 
 * Set Scratch space
 */

#ifdef __STDC__
int set_scratchspace(char *filename)
#else
int set_scratchspace(filename)
char *filename;
#endif
{
  FILE *file;
  char line[MAX_LINE_LENGTH];
  int linecount = 0;
  char *token1,*token2;

  /* Opening the file */
  file = fopen(filename,"r");

  if (file == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  while(read_line(file,line,&linecount) != NULL)
  {
    if (strncmp(line,"@SCRATCH:",9))
      continue;
    fclose(file);
    token1 = (char *)strtok(line,":\t\n");
    token2 = (char *)strtok(NULL," \t\n");
    if (token2 == NULL)
      return 0;

    scratch_path = (char *)strdup(token2);
    scratch_path[strlen(scratch_path)] = '\0';
    return 1;
  }
  fclose(file);
  scratch_path = (char *)strdup("/tmp/");
  return 0;
}

/*
 * set_MPI()
 */
#ifdef __STDC__
int setMPI(char *filename)
#else
int setMPI(filename)
char *filename;
#endif
{
  FILE *file;
  char line[MAX_LINE_LENGTH];
  int linecount = 0;
  char *token1,*token2,*token3;

  /* Opening the file */
  file = fopen(filename,"r");

  if (file == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  while(read_line(file,line,&linecount) != NULL)
  {
    if (strncmp(line,"@MPIHOSTS",7))
      continue;
    fclose(file);
    token1 = (char *)strtok(line," \t\n");
    token2 = (char *)strtok(NULL," \t\n");
    token3 = (char *)strtok(NULL," \t\n");
    if ((token2 == NULL)||(token3 == NULL))
    {
      MPInodes = 0;
      MPImachinefile[0] = '\0';
      return 1;
    }
    MPInodes = atoi(token3);
    strcpy(MPImachinefile,token2);
    fclose(file);
    return 1;
  }
  MPInodes = 0;
  return 1;
}


/*
 * set_number_processors()
 */
#ifdef __STDC__
int set_number_processors(char *filename)
#else
int set_number_processors(filename)
char *filename;
#endif
{
  FILE *file;
  char line[MAX_LINE_LENGTH];
  int linecount = 0;

  /* Opening the file */
  file = fopen(filename,"r");

  if (file == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  while(read_line(file,line,&linecount) != NULL)
  {
    if (strncmp(line,"@PROC:",6) == 0)
    {
      int n;
      
      n = atoi(&(line[6]));
      if (n <= 0)
      {
        fprintf(stderr,"%s, line %d:\n'%s' is an invalid number of processors\n",
                filename,linecount,&(line[5]));
        fclose(file);
        return -1;
      }
      number_processors = n;
      fclose(file);
      return 1;
    }
  }

  fprintf(stderr,
    "Missing line '@PROC:<number of processors>' in '%s'\n",
    filename);
  fclose(file);
  return -1;
}

/*
 * set_master()
 */
#ifdef __STDC__
int set_master(char *filename)
#else
int set_master(filename)
char *filename;
#endif
{
  FILE *file;
  char line[MAX_LINE_LENGTH];
  int linecount = 0;

  /* Opening the file */
  file = fopen(filename,"r");

  if (file == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  while(read_line(file,line,&linecount) != NULL)
  {
    if (strncmp(line,"@AGENT:",7) == 0)
    {
      line[strlen(line)-1]  = '\0';
      if (line[strlen(line)-1] == '*')
      {
        line[strlen(line)-1] = '\0';
        if (contact_master(&(line[7]),'*') == -1)
        {
          fclose(file);
          return -1;
        }
        else
        {
          fclose(file);
          return 1;
        }
      }
      else
      {
        if (contact_master(&(line[7]),'-') == -1)
        {
          fclose(file);
          return -1;
        }
        else
        {
          fclose(file);
          return 1;
        }
      }
    }
  }

  fprintf(stderr,"Missing line '@AGENT:<agent name>[*]' in '%s'\n",filename);
  fclose(file);
  return -1;
}

/*
 * set_restrictions()
 */
#ifdef __STDC__
int set_restrictions(char *filename)
#else
int set_restrictions(filename)
char *filename;
#endif
{
  FILE *file;
  char line[MAX_LINE_LENGTH];
  int linecount = 0;

  /* Opening the file */
  file = fopen(filename,"r");

  if (file == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  while(read_line(file,line,&linecount) != NULL)
  {
    if (strncmp(line,"@RESTRICTIONS:\n",15) == 0)
    {
      nb_restrictions = 0;
      while(read_line(file,line,&linecount) != NULL)
      {
        char domain[128];
        int max_pending;
        if (line[0] == '@')
          break;
        line[strlen(line)-1] = '\0';
        if (sscanf(line,"%s %d",domain,&max_pending) < 2)
        {
          fprintf(stderr,"'%s, line %d:\n",filename,linecount);
          fprintf(stderr,
            "line should be of the form : <domain> <number of requests>\n");
          fclose(file);
          return -1;
        }
        if (nb_restrictions >= NB_RESTRICTIONS)
        {
          fprintf(stderr,"'%s, line %d:\n",filename,linecount);
          fprintf(stderr,"Only %d restrictions are allowed for an agent\n");
          fclose(file);
          return -1;
        }
        if (domain[0] == '.')
        {
          fprintf(stderr,"'%s, line %d:\n",filename,linecount);
          fprintf(stderr,"domain name cannot start with a '.'\n");
          fclose(file);
          return -1;
        }
        strcpy(restrictions[nb_restrictions].domain,domain);
        restrictions[nb_restrictions].max_pending = max_pending;
        nb_restrictions++;
      }
    }
  }
  if (nb_restrictions == 0)
  {
    fprintf(stderr,"'%s, line %d:\n",filename,linecount);
    fprintf(stderr,"No restriction specified.\n");
    fclose(file);
    return -1;
  }
  fclose(file);
  return 1;
}
