/*****************************************************************/
/*      server_proc_mes.c                                        */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  proc_mes()                                                   */
/*  gen_pb_refused()                                             */
/*---------------------------------------------------------------*/
/* WHO_ARE_YOU                                                   */
/* PONG_LATENCY                                                  */
/* PONG_BANDWIDTH                                                */
/* ARE_YOU_ALIVE                                                 */
/* POLL_KILL							 */
/* I_EXIST                                                       */
/* PROBLEM_SOLVE                                                 */
/* I_DIE                                                         */
/*****************************************************************/

#include "server.h"
#include "server_prot.h"
#include "shared_prot.h"
#ifdef HBM
#include "hbm_client.h"
#endif

/*
 * proc_mes : the server/agent received a message on my remote socket
 */

#ifdef __STDC__
int proc_mes()
#else
int proc_mes()
#endif
{

  int message_type;
  int encoding;
  FILE *stream;

  encoding = init_recv(client_desc);
  read_int(&message_type);

  switch(message_type){

/***********************/
  case WHO_ARE_YOU:
/***********************/
{
  end_message();
  end_com();
  if (contact_master("again",'-') == -1)
  {
#ifdef VIEW
      fprintf(stderr,"Cannot contact agent for check-in\n");
#endif
  }
  return 1;
}

/***********************/
  case PONG_LATENCY:
/***********************/
{
  end_message();
  end_com();
  return 1;
}

/***********************/
  case PONG_BANDWIDTH:
/***********************/
{
  int size;
  int *buf;
  int pid;

  {
    close(rmt_sock); /* Close the listening socket */
  
    read_int(&size);
    if (size > 0)
      buf = (int *)malloc(sizeof(int)*size);
    while(1)
    {
      if (size == -1)
      {
        size = -1;
        end_message();
        break;
      } 
      read_int_array(buf,size);
      end_message();
      init_send(encoding,client_desc);
      write_int_array(buf,size);
      end_message();
      init_recv(client_desc);
      read_int(&size);
    }
    end_com();
  }
  exit(0);
}



/*******************/
  case POLL_KILL:
/*******************/
{
  end_message();
  end_com();
  bailout();
  term();
  return 1;
}

/***********************/
  case ARE_YOU_ALIVE:      
/***********************/
{
  int answer = ALIVE;

  end_message();
  init_send(encoding,client_desc);
  write_int(&answer);
  end_message();
  end_com();
  return 1;
}

/***********************/
  case I_EXIST:        /* This is a request from a newly accepted server */
/***********************/
{
  server_desc sv_desc;
  server_item *server;
  int number_of_pb;
  int j;
  problem_desc pb_desc;
  problem_item *problem;
  int delay;

  read_int(&delay); /* delay for the pong */
  read_string(sv_desc.hostname);
  read_IPaddr(&(sv_desc.IPaddr));
  read_int(&(sv_desc.data_format));
  read_int(&(sv_desc.speed));
  read_int(&(sv_desc.workload));
  read_int(&(sv_desc.number_proc));
  read_int(&(sv_desc.mpp));
 /*
  * From here we deal with the message received
  */
  sv_desc.host_status = UP;
  sv_desc.server_status = RUNNING;
  sv_desc.connection = TO_BE_A; /* Waiting a value */
  sv_desc.transfer   = TO_BE_A; /* Waiting a value */
  read_int(&number_of_pb);
  /* We remove if it already exists */
  if (number_of_pb != 0)
  {
    if ((server = (server_item *)find_server(sv_desc.IPaddr)) != 0)
    {
      if (remove_server(server) == -1)
      {
        perror("proc_mes() remove_server()");
        end_message(); /* END of a receive Phase */
        return -1;
      }
    }
  }
  else
  {
    if ((server = (server_item *)find_agent(sv_desc.IPaddr)) != 0)
    {
      if (remove_agent(server) == -1)
      {
        perror("proc_mes() remove_agent()");
        end_message(); /* END of a receive Phase */
        return -1;
      }
    }
  }

  if (number_of_pb != 0)
  {
    if ((server =(server_item *)new_server(sv_desc)) == 0)
    {
      perror("proc_mes() new_server()");
      end_message(); /* END OF A RECEIVE PHASE */
      return -1;
    }
  }
  else
  {
    if ((server =(server_item *)new_agent(sv_desc)) == 0)
    {
      perror("proc_mes() new_agent()");
      end_message(); /* END OF A RECEIVE PHASE */
      return -1;
    }
  }
  for (j=0;j<number_of_pb;j++)
  {
    read_problem(&pb_desc);
    problem = (problem_item *)find_problem(pb_desc.code);
    if (problem == 0) /* THE PROBLEM DOESN'T EXIST YET */
    { 
      problem = (problem_item *)new_problem(pb_desc);
      if (problem == 0)
      {
        perror("proc_mes() new_problem()");
        end_message(); /* END OF A RECEIVE PHASE */
        return -1;
      }
      add_pb_to_server(problem,server); 
      add_server_to_pb(problem,server); 
      free_pb_desc(&pb_desc);
    }
    else /* THE PROBLEM ALREADY EXISTS */
    {
      add_pb_to_server(problem,server);
      add_server_to_pb(problem,server);
      /* Skip the problem */
    }
  }
  /* New workload management process ??? */
  if ((my_self->number_problems != 0)&&(server->number_problems == 0))
  {
    (void)kill(work_pid,SIGTERM);
    work_pid = run_workload_process();
  }
  end_message(); /* END OF A READ PHASE */
  end_com();
#ifdef DEBUG
  fprintf(stderr,"I added %s to my view of the configuration\n",sv_desc.hostname);
#endif
  return 1; 
}
  /*******************/
  case PROBLEM_SOLVE: 
  /*******************/
{
  char clientagent[256];
  char code[CODE_LENGTH];
  problem_item *problem;
  char logbuf[80];
  int ack;
  int pid;
  int client_type;
  problem_desc pb_desc;
  int i;
  int status;
  int restriction_group;


  read_string(pb_desc.code);
#ifdef VIEW
  fprintf(stderr,"Asked to solve problem '%s'\n",pb_desc.code);
#endif
  read_string(clientagent);
  read_int(&client_type);
  read_int(&(pb_desc.nb_obj_input));
  pb_desc.input_obj = (object *)calloc(pb_desc.nb_obj_input,sizeof(object));
  for (i=0;i<pb_desc.nb_obj_input;i++)
  {
    read_int(&(pb_desc.input_obj[i].data_type));
    read_int(&(pb_desc.input_obj[i].object_type));
  }
  read_int(&(pb_desc.nb_obj_output));
  pb_desc.output_obj = (object *)calloc(pb_desc.nb_obj_output,sizeof(object));
  for (i=0;i<pb_desc.nb_obj_output;i++)
  {
    read_int(&(pb_desc.output_obj[i].data_type));
    read_int(&(pb_desc.output_obj[i].object_type));
  }
  end_message();

  /* I CHECK THE RESTRICTIONS */
  restriction_group = get_restriction_group(client_desc);
  if (restriction_group == -1)
  {
    status = NOT_AUTHORIZED;
#ifdef VIEW
    fprintf(stderr,"Unauthorized client !! Rejected\n");
#endif
    init_send(encoding,client_desc);
    write_int(&status);
    end_message();
    end_com();
    return 1;
  }
  if (restrictions[restriction_group].pending  >=
      restrictions[restriction_group].max_pending)
  {
    status = TOKENS_OVERFLOW;
    init_send(encoding,client_desc);
#ifdef VIEW
    fprintf(stderr,"Client exceeding quotas !! Rejected\n");
#endif
    write_int(&status);
    end_message();
    end_com();
    return 1;
  }

  /* At this point, there is no restriction problem */

  /* I check the problem */
  problem = (problem_item *)find_problem(pb_desc.code);
  if (problem == 0)
  { /* I don't solve the problem any more */
#ifdef VIEW
    fprintf(stderr,"\tProblem unknown\n");
#endif
    if (gen_pb_refused(encoding) == -1)
    {
      perror("proc_mes() gen_pb_refused()");
      free(pb_desc.output_obj);
      free(pb_desc.input_obj);
      return -1;
    }
    end_com();
    free(pb_desc.output_obj);
    free(pb_desc.input_obj);
    return 1;
  }
  pb_desc.complexity = NULL;
  status = problem_cmp(pb_desc,problem->problem);
  if ((status != -5)&&(status <0))
  {   /* No longer this specification */
#ifdef DEBUG
    fprintf(stderr,"\t Problem Specification error\n");
#endif
    if (gen_pb_refused(encoding) == -1)
    {
      perror("proc_mes() gen_pb_refused()");
      free(pb_desc.output_obj);
      free(pb_desc.input_obj);
      return -1;
    }
    end_com();
    free(pb_desc.output_obj);
    free(pb_desc.input_obj);
    return 1;
  }
  
  /* THE PROBLEM IS CORRECT */
  nb_pb_solved++;
  sprintf(logbuf,"Request #%d : problem='%s'",nb_pb_solved,pb_desc.code); 
  append_to_log(logbuf);
  add_request_to_log(client_desc);

  /* I ACCEPT THE PROBLEM */
  init_send(encoding,client_desc);
  ack = PB_ACCEPTED;
  write_int(&ack);
  end_message();
  (restrictions[restriction_group].pending)++;
  {
    int param_index = 0;
    int nb_param;
    char path[128];
    char **args;
    int i;
    int nb_in,nb_out;
    int upf_ok=1;
    int nb_upf=0;
    char **newargs;
    int FILE_PROTOCOL;
    int retcd;

    /*
     *  UPF ?
     */
    for (i=0;i<pb_desc.nb_obj_input;i++)
      if (pb_desc.input_obj[i].object_type == NETSOLVE_UPF)
        nb_upf++;
  
    if (nb_upf == 0) 
    {
       /* No problems here, we just fork the usual thing */
       sprintf(path,"%s/bin/%s/service-%s",
           root_path,my_arch,problem->problem.file);
    }
    else
    { /* UPFs needs to be taken into account */
  
      int i;
      int length;
      FILE *upf;
      FILE *script;
      FILE *stream;
      char *buf; 
      char *upfbuf; 
      char dirname[128];
      char upfname[128];
      char scriptname[128];
      char command_line[128];
      char c;
      struct stat st;
      int mode;
      int upf_type;
      int upf_index = 0;
      int unsafe_upf_index = 0;
  
      char *error_report;
  
      error_report = (char *)calloc(256,sizeof(char));
    
     /* Init_recv  */
      init_recv(client_desc);
  
     /* Make the temporary directory */
      sprintf(dirname,"/tmp/NetSolve_upf.%d",getpid());
      if (mkdir(dirname,0777))
      {
        perror("mkdir()");
        upf_ok = 0;
      }
  
      /* Get the script in buf */
      sprintf(scriptname,"%s/src/Server/upf_script",root_path);
      script = fopen(scriptname,"r");
      if (script == NULL)
      {
        perror("fopen()");
        upf_ok = 0;
      }
      stat(scriptname,&st);
      length = (int)st.st_size;
      buf = (char *)malloc(sizeof(char)*(length+1));
      if (fread(buf,1,st.st_size,script) != st.st_size)
      {
        perror("fread()");
        upf_ok = 0;
      }
      buf[length] = '\0';
      fclose(script);
  
     /* Open the new script */
      sprintf(scriptname,"%s/upf_script",dirname);
      script = fopen(scriptname,"w");
      if (script == NULL)
      {
        perror("fopen()");
        upf_ok = 0;
      }
      fprintf(stderr,"NETSOLVE_SERVER_ROOT=%s\n",root_path);
      fprintf(script,"NETSOLVE_SERVER_ROOT=%s\n",root_path);
      fprintf(script,"NETSOLVE_ARCH=%s\n",my_arch);
  
#ifdef ADD_
      fprintf(script,"FORTRAN_C_FLAG=ADD_\n");
#endif
#ifdef F2C
      fprintf(script,"FORTRAN_C_FLAG=F2C\n");
#endif
#ifdef NOCHANGE
      fprintf(script,"FORTRAN_C_FLAG=NOCHANGE\n");
#endif
#ifdef UPPCASE
      fprintf(script,"FORTRAN_C_FLAG=UPCASE\n");
#endif
  
      fprintf(script,"WORKING_DIR=%s\n",dirname);
#ifndef PLAT_SUN4
#ifndef PLAT_SUN4SOL2
      fprintf(stderr,"**Using default for this platform**\n");
      fprintf(stderr,"**Probability of success = 0.001 **\n");
      fprintf(script,"F77=f77\n");
      fprintf(script,"CC=cc\n");
      fprintf(script,"LIBS=-lm\n");
#endif
#endif

#ifdef PLAT_SUN4
      fprintf(script,"F77=f77\n");
      fprintf(script,"CC=cc\n");
      fprintf(script,"LIBS=-lm\n");
#endif
#ifdef PLAT_SUN4SOL2
      fprintf(script,"F77=/opt/SUNWspro/bin/f77\n");
      fprintf(script,"CC=/opt/SUNWspro/bin/cc\n");
      fprintf(script,"LIBS=-lm\n");
      fprintf(script,"LIBS1=-lnsl\n");
      fprintf(script,"LIBS2=-lsocket\n");
      fprintf(script,"LIBS3=-lrpcsvc\n");
#endif
  
     /* Receive the files */
      for (i=0;i<pb_desc.nb_obj_input;i++)
      {
        if (pb_desc.input_obj[i].object_type == NETSOLVE_UPF)
        {
          char upf_name[64];
          char upf_crap[64];
  
          read_int(&upf_type);
          read_string(upf_name);
          read_int(&length);
          upfbuf = (char *)malloc(sizeof(char)*(length+1));
          read_char_array(upfbuf,length);
          upfbuf[length] = '\0';
/*
          end_message();
*/
      
          /* Write the upf in the directory */
          if (upf_type == LANG_FORTRAN)
            sprintf(upfname,"%s/upf%d.f",dirname,upf_index);
          else
            sprintf(upfname,"%s/upf%d.c",dirname,upf_index);
          upf = fopen(upfname,"w");
          if (upf == NULL)
          {
            perror("fopen()");
            upf_ok = 0;
          }
          fprintf(upf,upfbuf);
          fclose(upf);
          free(upfbuf);
  
          /* Update the script */
          fprintf(script,"#\n# Making %s\n#\n",upfname);
          fprintf(script,"echo \"Compiling %s\"\n",upfname);
          if (upf_type == LANG_FORTRAN)
          {
            fprintf(script,"$F77 -c %s -o upf%d.o\n",upfname,upf_index);
          }
          else
          {
            fprintf(script,"$CC -c %s -o upf%d.o\n",upfname,upf_index);
          }
          fprintf(script,"UPF%d=$WORKING_DIR/upf%d.o\n",upf_index,upf_index);
          if (upf_type == LANG_FORTRAN)
          {
#ifdef NOCHANGE
            sprintf(upf_crap,"upf%d\n",upf_index);
            if (strcmp(upf_crap,upf_name))
              fprintf(script,"UPFCLAUSE%d=\"#define upf%d %s\"\n",
                       upf_index,upf_index,upf_name);
            else
              fprintf(script,"UPFCLAUSE%d=\" \"\n",upf_index);
#endif
#ifdef ADD_
            fprintf(script,"UPFCLAUSE%d=\"#define upf%d %s_\"\n",
                     upf_index,upf_index,upf_name);
#endif
#ifdef F2C
            fprintf(script,"UPFCLAUSE%d=\"#define upf%d %s__\"\n",
                     upf_index,upf_index,upf_name);
#endif
#ifdef UPCASE
            uppercase(upf_name);
            fprintf(script,"UPFCLAUSE%d=\"#define upf%d %s\"\n",
                     upf_index,upf_index,upf_name);
#endif
          }
          else
          {
            sprintf(upf_crap,"upf%d",upf_index);
            if (strcmp(upf_crap,upf_name))
              fprintf(script,"UPFCLAUSE%d=\"#define upf%d %s\"\n",
                       upf_index,upf_index,upf_name);
            else
              fprintf(script,"UPFCLAUSE%d=\" \"\n",upf_index);
  
          }
          upf_index++;
        }
      }
      end_message();
      /* Complete the script in the directory */
      for (i=0;i<upf_index;i++)
        fprintf(script,"echo $UPFCLAUSE%d >> $WORKING_DIR/defines\n",i);    
      fprintf(script,"UPF=\""); 
      for (i=0;i<upf_index;i++)
        fprintf(script,"$UPF%d ",i);
      fprintf(script,"\"\n");
      
      /* Put the base script at the end */ 
      fprintf(script,buf);
      fflush(script);
      fclose(script);
      free(buf);
  
      /* Make the script executable */
      if (chmod(scriptname,0777))
      {
        perror("chmod()");
        upf_ok = 0;
      }
     
      /* I change my directory */
      if (chdir(dirname))
      {
        perror("chdir()");
        upf_ok = 0;
      }
  
      /* I run the script */
      sprintf(command_line,"./upf_script %s %s",
            problem->problem.code,problem->problem.file); 
      stream=popen(command_line,"r");
      if (stream == NULL)
      {
        perror("popen()");
        upf_ok = 0;
      }
#ifdef DEBUG
      while( (c = getc(stream)) != EOF)
      {
        fprintf(stderr,"%c",c);
      }
#endif
      pclose(stream); /* should wait for the completion */
  
      /* Checking that service.upf has been created */
      sprintf(path,"%s/service.upf",dirname);
      upf_ok = 1;
      if (access(path,F_OK))
      {
#ifdef VIEW
        fprintf(stderr,"\tUPF Compilation has not been successful\n");
#endif
        upf_ok = UPF_FAILED;
      }
      else
      { 
        /* Checking the security */
        for (i=0;i<upf_index;i++)
        {
          char name[256];
          int debug;
  
          sprintf(name,"upf%d.o",i);
          if (check_security(error_report,name) != 1)
          {
#ifdef VIEW
            fprintf(stderr,"\t\tUPF unsafe\n");
#endif
            upf_ok = UPF_UNSAFE;
            unsafe_upf_index = i;
            break;
          }
        }
        if (upf_ok == 1) 
        {
#ifndef PLAT_SUN4SOL2
          if (are_undefined("service.upf"))
          {
#ifdef VIEW
            fprintf(stderr,"\t\tUPF failed\n");
#endif
            upf_ok = UPF_FAILED;
          }
#endif
        }
      }
      init_send(encoding,client_desc);
      if (upf_ok!= 1) /* UPF FAILED  or UPF UNSAFE*/
      {
        int message;
        FILE *stream;
        char command_line[256];
    
        message = upf_ok;
        write_int(&message);
        if (message == UPF_UNSAFE)
        {
          write_int(&unsafe_upf_index);
          write_string(error_report);
          free(error_report);
        }
        end_message();
        end_com();
        /* change my working dir */
        chdir("/");
        sleep(200);
        sprintf(command_line,"rm -rf /tmp/NetSolve_upf.%d",getpid());
        system(command_line);
        exit(0);
      }
      else
      {
        int message;
  
        message = UPF_OK;
        write_int(&message);
        end_message();
      }
    }
  
#ifdef DEBUG
    fprintf(stderr,"Forking %s\n",path);
#endif
    nb_in = problem->problem.nb_obj_input;
    nb_out = problem->problem.nb_obj_output;
  
    nb_param=1;                          /* EXECUTABLE NAME    */
    nb_param++;                          /* HBMDC TO REPORT TO */
    nb_param++;                          /* CODE               */
    nb_param +=  2+ (nb_in)*2+ nb_out*2; /* I/O                */
    nb_param++;                          /* sock               */
    nb_param++;                          /* encoding           */
    nb_param++;			         /* executable name    */
    nb_param++;                          /* CONDOR ?           */
    nb_param++;                          /* major              */
  
    args = (char **)calloc(nb_param+1,sizeof(char *));
    for (i=0;i<nb_param;i++)
      args[i] = (char *)calloc(80,sizeof(char));
  
    args[nb_param] = NULL;
  
    sprintf(args[param_index++],"service-%s",problem->problem.file);
    sprintf(args[param_index++],"%s",clientagent);
    sprintf(args[param_index++],"%s",problem->problem.code);
    sprintf(args[param_index++],"%d",nb_in);
    for (i=0;i<nb_in;i++)
    {
      sprintf(args[param_index++],"%d",problem->problem.input_obj[i].data_type);
      sprintf(args[param_index++],"%d",problem->problem.input_obj[i].object_type);
    } 
    sprintf(args[param_index++], "%d",nb_out);
  
    for (i=0;i<nb_out;i++)
    {
      sprintf(args[param_index++],
              "%d",problem->problem.output_obj[i].data_type);
      sprintf(args[param_index++],
              "%d",problem->problem.output_obj[i].object_type);
    }
    sprintf(args[param_index++],"%d",client_desc);
    sprintf(args[param_index++],"%d",encoding);
  
    /* Additional stuff for BLACS */
    sprintf(args[param_index++],"service-%s",problem->problem.file);

    if (CONDOR)
      FILE_PROTOCOL = 1;
    else if (problem->problem.parallel != 0)
      FILE_PROTOCOL = 1;
    else
      FILE_PROTOCOL = 0;

    if (!FILE_PROTOCOL)
    {
      /* Do the regular thing */

      /* Additional argument for CONDOR stuff */
       args[param_index++] = (char *)strdup("-");

      /* Additional argument for the major */
       sprintf(args[param_index++],"%d",problem->problem.maj);


      execv(path,args);  /* The child will register and 
                             de-register from the HBM */
    }
  
    if (FILE_PROTOCOL)
    {
      /* CONDOR, SCALAPACK, ETC..... */
      char commandline[512];
      char dirname[128];
      char commandfilename[128];
      char datafilename[128];
      char donefilename[128];
      char typefilename[128];
      char stdoutfilename[128];
      char result[10];
      int mypid;
      int loop;
     

      int nb_bytes;
      FILE *file;
      int fd;
      int type,m,n,maj;
      void *tmp;
      int stdoutlength;
      char *stdoutbuf;
      FILE *stream;
      int dummy;
  
#ifdef HBM
      /*
       * Register to Globus the HBM
       */
      if ((retcd = hbm_register("", 0, NULL,clientagent)) < 0) {
#ifdef VIEW
        fprintf(stderr,"Cannot register to the HBM\n");
#endif
        gen_sv_failure(client_desc);
        end_com();
        exit(0);
      }
#endif

      if (CONDOR)
      {
        sprintf(dirname,"%s/Netsolve_CONDOR.%d",scratch_path,getpid());
      }
      else
      {
        sprintf(dirname,"%s/Netsolve_SCALAPACK.%d",scratch_path,getpid());
      }

     fprintf(stderr,"Making the directory\n");

      if (mkdir(dirname,S_IRUSR|S_IWUSR|S_IXUSR|S_IROTH|S_IXOTH) == -1)
      {
        fprintf(stderr,"Unable to create directory '%s'\n",dirname);
        gen_sv_failure(client_desc);
        end_com();
#ifdef HBM
        if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
          fprintf(stderr, "\n:  Error on shutdown.\n");
#endif
        }
#endif
        exit(0);
      }

#ifdef DEBUG
      fprintf(stderr,"Creating directory '%s'\n",dirname);
#endif

      init_recv(client_desc);
    
      fprintf(stderr,"Reading the OBJECTS\n");
      for (i=0;i<pb_desc.nb_obj_input;i++)
      {
        int total_size;
#ifdef DEBUG
        fprintf(stderr,"Object #%d \n",i);
#endif
        sprintf(logbuf,"Object #%d ",i);
        append_to_log(logbuf);
        /* Open the file */
        sprintf(datafilename,"%s/input%d",dirname,i);
        fd = open(datafilename,O_WRONLY|O_CREAT,0644);
        if (fd <0)
        {
          gen_sv_failure(client_desc);
          end_com();
#ifdef HBM
          if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
            fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
          }
#endif
          exit(0);
        }
        switch(pb_desc.input_obj[i].object_type)
        {
          case NETSOLVE_MATRIX :
            read_int(&dummy);
            read_int(&dummy);
            read_int(&maj); 
            read_int(&m); 
            read_int(&n);
            /* Put maj in the file */
            if (write(fd,&maj,sizeof(int)) != sizeof(int))
            {
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            /* Put m and n in the file */
            if (write(fd,&m,sizeof(int)) != sizeof(int))
            {
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            if (write(fd,&n,sizeof(int)) != sizeof(int))
            {
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
#ifdef DEBUG
            fprintf(stderr,"MATRIX %d %d \n",
                   m,n);
#endif
            sprintf(logbuf,"MATRIX %d %d : type %d",m,
                    n,pb_desc.input_obj[i].data_type);
            append_to_log(logbuf);
            break;
          case NETSOLVE_VECTOR :
          {
            read_int(&dummy);
            read_int(&dummy);
            read_int(&m);
            n = 1;
            /* Put m in the file */
            if (write(fd,&m,sizeof(int)) != sizeof(int))
            {
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
#ifdef DEBUG
            fprintf(stderr,"VECTOR %d\n",m);
#endif
            sprintf(logbuf,"VECTOR %d : type %d",m,
                    pb_desc.input_obj[i].data_type);
            append_to_log(logbuf);
            break;
          }
          case NETSOLVE_FILE :
          {
            read_int(&dummy);
            read_int(&dummy);
            read_int(&m);
            n = 1;
#ifdef DEBUG
            fprintf(stderr,"FILE %d\n",m);
#endif
            sprintf(logbuf,"FILE %d : type %d",m,
                    pb_desc.input_obj[i].data_type);
            append_to_log(logbuf);
            break;
          }
          case NETSOLVE_SCALAR :
          {
#ifdef DEBUG
        fprintf(stderr,"SCALAR\n");
#endif
            read_int(&dummy);
            read_int(&dummy);
            m = 1;
            n = 1;
            break;
          }
          case NETSOLVE_UPF :
          {
             /* Do nothing */
             break;
          }
          default :
          {
#ifdef DEBUG
            fprintf(stderr,"Unknown object type ??\n");
#endif
            sprintf(logbuf,"Unknown object type ??");
            append_to_log(logbuf);
            gen_sv_failure(client_desc);
            end_com();
#ifdef HBM
            if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
              fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
            }
#endif
            exit(0);
          }
        }
        total_size = m*n;
        switch(pb_desc.input_obj[i].data_type)
        {
          case NETSOLVE_I :
          {
            nb_bytes = total_size*INTL;
            if (read_int_array_file(fd,total_size) != nb_bytes)
            {
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            break;
          }
          case NETSOLVE_B :
          {
            int kk, num_vec, vec_len;
            char *buffer;

            nb_bytes = total_size*CHARL;
            buffer=(char *)calloc(nb_bytes,sizeof(char));

            if( (pb_desc.input_obj[i].object_type == NETSOLVE_MATRIX)  &&
                (maj == ROW_MAJOR))
            {
              num_vec = m;
              vec_len = n;
            }
            else
            {
              num_vec = n;
              vec_len = m;
            }
    
            for (kk=0;kk< num_vec;kk++)
            {
              read_byte_array((char *)((char *)buffer+
                       kk*(vec_len)),vec_len);
            }
            /* Now I have the buffer */
            if (write(fd,buffer,nb_bytes) != nb_bytes)
            {
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }            
            free(buffer);
            break;
          }
          case NETSOLVE_CHAR :
          {
            nb_bytes = total_size*CHARL;
            if (read_char_array_file(fd,total_size) != nb_bytes)
            { 
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            break;
          }
          case NETSOLVE_ASCII :
          {
            nb_bytes = total_size*CHARL;
            if (read_char_array_file(fd,total_size) != nb_bytes)
            {
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            break;
          }
          case NETSOLVE_S :
          {
            nb_bytes = total_size*FLOATL;
            if (read_single_array_file(fd,total_size) != nb_bytes)
            { 
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            break;
          }
          case NETSOLVE_D :
          {
            nb_bytes = total_size*DOUBL;
            if (read_double_array_file(fd,total_size) != nb_bytes)
            { 
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            break;
          }
          case NETSOLVE_C :
          {
            nb_bytes = total_size*SCOMPL;
            if (read_scomplex_array_file(fd,total_size) != nb_bytes)
            { 
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            break;
          }
          case NETSOLVE_Z :
          {
            nb_bytes = total_size*DCOMPL;
            if (read_dcomplex_array_file(fd,total_size) != nb_bytes)
            { 
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            break;
          }
          case NETSOLVE_EXTERNAL :
          {
             /* Do nothing */
             break;
          }
          default :
          {
#ifdef DEBUG
            fprintf(stderr,"Unknown data type ??\n");
#endif
            sprintf(logbuf,"Unknown data type ??");
            append_to_log(logbuf);
            gen_sv_failure(client_desc);
            end_com();
#ifdef HBM
            if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
              fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
            }
#endif
            exit(0);
          }
        }
        close(fd);
      }
      end_message();

      mypid = getpid();
      init_send(encoding,client_desc);
      write_int(&mypid);
      end_message();
   
      sprintf(args[param_index++],"%s",dirname);
      sprintf(args[param_index++],"%d",problem->problem.maj);
   
      if (CONDOR)
      {
        /* Forking to CONDOR */
        /* Creating the command file for CONDOR */
        sprintf(commandfilename,"%s/do.cmd",dirname);
        file = fopen(commandfilename,"w");
        if (file == NULL)
        {
#ifdef DEBUG
          fprintf(stderr,"Unable to create file '%s'\n",commandfilename);
#endif
          gen_sv_failure(client_desc);
          end_com();
#ifdef HBM
          if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
            fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
          }
#endif
          exit(0);
        }
        fprintf(file,"notify = NEVER\n");
        fprintf(file,"executable = %s\n",path);
        fprintf(file,"output = %s/condor.stdout\n",dirname);
        fprintf(file,"error = %s/condor.stderr\n",dirname);
        fprintf(file,"arguments =");
        for (loop=0;loop<nb_param;loop++)
          fprintf(file," %s",args[loop]);
        fprintf(file,"\n");
        fprintf(file,"queue\n");
        
        fclose(file);

        sprintf(commandline,"%s/bin/condor_submit %s",
                CONDOR_path,commandfilename);
        if ((stream = popen(commandline,"r")) == NULL)
        {
          perror("popen");
          gen_sv_failure(client_desc);
          end_com();
#ifdef HBM
          if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
            fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
          }
#endif
          exit(0);
        }
      }
      else
      {
        switch(problem->problem.parallel)
        {
          case PVM:
            /*
             * right now, we assume that the PVM virtual machine is
             * running, contains enough hosts and one of these hosts
             * is the server's host
             */
            newargs = (char **)calloc(nb_param+2,sizeof(char *));
            newargs[0] = (char *)strdup("service-scalapack_two");
            for (i=0;i<nb_param;i++)
              newargs[i+1] = args[i];
            newargs[nb_param+1] = NULL;
  
            execv(path,newargs);
            break;

          case MPI:
            sprintf(commandline,"mpirun -machinefile %s/%s -np %d %s ",
                          root_path,MPImachinefile,MPInodes,
                          path);
            for (i=0;i<nb_param;i++)
            {
              if (args[i] != NULL)
                strcat(commandline,args[i]);
              strcat(commandline," ");
            }
#ifdef VIEW
            fprintf(stderr,"Starting up a parallel machine....\n");
#endif
            if (system(commandline) == -1)
            {
              perror("system");
              fprintf(stderr,"Don't know what to do.....aborting\n");
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            /* Things should be on their way now */
            break;
  
          default:
            fprintf(stderr,"Type of parallel machine not supported\n");
            fprintf(stderr,"Don't know what to do.....aborting\n");
            gen_sv_failure(client_desc);
            end_com();
#ifdef HBM
            if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
              fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
            }
#endif
            exit(0);
        }
      }

      /* Waiting for the 'done' file to be created */
  
#ifdef DEBUG
      fprintf(stderr,"Waiting for the job completion\n");
#endif
      sprintf(donefilename,"%s/done",dirname);
      while ((file = fopen(donefilename,"r")) == NULL)
      {
        sleep(1);
      }
   
      /* What kind of result */
      fscanf(file,"%s",result);
      if (!strcmp(result,"ERROR"))
      {
#ifdef VIEW
        fprintf(stderr,"NetSolve : ERROR from Job\n");
#endif
        gen_sv_failure(client_desc);
        end_com();
#ifdef HBM
        if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
          fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
        }
#endif
        exit(0);
      } 
      fclose(file);
      /* From here, the result is DONE */

      sprintf(typefilename,"%s/type",dirname);
      file = fopen(typefilename,"r");
      if (file == NULL)
      {
#ifdef DEBUG
        fprintf(stderr,"Cannot open file '%s'\n",typefilename);
#endif
        gen_sv_failure(client_desc);
        end_com();
#ifdef HBM
        if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
          fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
        }
#endif
        exit(0);
      }
      if (fread(&type,sizeof(int),1,file) != 1)
      {
        gen_sv_failure(client_desc);
        end_com();
#ifdef HBM
        if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
          fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
        }
#endif
        exit(0);
      }
      fclose(file);

      sprintf(stdoutfilename,"%s/stdout",dirname);
      file = fopen(stdoutfilename,"r");
      if (file == NULL)
      {
#ifdef DEBUG
        fprintf(stderr,"Cannot open file '%s'\n",stdoutfilename);
#endif
        gen_sv_failure(client_desc);
        end_com();
#ifdef HBM
        if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
          fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
        }
#endif
        exit(0);
      }
      if (fread(&stdoutlength,sizeof(int),1,file) != 1)
      {
        gen_sv_failure(client_desc);
        end_com();
#ifdef HBM
        if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
          fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
        }
#endif
        exit(0);
      }
      stdoutbuf = (char *)malloc(sizeof(char)*stdoutlength);
      stdoutbuf[stdoutlength-1] = '\0';
      if (fread(stdoutbuf,sizeof(char),stdoutlength-1,file) != stdoutlength-1)
      {
        gen_sv_failure(client_desc);
        end_com();
#ifdef HBM
        if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
          fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
        }
#endif
        exit(0);
      }
      fclose(file);

      /* Forward the type, stdout and stdoutlength anyway */
      init_send(encoding,client_desc);
      write_int(&type);
      write_int(&stdoutlength);
      write_char_array(stdoutbuf,stdoutlength);

      if (type != SOLUTION)
      {
        end_message();
        end_com();
#ifdef HBM
        if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
          fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
        }
#endif
        exit(0);
      }
      /* Read in the data files and forward them to the client */

      for (i=0;i<pb_desc.nb_obj_output;i++)
      {
        sprintf(datafilename,"%s/output%d",dirname,i);
        fd = open(datafilename,O_RDONLY,0644);
        if (fd <0)
        {
#ifdef DEBUG
          fprintf(stderr,"Unable to open file '%s'\n",datafilename);
#endif
          gen_sv_failure(client_desc);
          end_com();
#ifdef HBM
          if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
            fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
          }
#endif
          exit(0);
        }
        switch(pb_desc.output_obj[i].object_type)
        {
          case NETSOLVE_MATRIX :
            if (read(fd,&maj,sizeof(int)) != sizeof(int))
            {
              close(fd);
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            if (read(fd,&m,sizeof(int)) != sizeof(int))
            {
              close(fd);
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            if (read(fd,&n,sizeof(int)) != sizeof(int))
            {
              close(fd);
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            write_int(&(pb_desc.output_obj[i].object_type));
            write_int(&(pb_desc.output_obj[i].data_type));
            write_int(&maj);
            write_int(&m);
            write_int(&n);
            fprintf(stderr,"Sending Back a %dx%d Matrix\n",m,n);
            break;
          case NETSOLVE_VECTOR :
            if (read(fd,&m,sizeof(int)) != sizeof(int))
            {
              close(fd);
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            write_int(&(pb_desc.output_obj[i].object_type));
            write_int(&(pb_desc.output_obj[i].data_type));
            write_int(&m);
            n = 1;
            break;

          case NETSOLVE_FILE :
          {
            struct stat st;
            stat(datafilename,&st);
            m = (int)(st.st_size);
            write_int(&(pb_desc.output_obj[i].object_type));
            write_int(&(pb_desc.output_obj[i].data_type));
            write_int(&m);
            n = 1;
            break;
          }
 
          case NETSOLVE_SCALAR :
            m = 1;
            n = 1;
            write_int(&(pb_desc.output_obj[i].object_type));
            write_int(&(pb_desc.output_obj[i].data_type));
            break;
        }
        /* Forward it */
        switch(pb_desc.output_obj[i].data_type)
        {
          case NETSOLVE_B :
          {
            int kk, num_vec, vec_len;
            char *buffer;

            nb_bytes = m*n*CHARL;
            buffer=(char *)calloc(nb_bytes,sizeof(char));

            if (read(fd,buffer,nb_bytes) != nb_bytes)
            {
              close(fd);
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }  

            if( (pb_desc.input_obj[i].object_type == NETSOLVE_MATRIX)  &&
                (maj == ROW_MAJOR))
            {
              num_vec = m;
              vec_len = n;
            }
            else
            {
              num_vec = n;
              vec_len = m;
            }
   
            for (kk=0;kk< num_vec;kk++)
            {
              write_byte_array((char *)((char *)buffer+
                       kk*(vec_len)),vec_len);
            }
            /* Now I have the buffer */
            free(buffer);

            break;
          }
          case NETSOLVE_CHAR :
            if (write_char_array_file(fd,m*n) != m*n*sizeof(char))
            {
              close(fd);
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            break;
          case NETSOLVE_ASCII :
            if (write_char_array_file(fd,m*n) != m*n*sizeof(char))
            {
              close(fd);
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            break;
          case NETSOLVE_I :
            if (write_int_array_file(fd,m*n) != m*n*sizeof(int))
            {
              close(fd);
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            break;
          case NETSOLVE_S :
            if (write_single_array_file(fd,m*n) != m*n*sizeof(float))
            {
              close(fd);
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            break;
          case NETSOLVE_D :
            if (write_double_array_file(fd,m*n) != m*n*sizeof(double))
            {
              close(fd);
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            break;
          case NETSOLVE_C :
            if (write_scomplex_array_file(fd,m*n) != m*n*sizeof(scomplex))
            {
              close(fd);
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            break;
          case NETSOLVE_Z :
            if (write_dcomplex_array_file(fd,m*n) != m*n*sizeof(dcomplex))
            {
              close(fd);
              gen_sv_failure(client_desc);
              end_com();
#ifdef HBM
              if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
                fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
              }
#endif
              exit(0);
            }
            break;
        }
        close(fd);
      }
      end_message();
      end_com(); 

      /* Clean the Directory */
#ifdef DEBUG
      fprintf(stderr,"Cleaning up the directory\n");
#endif
      sprintf(commandline,"rm -rf %s",dirname);
      system(commandline);
#ifdef HBM
      if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
        fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
      }
#endif
    }
    free(pb_desc.output_obj);
    free(pb_desc.input_obj);
#ifdef HBM
    if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
      fprintf(stderr, "\n%s:  Error on shutdown.\n");
#endif
    }  
#endif
    exit(0);
  }
}
  /*******************/
  case I_DIE:
  /*******************/
{
  unsigned int IPaddr;
  server_item *server;

  read_IPaddr(&IPaddr);
  end_message();
  end_com();
  server = (server_item *)find_server(IPaddr);
  if (server == 0)
  {
    server = (server_item *)find_agent(IPaddr);
    if (server == 0) /* The server is unknown */
      break;
    else
    {
#ifdef DEBUG
      fprintf(stderr,"I remove %s\n",server->server.hostname);
#endif
      remove_agent(server);
      return 1;
    }
  }
  /* The server is known */
#ifdef DEBUG
  fprintf(stderr,"I remove %s\n",server->server.hostname);
#endif
  remove_server(server);
  return 1;
}

/*****************/
  }  /* END CASE */
/*****************/

  return -1;
}

/*
 * gen_pb_refused() : The problem doesn't correspond to the specification
 *                    I can't solve it...
 */

#ifdef __STDC__
int gen_pb_refused(int encoding)
#else
int gen_pb_refused(encoding)
int encoding;
#endif
{
  int message_type;

  message_type = PB_REFUSED;
  init_send(encoding,client_desc);
  write_int(&message_type);
  /* END OF A SEND PHASE */
  end_message();
  return 1;
}
