/*****************************************************************/
/*      server_util.c                                            */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  my_speed()                                                   */
/*  server_init()                                                */
/*  contact_master()                                             */
/*  print_cache()                                                */
/*  check_security()                                             */
/*  is_unsafe()                                                  */
/*****************************************************************/

#include "server.h"
#include "server_prot.h"
#include "shared_prot.h"
#include "security.h"

/*
 * my_speed()
 *
 *  Returns the Mflops of the host
 */
#ifdef __STDC__
int my_speed()
#else
int my_speed()
#endif
{
 int result;
#ifdef MPP
  FILE *stream;
  char buf[10];
  char buf1[10];
#endif /* MPP */
#ifdef VIEW
  fprintf(stderr,"running LINPACK benchmark......");
  fflush(stderr);
#endif /* VIEW */
#ifndef MPP
  mpp =0;
  result = kflops();
#else /* MPP */
  mpp =1;
  stream = popen("mpp_kflops","r");
  if (stream == 0)
    return -1;
  while(fgets(buf,100,stream) == 0);
  while(fgets(buf1,100,stream) == 0);
  pclose(stream);
  number_processors = atoi(buf1);
  result =  atoi(buf);
#endif
  fprintf(stderr," %d kflops\n",result);
  return result;
}



/*
 * server_init()
 *
 *  Initialize the server
 */

#ifdef __STDC__
int server_init()
#else
int server_init()
#endif
{
  server_desc descriptor;
  FILE *stream;
  int i;

  nb_restrictions = 0;
  for(i=0;i<NB_RESTRICTIONS;i++)
  {
    restrictions[i].max_pending = 0;
    (restrictions[i].domain)[0] = '\0';
  }

  strcpy(descriptor.hostname,my_host);
  descriptor.IPaddr = my_address;
  my_data_format = data_format();
  descriptor.data_format = my_data_format;

  descriptor.speed = my_speed(); 
  descriptor.workload = my_workload();
  descriptor.number_proc = number_processors;
  descriptor.mpp = mpp;
  descriptor.connection = 0;
  descriptor.transfer = 1000000;
  descriptor.host_status   = UP;
  descriptor.server_status = RUNNING;
#ifdef DEBUG
  fprintf(stderr,"my address is %u.%u.%u.%u\n",BYTE(&my_address,0),
                                       BYTE(&my_address,1),
                                       BYTE(&my_address,2),
                                       BYTE(&my_address,3));
#endif
  if ((my_self = (server_item *)new_server(descriptor)) == 0)
  {
    perror("init_server() new_server()");
    return -1;
  }
  nb_pb_solved = 0;
  nb_lcl_request = 0; 
  time(&up_date);
  strcpy(char_up_date,ctime(&up_date));
  
  return 1;
}

/*
 *  contact_master()
 *
 *   contact the master server
 */

#ifdef __STDC__
int contact_master(char *name,char all)
#else
int contact_master(name,all)
char *name;
char all;
#endif
{
int sock;
struct sockaddr_in server_INET_add; /* Server INET socket address  */
struct hostent *hp;                 /* host descriptor             */
int connection;
pb_list *list;                /* List of problems                */
problem_desc *problem;        /* pointer to a problem descriptor */
int type;                     /* Message type                    */
int nb_problems=0;           
int status;                   /* Reply status                    */
int nb_failed_pb;             /* Nb of failed problems           */
problem_desc pb_desc;         /* Problem descriptor              */
server_item *server;
int i,j;
char master_host[256];
static unsigned int master_address=0;


  if (master_address == 0)
  {
    if ((hp = gethostbyname(name)) == NULL)
    {
      fprintf(stderr,"Cannot resolve name '%s'\n",name);
      return -1;
    }
    strcpy(master_host,hp->h_name);
    BCOPY(hp->h_addr_list[0],&master_address,sizeof(unsigned int));
#ifdef VIEW
    fprintf(stderr,"contacting '%s' ...\n",master_host);
#endif
  }
  sock = connect_to_agent(master_host,master_address);
  if (sock == -3)
    return -1;
  if (sock < 0)
  {
    fprintf(stderr,"\tImpossible to contact agent on %s\n",master_host);
    return -1;
  }
  nb_problems = my_self->number_problems;
  list = my_self->first_pb;
  type = I_AM_NEW;
  init_send(DATA_XDR,sock);
  write_int(&type);
  write_string(my_host);
  write_IPaddr(&((my_self->server).IPaddr));
  write_int(&((my_self->server).data_format));
  write_int(&((my_self->server).speed));
  write_int(&((my_self->server).workload));
  write_int(&((my_self->server).number_proc));
  write_int(&((my_self->server).mpp));
  write_int(&nb_problems);
  list = my_self->first_pb;
  for (j=0;j<nb_problems;j++)
  {
    problem = &((list->problem)->problem);
    write_problem(*problem);
    list = list->next;
  }
  end_message();

  /* Get the answer */
  /* BEGINNING OF A RECEIVE PHASE */
  init_recv(sock);  
  read_int(&status);
  if (status == WRONG_PROBLEM) /* Report errors and exit */
  {
    read_int(&nb_failed_pb);
    fprintf(stderr,"----------------------------\n");
    for (j=0;j<nb_failed_pb;j++)
    {
      int i;

      read_problem(&pb_desc);
      fprintf(stderr,"Problem %s badly specified - Should be :\n",pb_desc.code);
      fprintf(stderr,"Nb of objects in input : %d\n",pb_desc.nb_obj_input);
      fprintf(stderr,"Nb of objects in output : %d\n",pb_desc.nb_obj_output);
      fprintf(stderr,"Complexity : %s\n",pb_desc.complexity);
      fprintf(stderr,"Input :\n");
      for (i=0;i< pb_desc.nb_obj_input;i++)
      {
        switch (pb_desc.input_obj[i].data_type)
        {
          case NETSOLVE_B :
          {
             fprintf(stderr," Byte ");
             break;
          } 
          case NETSOLVE_CHAR :
          {
             fprintf(stderr," Character ");
             break;
          }
          case NETSOLVE_I :
          {
             fprintf(stderr," Integer ");
             break;
          }
          case NETSOLVE_S :
          {
             fprintf(stderr," Single precision real ");
             break;
          }
          case NETSOLVE_D :
          {
             fprintf(stderr," Double precision real ");
             break;
          }
          case NETSOLVE_C :
          {
             fprintf(stderr," Single precision complex ");
             break;
          }
          case NETSOLVE_Z :
          {
             fprintf(stderr," Double precision complex ");
             break;
          }
          case NETSOLVE_ASCII :
          {
             fprintf(stderr," ASCII");
             break;
          }
          case NETSOLVE_EXTERNAL :
          {
             fprintf(stderr," External");
             break;
          }
          default :
          {
            fprintf(stderr,"Unknown data type ???");
            return -1;
          }
        }
        switch (pb_desc.input_obj[i].object_type)
        {

          case NETSOLVE_FILE :
          {
             fprintf(stderr," File \n");
             break;
          }
          case NETSOLVE_MATRIX :
          {
             fprintf(stderr," Matrix \n");
             break;
          }
          case NETSOLVE_VECTOR :
          {
             fprintf(stderr," Vector \n");
             break;
          }
          case NETSOLVE_SCALAR :
          {
             fprintf(stderr," Scalar \n");
             break;
          }
          case NETSOLVE_UPF :
          {
            fprintf(stderr,"User provided function \n");
            break;
          }
          default :
          {
            fprintf(stderr,"Unknown object type ???");
            return -1;
          }
        }
      }
      fprintf(stderr,"Output : \n");
      for (i=0;i< pb_desc.nb_obj_output;i++)
      {
        switch (pb_desc.output_obj[i].data_type)
        {

          case NETSOLVE_B :
          {
             fprintf(stderr," Byte ");
             break;
          }
          case NETSOLVE_CHAR :
          {
             fprintf(stderr," Character ");
             break;
          }
          case NETSOLVE_I :
          {
             fprintf(stderr," Integer ");
             break;
          }
          case NETSOLVE_S :
          {
             fprintf(stderr," Single precision real ");
             break;
          }
          case NETSOLVE_D :
          {
             fprintf(stderr," Double precision real ");
             break;
          }
          case NETSOLVE_C :
          {
             fprintf(stderr," Single precision complex ");
             break;
          }
          case NETSOLVE_ASCII :
          {
             fprintf(stderr," ASCII ");
             break;
          }
          case NETSOLVE_Z :
          {
             fprintf(stderr," Double precision complex ");
             break;
          }
          default :
          {
            fprintf(stderr,"Unknown data type ???");
            return -1;
          }
        }
        switch (pb_desc.output_obj[i].object_type)
        {

          case NETSOLVE_FILE :
          {
             fprintf(stderr," File \n");
             break;
          }
          case NETSOLVE_MATRIX :
          {
             fprintf(stderr," Matrix \n");
             break;
          }
          case NETSOLVE_VECTOR :
          {
             fprintf(stderr," Vector \n");
             break;
          }
          case NETSOLVE_SCALAR :
          {
             fprintf(stderr," Scalar \n");
             break;
          }
          default :
          {
            fprintf(stderr,"Unknown object type ???");
            return -1;
          }
        }
      }
      fprintf(stderr,"Formats : %d\n",pb_desc.nb_formats);
      for (i=0;i<pb_desc.nb_formats;i++)
      {
        int j;
        
        fprintf(stderr,"format #%d - %d argument\n",i,pb_desc.formats[i].nb_args);
        for (j=0;j<pb_desc.formats[i].nb_args;j++)
        {
          fprintf(stderr,"--> %s\n",pb_desc.formats[i].args_desc[j]);
        }
      }
     
      fprintf(stderr,"----------------------------\n");
    } 
    end_message();
    return -1;
  }
  if (status == SV_FAILURE) /* The master has failed !!! */
  {
    fprintf(stderr,"The master %s failed\n",master_host);
    fprintf(stderr,"Sorry, impossible to check in\n");
    end_message();
  }
  /* From now, status = YOU_ARE_NEW */
#ifdef VIEW
  fprintf(stderr,"successfully checked in\n");
#endif
  if (get_all_struct() == -1)
  {
    perror("contact_master() get_all_struct()");
    end_com();
    return -1;
  }

  end_message();
  end_com();

/*****************************************/
/* Possible broadcast to other agents    */
/*****************************************/
  if (all == '*')
  { 
    time_t date;
    int rest_delay=5;

    type = I_EXIST;
    for (i=0;i<HASH_SV;i++)
    {

      server = ag_hash_table[i];
      while(server != 0)
      {
        if ((server->server.IPaddr != master_address))
        {
#ifdef DEBUG
          fprintf(stderr,"I report my existence to %s\n",server->server.hostname);
#endif
          sock = connect_to_agent(server->server.hostname,
                                   server->server.IPaddr);
          if (sock == -3)
          {
            perror("contact_master() socket()");
            exit(0);
          }
          if (sock < 0)
          {
#ifdef VIEW
            fprintf(stderr,"WARNING :Agent %s unreachable\n",server->server.hostname);
#endif
            server->server.server_status = STOPPED;
            server->server.host_status = DOWN;
            server=server->next;
            continue;
          }
          /* XDR or RAW */
          if (my_data_format == (server->server.data_format))
            init_send(DATA_RAW,sock);
          else
            init_send(DATA_XDR,sock);
          write_int(&type);
          write_int(&rest_delay);
          rest_delay+= REST_DELAY_INCR;
          write_string(my_host);
          write_IPaddr(&((my_self->server).IPaddr));
          write_int(&((my_self->server).data_format));
          write_int(&((my_self->server).speed));
          write_int(&((my_self->server).workload));
          write_int(&((my_self->server).number_proc));
          write_int(&((my_self->server).mpp));
          write_int(&nb_problems);
          list = my_self->first_pb;
          for (j=0;j<nb_problems;j++)
          {
            problem = &((list->problem)->problem);
            write_problem(*problem);
            list = list->next;
          }
          end_message();
          end_com();
        }
        server = server->next;
      }
    }
  }
  return 1;
}

/*
 * get_all_struct() : get all the data structure from the master
 */

#ifdef __STDC__
int get_all_struct()
#else
int get_all_struct()
#endif
{
  int number_of_pb;
  int number_of_sv;

  problem_desc *pb_desc;
  server_desc sv_desc;
  server_item *server;
  problem_item *problem;

  int nb_of_problems;
  char code[CODE_LENGTH];

  int i,j;

  read_int(&number_of_pb);
  for (j=0;j<number_of_pb;j++)
  {
    pb_desc = (problem_desc *)calloc(1,sizeof(problem_desc));
    read_problem(pb_desc);
    if ((problem_item *)find_problem(pb_desc->code) == NULL)
    {
      if ((problem_item *)new_problem(*pb_desc) == NULL)
      {
        perror("get_all_struct() new_problem()");
        return -1;
      }
    }
    free_pb_desc(pb_desc);
  }
  read_int(&number_of_sv);
#ifdef DEBUG
  fprintf(stderr,"I received %d servers\n",number_of_sv);
#endif
  for (j=0;j<number_of_sv;j++)
  {
    read_string(sv_desc.hostname);
    read_IPaddr(&(sv_desc.IPaddr));
    read_int(&(sv_desc.data_format));
    read_int(&(sv_desc.speed));
    read_int(&(sv_desc.host_status));
    read_int(&(sv_desc.server_status));
    read_int(&(sv_desc.workload));
    read_int(&(sv_desc.number_proc));
    read_int(&(sv_desc.mpp));
    sv_desc.connection = TO_BE_A; /* Waiting a value */
    sv_desc.transfer   = TO_BE_A; /* Waiting a value */
    read_int(&nb_of_problems);
    if ((strcmp(sv_desc.hostname,my_host) == 0)&&
        (nb_of_problems == my_self->number_problems))
    {
        /* It is myself !!! */
    }
    else
    {
      if (nb_of_problems != 0)
      {
        if ((server =(server_item *)new_server(sv_desc)) == NULL)
        {
          perror("get_all_struct() new_server()");
          return -1;
        }
      }
      else
      {
        if ((server =(server_item *)new_agent(sv_desc)) == NULL)
        {
          perror("get_all_struct() new_agent()");
          return -1;
        }
      }
    }
    for (i=0;i<nb_of_problems;i++)
    {
      read_string(code);
      if ((strcmp(sv_desc.hostname,my_host) == 0)&&
         (nb_of_problems == my_self->number_problems))
      {
         /* It's myself */
      }
      else
      { 
        problem =(problem_item *)find_problem(code);
        add_pb_to_server(problem,server);
        add_server_to_pb(problem,server);
      }
    }
  }
  return 1;
}

  
/*
 * check_security()
 */
#ifdef __STDC__
int check_security(char *error_report,char *filename)
#else
int check_security(error_report,filename)
char *error_report;
char *filename;
#endif
{
  FILE *stream;
  char buf[1];
  int info;
  char line[256];
  char command_line[256];

  char *start = error_report;

  int result = 1;

  sprintf(command_line,"nm -u ./%s",filename);
  stream = popen(command_line,"r");
  if (stream == NULL)
  {
    perror("popen()");
    return 0;
  }

  while( fgets(line,256,stream) != NULL)
  {
    char *begin = line;
    char *end;

    while((*begin == '_') || (*begin == ' '))
      begin++;

    end = (char *)((char *)line + strlen(line))-1;
    while ((*end == '\n')||(*end == ' ')||(*end == '_'))
      *(end--)='\0';
    
    if (is_unsafe(begin))
    {
      if (start+strlen(begin)+1 < error_report+256)
      {
        strncpy(start,begin,strlen(begin));
        start+=strlen(begin);
        *start=',';
        start++;
      }
      result = UPF_UNSAFE;
    }
  } 
  pclose(stream);
  return result;
}

/*
 * is_unsafe()
 */
#ifdef __STDC__
int is_unsafe(char *name)
#else
int is_unsafe(name)
char *name;
#endif
{
  int i;

  uppercase(name);
#ifdef PLAT_SUN4SOL2
  if (strncmp(name,"UNDEFINED SYMBOLS FROM",22) == 0)
    return 0;
  if (name[0] == '\0')
    return 0;
#endif
  for(i=0;i<NB_ALLOWED;i++)
  {
    if (!strcmp(name,allowed[i]))
      return 0;
  }
  return 1;
}

/*
 * are_undefined()
 */
#ifdef __STDC__
int are_undefined(char *filename)
#else
int are_undefined(filename)
char *filename;
#endif
{
  FILE *stream;
  char line[256];
  char command_line[256];

  sprintf(command_line,"nm -u ./%s",filename);
  stream = popen(command_line,"r");
  if (stream == NULL)
  {
    perror("popen()");
    return 0;
  }

  while ( fgets(line,256,stream) != NULL)
  {
#ifdef PLAT_SUN4SOL2
    if (strncmp(line,"Undefined symbols from",22) == 0)
      continue;
    if (strcmp(line,"\n") == 0)
      continue;
#endif
    fprintf(stderr,"===> '%s'\n",line);
    return 1;
  }
  pclose(stream);
  return 0;
}

