/*****************************************************************/
/*      service.c                                                */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  service()                                                    */
/*  transpose_matrix()                                           */
/*****************************************************************/

#include "server.h"
#include "server_prot.h"
#include "shared_prot.h"
#ifdef HBM
#include "hbm_client.h"
#endif

/*
 * LINUX :
 *
 *  Faking a Fortran-like main routine for Linux !!!
 */
#ifdef PLAT_LINUX
MAIN__(){}
#endif

/*
 * service 
 *
 *  Process in charge with a problem solve
 */

#ifdef __STDC__
int main(int argc,char **argv)
#else
int main(argc,argv)
int argc;
char **argv;
#endif
{
  char hbmdchost[256];
  int mypid;
  int i,dummy;
  int total_size;
  int mode;
  int sock;
  problem_desc pb_desc;
  input_sizes *sizes_input;
  input_sizes *sizes_output;
  void **data_input;
  void **data_output;
  FILE *file;
  char filename[128];
  char command_line[256];
  char logbuf[256];
  time_t current;
  char *condordir;
  int *majors;
  int client_major;
  int retcd;
  int param_index=1;

#ifndef ISCONDOR
#ifndef NOLOG
  set_server_log();
#endif
#endif
  time(&current);

  /* Where to put the files if any */

  scratch_path = (char*)strdup("/tmp/");

#ifdef VIEW
  fprintf(stderr,"Computing process started...\n");
#endif

#ifdef ISCONDOR
  argv = &(argv[1]);
#endif
  strcpy(hbmdchost,argv[param_index++]);
#ifndef ISCONDOR

#ifdef HBM
  /*
   * Register to Globus the HBM
   */
  if ((retcd = hbm_register("", 0, NULL,hbmdchost)) < 0) {
    gen_sv_failure(client_desc);
    end_com();
    exit(0);
  }
#ifdef DEBUG
  fprintf(stderr,"Computing process '%d' just registered with the HBMDC\n",
            getpid());
#endif
#endif

#endif

  strcpy(pb_desc.code,argv[param_index++]);
  pb_desc.nb_obj_input = atoi(argv[param_index++]);
  if ((pb_desc.input_obj = 
           (object *)calloc(pb_desc.nb_obj_input,sizeof(object))) == NULL)
  {
#ifdef ISCONDOR
    create_done_file(condordir,-1);
#else
    gen_sv_failure(sock);
    end_com();
#ifdef HBM
    if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
      fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
    }
#endif
#endif
    exit(0);
  }
  for (i = 0;i< pb_desc.nb_obj_input; i++)
  {
    pb_desc.input_obj[i].data_type = atoi(argv[param_index++]);
    pb_desc.input_obj[i].object_type = atoi(argv[param_index++]);
  }
  pb_desc.nb_obj_output = atoi(argv[param_index++]);

  if ((pb_desc.output_obj =
          (object *)calloc(pb_desc.nb_obj_output,sizeof(object))) == NULL)
  {
#ifdef ISCONDOR
    create_done_file(condordir,-1);
#else
    gen_sv_failure(sock);
    end_com();
#ifdef HBM
    if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
      fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
    }
#endif
#endif
    exit(0);
  }
  for (i = 0;i< pb_desc.nb_obj_output; i++)
  {
    pb_desc.output_obj[i].data_type = 
                    atoi(argv[param_index++]);
    pb_desc.output_obj[i].object_type = 
                    atoi(argv[param_index++]);
  }
  sock = atoi(argv[param_index++]);
  mode = atoi(argv[param_index++]);
  
  /* I skip the BLACS stuff */
  param_index++;

  /* I get the CONDOR stuff */

  condordir = argv[param_index++];
  if (strcmp(condordir,"-") == 0)
    condordir = NULL;

  pb_desc.maj = atoi(argv[param_index++]);

  /* Allocate the sizes arrays */

  if ((sizes_input = 
     (input_sizes *)calloc(pb_desc.nb_obj_input,sizeof(input_sizes))) == NULL)
  {
#ifdef ISCONDOR
    create_done_file(condordir,-1);
#else
    gen_sv_failure(sock);
    end_com();
#ifdef HBM
    if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
      fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
    }
#endif
#endif
    exit(0);
  }
 
  if ((majors = (int *)calloc(pb_desc.nb_obj_input,sizeof(int))) == NULL)
  {
#ifdef ISCONDOR
    create_done_file(condordir,-1);
#else 
    gen_sv_failure(sock);
    end_com();
#ifdef HBM
    if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
      fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
    }
#endif
#endif
    exit(0);
  }
  for (i=0;i<pb_desc.nb_obj_input;i++)
    majors[i] = -1;

  if ((sizes_output = 
    (input_sizes *)calloc(pb_desc.nb_obj_output,sizeof(input_sizes))) == NULL)
  {
#ifdef ISCONDOR
    create_done_file(condordir,-1);
#else
    gen_sv_failure(sock);
    end_com();
#ifdef HBM
    if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
      fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
    }
#endif
#endif
    exit(0);
  }

  /* Allocate the spaces */

  if ((data_input = 
         (void **)calloc(pb_desc.nb_obj_input,sizeof(void *))) == NULL)
  {
#ifdef ISCONDOR
    create_done_file(condordir,-1);
#else
    gen_sv_failure(sock);
    end_com();
#ifdef HBM
    if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
      fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
    }
#endif
#endif
    exit(0);
  }

  if ((data_output = 
         (void **)calloc(pb_desc.nb_obj_output,sizeof(void *)) ) == NULL)
  {
#ifdef ISCONDOR
    create_done_file(condordir,-1);
#else
    gen_sv_failure(sock);
    end_com();
#ifdef HBM
    if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
      fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
    }
#endif
#endif
    exit(0);
  }

  /* Storing the input parameters */

#ifndef ISCONDOR
  if (!condordir)
  {
    init_recv(sock);
  }
#endif

  for (i=0;i<pb_desc.nb_obj_input;i++)
  {
#ifdef DEBUG
    fprintf(stderr,"Object #%d \n",i);
#endif
#ifndef NOLOG
    sprintf(logbuf,"Object #%d ",i);
#endif
#ifndef ISCONDOR
#ifndef NOLOG
    append_to_log(logbuf);
#endif
#endif

#ifdef ISCONDOR
    sprintf(filename,"%s/input%d",condordir,i);
    file = fopen(filename,"r");
    if (file == NULL)
    {
#ifdef DEBUG
      fprintf(stderr,"Cannot open file '%s'\n",filename);
#endif
      create_done_file(condordir,-1);
      exit(0);
    }
#endif
    switch(pb_desc.input_obj[i].object_type) 
    {
      case NETSOLVE_MATRIX :
      {
#ifdef ISCONDOR
        if (fread(&(majors[i]),sizeof(int),1,file) != 1)
        {
          fclose(file);
          create_done_file(condordir,-1);
          exit(0);
        }
        if (fread(&(sizes_input[i].m),sizeof(int),1,file) != 1)
        {
          fclose(file);
          create_done_file(condordir,-1);
          exit(0);
        }
        if (fread(&(sizes_input[i].n),sizeof(int),1,file) != 1)
        {
          fclose(file);
          create_done_file(condordir,-1);
          exit(0);
        }
#else
        read_int(&dummy);
        read_int(&dummy);
        read_int(&(majors[i]));
        read_int(&(sizes_input[i].m));
        read_int(&(sizes_input[i].n));
        
#endif
     
#ifdef DEBUG
        fprintf(stderr,"MATRIX %d %d \n",sizes_input[i].m,sizes_input[i].n);
#endif
#ifndef NOLOG
        sprintf(logbuf,"MATRIX %d %d : type %d",sizes_input[i].m,
                sizes_input[i].n,pb_desc.input_obj[i].data_type);
#endif
#ifndef ISCONDOR
#ifndef NOLOG
        append_to_log(logbuf);
#endif
#endif
        break;
      }

      case NETSOLVE_VECTOR :
      {
#ifdef ISCONDOR
        if (fread(&(sizes_input[i].m),sizeof(int),1,file) != 1)
        {
          fclose(file);
          create_done_file(condordir,-1);
          exit(0);
        }
#else
        read_int(&dummy);
        read_int(&dummy);
        read_int(&(sizes_input[i].m));
#endif
#ifdef DEBUG
        fprintf(stderr,"VECTOR %d\n",sizes_input[i].m);
#endif
#ifndef NOLOG
        sprintf(logbuf,"VECTOR %d : type %d",sizes_input[i].m,
                pb_desc.input_obj[i].data_type);
#endif
#ifndef ISCONDOR
#ifndef NOLOG
        append_to_log(logbuf);
#endif
#endif
        sizes_input[i].n = 1;
        break;
      }
      case NETSOLVE_FILE :
      {
#ifdef ISCONDOR
        /* NOTHING */
#else
        read_int(&dummy);
        read_int(&dummy);
        read_int(&(sizes_input[i].m));
#endif
#ifdef DEBUG
        fprintf(stderr,"FILE %d\n",sizes_input[i].m);
#endif
#ifndef NOLOG
        sprintf(logbuf,"VECTOR %d : type %d",sizes_input[i].m,
                pb_desc.input_obj[i].data_type);
#endif
#ifndef ISCONDOR
#ifndef NOLOG
        append_to_log(logbuf);
#endif
#endif
        sizes_input[i].n = 1;
        break;
      }
      case NETSOLVE_SCALAR :
      {
#ifdef DEBUG
        fprintf(stderr,"SCALAR\n");
#endif
#ifndef NOLOG
        sprintf(logbuf,"SCALAR : type %d",pb_desc.input_obj[i].data_type);
#endif
#ifndef ISCONDOR
#ifndef NOLOG
        append_to_log(logbuf);
#endif
#endif
        sizes_input[i].m = 1;
        sizes_input[i].n = 1;
#ifdef ISCONDOR

#else
        read_int(&dummy);
        read_int(&dummy);
#endif
        break;
      }

      case NETSOLVE_UPF :
      {
         /* Do nothing */
         break;
      }
      default :
      {
#ifdef DEBUG
        fprintf(stderr,"Unknown object type ??\n");
#endif
#ifndef NOLOG
        sprintf(logbuf,"Unknown object type ??");
#endif
#ifndef ISCONDOR
#ifndef NOLOG
        append_to_log(logbuf);
#endif
#endif
#ifdef ISCONDOR
        create_done_file(condordir,-1);
#else
        gen_sv_failure(sock);
        end_com();
#ifdef HBM
        if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
          fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
        }
#endif
#endif
        exit(0);
      }
    }
    total_size = (sizes_input[i].m)*(sizes_input[i].n);
    switch(pb_desc.input_obj[i].data_type)
    {
      case NETSOLVE_I :
      {
        if ((data_input[i] = (int *)calloc(total_size,INTL)) == NULL)
        {
#ifdef ISCONDOR
          create_done_file(condordir,-1);
#else
          gen_sv_failure(sock);
          end_com();
#ifdef HBM
          if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
            fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
          }
#endif
#endif
          exit(0);
        }
#ifdef ISCONDOR
        fread(data_input[i],sizeof(int),total_size,file);
#else
        read_int_array(data_input[i],total_size);
#endif
        break;
      }
      case NETSOLVE_CHAR :
      {
        if ((data_input[i] = (char *)calloc(total_size,sizeof(char))) == NULL)
        {
#ifdef ISCONDOR
          create_done_file(condordir,-1);
#else
          gen_sv_failure(sock);
          end_com();
#ifdef HBM
          if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
            fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
          }
#endif
#endif
          exit(0);
        }
#ifdef ISCONDOR
        fread(data_input[i],sizeof(char),total_size,file);
#else
        read_char_array(data_input[i],total_size);
#endif
        break;
      }
      case NETSOLVE_ASCII :
      {
#ifdef ISCONDOR
         data_input[i] = (char *)strdup(filename);
#else
        int fd;
        char filename[256];
        sprintf(filename,"%s/input%d.%d",scratch_path,i,getpid());
        fd = open(filename,O_RDWR|O_CREAT,0600);
        if (fd <0)
        {
          fprintf(stderr,"Impossible to create file '%s'\n",filename);
          gen_sv_failure(sock);
          end_com();
#ifdef HBM
          if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) 
          {
#ifdef DEBUG
            fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
          }
#endif
          exit(0);
        }
        read_char_array_file(fd,total_size);
        data_input[i] = (char *)strdup(filename);
        close(fd);
#endif
        break;
      }

     case NETSOLVE_B :
      {
        if ((data_input[i] = 
           (char *)calloc(total_size,sizeof(char))) == NULL)
        {
#ifdef ISCONDOR
          create_done_file(condordir,-1);
#else
          gen_sv_failure(sock);
          end_com();
#ifdef HBM
          if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
            fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
          }
#endif
#endif
          exit(0);
        }
#ifdef ISCONDOR
        fread(data_input[i],sizeof(char),total_size,file);
#else
        {
        int kk, num_vec, vec_len;

        if( (pb_desc.input_obj[i].object_type == NETSOLVE_MATRIX)  &&
            (majors[i] == ROW_MAJOR)) 
        {
          num_vec = sizes_input[i].m;
          vec_len = sizes_input[i].n;
        } 
        else 
        {
          num_vec = sizes_input[i].n;
          vec_len = sizes_input[i].m;
        }

        for (kk=0;kk< num_vec;kk++)
        {
          read_byte_array((char *)((char *)(data_input[i])+
                   kk*(vec_len)),vec_len);
        }
        }
#endif
        break;
      }

      case NETSOLVE_S :
      {
        if ((data_input[i] = (float *)calloc(total_size,FLOATL)) == NULL)
        {
#ifdef ISCONDOR
          create_done_file(condordir,-1);
#else
          gen_sv_failure(sock);
          end_com();
#ifdef HBM
          if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
            fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
          }
#endif
#endif
          exit(0);
        }
#ifdef ISCONDOR
        fread(data_input[i],sizeof(float),total_size,file);
#else
        read_single_array(data_input[i],total_size);
#endif
        break;
      }
      case NETSOLVE_D :
      {
        if ((data_input[i] = (double *)calloc(total_size,DOUBL)) == NULL)
        {
#ifdef ISCONDOR
          create_done_file(condordir,-1);
#else
          gen_sv_failure(sock);
          end_com();
#ifdef HBM
          if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
            fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
          }
#endif
#endif
          exit(0);
        }
#ifdef ISCONDOR
        fread(data_input[i],sizeof(double),total_size,file);
#else
        read_double_array(data_input[i],total_size);
#endif
        break;
      }
      case NETSOLVE_C :
      {
        if ((data_input[i] = (scomplex *)calloc(total_size,SCOMPL)) == NULL)
        {
#ifdef ISCONDOR
          create_done_file(condordir,-1);
#else
          gen_sv_failure(sock);
          end_com();
#ifdef HBM
          if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
            fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
          }
#endif
#endif
          exit(0);
        }
#ifdef ISCONDOR
        fread(data_input[i],sizeof(scomplex),total_size,file);
#else
        read_scomplex_array(data_input[i],total_size);
#endif
        break;
      }
      case NETSOLVE_Z :
      {
        if ((data_input[i] = (dcomplex *)calloc(total_size,DCOMPL)) == NULL)
        {
#ifdef ISCONDOR
          create_done_file(condordir,-1);
#else
          gen_sv_failure(sock);
          end_com();
#ifdef HBM
          if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
            fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
          }
#endif
#endif
          exit(0);
        }
#ifdef ISCONDOR
        fread(data_input[i],sizeof(dcomplex),total_size,file);
#else
        read_dcomplex_array(data_input[i],total_size);
#endif
        break;
      }
      case NETSOLVE_EXTERNAL :
      {  
         /* Do nothing */
         break;
      }
      default :
      {
#ifdef DEBUG
        fprintf(stderr,"Unknown data type ??\n");
#endif
#ifndef NOLOG
        sprintf(logbuf,"Unknown data type ??");
#endif
#ifndef ISCONDOR
#ifndef NOLOG
        append_to_log(logbuf);
#endif
#endif
#ifdef ISCONDOR
        create_done_file(condordir,-1);
#else
        gen_sv_failure(sock);
        end_com();
#ifdef HBM
        if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
          fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
        }
#endif
#endif
        exit(0);
      }
    }
#ifdef ISCONDOR
    fclose(file);
#endif
  }
#ifndef ISCONDOR
  end_message();
#endif

#ifdef VIEW
  fprintf(stderr,"Data obtained...\n");
#endif

  /* Sending my pid back to the client for the HBM */
  init_send(mode,sock);
  mypid = getpid();
  write_int(&mypid);
  end_message();

  /* DUMMY STUFF FOR DEBUG */
  init_recv(sock);
  read_int(&dummy);
  end_message();


#ifdef VIEW
  fprintf(stderr,"Re-majoring the matrices ?\n");
#endif
  client_major = -1;
  for (i=0;i<pb_desc.nb_obj_input;i++)
  {
    if (majors[i] == -1)
      continue;
    if (client_major == -1)
    {
      client_major = majors[i];
    }
    else
    {
      if (client_major != majors[i])
      { 
#ifdef VIEW
        fprintf(stderr,"Incompatible majors in input !!\n");
#endif
#ifdef ISCONDOR
        create_done_file(condordir,-1);
#else
        gen_sv_failure(sock);
        end_com();
#ifdef HBM
        if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
          fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
        }
#endif
#endif
        exit(0);
      }
    }
  }
  /* At this point, the majors are consistent   */
  /* I can do the real re-majoring if necessary */
  if (pb_desc.maj != client_major)
  {
    /* Transpose everything */
    for (i=0;i<pb_desc.nb_obj_input;i++)
    {
      if (pb_desc.input_obj[i].object_type == NETSOLVE_MATRIX) 
      {  
        /* Need to transpose it */
        transpose_matrix(client_major,
                         pb_desc.input_obj[i].data_type,&(data_input[i]),
                         sizes_input[i].m,sizes_input[i].n);
      } 
    }
  }

  /* Assigning a name to the files in output if any */
#ifdef VIEW
  fprintf(stderr,"Files in output ?\n");
#endif
  for (i=0;i<pb_desc.nb_obj_output;i++)
  {
    char filename[256];

    if (pb_desc.output_obj[i].object_type != NETSOLVE_FILE)
      continue;
    sprintf(filename,"%s/output%d.%d",scratch_path,i,getpid());
    data_output[i] = (char*)strdup(filename);
  }

#ifdef VIEW
  fprintf(stderr,"Solving '%s'....\n",pb_desc.code);
#endif
  time(&current);
#ifndef NOLOG
  sprintf(logbuf,"Compute() ... '%s' at %s....",pb_desc.code,ctime(&current));
#endif
#ifndef ISCONDOR
#ifndef NOLOG
  append_to_log(logbuf);
#endif
#endif

/*
 * Perform the computations
 */

  if  (compute(&pb_desc,sizes_input,sizes_output,
                   data_input,data_output,sock,
                   mode,condordir,client_major) == -1)
  {
    char *stdoutbuf;
    int stdoutlength;
    int type = SV_FAILURE;

    time(&current);
#ifndef NOLOG
    sprintf(logbuf,"'%s' Failed at %u",pb_desc.code,ctime(&current));
#endif
#ifndef ISCONDOR
#ifndef NOLOG
    append_to_log(logbuf);
#endif
#endif

    stdoutbuf = (char *)strdup("Server Error : No stdout available");
    stdoutlength = strlen(stdoutbuf)+1;

#ifdef ISCONDOR
     sprintf(filename,"%s/type",condordir);
     file = fopen(filename,"w");
     if (file == NULL)
     {
#ifdef DEBUG
       fprintf(stderr,"Cannot create file '%s'\n",filename);
#endif
       create_done_file(condordir,-1);
       exit(0);
     } 
     if (fwrite(&type,sizeof(int),1,file) != 1)
     {
       create_done_file(condordir,-1);
       exit(0);
     }
     fclose(file);
     sprintf(filename,"%s/stdout",condordir);
     file = fopen(filename,"w");
     if (file == NULL)
     {
#ifdef DEBUG
       fprintf(stderr,"Cannot create file '%s'\n",filename);
#endif
       create_done_file(condordir,-1);
       exit(0);
     }
     if (fwrite(&stdoutlength,sizeof(int),1,file) != 1)
     {
       create_done_file(condordir,-1);
       exit(0);
     }
     if (fwrite(&stdoutbuf,sizeof(char),stdoutlength,file) != stdoutlength)
     {
       create_done_file(condordir,-1);
       exit(0);
     }
     fclose(file);
#else
    init_send(mode,sock);
    write_int(&type);
    write_int(&stdoutlength);
    write_char_array(stdoutbuf,stdoutlength);
    end_message();
    end_com();
#endif
    free(stdoutbuf);
  }
  
#ifdef VIEW
  fprintf(stderr,"'%s' Completed\n",pb_desc.code);
#endif
  time(&current);
#ifndef NOLOG
  sprintf(logbuf,"'%s' Completed at %s",pb_desc.code,ctime(&current));
#endif
#ifndef ISCONDOR
#ifndef NOLOG
  append_to_log(logbuf);
#endif
#endif

 /*
  *    Possible clean-up for UPF, FILES, etc..
  */

#ifndef ISCONDOR
  for (i=0;i<pb_desc.nb_obj_input;i++)
  {
    char command[256];

    if (pb_desc.input_obj[i].object_type != NETSOLVE_FILE)
      continue;

    sprintf(command,"rm -f %s",data_input[i]);
    system(command);
  }
  for (i=0;i<pb_desc.nb_obj_output;i++)
  {
    char command[256];

    if (pb_desc.output_obj[i].object_type != NETSOLVE_FILE)
      continue;

    sprintf(command,"rm -f %s",data_input[i]);
    system(command);
  }
#endif 

#ifndef ISCONDOR
  chdir("/");
  sprintf(command_line,"rm -rf /tmp/NetSolve_upf.%d",getpid());
  system(command_line); 
#endif

#ifdef ISCONDOR
  create_done_file(condordir,1);
#else
#ifdef HBM
  if (hbm_shutdown("", 0, NULL, HBM_MSGTYPE_SHUTDOWN_NORMAL)) {
#ifdef DEBUG
    fprintf(stderr, "\n%s:  HBM: cannot shutdown.\n");
#endif
  }
#endif
#endif
  exit(0);
} 

#ifdef __STDC__
int create_done_file(char *dirname,int code)
#else
int create_done_file(dirname,code)
char *dirname;
int code;
#endif
{
  FILE *file;
  char name[128];

  sprintf(name,"%s/done",dirname);
  file = fopen(name,"w");
  if (file == NULL)
  {
#ifdef DEBUG
     fprintf(stderr,"Cannot create file '%s'\n",name);
#endif
     return -1; 
  } 
  if (code == -1)
    fprintf(file,"ERROR");
  else
    fprintf(file,"DONE");

  fclose(file);
  return 1;
}


/*
 * transpose_matrix()
 */
#ifdef __STDC__
int transpose_matrix(int original_major,int data_type,
                     void **ptr,int m,int n)
#else
int transpose_matrix(original_major,data_type,ptr,m,n)
int original_major;
int data_type;
void **ptr;
int m,n;
#endif
{
  int block_size;
  int block_number;
  void *tmp;
  void *new;
  int offset;
  int i,j;
  int count;

  switch(original_major)
  {
    case COL_MAJOR:
      block_size = m;
      block_number = n;
      break;
    case ROW_MAJOR:
      block_size = n;
      block_number = m;
      break;
  }

  switch(data_type)
  {
    case NETSOLVE_CHAR:
      new = (char *)calloc(m*n,sizeof(char));
      break;

    case NETSOLVE_B:
      new = (char *)calloc(m*n,sizeof(char));
      break;

    case NETSOLVE_I:
      new = (int *)calloc(m*n,sizeof(int));
      break;

    case NETSOLVE_S:
      new = (float *)calloc(m*n,sizeof(float));
      break;

    case NETSOLVE_D:
      new = (double *)calloc(m*n,sizeof(double));
      break;

    case NETSOLVE_C:
      new = (scomplex *)calloc(m*n,sizeof(scomplex));
      break;

    case NETSOLVE_Z:
      new = (dcomplex *)calloc(m*n,sizeof(dcomplex));
      break;

    case NETSOLVE_EXTERNAL:
      /* Does not happen */
      return -1;
      break;
  }

  count = 0;
  for (i=0;i<block_size;i++)
  {
    for (j=0;j<block_number;j++)
    {
      switch(data_type)
      {
        case NETSOLVE_CHAR:
          ((char *)new)[count] = ((char *)*ptr)[j*block_size+i];
          break;

        case NETSOLVE_B:
          ((char *)new)[count] = ((char *)*ptr)[j*block_size+i];
          break;

        case NETSOLVE_I:
          ((int *)new)[count] = ((int *)*ptr)[j*block_size+i];
          break;

        case NETSOLVE_S:
          ((float *)new)[count] = ((float *)*ptr)[j*block_size+i];
          break;

        case NETSOLVE_D:
          ((double *)new)[count] = ((double *)*ptr)[j*block_size+i];
          break;

        case NETSOLVE_C:
          ((scomplex *)new)[count] = ((scomplex *)*ptr)[j*block_size+i];
          break;

        case NETSOLVE_Z:
          ((dcomplex *)new)[count] = ((dcomplex *)*ptr)[j*block_size+i];
          break;
      }
      count++;
    }
  }
  tmp = *ptr;
  *ptr = new;
  free(tmp);
  
  return 1;
}

#ifdef ISCONDOR
/*
 * uppercase() : change a string in uppercase
 */
#ifdef __STDC__
int uppercase(char *string)
#else
int uppercase(string)
char *string;
#endif
{
  int length,i;

  length = strlen(string);
  for (i=0;i<length;i++)
    if ((string[i] >= 97)&&(string[i]<=122))
      string[i] = string[i]-32;
}

#endif
