/*****************************************************************/
/*      workload_util.c                                          */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  alarm_handler()                                              */
/*  workload_death()                                             */
/*  run_workload_process()                                       */
/*  my_workload()                                                */
/*****************************************************************/

#include "server.h"
#include "server_prot.h"
#include "shared_prot.h"

/*
 * workload_death()
 */
#ifdef __STDC__
void workload_death(int sig)
#else
void workload_death(sig)
int sig;
#endif
{
#ifdef DEBUG
  fprintf(stderr,"End of workload management process : %d\n",getpid());
#endif
  end_com();
  exit(0);
}

/*
 * workload_handler()
 */
#ifdef __STDC__
void workload_handler(int sig)
#else
void workload_handler(sig)
int sig;
#endif
{
#ifdef VIEW
  fprintf(stderr,"Warning : Connect timeout expired\n");
#endif
#ifdef SYSVSIGNAL
    signal(SIGALRM,workload_handler);
#endif
}

/*
 * run_workload_process()
 */
#ifdef __STDC__
int run_workload_process()
#else
int run_workload_process()
#endif
{
  int pid;
  FILE *stream;
  int i;
  int sock;
  int type;
  struct sockaddr_in server_INET_add;
  int nb_server;
  struct hostent *hp;
  int connection;
  int diff;
  int workload;
  server_item *server;
  time_t date,date1,date2;
  int *last_broadcasted;
  int index;
  void workload_handler();
  void workload_death();
  int *force;


  pid = fork();

  if (pid != 0) /* I am the father */
    return pid;

  /* I am the son */
  close (rmt_sock);
  signal(SIGINT,workload_death);
  signal(SIGKILL,workload_death);
  signal(SIGTERM,workload_death);
  alarm(0);
  signal(SIGALRM,workload_handler);
  nb_server = 0;

  for (i=0;i<HASH_SV;i++)
  {
    server = ag_hash_table[i];
    while(server != NULL)
    {
      nb_server++;
      server = server->next;
    }   
  }
  last_broadcasted = (int *)calloc(nb_server,INTL);
  force = (int *)calloc(nb_server,INTL);
  for (i=0;i<nb_server;i++) {
    last_broadcasted[i] = -1;
    force[i] = 0;
  }
  
  while(1)
  {
    time(&date1);
    index = -1;
    for(i=0;i<HASH_SV;i++)
    {
      server = ag_hash_table[i];
      while(server != 0)
      {
        index++;
        if ((server->server.host_status == DOWN) ||
            (server->number_problems != 0)       ||
            (server->server.server_status == STOPPED))
        {
          server = server->next;
          continue;
        }
        /* Send my workload to the server */
#ifdef DEBUG
        fprintf(stderr,"I broadcast my workload to %s\n",server->server.hostname);
#endif
        workload = my_workload();
        if (last_broadcasted[index] == -1)
        {
          diff = 0;
          last_broadcasted[index] = workload;
        }
        else
          diff = last_broadcasted[index]-workload;

#ifdef DEBUG
         fprintf(stderr,"diff = %d\n",diff);
#endif
        if ((diff<=WORKLOAD_CONFIDENCE)&&
            (diff>=-WORKLOAD_CONFIDENCE) &&
            (force[index] <= FORCE_BROADCAST))
        {
          force[index]= force[index] + 1;
          server = server->next;
          break;
        }
        force[index] = 0;
        time(&date);
        alarm(CONNECT_TIMEOUT);
        sock = connect_to_agent(server->server.hostname,
                                 server->server.IPaddr);
        alarm(0);
        if (sock == -3)
        {
          perror("run_workload_process() socket()");
          exit(0);
        }
        if (sock < 0)
        {
#ifdef VIEW
          fprintf(stderr,"WARNING :Agent %s unreachable\n",server->server.hostname);
#endif
          /* Report the failure to the master */
          server = server->next;
          break;
        }
#ifdef DEBUG
         fprintf(stderr,"I send my workload to %s : %d\n",
                server->server.hostname,workload);
#endif
        if (my_data_format == server->server.data_format)
          init_send(DATA_RAW,sock); 
        else
        init_send(DATA_XDR,sock);
        type = WORKLOAD;
        write_int(&type);
        write_IPaddr(&my_address);
        write_int(&workload);
        end_message();
        end_com();
        last_broadcasted[index] = workload;
        server = server->next;
      }
    }
    time(&date2);
    sleep((WORKLOAD_TIME - (date2-date1))<0 ? 0 : (WORKLOAD_TIME-(date2-date1)));
  }
} 


/*
 * my_workload()
 */
#ifdef __STDC__
int my_workload()
#else
int my_workload()
#endif
{
  char buf[200];
  char *tmp;
  char *tmp2;
  float essai;
  FILE *stream;
#ifdef PLAT_SGI64
  char *mystrstr();
#endif

#ifdef PLAT_SGI5
  return 58;
#endif
#ifndef PLAT_T3E
  stream = popen("uptime","r");
  if (stream == 0)
    return (my_self->server.workload);
  while(fgets(buf,100,stream) == 0);

  pclose(stream);

#ifdef PLAT_SGI64
  tmp = (char *)mystrstr(buf,"average"); 
#else
  tmp = (char *)strstr(buf,"average"); 
#endif

  tmp+=7;
  
  while(((*tmp) > '9') || ((*tmp)<'0'))
  {
   (char *) tmp++;
  }
  tmp2 = (char *)tmp;
  while( ((*tmp2 <='9')&&(*tmp2 >='0'))||(*tmp2 == '.'))
    (char *)tmp2++;
  *tmp2 = '\0';
  sscanf(tmp,"%f",&essai);
  return ((int)(essai*100));
#else
  return 50;
#endif

}

#ifdef PLAT_SGI64
#ifdef __STDC__
char *mystrstr(char*s1,char*s2)
#else
char *mystrstr(s1,s2)
char *s1;
char *s2;
#endif
{
  int l1 = strlen(s1);
  int l2 = strlen(s2);

  int i,j;

  for (i=0;i<l1-l2;i++)
  {
    int equal = 1;
    for (j=0;j<l2;j++)
    {
      if (s1[i+j] != s2[j])
        equal = 0;
    }
    if (equal == 1)
      return (char *)&s1[i];
  }
  return NULL;
}
#endif
