/*****************************************************************/
/*      gen_mes_all.c                                            */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  gen_sv_failure()                                             */
/*  gen_failure_report()                                         */
/*  gen_distance_report()                                        */
/*****************************************************************/

#include "server.h"
#include "shared_prot.h"

/*
 * gen_sv_failure()
 *
 * This function sends back to a client a failure notification,
 * that is a message containing one integer : SV_FAILURE
 *
 * The end_com() is done by the calling routine.
 */

#ifdef __STDC__
int gen_sv_failure(int sock)
#else
int gen_sv_failure(sock)
int sock;
#endif
{
  int message_type=SV_FAILURE;

  init_send(DATA_XDR,sock);
  write_int(&message_type);
  end_message();
  return 1;
}

/*
 * gen_failure_report()
 *
 * This function generates a failure report message and send it
 * to the local server. The message contains the 'date' of the
 * failure and the 'name' of the server that failed.
 * This mesage can be generated by a
 * forked process in charge of the workload-broadcast for instance
 *
 */

#ifdef __STDC__
int gen_failure_report(time_t date,unsigned int IPaddr)
#else
int gen_failure_report(date,IPaddr)
time_t date;
unsigned int IPaddr;
#endif
{
  int sock;                      /* socket descriptor   */
  int message_type;

  sock = connect_to_agent(my_host,my_address);
  if(sock == -3)
  {
    perror("gen_failure_report() socket()");
    return -1;
  }
  if (sock < 0)
  {
#ifdef DEBUG
    fprintf(stderr,"No netsolve agent running on %s\n",my_host);
#endif
    return -1;
  }

  message_type = FAILURE_REPORT;
  /* BEGINNING OF A SEND PHASE */
  /* Always XDR */
  init_send(DATA_XDR,sock);
  write_int(&message_type);
  write_long(&date);
  write_IPaddr(&IPaddr);
  /* END OF A SEND PHASE */
  end_message();
  end_com();
  return 1;
}



/*
 * gen_distance_report()
 *
 * This function generates a distance report message and send it
 * to the local server. The message contains the 'delay' of the
 * distance test and the 'name' of the server that failed.
 * This mesage is generated by 
 * forked process in charge of the distance requests
 */

#ifdef __STDC__
int gen_distance_report(unsigned int IPaddr,int delay1,int delay2)
#else
int gen_distance_report(IPaddr,delay1,delay2)
unsigned int IPaddr;
int delay1;
int delay2;
#endif
{
  int sock;                      /* socket descriptor   */
  int message_type;

  sock = connect_to_agent(my_host,my_address);
  if (sock == -2)
    return -1;
  if (sock == -1)
  {
#ifdef DEBUG
    fprintf(stderr," No netsolve agent on this machine\n");
#endif
    return -1;
  }

  message_type = DISTANCE_REPORT;
  /* BEGINNING OF A SEND PHASE */
  /* Always RAW */
  init_send(DATA_RAW,sock);
  write_int(&message_type);
  write_IPaddr(&IPaddr);
  write_int(&delay1);
  write_int(&delay2);
  /* END OF A SEND PHASE */
  end_message();
  end_com();
  return 1;
}
