/*****************************************************************/
/*      gen_mes_poll.c                                           */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  poll_gen_pb_list()                                           */
/*  poll_gen_sv_list()                                           */
/*  poll_gen_get_pb()                                            */
/*  poll_gen_get_sv()                                            */
/*  poll_gen_stat()                                              */
/*****************************************************************/

#include "server.h"
#include "shared_prot.h"

/*
 * poll_gen_pb_list()
 *
 * The server has been contacted by the interractive tool
 * It must send the list of its problems.
 *
 * The end_com() is done by the calling function : proc_rmt_mes()
 */

#ifdef __STDC__
int poll_gen_pb_list(int encoding,int detail)
#else
int poll_gen_pb_list(encoding,detail)
int encoding;
int detail;
#endif
{
  int number_of_pb;
  int j;
  problem_item *problem;
  format_spec *format_descriptor;
  constant *current_constant;
  formula *current_form;

  number_of_pb = 0;
  for (j=0;j<HASH_PB;j++)
  {
    problem = pb_hash_table[j];
    while(problem != 0)
    {
      number_of_pb++;
      problem=problem->next;
    }
  }
  init_send(encoding,client_desc);
  write_int(&number_of_pb);
  for (j=0;j<HASH_PB;j++)
  {
    int i;

    problem = pb_hash_table[j];
    while(problem != 0)
    {
      write_string((problem->problem).code);
      write_string((problem->problem).path);
      if (detail >= 2)
      {
        write_string((problem->problem).desc);
      }
      if (detail >= 1)
      {
        write_int(&((problem->problem).nb_obj_input));
        for (i=0;i<(problem->problem).nb_obj_input;i++)
        {
          write_int(&((problem->problem).input_obj[i].data_type));
          write_int(&((problem->problem).input_obj[i].object_type));
          if (detail >=3)
          {
            write_string(problem->problem.input_obj[i].object_desc); 
          }
        }
        write_int(&((problem->problem).nb_obj_output));
        for (i=0;i<(problem->problem).nb_obj_output;i++)
        {
          write_int(&((problem->problem).output_obj[i].data_type));
          write_int(&((problem->problem).output_obj[i].object_type));
          if (detail >= 3)
            write_string(problem->problem.output_obj[i].object_desc);
        }
        write_string((problem->problem).complexity);
      }
      if (detail >= 4)
      {
        /* Formats */
        write_int(&((problem->problem).nb_formats));
        for (i=0;i<(problem->problem).nb_formats;i++)
        {
          int j;
          write_int(&((problem->problem).formats[i].nb_args));
          for (j=0;j<((problem->problem).formats[i].nb_args);j++)
            write_string((problem->problem).formats[i].args_desc[j]);
          write_int(&((problem->problem).formats[i].nb_constants));
          current_constant = (problem->problem).formats[i].consts;
          for (j=0;j<((problem->problem).formats[i].nb_constants);j++)
          {
            write_string(current_constant->mnemonic);
            write_int(&(current_constant->value));
            current_constant = current_constant->next;
          }
          write_int(&((problem->problem).formats[i].nb_formulae));
          current_form = (problem->problem).formats[i].forms;
          for (j=0;j<((problem->problem).formats[i].nb_formulae);j++)
          {
            write_string(current_form->mnemonic);
            write_string(current_form->expression);
            current_form = current_form->next;
          }
        }
      }
    problem=problem->next;
    }
  }
  end_message();
  return 1;
}

/*
 * poll_gen_sv_list()
 *
 * The server has been contacted by the interractive tool
 * It must send the list of its servers.
 *
 * The end_com() is done by the calling function : proc_rmt_mes()
 */

#ifdef __STDC__
int poll_gen_sv_list(int encoding,int detail)
#else
int poll_gen_sv_list(encoding,detail)
int encoding;
int detail;
#endif
{
  server_item *server;
  int number_of_sv;
  int j;

  number_of_sv = 0;
  for (j=0;j<HASH_SV;j++)
  {
    server = sv_hash_table[j];
    while(server != 0)
    {
      number_of_sv++;
      server=server->next;
    }
  }
  for (j=0;j<HASH_SV;j++)
  {
    server = ag_hash_table[j];
    while(server != 0)
    {
      number_of_sv++;
      server=server->next;
    }
  }

  init_send(encoding,client_desc);
  write_int(&number_of_sv);
  for (j=0;j<HASH_SV;j++)
  {
    server = sv_hash_table[j];
    while(server != 0)
    {
      write_string((server->server).hostname);
      write_IPaddr(&((server->server).IPaddr));
      if (detail >= 1)
      {
        write_int(&((server->server).data_format));
        write_int(&((server->server).speed));
        write_int(&((server->server).host_status));
        write_int(&((server->server).server_status));
        write_int(&((server->server).number_proc));
        write_int(&((server->server).mpp));
        write_int(&((server->server).workload));
        write_int(&(server->number_problems));
      }
      server=server->next;
    }
  }
  for (j=0;j<HASH_SV;j++)
  {
    server = ag_hash_table[j];
    while(server != 0)
    {
      write_string((server->server).hostname);
      write_IPaddr(&((server->server).IPaddr));
      if (detail >= 1)
      {
        write_int(&((server->server).data_format));
        write_int(&((server->server).speed));
        write_int(&((server->server).host_status));
        write_int(&((server->server).server_status));
        write_int(&((server->server).number_proc));
        write_int(&((server->server).mpp));
        write_int(&((server->server).workload));
        write_int(&(server->number_problems));
      }
      server=server->next;
    }
  }

  end_message();
  return 1;
}

/*
 * poll_gen_get_pb()
 *
 * The server has been contacted by the interractive tool
 * It must send detailed information on a specific problem.
 *
 * The end_com() is done by the calling function : proc_rmt_mes()
 */
#ifdef __STDC__
int poll_gen_get_pb()
#else
int poll_gen_get_pb()
#endif
{
  int i;
  char code[CODE_LENGTH];
  problem_item *problem;
  sv_list *current;
  int status;

  read_string(code);
  end_message();

  problem = (problem_item *)find_problem(code);
  if (problem == 0)
  {
    init_send(DATA_XDR,client_desc); 
    status = -1;
    write_int(&status);
    end_message();
    return 1;
  }

  /* The problem has been recognized */
  init_send(DATA_XDR,client_desc);
  status = 1;
  write_int(&status);
  write_int(&((problem->problem).nb_obj_input));
  for (i=0;i<(problem->problem).nb_obj_input;i++)
  {
    write_int(&((problem->problem).input_obj[i].data_type));
    write_int(&((problem->problem).input_obj[i].object_type));
  }
  write_int(&((problem->problem).nb_obj_output));
  for (i=0;i<(problem->problem).nb_obj_output;i++)
  {
    write_int(&((problem->problem).output_obj[i].data_type));
    write_int(&((problem->problem).output_obj[i].object_type));
  }
  write_string((problem->problem).complexity);
  write_int(&(problem->problem.nb_formats));
  /* Format */
  for (i=0;i< problem->problem.nb_formats;i++)
  {
    int k;
    write_int(&(problem->problem.formats[i].nb_args));
    for(k=0;k< problem->problem.formats[i].nb_args;k++)
    {
      write_string(problem->problem.formats[i].args_desc[k]);
    }
  }
  write_int(&(problem->number_servers));

  current = (sv_list *)problem->first_server;
  for (i=0;i<(problem->number_servers);i++)
  {
    write_string(((current->server)->server).hostname);
    write_IPaddr(&(((current->server)->server).IPaddr));
    current = current->next;
  }
  end_message();
  return 1;
}


/*
 * poll_gen_get_sv()
 *
 * The server has been contacted by the interractive tool
 * It must send detailed information on a specific server.
 *
 * The end_com() is done by the calling function : proc_rmt_mes()
 */
#ifdef __STDC__
int poll_gen_get_sv()
#else
int poll_gen_get_sv()
#endif
{
  int i;
  unsigned int IPaddr;
  server_item *server;
  pb_list *current;
  int status;

  read_IPaddr(&IPaddr);
  end_message();

  server = (server_item *)find_server(IPaddr);
  if (server == 0)
  {
    init_send(DATA_XDR,client_desc); 
    status = -1;
    write_int(&status);
    end_message();
    return 1;
  }

  /* The problem has been recognized */
  init_send(DATA_XDR,client_desc);
  status = 1;
  write_int(&status);
  write_string((server->server).hostname);
  write_int(&((server->server).data_format));
  write_int(&((server->server).speed));
  write_int(&((server->server).host_status));
  write_int(&((server->server).server_status));
  write_int(&((server->server).number_proc));
  write_int(&((server->server).mpp));
  write_int(&((server->server).workload));
  write_int(&(server->number_problems));

  current = (pb_list *)server->first_pb;
  for (i=0;i<(server->number_problems);i++)
  {
    write_string(((current->problem)->problem).code);
    current = current->next;
  }
  end_message();
  return 1;
}

/*
 * poll_gen_stat()
 *
 * The server has been contacted by the interractive tool
 * It must send statistics about himself.
 *
 * The end_com() is done by the calling function : proc_rmt_mes()
 */
#ifdef __STDC__
int poll_gen_stat()
#else
int poll_gen_stat()
#endif
{
  char date[200];

  init_send(DATA_XDR,client_desc);
  write_int(&nb_pb_solved);
  write_int(&nb_lcl_request);
  strcpy(date,(char *)ctime(&up_date));
  write_string(date);
  end_message();
  return 1;
}
