/*****************************************************************/
/*      list_processing.c                                        */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*      Internal Level                                           */
/*      --------------                                           */
/*       int uppercase()                                         */
/*       int hash_sv()                                           */
/*       int hash_pb()                                           */
/*       server_item *new_server_item()                          */
/*       problem_item *new_problem_item()                        */
/*                                                               */
/*      External Level                                           */
/*      --------------                                           */
/*       int add_server_to_pb()                                  */
/*       int add_pb_to_server()                                  */
/*       problem_item *find_problem()                            */
/*       int *give_server_list()                                 */
/*       server_item *find_server()                              */
/*       problem_item *new_problem()                             */
/*       server_item *new_server()                               */
/*       int remove_server()                                     */
/*****************************************************************/

#include "server.h"
#include "shared_prot.h"

/******************/
/* Internal level */
/******************/

/*
 * uppercase() : change a string in uppercase
 */
#ifdef __STDC__
int uppercase(char *string)
#else
int uppercase(string)
char *string;
#endif
{
  int length,i;

  length = strlen(string);
  for (i=0;i<length;i++)
    if ((string[i] >= 97)&&(string[i]<=122))
      string[i] = string[i]-32;
}

/*
 * lowercase() : change a string in uppercase
 */
#ifdef __STDC__
int lowercase(char *string)
#else
int lowercase(string)
char *string;
#endif
{
  int length,i;

  length = strlen(string);
  for (i=0;i<length;i++)
    if ((string[i] >= 65)&&(string[i]<=90))
      string[i] = string[i]+32;
}


/*
 * hash function()
 *  returns a number between 0 and HASH_SV-1
 *  given a server IPaddr
 */

#ifdef __STDC__
int hash_sv(unsigned int IPaddr)
#else
int hash_sv(IPaddr)
unsigned int IPaddr;
#endif
{
  return((int)(IPaddr % HASH_SV));
}

/*
 * hash function()
 *  returns a number between 0 and HASH_PB-1
 *  given a problem code
 */

#ifdef __STDC__
int hash_pb(char *code)
#else
int hash_pb(code)
char* code;
#endif
{
  int result=0;
  int i;
  char *buf;
  
  buf = (char *)strdup(code);
  uppercase(buf);
  for(i=0;i<strlen(code);i++)
    result += (int)code[i];
  free(buf);
  return(result % HASH_PB);
}

/*
 * New server_item()
 *  Return a pointer to a new server item
 */

#ifdef __STDC__
server_item *new_server_item(unsigned int IPaddr)
#else
server_item *new_server_item(IPaddr)
unsigned int IPaddr;
#endif
{
  server_item *current;
  int hash_code;

  hash_code = hash_sv(IPaddr);
  if ((current = (server_item *)calloc(1,sizeof(server_item))) == 0)
  {
    perror("new_server_item() calloc()");
    return 0;
  }
  current->next = (server_item *)sv_hash_table[hash_code];
  if (current->next != 0)
    (current->next)->previous = current;
  current->number_problems = 0;
  current->number_failed = 0;
  current->first_stopped = 0;      /* RUNNING now we hope !! */
  current->last_pong = 0;
  current->time_ponged = 0;
  current->dirty = 0;
  time(&(current->creation));
  current->previous = 0;
  current->first_pb = 0;
  sv_hash_table[hash_code] = current;
  return (sv_hash_table[hash_code]);
}

/*
 * New agent_item()
 *  Return a pointer to a new server item
 */

#ifdef __STDC__
server_item *new_agent_item(unsigned int IPaddr)
#else
server_item *new_agent_item(IPaddr)
unsigned int IPaddr;
#endif
{
  server_item *current;
  int hash_code;

  hash_code = hash_sv(IPaddr);
  if ((current = (server_item *)calloc(1,sizeof(server_item))) == 0)
  {
    perror("new_server_item() calloc()");
    return 0;
  }
  current->next = (server_item *)ag_hash_table[hash_code];
  if (current->next != 0)
    (current->next)->previous = current;
  current->number_problems = 0;
  current->number_failed = 0;
  current->first_stopped = 0;
  current->last_pong = 0;
  current->time_ponged = 0;
  current->dirty = 0;
  time(&(current->creation));
  current->previous = 0;
  current->first_pb = 0;
  ag_hash_table[hash_code] = current;
  return (ag_hash_table[hash_code]);
}


/*
 * New Problem 
 *  Return a pointer to a new problem item
 */

#ifdef __STDC__
problem_item *new_problem_item(char *code)
#else
problem_item *new_problem_item(code)
char *code;
#endif
{
  problem_item *current;
  int hash_code;
  char *buf;

  buf = (char *)strdup(code);
  uppercase(buf);
  hash_code = hash_pb(buf);
  if ((current = (problem_item *)calloc(1,sizeof(problem_item))) == 0)
  {
    perror("new_problem_item() calloc()");
    free(buf);
    return 0;
  }
  current->next = pb_hash_table[hash_code];
  if (current->next != 0)
    (current->next)->previous = current;
  current->previous = 0;
  current->number_servers = 0;
  current->first_server =0;
  pb_hash_table[hash_code] = current;
  free(buf);
  return (pb_hash_table[hash_code]);
}

/*********************/
/* External level    */
/*********************/

/*
 * add_server_to_pb()
 *   add a server to a problem's server list
 */
 
#ifdef __STDC__
int add_server_to_pb(problem_item *problem,server_item *server)
#else
int add_server_to_pb(problem,server)
problem_item *problem;
server_item *server;
#endif
{
  sv_list *current;
 
  if (problem->number_servers == 0)
  {  /* Creation of the first server */
    if ((problem->first_server =(sv_list *)calloc(1,sizeof(sv_list))) == 0)
    {
      perror("add_server_to_pb() calloc()");
      return -1;
    }
    (problem->first_server)->server = server;
    (problem->first_server)->next = 0;
    (problem->first_server)->previous = 0;
    problem->number_servers++;
    return 1;
  }
  /* From here, there is already a server */
  if ((current = (sv_list *)calloc(1,sizeof(sv_list))) == 0)
  {
    perror("add_server_to_pb() calloc()");
    return -1;
  }
  current->server = server;
  current->next = problem->first_server;
  (current->next)->previous = current;
  current->previous = 0;
  problem->first_server = current;
  problem->number_servers++;
  return 1;
}

/*
 * add_pb_to_server()
 *   add a problem to server's problem list
 */

#ifdef __STDC__
int add_pb_to_server(problem_item *problem,server_item *server)
#else
int add_pb_to_server(problem,server)
problem_item *problem;
server_item *server;
#endif
{
  pb_list *current;

  if (server->number_problems == 0)
  { /* Creation of the first problem */
    if ((server->first_pb = 
           (pb_list *)calloc(1,sizeof(pb_list))) == 0)
    {
      perror("add_pb_to_server() calloc()");
      return -1;
    }
    (server->first_pb)->problem = problem;
    (server->first_pb)->next = 0;
    (server->first_pb)->previous = 0;
    server->number_problems++;
    return 1;
  }
  /* From here, there is already a problem */
  if ((current = (pb_list *)calloc(1,sizeof(pb_list))) == 0)
  {
    perror("add_pb_to_server() calloc()");
    return -1;
  }  
  current->next = server->first_pb;
  (current->next)->previous = current;
  current->previous = 0;
  current->problem = problem;
  server->first_pb = current;
  server->number_problems++;
  return 1;
}
    
/*
 * find_problem()
 *   finds a problem item given its code
 *   returns (problem_item *)0 if not found
 */

#ifdef __STDC__
problem_item *find_problem(char *code)
#else
problem_item *find_problem(code)
char *code;
#endif
{
  problem_item *current;
  char *buf1,*buf2;
 
  buf1 = (char *)strdup(code);
  uppercase(buf1);
  current = pb_hash_table[hash_pb(buf1)];
  while(current != 0)
  {
    buf2 = (char *)strdup((current->problem).code);
    uppercase(buf2);
    if (strcmp(buf1,buf2) == 0)
    {
      free(buf1);free(buf2);
      return current;
    }
    free(buf2);
    current=current->next;
  }
  free(buf1);
  return 0;
}
 
/*
 * give_server_list()
 *   Return a pointeur to a list of addresses, which are the
 *   servers_item locations
 *   (cryptic, but efficient.....sorry about that)
 */

#ifdef __STDC__
server_item **give_server_list(problem_item *problem,int *length)
#else
server_item **give_server_list(problem,length)
problem_item *problem;
int *length;
#endif
{
  server_item **list;
  sv_list *server;
  int i;

  *length = problem->number_servers;
  if (*length == 0)
    return 0;
  if ((list = (server_item **)calloc(*length,sizeof(server_item *))) == 0)
  {
    perror("give_server_list() calloc()");
    return 0;
  }
  server = (sv_list *)(problem->first_server);
  for (i=0;i<*length;i++)
  {
    list[i]=server->server;
    server = server->next;
  } 
  return list;
}

/*
 * Find Server()
 *   finds a server item given an IPaddr
 *   returns (sever_item *)0 if not found
 */

#ifdef __STDC__
server_item *find_server(unsigned int IPaddr)
#else
server_item *find_server(IPaddr)
unsigned int IPaddr;
#endif
{
  server_item *current;

  current = sv_hash_table[hash_sv(IPaddr)];
  while(current != 0)
  {
    if (IPaddr == (current->server).IPaddr)
      return current;
    current=current->next;
  }
  return 0;
}

/*
 * Find Agent()
 *   finds a agent item given an IPaddr
 *   returns (sever_item *)0 if not found
 */

#ifdef __STDC__
server_item *find_agent(unsigned int IPaddr)
#else
server_item *find_agent(IPaddr)
unsigned int IPaddr;
#endif
{
  server_item *current;

  current = ag_hash_table[hash_sv(IPaddr)];
  while(current != 0)
  {
    if (IPaddr == (current->server).IPaddr)
      return current;
    current=current->next;
  }
  return 0;
}


/*
 * new_problem()
 *  Insert a new Problem in the hash table given
 *  the problem descriptor. 
 */
#ifdef __STDC__
problem_item *new_problem(problem_desc descriptor)
#else
problem_item *new_problem(descriptor)
problem_desc descriptor;
#endif
{
  problem_item *problem;

  if ((problem = (problem_item *)new_problem_item(descriptor.code)) == 0)
  {
    perror("new_problem() new_problem_item()");
    return 0;
  }
  BCOPY((char *)&descriptor,(char *)(&(problem->problem)),sizeof(descriptor));
  return problem;
}

/*
 * new_server()
 *  Insert a new Server in the Server Hash Table
 *  given the server descriptor
 */
#ifdef __STDC__
server_item *new_server(server_desc descriptor)
#else
server_item *new_server(descriptor)
server_desc descriptor;
#endif
{
  server_item *server;

  if ((server = (server_item *)new_server_item(descriptor.IPaddr)) == 0)
  {
    perror("new_server() new_server_item()");
    return 0;
  }
  BCOPY((char *)(&descriptor),(char *)(&(server->server)),sizeof(server_desc));
  return server;
}

/*
 * new_agent()
 *  Insert a new Server in the Server Hash Table
 *  given the server descriptor
 */
#ifdef __STDC__
server_item *new_agent(server_desc descriptor)
#else
server_item *new_agent(descriptor)
server_desc descriptor;
#endif
{
  server_item *server;

  if ((server = (server_item *)new_agent_item(descriptor.IPaddr)) == 0)
  {
    perror("new_server() new_agent_item()");
    return 0;
  }
  BCOPY((char *)(&descriptor),(char *)(&(server->server)),sizeof(server_desc));
  return server;
}


/*
 * remove_server()
 *   Remove a server item and its problems from the hash_tables
 */
#ifdef __STDC__
int remove_server(server_item *server)
#else
int remove_server(server)
server_item *server;
#endif
{
  pb_list *current_pb=NULL;
  sv_list *current_sv=NULL;
  problem_item *problem=NULL;

/* First, we remove this server in all the lists */
  current_pb = server->first_pb;
  while(current_pb != 0)
  {
    pb_list *tmp;

  /* We cruise to the problem's server list */
    current_sv = current_pb->problem->first_server;
    while(current_sv != 0)
    {
      if (current_sv->server == server)
      { /* We want to remove current_sv */
        if (current_sv->previous == 0)
        {
          current_pb->problem->first_server = current_sv->next; 
          if (current_sv->next != 0)
            current_sv->next->previous = 0;
        }
        else
        {
          current_sv->previous->next = current_sv->next;
          if (current_sv->next != 0)
            current_sv->next->previous = current_sv->previous;   
        }
        if (current_sv != NULL)
          free(current_sv);
        ((current_pb->problem)->number_servers)--;
        if (current_pb->problem->number_servers == 0)
        { /* Remove the problem */
          char *buf;

          problem = current_pb->problem;
          buf = (char *)strdup(problem->problem.code);
          uppercase(buf);
          if (problem->previous == 0)
          {
            pb_hash_table[hash_pb(buf)] = problem->next;
            if (problem->next != 0)
            {
              problem->next->previous = 0;
            }
          }
          else
          {
            problem->previous->next = problem->next;
            if (problem->next != 0)
            {
              problem->next->previous = problem->previous;
            }
          }
          free_pb_desc(&(problem->problem));
/*
          free(problem);
*/
          free(buf);
        }
        break;
      }
      current_sv=current_sv->next;
    }
    tmp = current_pb;
    current_pb=current_pb->next;
    free(tmp);
  }

/* Then we remove the server itself */
  if (server->previous == 0)
  {
    sv_hash_table[hash_sv(server->server.IPaddr)] = server->next;
    if (server->next != 0)
      server->next->previous = 0;
  }
  else
  {
    server->previous->next = server->next;
    if (server->next != 0)
      server->next->previous = server->previous;
  }
  if (server != NULL)
    free(server);
  return 1;
}

/*
 * remove_agent()
 *   Remove a agent item from the hash_tables
 */
#ifdef __STDC__
int remove_agent(server_item *server)
#else
int remove_agent(server)
server_item *server;
#endif
{
  pb_list *current_pb=NULL;
  sv_list *current_sv=NULL;
  problem_item *problem=NULL;

/* First, we remove this server in all the lists */
  if (server->previous == 0)
  {
    ag_hash_table[hash_sv(server->server.IPaddr)] = server->next;
    if (server->next != 0)
      server->next->previous = 0;
  }
  else
  {
    server->previous->next = server->next;
    if (server->next != 0)
      server->next->previous = server->previous;
  }
  if (server != NULL)
    free(server);
  return 1;
}

#ifdef __STDC__
Print_cache(int detail)
#else
Print_cache(detail)
int detail;
#endif
{
  int i;
  server_item *server;
  problem_item *problem;
  pb_list *current;
  sv_list *cur;

  fprintf(stderr,"--------------------------------------------\n");
  for (i=0;i<HASH_PB;i++)
  {
    problem = pb_hash_table[i];
    while(problem!=0)
    {
      int j;

      fprintf(stderr,"HASH = %d\n",i);
      fprintf(stderr,"Problem : %s\n",problem->problem.code);
      fprintf(stderr,"\t input  %d %d compl : %s\n",
             problem->problem.nb_obj_input,
             problem->problem.nb_obj_output,
             problem->problem.complexity);
      problem=problem->next;
    }
  }
  fprintf(stderr,"--------------------------------------------\n");
}

