/********************************************************/
/* log_file.c                                           */
/* modified to communicate logfile statistics           */
/* July 1997 ...    Kim Buckner                         */
/* refer to file stat_comm.c for actual communication   */
/*------------------------------------------------------*/
/* new_agent_log()                                      */
/* new_server_log()                                     */
/* set_server_log()                                     */
/* log_truncate()                                       */
/* log_term()                                           */
/* new_log()                                            */
/* set_date()                                           */
/* add_request_to_log()                                 */
/* append_to_log()                                      */
/********************************************************/
#include "server.h" 
#include "shared_prot.h"
#include "bfunc.h"
/*
#include <errno.h>
#include <sys/time.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <sys/socket.h>
#include <stdio.h>
#include <sys/types.h>
*/

#define BYTE(x,y) (((unsigned char *)x)[(y)]&0xff)

/* Date */
char date[50];

/* File descriptor */
FILE *log_file;
char filename[256];
/*int linecount;*/

/* Maximum size of the log-file */
#define LOG_MAX 4096
/* Truncate minimum */
#define LOG_MIN 10

/* Default log name */
#define SV_NAME "/var/tmp/netsolve_server.log"
#define AG_NAME "/var/tmp/netsolve_agent.log"

/*
 *  new_log()
 *  this routine tries to open the file if it exists, otherwise
 *  it creates it. It then inserts a line about the current session 
 * 
 */

#ifdef __STDC__
int new_server_log() 
#else
int new_server_log() 
#endif
{ 
return new_log(SV_NAME); 
}

#ifdef __STDC__
int new_agent_log() 
#else
int new_agent_log() 
#endif
{ 
return new_log(AG_NAME); 
}

/* this sets the logfile name that is lost among the fork and exec calls */

#ifdef __STDC__
int set_server_log() 
#else
int set_server_log() 
#endif
{ 
return (int)strcpy(filename,SV_NAME); 
}

#ifdef __STDC__
int new_log(char *string) 
#else
int new_log(string) 
char *string;
#endif
{
  strcpy(filename,string);
  /*linecount = 0;*/

  log_file = fopen(filename,"a");
  if (log_file == NULL)
  { 
#ifdef DEBUG
    fprintf(stderr,"Cannot open log file '%s'\n",filename);
#endif
    return 1;
  }
 
  /* the following ensures that the correct data sizes are
   * available for reference
   */
  fprintf(log_file,"***** \n");
  set_date();
  fprintf(log_file,"* New Session : pid = %d, date = %s\n",getpid(),date);
  fprintf(log_file,"%u.%u.%u.%u ( %s )\n",BYTE(&my_address,0),
			 BYTE(&my_address,1),               
			 BYTE(&my_address,2),               
			 BYTE(&my_address,3),my_host);
  fprintf(log_file,"\nData sizes:\tTYPE\tCODE\tSIZE\n");
  fprintf(log_file,"\t\t INT\t %d\t %d\n",NETSOLVE_I,sizeof(int));
  fprintf(log_file,"\t\t FLOAT\t %d\t %d\n",NETSOLVE_S,sizeof(float));
  fprintf(log_file,"\t\t DOUBLE\t %d\t %d\n",NETSOLVE_D,sizeof(double));
  fprintf(log_file,"\t\t S_COMP\t %d\t %d\n",NETSOLVE_C,sizeof(scomplex));
  fprintf(log_file,"\t\t D_COMP\t %d\t %d\n",NETSOLVE_Z,sizeof(dcomplex));
  fprintf(log_file,"\t\t U_CHAR\t %d\t %d\n",NETSOLVE_CHAR,sizeof(unsigned char));
  fprintf(log_file,"\t\t BYTE\t %d\t %d\n",NETSOLVE_B,sizeof(char));
  fprintf(log_file,"***** \n");
  fclose(log_file);
  return 1;
}

/*
 * Sets the date to the present date
 */

#ifdef __STDC__
int set_date()
#else
int set_date()
#endif
{
  time_t now; 

  time(&now);
  strcpy(date,(char *)ctime(&now));
}

/*
 * add_request_to_log()
 *  Appends a Request Line to the log file
 */
#ifdef __STDC__
int add_request_to_log(int s)
#else
int add_request_to_log(s)
int s;
#endif
{
  struct sockaddr_in name;
  int namelen;
  unsigned int IPaddr=0;
  struct stat s_buf;
  time_t current;
  
  log_file = fopen(filename,"a");
  if (log_file == NULL)
  {
#ifdef DEBUG
    fprintf(stderr,"Cannot open log file '%s'\n",filename);
#endif
    return 1;
  }


  namelen = sizeof(name);
  if (getpeername(s,(struct sockaddr *)&name,&namelen) != 0)
  {
    fprintf(log_file,"Unknown peer : %s",date);
  }
  else
  {
    struct hostent *hp;
    char hostname[100];

    bcopy(&(name.sin_addr),&IPaddr,sizeof(unsigned int));
    hp = gethostbyaddr((char *)&IPaddr,sizeof(unsigned int),AF_INET);
    if (hp != NULL)
       strcpy(hostname,hp->h_name);
    else
       strcpy(hostname,"No name in DNS");
    time(&current);
    fprintf(log_file,"%u.%u.%u.%u ( %s ) at %s\n",BYTE(&IPaddr,0),
                           BYTE(&IPaddr,1),               
                           BYTE(&IPaddr,2),               
                           BYTE(&IPaddr,3),
                           hostname,
                           ctime(&current));
  }
  fclose(log_file);
  stat(filename,&s_buf);
  if (s_buf.st_size > LOG_MAX)
    log_truncate();

  return 1; 
}

/*
 * append_to_log()
 *  Just appends a line to the log file
 */

#ifdef __STDC__
int append_to_log(char *string)
#else
int append_to_log(string)
char *string;
#endif
{
  struct stat s_buf;

  log_file = fopen(filename,"a");
  if (log_file == NULL)
  {
#ifdef DEBUG
    fprintf(stderr,"Cannot open log file '%s'\n",filename);
#endif
    return 1;
  }

  fprintf(log_file,"%s\n",string);
  fclose(log_file);
  stat(filename,&s_buf);
  if (s_buf.st_size > LOG_MAX)
    log_truncate();
  return 1;
}

/*
 * log_truncate()
 *  the file has gotten too big
 */

#ifdef __STDC__
int log_truncate()
#else
int log_truncate()
#endif
{
  /* this will now trucate the file, rename it and fork off a new
   * process to comunicate the file.  I will need to do this quickly
   * because there are multiple processes writing to the file.  Would be
   * nice to use a lock but that is impractical.  Just have to deal 
   * with it.
  */
  int pid;
  int rtn;
  char command[256];
  char new_file[256];
  char *tmp_arch;
  char *tmp_path;
  char **args;
  time_t current;

  time(&current);
  sprintf(new_file,"/var/tmp/netsolve_old.log.%d",current%10000);
  sprintf(command,"head -13 %s > /tmp/netsolvehead.%d\n",filename,getpid());
  system(command);
  sprintf(command,"tail -%d %s > /tmp/netsolvetail.%d\n",LOG_MIN,filename,getpid());
  system(command);
  rtn=rename(filename,new_file);
  if(rtn <0)
    perror("log_truncate() file rename");
  sprintf(command,"cat /tmp/netsolvehead.%d /tmp/netsolvetail.%d > %s\n",
                  getpid(),getpid(),filename);
  system(command);
  sprintf(command,"/usr/bin/rm /tmp/netsolvehead.%d /tmp/netsolvetail.%d \n",
                   getpid(),getpid());
  system(command);
  if((pid=fork())==0) {
    if ((tmp_path=(char *)getenv("NETSOLVE_SERVER_ROOT")) == NULL) {
#ifdef VIEW
      perror("log_truncate() at getenv root");
#endif
      remove(new_file);
      exit(1);
      }
    if ((tmp_arch=(char *)getenv("NETSOLVE_ARCH")) == NULL) {
#ifdef VIEW
      perror("log_truncate() at getenv arch");
#endif
      remove(new_file);
      exit(1);
      }
    args=(char **)malloc(3*sizeof(char *));
    if(args==NULL) {
#ifdef VIEW
      perror("log_truncate()");
#endif
      remove(new_file);
      exit(1);
      }
    args[0]=(char *)strdup("stat_comm");
    args[1]=(char *)strdup(new_file);
    args[2]=NULL;
    sprintf(command,"%s/bin/%s/stat_comm",tmp_path,tmp_arch);
    execv(command,args);
#ifdef VIEW
    fprintf(stderr,"******!!!!%s\n",command);
    perror("log_truncate()");
#endif
    remove(new_file);
    exit(1);
    }
  /*linecount = LOG_MIN;*/
}

#ifdef __STDC__
int log_term()
#else
int log_term()
#endif
{
  /* this will now trucate the file, rename it and fork off a new
   * process to comunicate the file.  I will need to do this quickly
   * because there are multiple processes writing to the file.  Would be
   * nice to use a lock but that is impractical.  Just have to deal 
   * with it.
  */
  int pid;
  int rtn;
  FILE *output;
  char command[256];
  char new_file[256];
  char *tmp_arch;
  char *tmp_path;
  char **args;
  time_t current;

  if(filename==NULL) {
    return 1;
    }
  time(&current);
  sprintf(new_file,"/var/tmp/netsolve_old.log.%d",current%10000);
  
  rtn=rename(filename,new_file);
  if(rtn <0) {
#ifdef VIEW
    perror("log_term() file rename");
#endif
    return 1;
    }
  if((pid=fork())==0) {
    if((output=fopen(new_file,"a"))==NULL){
#ifdef VIEW
      perror("log_term() at fopen");
#endif
      remove(new_file);
      exit(1);
      }
    set_date();
    fprintf(output,"\n***Server Terminated at %s***\n",date);
    fclose(output);
    if ((tmp_path=(char *)getenv("NETSOLVE_SERVER_ROOT")) == NULL) {
#ifdef VIEW
      perror("log_term() at getenv root");
#endif
      remove(new_file);
      exit(1);
      }
    if ((tmp_arch=(char *)getenv("NETSOLVE_ARCH")) == NULL) {
#ifdef VIEW
      perror("log_term() at getenv arch");
#endif
      remove(new_file);
      exit(1);
      }
    args=(char **)malloc(3*sizeof(char *));
    if(args==NULL) {
#ifdef VIEW
      perror("log_term()");
#endif
      remove(new_file);
      exit(1);
      }
    args[0]=(char *)strdup("stat_comm");
    args[1]=(char *)strdup(new_file);
    args[2]=NULL;
    sprintf(command,"%s/bin/%s/stat_comm",tmp_path,tmp_arch);
    execv(command,args);
#ifdef VIEW
    fprintf(stderr,"******!!!!%s\n",command);
    perror("log_term()");
#endif
    remove(new_file);
    exit(1);
    }
}
