/*****************************************************************/
/*      pack_util.c                                              */
/*      Version 1.0                               Henri Casanova */
/*      Addition of _file function                Kim Buckner    */
/*      Modif. and Addition of _file functions    Henri Casanova */
/*---------------------------------------------------------------*/
/*  free_pb_desc()                                               */ 
/*  arch_to_format()                                             */ 
/*  is_number()                                                  */
/*  tread()                                                      */
/*  xdr_scomplex()                                               */
/*  xdr_dcomplex()                                               */
/*  init_send()                                                  */
/*  init_recv()                                                  */
/*  end_message()                                                */
/*  end_com()                                                    */
/*  read_string()                                                */
/*  read_int()                                                   */
/*  read_int_array()                                             */
/*  read_long()                                                  */
/*  read_single_array()                                          */
/*  read_double_array()                                          */
/*  read_u_char_array()                                          */
/*  read_scomplex_array()                                        */
/*  read_dcomplex_array()                                        */
/*  read_bytes()                                                 */
/*  write_string()                                               */
/*  write_int()                                                  */
/*  write_int_array()                                            */
/*  write_long()                                                 */
/*  write_single_array()                                         */
/*  write_double_array()                                         */
/*  write_scomplex_array()                                       */
/*  write_dcomplex_array()                                       */
/*  write_u_char_array()                                         */
/*  writebytes()                                                 */
/*  read_char_array_file()                                       */
/*  read_int_array_file()                                        */
/*  read_single_array_file()                                     */
/*  read_double_array_file()                                     */
/*  read_scomplex_array_file()                                   */
/*  read_dcomplex_array_file()                                   */
/*  write_char_array_file()                                      */
/*  write_int_array_file()                                       */
/*  write_single_array_file()                                    */
/*  write_double_array_file()                                    */
/*  write_scomplex_array_file()                                  */
/*  write_dcomplex_array_file()                                  */
/*****************************************************************/

#include "server.h"
#include "shared_prot.h"

/*
 * These are the global communication variables.
 * They are never used in the above level
 */
int current_mode;     /* DATA_RAW or DATA_XDR         */
int current_desc;     /* Descriptor of current socket */
XDR current_xdr;      /* Current XDR objet            */
FILE *current_stream; /* Current Stream               */

/* Buffer size for the file operations */
#define BUFFSIZE 8192
#define min(x,y) (x>y ? y : x)

/*
 * free_pb_desc()
 *   Release memory for a problem descriptor
 */
#ifdef __STDC__
int free_pb_desc(problem_desc *pb_desc)
#else
int free_pb_desc(pb_desc)
problem_desc *pb_desc;
#endif
{
  matlab_merge *tmp_matlab_merge;
  format_spec *tmp_format;
  int i,j;

  if (pb_desc == NULL)
    return;
  /* MATLAB_MERGE */
  tmp_matlab_merge = pb_desc->matlab;
  while(tmp_matlab_merge != NULL)
  {
    matlab_merge *tmp;
    tmp = tmp_matlab_merge->next;
    if (tmp_matlab_merge != NULL)
      free(tmp_matlab_merge);
    tmp_matlab_merge = tmp;
  }
  /* Complexity */
  if (pb_desc->complexity != NULL)
    free(pb_desc->complexity);
  /* Formats */
  for (j=0;j<pb_desc->nb_formats;j++)
  {
    format_spec *tmp;
    constant *tmp_const;
    formula *tmp_formula;

    tmp_format = &((pb_desc->formats)[j]);
    for (i=0;i<tmp_format->nb_args;i++)
        free((tmp_format->args_desc)[i]);
    free(tmp_format->args_desc); 
    tmp_const = tmp_format->consts;
    while(tmp_const != NULL)
    {
      constant *tmp;
      tmp = tmp_const->next;
      if (tmp_const->mnemonic != NULL)
        free(tmp_const->mnemonic);
      if (tmp_const != NULL)     
        free(tmp_const);
      tmp_const = tmp;
    }
    tmp_formula = tmp_format->forms;
    while(tmp_formula != NULL)
    {
      formula *tmp;
      tmp = tmp_formula->next;
      if (tmp_formula->mnemonic != NULL)
        free(tmp_formula->mnemonic);
      if (tmp_formula->expression != NULL) 
        free(tmp_formula->expression);
      if (tmp_formula != NULL)
        free(tmp_formula);
      tmp_formula = tmp;
    }
  } 
  if (pb_desc->formats != NULL)
    free(pb_desc->formats);

  /* Input */
  if (pb_desc->input_obj !=NULL)
    free(pb_desc->input_obj);

  /* Output */
  if (pb_desc->output_obj != NULL)
   free(pb_desc->output_obj);

  /* Pb_DESC */
  free(pb_desc);
}

/*
 * arch_to_format() : SUN4 ----> 11
 */
#ifdef __STDC__
int arch_to_format(char *string)
#else
int arch_to_format(string)
char *string;
#endif
{
  int i = 1;
  struct archtypes arch;

  arch=arches[i];
  while (strcmp(arch.archname,"") != 0)
  {
    if (strcmp(arch.archname,string) == 0)
      return (arch.archnum);
    i++;
    arch = arches[i];
  }
  return -1;
}


/*
 * is_number()
 *
 * given a string, return 0 if the string is an integer, -1 otherwise
 */
#ifdef __STDC__
int is_number(char *string)
#else
int is_number(string)
char *string;
#endif
{
  int i=0;

  while((string[i] != '\0')&&(string[i] != '\n'))
  {
    if ((string[i]<'0')||(string[i]>'9'))
      return -1;
    i++;
  }
  return 0;
}

/*
 * tread()
 * 
 * reads n bytes from the socket sock. blocks until the n bytes are
 * received.
 * returns the number of bytes read, or -1 if failure.
 */
#ifdef __STDC__
int tread(int d,char *b,int n)
#else
int tread(d, b, n)
int d;
char *b;
int n;
#endif
{
  int e;
  int x = n;

  while (x > 0) {
    if ((e = read(d, b, x)) == -1)
      return e;
    if (!e)
      return n - x;
    b += e;
    x -= e;
  }
  return n - x;
}

/*
 * init_send()
 *
 * initiates a sending phase of XDR or RAW protocol.
 * 
 * sends the protocol to the receving end to inform it
 * about the protocol. This integer is sent under the
 * network format .
 *
 * In the case of XDR, the XDR object is created and associated
 * with a stream. This stream is in fact a high level view 
 * of the current socket.
 * This trick allows the use of xdrstdio_create(), and thus freed us from
 * dealing with memory (is this cool or what ??)
 */
#ifdef __STDC__
int init_send(int mode,int descriptor)
#else
int init_send(mode,descriptor)
int mode;
int descriptor;
#endif
{
  u_short net_short;
 
  net_short = htons((u_short)mode);

  if (mode == DATA_RAW)
  {
    current_mode = DATA_RAW;
    current_desc = descriptor;
    current_stream = (FILE *)fdopen(dup(descriptor),"w");
    write(descriptor,&net_short,sizeof(u_short));
  }
  if (mode == DATA_XDR)
  {
    current_mode = DATA_XDR;
    current_desc = descriptor;
    current_stream = (FILE *)fdopen(dup(descriptor),"w");
    xdrstdio_create(&current_xdr,current_stream,XDR_FREE);
    xdrstdio_create(&current_xdr,current_stream,XDR_ENCODE);
    write(descriptor,&net_short,sizeof(u_short));
  }
  return 1;
}

/*
 * init_recv()
 *
 * initiates a receiving phase
 *
 * begins with reading the protocol from the current socket (see init_send())
 *
 * If XDR, the XDR object is created and 'streamed' to the current socket.
 *        (see init_send())
 */
#ifdef __STDC__
int init_recv(int descriptor)
#else
int init_recv(descriptor)
int descriptor;
#endif
{
  u_short net_short;
  int cc;

  cc = tread(descriptor,(char *)(&net_short),sizeof(u_short));
  if (cc != sizeof(u_short))
    return -1;
  current_mode = (int)ntohs(net_short);
  if (current_mode == DATA_RAW)
  {
    current_desc = descriptor;
    current_stream = (FILE *)fdopen(dup(descriptor),"r");
  }
  if (current_mode == DATA_XDR)
  {
    current_desc = descriptor;
    current_stream = (FILE *)fdopen(dup(descriptor),"r");
    xdrstdio_create(&current_xdr,current_stream,XDR_FREE);
    xdrstdio_create(&current_xdr,current_stream,XDR_DECODE);
  }
  return current_mode;
}

/*
 * end_message()
 * 
 * Terminates a message.
 *
 * In fact does nothing in the RAW protocol, and destroys
 * the XDR object while using the XDR protocol.
 */
#ifdef __STDC__
int end_message()
#else
int end_message()
#endif
{
  if (current_mode == DATA_XDR)
  {
      xdr_destroy(&current_xdr);
      fclose(current_stream);
  }
  if (current_mode == DATA_RAW)
  {
      fflush(current_stream);
      fclose(current_stream);
  }
  return 1;
}

/*
 * end_com()
 *
 * Terminate a communication phase
 *
 * Shutdowns the current socket and closes it.
 */
#ifdef __STDC__
int end_com()
#else
int end_com()
#endif
{
  if (current_mode == DATA_XDR)
  {
    shutdown(current_desc,2);
    close(current_desc);
  }
  if (current_mode == DATA_RAW)
  {
    shutdown(current_desc,2);
    close(current_desc);
  }
  return 1;
}


/*
 * xdr_scomplex()
 *
 * Wrapper to create a simple complex xdr encoding-decoding function
 */

#ifdef __STDC__
int xdr_scomplex(XDR *xdrptr,scomplex *ptr)
#else
int xdr_scomplex(xdrptr,ptr)
XDR *xdrptr;
scomplex *ptr;
#endif
{
  return (xdr_float(xdrptr,&ptr->r) &&
          xdr_float(xdrptr,&ptr->i));
}

/*
 * xdr_dcomplex()
 *
 * Wrapper to create a double complex xdr encoding-decoding function
 */

#ifdef __STDC__
int xdr_dcomplex(XDR *xdrptr,dcomplex *ptr)
#else
int xdr_dcomplex(xdrptr,ptr)
XDR *xdrptr;
dcomplex *ptr;
#endif
{
  return (xdr_double(xdrptr,&ptr->r) &&
          xdr_double(xdrptr,&ptr->i));
}


/*
 * read_string()
 * 
 * reads a string from the current socket
 */
#ifdef __STDC__
int read_string(char *data)
#else
int read_string(data)
char *data;
#endif
{
  char c = 'a';
  char *tmp = data;

  if (current_mode == DATA_RAW)
  {
    while (c != '\0')
    {
      tread(current_desc,&c,CHARL);
      *(data++) = c;
    }
    return (data-tmp);
  }
  if (current_mode == DATA_XDR)
  {
    if (xdr_string(&current_xdr,&data,MAX_STRING) == 1)
      return(1+strlen(data));
    else
      return -1;
  }
} 

/*
 * read_int()
 *
 * reads an integer from the current socket
 */
#ifdef __STDC__
int read_int(int *data)
#else
int read_int(data)
int *data;
#endif
{
  if (current_mode == DATA_RAW)
    return (tread(current_desc,(char*)data,INTL));

  if (current_mode == DATA_XDR)
  {
    if (xdr_int(&current_xdr,data) == 1)
    {
      return INTL;
    }
    else
      return -1;
  }
}


/*
 * read_byte_array()
 *
 * reads an array of byte from the current socket
 */
#ifdef __STDC__
int read_byte_array(void *data,int number)
#else
int read_byte_array(data,number)
void *data;
int number;
#endif
{
  int i;
  int tmp=1;

  if (current_mode == DATA_RAW)
    return (tread(current_desc,(char *)data,CHARL*number));

  if (current_mode == DATA_XDR)
  {
/*
    for (i=0;i<number;i++)
    {
      if (xdr_opaque(&current_xdr,&(((char *)data)[i]),1) != 1)
        tmp=-1;
    }
*/
    if (xdr_opaque(&current_xdr,(((char *)data)),number) != 1)
      tmp=-1;
    if (tmp == -1)
      return -1;
    else
      return (number*CHARL);
  }

}

/*
 * read_char_array()
 *
 * reads an array of char from the current socket
 */
#ifdef __STDC__
int read_char_array(char *data,int number)
#else
int read_char_array(data,number)
char *data;
int number;
#endif
{
  int i;
  int tmp=1;

  if (current_mode == DATA_RAW)
    return (tread(current_desc,data,CHARL*number));

  if (current_mode == DATA_XDR)
  {
    for (i=0;i<number;i++)
    {
      if (xdr_char(&current_xdr,&(data[i])) != 1)
        tmp=-1;
    }
    if (tmp == -1)
      return -1;
    else
      return (number*CHARL);
  }
}



/*
 * read_int_array()
 *
 * reads an array of ints from the current socket
 */
#ifdef __STDC__
int read_int_array(int *data,int number)
#else
int read_int_array(data,number)
int *data;
int number;
#endif
{
  int tmp = 0;
  int i;

  if (current_mode == DATA_RAW)
    return (tread(current_desc,(char*)data,INTL*number));

  if (current_mode == DATA_XDR)
  {
    for (i=0;i<number;i++)
    {
      if (xdr_int(&current_xdr,&(data[i])) != 1)
        tmp =-1;
    }
    if (tmp == -1)
      return -1;
    else
      return (number*INTL); 
  }
}



/*
 * read_IPaddr()
 *
 * Read an IPaddr from the current socket
 */

#ifdef __STDC__
int read_IPaddr(unsigned int *IPaddr)
#else
int read_IPaddr(IPaddr)
unsigned int *IPaddr;
#endif
{
  if (current_mode == DATA_RAW)
    return (tread(current_desc,(char*)IPaddr,sizeof(unsigned int)));
  if (current_mode == DATA_XDR)
  {
    unsigned char tmp;

#ifndef PLAT_T3E
    if (xdr_u_char(&current_xdr,&tmp) != 1)
      return -1;
    BCOPY(&tmp,(unsigned char *)IPaddr,1);
    if (xdr_u_char(&current_xdr,&tmp) != 1)
      return -1;
    BCOPY(&tmp,(unsigned char *)((unsigned char *)IPaddr+1),1);
    if (xdr_u_char(&current_xdr,&tmp) != 1)
      return -1;
    BCOPY(&tmp,(unsigned char *)((unsigned char *)IPaddr+2),1);
    if (xdr_u_char(&current_xdr,&tmp) != 1)
      return -1;
    BCOPY(&tmp,(unsigned char *)((unsigned char *)IPaddr+3),1);
#else
    if (xdr_u_char(&current_xdr,(char*)&tmp) != 1)
      return -1;
    BCOPY(&tmp,(unsigned char *)IPaddr,1);
    if (xdr_u_char(&current_xdr,(char*)&tmp) != 1)
      return -1;
    BCOPY(&tmp,(unsigned char *)((unsigned char *)IPaddr+1),1);
    if (xdr_u_char(&current_xdr,(char*)&tmp) != 1)
      return -1;
    BCOPY(&tmp,(unsigned char *)((unsigned char *)IPaddr+2),1);
    if (xdr_u_char(&current_xdr,(char*)&tmp) != 1)
      return -1;
    BCOPY(&tmp,(unsigned char *)((unsigned char *)IPaddr+3),1);
#endif


    return sizeof(unsigned int);
  }
}

/*
 * read_long()
 *
 * reads a long from the current socket
 */
#ifdef __STDC__
int read_long(long *data)
#else
int read_long(data)
long *data;
#endif
{
  if (current_mode == DATA_RAW)
    return(tread(current_desc,(char*)data,LONGL));

  if (current_mode == DATA_XDR)
  {
    if (xdr_long(&current_xdr,data) == 1)
      return INTL;
    else
      return -1;
  }
}


/*
 * read_single_array()
 *
 * reads an array of floats from the current socket
 */
#ifdef __STDC__
int read_single_array(float *data,int number)
#else
int read_single_array(data,number)
float *data;
int number;
#endif
{
  int i;
  int tmp=0;

  if (current_mode == DATA_RAW)
    return (tread(current_desc,(char*)data,FLOATL*number));

  if (current_mode == DATA_XDR)
  {
    for (i=0;i<number;i++)
    {
      if (xdr_float(&current_xdr,&(data[i])) != 1)
        tmp = -1;
    }
    if (tmp == -1)
      return -1;
    else
      return (number*FLOATL);
  }
}

/*
 * read_double_array()
 *
 * reads an array of doubles from the current socket
 */
#ifdef __STDC__
int read_double_array(double *data,int number)
#else
int read_double_array(data,number)
double *data;
int number;
#endif
{
  int tmp=0;
  int i;
  if (current_mode == DATA_RAW)
  {
    return (tread(current_desc,(char*)data,DOUBL*number));
  }
   
  if (current_mode == DATA_XDR)
  {
    for (i=0;i<number;i++)
    {
      if (xdr_double(&current_xdr,&(data[i])) != 1)
        tmp = -1;
    }
    if (tmp == -1)
      return -1;
    else
      return (number*DOUBL);
  }
}

/*
 * read_u_char_array()
 *
 * reads an array of unsigned char from the current socket
 */
#ifdef __STDC__
int read_u_char_array(unsigned char *data,int number)
#else
int read_u_char_array(data,number)
unsigned char *data;
int number;
#endif
{
  int i;
  int tmp = 1;

  if (number == 0)
    return 0;

  if (current_mode == DATA_RAW)
    return (tread(current_desc,(char*)data,UCHARL*number));
   
  if (current_mode == DATA_XDR)
  {
    for (i=0;i<number;i++)
    {
#ifdef PLAT_T3E
      if (xdr_u_char(&current_xdr,(char*)&(data[i])) != 1)
        tmp = -1;
#else
      if (xdr_u_char(&current_xdr,&(data[i])) != 1)
        tmp = -1;
#endif
    }
    if (tmp == -1)
      return -1;
    else
      return (number*CHARL);
  }
}

/*
 * read_scomplex_array()
 *
 * reads an array of single complexes from the curent socket
 */
#ifdef __STDC__
int read_scomplex_array(scomplex *data,int number)
#else
int read_scomplex_array(data,number)
scomplex *data;
int number;
#endif
{
  int i;
  int tmp = 1;

  if (current_mode == DATA_RAW)
    return(tread(current_desc,(char*)data,SCOMPL*number));

  if (current_mode == DATA_XDR)
  {
    for (i=0;i<number;i++)
    {
      if (xdr_scomplex(&current_xdr,&(data[i])) != 1)
        tmp = -1;
    }
    if (tmp == -1)
      return -1;
    else
      return (number*SCOMPL);
  }
}

/*
 * read_dcomplex_array()
 *
 * reads an array of double complexes from the curent socket
 */
#ifdef __STDC__
int read_dcomplex_array(dcomplex *data,int number)
#else
int read_dcomplex_array(data,number)
dcomplex *data;
int number;
#endif
{
  int i;
  int tmp = 1;

  if (current_mode == DATA_RAW)
  {
    return  tread(current_desc,(char*)data,DCOMPL*number);
  }

  if (current_mode == DATA_XDR)
  {
    for (i=0;i<number;i++)
    {
      if (xdr_dcomplex(&current_xdr,&(data[i])) != 1)
        tmp = -1;
    }
    if (tmp == -1)
      return -1;
    else
      return (number*DCOMPL);
  }
}

/*
 * write_string()
 *
 * writes a string to the current socket
 */
#ifdef _STDC__
int write_string(char *data)
#else
int write_string(data)
char *data;
#endif
{
  if (current_mode == DATA_RAW)
  {
    return (write(current_desc,data,1+strlen(data)));
  }

  if (current_mode == DATA_XDR)
  { 
    if (xdr_string(&current_xdr,&data,MAX_STRING) == 1)
      return (1+strlen(data));
    else
      return -1;
  }
}

/*
 * write_int()
 *
 * writes an integer to the current socket
 */

#ifdef __STDC__
int write_int(int *data)
#else
int write_int(data)
int *data;
#endif
{
  if (current_mode == DATA_RAW)
  {
    return (write(current_desc,data,INTL));
  }
   
  if (current_mode == DATA_XDR)
  {
    if (xdr_int(&current_xdr,data) == 1)
      return INTL;
    else
      return -1;
  }
}

/*
 * write_int_array()
 *
 * write an array of ints to the current socket
 */
#ifdef __STDC__
int write_int_array(int *data,int number)
#else
int write_int_array(data,number)
int *data;
int number;
#endif
{
  int i;
  int tmp = 1;
  if (current_mode == DATA_RAW)
    return(write(current_desc,data,number*INTL));

  if (current_mode == DATA_XDR)
  {
    for (i=0;i<number;i++)
    {
      if (xdr_int(&current_xdr,&(data[i])) != 1)
        tmp = -1;
    }
    if (tmp == -1)
      return -1;
    else
      return (number*INTL);
  }
}

/*
 * write_byte_array()
 *
 * write an array of bytes to the current socket
 */
#ifdef __STDC__
int write_byte_array(void *data,int number)
#else
int write_byte_array(data,number)
void *data;
int number;
#endif
{
  int i;
  int tmp = 1;
  if (current_mode == DATA_RAW)
    return(write(current_desc,(char *)data,number*CHARL));

  if (current_mode == DATA_XDR)
  {
/*
    for (i=0;i<number;i++)
    {
      if (xdr_opaque(&current_xdr,&(((char *)data)[i]),1) != 1)
        tmp = -1;
    }
*/
    if (xdr_opaque(&current_xdr,(((char *)data)),number) != 1)
      tmp = -1;
    if (tmp == -1)
      return -1;
    else
      return (number*CHARL);
  }

}


/*
 * write_char_array()
 *
 * write an array of chars to the current socket
 */
#ifdef __STDC__
int write_char_array(char *data,int number)
#else
int write_char_array(data,number)
char *data;
int number;
#endif
{
  int i;
  int tmp = 1;
  if (current_mode == DATA_RAW)
    return(write(current_desc,data,number*CHARL));

  if (current_mode == DATA_XDR)
  {
    for (i=0;i<number;i++)
    {
      if (xdr_char(&current_xdr,&(data[i])) != 1)
        tmp = -1;
    }
    if (tmp == -1)
      return -1;
    else
      return (number*CHARL);
  }
}


/*
 * write_IPaddr()
 *
 * Write an IPaddr to the current socket
 */

#ifdef __STDC__
int write_IPaddr(unsigned int *IPaddr)
#else
int write_IPaddr(IPaddr)
unsigned int *IPaddr;
#endif
{
  if (current_mode == DATA_RAW)
    return (write(current_desc,IPaddr,sizeof(unsigned int)));
  if (current_mode == DATA_XDR)
  {
    unsigned char tmp;

#ifdef PLAT_T3E
    BCOPY((unsigned char *)IPaddr,&tmp,1);
    if (xdr_u_char(&current_xdr,(char*)&tmp) != 1)
      return -1;
    BCOPY((unsigned char *)((unsigned char *)IPaddr+1),&tmp,1);
    if (xdr_u_char(&current_xdr,(char*)&tmp) != 1)
      return -1;
    BCOPY((unsigned char *)((unsigned char *)IPaddr+2),&tmp,1);
    if (xdr_u_char(&current_xdr,(char*)&tmp) != 1)
      return -1;
    BCOPY((unsigned char *)((unsigned char *)IPaddr+3),&tmp,1);
    if (xdr_u_char(&current_xdr,(char*)&tmp) != 1)
      return -1;
#else
    BCOPY((unsigned char *)IPaddr,&tmp,1);
    if (xdr_u_char(&current_xdr,&tmp) != 1)
      return -1;
    BCOPY((unsigned char *)((unsigned char *)IPaddr+1),&tmp,1);
    if (xdr_u_char(&current_xdr,&tmp) != 1)
      return -1;
    BCOPY((unsigned char *)((unsigned char *)IPaddr+2),&tmp,1);
    if (xdr_u_char(&current_xdr,&tmp) != 1)
      return -1;
    BCOPY((unsigned char *)((unsigned char *)IPaddr+3),&tmp,1);
    if (xdr_u_char(&current_xdr,&tmp) != 1)
      return -1;
#endif
    return sizeof(unsigned int);
  }
}

/*
 * write_long()
 *
 * write a long to the current socket
 */
#ifdef __STDC__
int write_long(long *data)
#else
int write_long(data)
long *data;
#endif
{
  if (current_mode == DATA_RAW)
    return (write(current_desc,data,LONGL));

  if (current_mode == DATA_XDR)
  {
    if (xdr_long(&current_xdr,data) == 1)
      return INTL;
    else
      return -1;
  }
}

/*
 * write_single_array()
 *
 * write an array of floats to the current socket
 */
#ifdef __STDC__
int write_single_array(float *data,int number)
#else
int write_single_array(data,number)
float *data;
int number;
#endif
{
  int i;
  int tmp = 1;
  if (current_mode == DATA_RAW)
    return(write(current_desc,data,number*FLOATL));
   
  if (current_mode == DATA_XDR)
  {
    for (i=0;i<number;i++)
    {
      if (xdr_float(&current_xdr,&(data[i])) != 1)
        tmp = -1;
    }
    if (tmp == -1)
      return -1;
    else
      return (number*FLOATL);
  }
}

/*
 * write_double_array()
 *
 * write an array of doubles to the current socket
 */
#ifdef __STDC__
int write_double_array(double *data,int number)
#else
int write_double_array(data,number)
double *data;
int number;
#endif
{
int tmp = 1;
int i;
  if (current_mode == DATA_RAW)
  {
    return (write(current_desc,data,number*DOUBL));
  }
   
  if (current_mode == DATA_XDR)
  {
    for (i=0;i<number;i++)
    {
      if (xdr_double(&current_xdr,&(data[i])) != 1)
        tmp = -1;
    }
    if (tmp == -1)
      return -1;
    else
      return (number*DOUBL);
  }
}

/*
 * write_u_char_array()
 *
 * write an array of unsigned char to the current socket
 */
#ifdef __STDC__
int write_u_char_array(unsigned char *data,int number)
#else
int write_u_char_array(data,number)
unsigned char *data;
int number;
#endif
{
  int i;
  int tmp = 1;

  if (number == 0)
    return -1;

  if (current_mode == DATA_RAW)
   return (write(current_desc,data,number*UCHARL));
   
  if (current_mode == DATA_XDR)
  {
    for (i=0;i<number;i++)
    {
#ifdef PLAT_T3E
      if (xdr_u_char(&current_xdr,(char *)&(data[i])) != 1)
        tmp = -1;
#else
      if (xdr_u_char(&current_xdr,&(data[i])) != 1)
        tmp = -1;
#endif
    }
    if (tmp == -1)
      return -1;
    else
      return (number*CHARL);
  }
}

/*
 * write_scomplex_array()
 *
 * writes an array of single complexes to the curent socket
 */
#ifdef __STDC__
int write_scomplex_array(scomplex *data,int number)
#else
int write_scomplex_array(data,number)
scomplex *data;
int number;
#endif
{
  int i;
  int tmp = 1;

  if (current_mode == DATA_RAW)
    return(write(current_desc,data,SCOMPL*number));

  if (current_mode == DATA_XDR)
  {
    for (i=0;i<number;i++)
    {
      if (xdr_scomplex(&current_xdr,&(data[i])) != 1)
        tmp = -1;
    }
    if (tmp == -1)
      return -1;
    else
      return (number*SCOMPL);
  }
}

/*
 * write_dcomplex_array()
 *
 * writes an array of double complexes to the curent socket
 */
#ifdef __STDC__
int write_dcomplex_array(dcomplex *data,int number)
#else
int write_dcomplex_array(data,number)
dcomplex *data;
int number;
#endif
{
  int i;
  int tmp = 1;
  if (current_mode == DATA_RAW)
  {
    return write(current_desc,data,DCOMPL*number);
  }

  if (current_mode == DATA_XDR)
  {
    for (i=0;i<number;i++)
    {
      if (xdr_dcomplex(&current_xdr,&(data[i])) != 1)
        tmp = -1;
    }
    if (tmp == -1)
      return -1;
    else
      return (number*DCOMPL);
  }
}

/*
 * reads an array of char from the current socket
 * to a file
 */
#ifdef __STDC__
int read_char_array_file(int fd,int size)
#else
int read_char_array_file(fd,size)
int fd;
int size;
#endif
{
  int tmp;
  int i;
  char buffer[BUFFSIZE];
  int count = size*CHARL;

  while(count != 0)
  {
    int subcount=min(BUFFSIZE,count);
    int subindex = subcount/CHARL;

    if (current_mode == DATA_RAW){
      tmp=tread(current_desc,buffer,subcount);
      if(tmp<0) {
        return -1;
        }
      tmp=write(fd,buffer,subcount);
      if (tmp<0)
        return -1;
      }
 
    if (current_mode == DATA_XDR)
    {
      for (i=0;i<subindex;i++)
      {
        if (xdr_char(&current_xdr,&(((char *)buffer)[i])) != 1)
          tmp =-1;
      }
      if (tmp == -1) {
        return -1;
        }
      tmp = write(fd,buffer,subcount);
      if (tmp<0)
        return -1;
    }
    count-=subcount;
  }
  return size*CHARL;
}

/*
 * reads an array of bytes from the current socket
 * to a file
 */
#ifdef __STDC__
int read_byte_array_file(int fd,int size)
#else
int read_byte_array_file(fd,size)
int fd;
int size;
#endif
{
  int tmp;
  int i;
  char buffer[BUFFSIZE];
  int count = size*CHARL;

  while(count != 0)
  {
    int subcount=min(BUFFSIZE,count);
    int subindex = subcount/CHARL;

    tmp=tread(current_desc,buffer,subcount);
    if(tmp<0) {
      return -1;
      }
    tmp=write(fd,buffer,subcount);
    if (tmp<0)
      return -1;
    count-=subcount;
  }
  return size*CHARL;
}


/*
 * reads an array of ints from the current socket
 * to a file
 */
#ifdef __STDC__
int read_int_array_file(int fd,int size)
#else
int read_int_array_file(fd,size)
int fd;
int size;
#endif
{
  int tmp;
  int i;
  char buffer[BUFFSIZE];
  int count = size*INTL;

  while(count != 0)
  {
    int subcount=min(BUFFSIZE,count);
    int subindex = subcount/INTL;

    if (current_mode == DATA_RAW){
      tmp=tread(current_desc,buffer,subcount);
      if(tmp<0) {
        return -1;
        }
      tmp=write(fd,buffer,subcount);
      if (tmp<0)
        return -1;
      }
  
    if (current_mode == DATA_XDR)
    {
      for (i=0;i<subindex;i++)
      {
        if (xdr_int(&current_xdr,&(((int *)buffer)[i])) != 1)
          tmp =-1;
      }
      if (tmp == -1) {
        return -1;
        }
      tmp = write(fd,buffer,subcount);
      if (tmp<0)
        return -1;
    }
    count-=subcount;
  }
  return size*INTL;
}

/*
 * reads an array of floats from the current socket
 * to a file
 */
#ifdef __STDC__
int read_single_array_file(int fd,int size)
#else
int read_single_array_file(fd,size)
int fd;
int size;
#endif
{
  int tmp;
  int i;
  char buffer[BUFFSIZE];
  int count = size*FLOATL;

  while(count != 0)
  {
    int subcount=min(BUFFSIZE,count);
    int subindex = subcount/FLOATL;

    if (current_mode == DATA_RAW){
      tmp=tread(current_desc,buffer,subcount);
      if(tmp<0) {
        return -1;
        }
      tmp=write(fd,buffer,subcount);
      if (tmp<0)
        return -1;
      }
  
  
    if (current_mode == DATA_XDR)
    {
      for (i=0;i<subindex;i++)
      {
        if (xdr_float(&current_xdr,&(((float *)buffer)[i])) != 1)
          tmp =-1;
      }
      if (tmp == -1) {
        return -1;
        }
      tmp = write(fd,buffer,subcount);
      if (tmp<0)
        return -1;
    }
    count -= subcount;
  }
  return size*FLOATL;
}
  
/*
 * reads an array of doubles from the current socket
 * to a file
 */
#ifdef __STDC__
int read_double_array_file(int fd,int size)
#else
int read_double_array_file(fd,size)
int fd;
int size;
#endif
{
  int tmp;
  int i;
  char buffer[BUFFSIZE];
  int count = size*DOUBL;

  while(count != 0)
  {
    int subcount=min(BUFFSIZE,count);
    int subindex = subcount/DOUBL;


    if (current_mode == DATA_RAW){
      tmp=tread(current_desc,buffer,subcount);
      if(tmp<0) {
        return -1;
        }
      tmp=write(fd,buffer,subcount);
      if (tmp<0)
        return -1;
      }
  
  
    if (current_mode == DATA_XDR)
    {
      for (i=0;i<subindex;i++)
      {
        if (xdr_double(&current_xdr,&(((double *)buffer)[i])) != 1)
          tmp =-1;
      }
      if (tmp == -1) {
        return -1;
        }
      tmp = write(fd,buffer,subcount);
      if (tmp<0)
        return -1;
    }
    count -= subcount;
  }
  return size*DOUBL;
}
/*
 * reads an array of single complex from the current socket
 * to a file
 */

#ifdef __STDC__
int read_scomplex_array_file(int fd,int size)
#else
int read_scomplex_array_file(fd,size)
int fd;
int size;
#endif
{
  int tmp;
  int i;
  scomplex *data;
  char buffer[BUFFSIZE];
  int count = size*SCOMPL;

  while(count != 0)
  {
    int subcount=min(BUFFSIZE,count);
    int subindex = subcount/SCOMPL;

    if (current_mode == DATA_RAW){
      tmp=tread(current_desc,buffer,subcount);
      if(tmp<0) {
        return -1;
        }
      tmp=write(fd,buffer,subcount);
      if (tmp<0)
        return -1;
      }
  
  
    if (current_mode == DATA_XDR)
    {
      for (i=0;i<subindex;i++)
      {
        if (xdr_scomplex(&current_xdr,&(((scomplex*)buffer)[i])) != 1)
          tmp =-1;
      }
      if (tmp == -1) {
        return -1;
        }
      tmp=write(fd,buffer,subcount);
      if (tmp<0)
        return -1;
    }
    count -= subcount;
  }
  return size*SCOMPL;
}
  
/*
 * reads an array of ints from the current socket
 * to a file
 */
#ifdef __STDC__
int read_dcomplex_array_file(int fd,int size)
#else
int read_dcomplex_array_file(fd,size)
int fd;
int size;
#endif
{
  int tmp;
  int i;
  char buffer[BUFFSIZE];
  int count = size*DCOMPL;

  while(count != 0)
  {
    int subcount=min(BUFFSIZE,count);
    int subindex = subcount/DCOMPL;

    if (current_mode == DATA_RAW){
      tmp=tread(current_desc,buffer,subcount);
      if(tmp<0) {
        return -1;
        }
      tmp=write(fd,buffer,subcount);
      if (tmp<0)
        return -1;
      }
  
  
    if (current_mode == DATA_XDR)
    {
      for (i=0;i<subindex;i++)
      {
        if (xdr_dcomplex(&current_xdr,&(((dcomplex *)buffer)[i])) != 1)
          tmp =-1;
      }
      if (tmp == -1) {
        return -1;
        }
      tmp =write(fd,buffer,subcount);
      if (tmp<0)
        return -1;
    }
    count -= subcount;
  }
  return size*DCOMPL;
}

#ifdef __STDC__
int write_char_array_file(int fd,int size)
#else
int write_char_array_file(fd,size)
int fd;
int size;
#endif
{
  int i;
  int tmp = 1;
  char buffer[BUFFSIZE];
  int count = size*CHARL;

  while(count != 0)
  {
    int subcount=min(BUFFSIZE,count);
    int subindex = subcount/CHARL;

    if (current_mode == DATA_RAW){
      tmp=read(fd,buffer,subcount);
      if(tmp<0) {
        return -1;
        }
      tmp=write(current_desc,buffer,subcount);
      if (tmp <0)
        return -1;
      }

    if (current_mode == DATA_XDR)
    {
      tmp = read(fd,buffer,subcount);
      if (tmp <0) {
        return -1;
        }
      for (i=0;i<subindex;i++)
      {
        if (xdr_char(&current_xdr,&(((char *)buffer)[i])) != 1)
          tmp =-1;
      }
      if (tmp == -1) {
        return -1;
        }
    }
    count-=subcount;
  }
  return size*CHARL;
}


#ifdef __STDC__
int write_byte_array_file(int fd,int size)
#else
int write_byte_array_file(fd,size)
int fd;
int size;
#endif
{
  int i;
  int tmp = 1;
  char buffer[BUFFSIZE];
  int count = size*CHARL;

  while(count != 0)
  {
    int subcount=min(BUFFSIZE,count);
    int subindex = subcount/CHARL;

    tmp=read(fd,buffer,subcount);
    if(tmp<0) {
      return -1;
      }
    tmp=write(current_desc,buffer,subcount);
    if (tmp <0)
      return -1;

    count-=subcount;
  }
  return size*CHARL;
}

#ifdef __STDC__
int write_int_array_file(int fd,int size)
#else
int write_int_array_file(fd,size)
int fd;
int size;
#endif
{
  int i;
  int tmp = 1;
  char buffer[BUFFSIZE];
  int count = size*INTL;

  while(count != 0)
  {
    int subcount=min(BUFFSIZE,count);
    int subindex = subcount/INTL;

    if (current_mode == DATA_RAW){
      tmp=read(fd,buffer,subcount);
      if(tmp<0) {
        return -1;
        }
      tmp=write(current_desc,buffer,subcount);
      if (tmp <0)
        return -1;
      }
 
    if (current_mode == DATA_XDR)
    {
      tmp = read(fd,buffer,subcount);
      if (tmp <0) {
        return -1;
        }
      for (i=0;i<subindex;i++)
      {
        if (xdr_int(&current_xdr,&(((int *)buffer)[i])) != 1)
          tmp =-1;
      }
      if (tmp == -1) {
        return -1;
        }
    }
    count-=subcount;
  }
  return size*INTL;
}
  
#ifdef __STDC__
int write_single_array_file(int fd,int size)
#else
int write_single_array_file(fd,size)
int fd;
int size;
#endif
{
  int i;
  int tmp = 1;
  char buffer[BUFFSIZE];
  int count = size*FLOATL;

  while(count != 0)
  {
    int subcount=min(BUFFSIZE,count);
    int subindex = subcount/FLOATL;

    if (current_mode == DATA_RAW){
      tmp=read(fd,buffer,subcount);
      if(tmp<0) {
        return -1;
        }
      tmp=write(current_desc,buffer,subcount);
      if (tmp <0)
        return -1;
      }

    if (current_mode == DATA_XDR)
    {
      tmp = read(fd,buffer,subcount);
      if (tmp <0) {
        return -1;
        }
      for (i=0;i<subindex;i++)
      {
        if (xdr_float(&current_xdr,&(((float *)buffer)[i])) != 1)
          tmp =-1;
      }
      if (tmp == -1) {
        return -1;
        }
    }
    count-=subcount;
  }
  return size*FLOATL;
}

#ifdef __STDC__
int write_double_array_file(int fd,int size)
#else
int write_double_array_file(fd,size)
int fd;
int size;
#endif
{
  int i;
  int tmp = 1;
  char buffer[BUFFSIZE];
  int count = size*DOUBL;

  while(count != 0)
  {
    int subcount=min(BUFFSIZE,count);
    int subindex = subcount/DOUBL;

    if (current_mode == DATA_RAW){
      tmp=read(fd,buffer,subcount);
      if(tmp<0) {
        return -1;
        }
      tmp=write(current_desc,buffer,subcount);
      if (tmp <0)
        return -1;
      }

    if (current_mode == DATA_XDR)
    {
      tmp = read(fd,buffer,subcount);
      if (tmp <0) {
        return -1;
        }
      for (i=0;i<subindex;i++)
      {
        if (xdr_double(&current_xdr,&(((double *)buffer)[i])) != 1)
          tmp =-1;
      }
      if (tmp == -1) {
        return -1;
        }
    }
    count-=subcount;
  }
  return size*DOUBL;
}

#ifdef __STDC__
int write_scomplex_array_file(int fd,int size)
#else
int write_scomplex_array_file(fd,size)
int fd;
int size;
#endif
{
  int i;
  int tmp = 1;
  char buffer[BUFFSIZE];
  int count = size*SCOMPL;

  while(count != 0)
  {
    int subcount=min(BUFFSIZE,count);
    int subindex = subcount/SCOMPL;

    if (current_mode == DATA_RAW){
      tmp=read(fd,buffer,subcount);
      if(tmp<0) {
        return -1;
        }
      tmp=write(current_desc,buffer,subcount);
      if (tmp <0)
        return -1;
      }

    if (current_mode == DATA_XDR)
    {
      tmp = read(fd,buffer,subcount);
      if (tmp <0) {
        return -1;
        }
      for (i=0;i<subindex;i++)
      {
        if (xdr_scomplex(&current_xdr,&(((scomplex *)buffer)[i])) != 1)
          tmp =-1;
      }
      if (tmp == -1) {
        return -1;
        }
    }
    count-=subcount;
  }
  return size*SCOMPL;
}

#ifdef __STDC__
int write_dcomplex_array_file(int fd,int size)
#else
int write_dcomplex_array_file(fd,size)
int fd;
int size;
#endif
{
  int i;
  int tmp = 1;
  char buffer[BUFFSIZE];
  int count = size*DCOMPL;

  while(count != 0)
  {
    int subcount=min(BUFFSIZE,count);
    int subindex = subcount/DCOMPL;

    if (current_mode == DATA_RAW){
      tmp=read(fd,buffer,subcount);
      if(tmp<0) {
        return -1;
        }
      tmp=write(current_desc,buffer,subcount);
      if (tmp <0)
        return -1;
      }

    if (current_mode == DATA_XDR)
    {
      tmp = read(fd,buffer,subcount);
      if (tmp <0) {
        return -1;
        }
      for (i=0;i<subindex;i++)
      {
        if (xdr_dcomplex(&current_xdr,&(((dcomplex *)buffer)[i])) != 1)
          tmp =-1;
      }
      if (tmp == -1) {
        return -1;
        }
    }
    count-=subcount;
  }
  return size*DCOMPL;
}
