/*****************************************************************/
/*      problem_util.c                                           */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  write_problem()                                              */
/*  read_problem()                                               */
/*  problem_cmp()                                                */
/*  problem_cmp_format()                                         */  
/*  arg_list_cmp()                                               */  
/*  power()                                                      */  
/*  read_line()                                                  */
/*****************************************************************/

#include "server.h"
#include "shared_prot.h"
#include "security.h"

/*
 * write_problem()
 * 
 *   write a problem on the current socket
 */ 
#ifdef __STDC__
int write_problem(problem_desc problem)
#else
int write_problem(problem)
problem_desc problem;
#endif
{
  int i;
  constant *current_constant;
  formula *current_form;
  matlab_merge *current_matlab;

  write_string(problem.code);
  write_string(problem.path);
  write_string(problem.desc);
  write_int(&(problem.nb_obj_input));
  for (i=0;i<problem.nb_obj_input;i++)
  {
    write_int(&(problem.input_obj[i].data_type));
    write_int(&(problem.input_obj[i].object_type));
    write_string(problem.input_obj[i].object_desc);
  }
  write_int(&(problem.nb_obj_output));
  for (i=0;i<problem.nb_obj_output;i++)
  {
    write_int(&(problem.output_obj[i].data_type));
    write_int(&(problem.output_obj[i].object_type));
    write_string(problem.output_obj[i].object_desc);
  }
  /* Matlab stuff */
  write_int(&(problem.nb_matlab));
  current_matlab = problem.matlab;
  for (i=0;i<problem.nb_matlab;i++)
  {
    write_int(&(current_matlab->output1));
    write_int(&(current_matlab->output2));
  }
  /* Complexity */
  write_string(problem.complexity);
  /* Formats */
  write_int(&(problem.nb_formats));
  for (i=0;i<problem.nb_formats;i++)
  {
    int j;
    write_int(&(problem.formats[i].nb_args));
    for (j=0;j<(problem.formats[i].nb_args);j++)
      write_string(problem.formats[i].args_desc[j]);
    write_int(&(problem.formats[i].nb_constants));
    current_constant = problem.formats[i].consts;
    for (j=0;j<(problem.formats[i].nb_constants);j++)
    {
      write_string(current_constant->mnemonic);
      write_int(&(current_constant->value));
      current_constant = current_constant->next; 
    }
    write_int(&(problem.formats[i].nb_formulae));
    current_form = problem.formats[i].forms;
    for (j=0;j<(problem.formats[i].nb_formulae);j++)
    {
      write_string(current_form->mnemonic);
      write_string(current_form->expression);
      current_form = current_form->next;
    }

  }

  return 1;
}

/*
 * read_problem()
 *
 * read a problem from the current socket
 */
#ifdef __STDC__
int read_problem(problem_desc *problem)
#else
int read_problem(problem)
problem_desc *problem;
#endif
{
  int i;
  char buf[15];
  constant *current_constant;
  formula *current_form;
  matlab_merge *current_matlab;

  read_string(problem->code);
  read_string(problem->path);
  read_string(problem->desc);
  read_int(&(problem->nb_obj_input));
  problem->input_obj = (object *)calloc(problem->nb_obj_input,sizeof(object));
  for (i=0;i<problem->nb_obj_input;i++)
  {
    read_int(&(problem->input_obj[i].data_type));
    read_int(&(problem->input_obj[i].object_type));
    read_string(problem->input_obj[i].object_desc);
  }
  read_int(&(problem->nb_obj_output));
  problem->output_obj = (object *)calloc(problem->nb_obj_output,sizeof(object));
  for (i=0;i<problem->nb_obj_output;i++)
  {
    read_int(&(problem->output_obj[i].data_type));
    read_int(&(problem->output_obj[i].object_type));
    read_string(problem->output_obj[i].object_desc);
  }
  /* Matlab Stuff */
  read_int(&(problem->nb_matlab));
  problem->matlab = 0;
  for (i=0;i<problem->nb_matlab;i++)
  {
    current_matlab = (matlab_merge *)calloc(1,sizeof(matlab_merge));
    read_int(&(current_matlab->output1));
    read_int(&(current_matlab->output2));
    current_matlab->next = problem->matlab;
    problem->matlab = current_matlab;
  }
  /* Complexity */
  read_string(buf);
  problem->complexity = (char *)strdup(buf);
 /* Formats */
  read_int(&(problem->nb_formats));
  problem->formats = (format_spec *)calloc(problem->nb_formats,sizeof(format_spec));
  for (i=0;i<problem->nb_formats;i++)
  {
    char buf[256];
    int j;
    read_int(&(problem->formats[i].nb_args));
    problem->formats[i].args_desc = 
         (char **)calloc(problem->formats[i].nb_args,sizeof(char *));
    for(j=0;j<(problem->formats[i].nb_args);j++)
    {
      read_string(buf);
      problem->formats[i].args_desc[j] = (char *)strdup(buf);
    }
    read_int(&(problem->formats[i].nb_constants));
    problem->formats[i].consts = 0;
    for (j=0;j<(problem->formats[i].nb_constants);j++)
    {
      char buf[256];

      read_string(buf);
      current_constant = (constant *)calloc(1,sizeof(constant));
      current_constant->mnemonic = (char *)strdup(buf);
      read_int(&(current_constant->value));
      current_constant->next = problem->formats[i].consts;
      problem->formats[i].consts = current_constant;
    }
    read_int(&(problem->formats[i].nb_formulae));
    problem->formats[i].forms = 0;
    for (j=0;j<(problem->formats[i].nb_formulae);j++)
    {
      char buf1[256];
      char buf2[256];

      read_string(buf1);
      current_form = (formula *)calloc(1,sizeof(formula));
      current_form->mnemonic = (char *)strdup(buf1);
      read_string(buf2);
      current_form->expression = (char *)strdup(buf2);
      current_form->next = problem->formats[i].forms;
      problem->formats[i].forms = current_form;
    }
  }    
 
  return 1;
}

/*
 * problem_cmp()
 *
 *   compares two pb_desc 
 */
#ifdef __STDC__
int problem_cmp(problem_desc desc1,problem_desc desc2)
#else
int problem_cmp(desc1,desc2)
problem_desc desc1,desc2;
#endif
{
  int i;

  if (desc1.nb_obj_input != desc2.nb_obj_input)
    return -1;
  if (desc1.nb_obj_output != desc2.nb_obj_output)
    return -2;
  for (i=0;i<desc1.nb_obj_input;i++)
  {
    if(desc1.input_obj[i].data_type != desc2.input_obj[i].data_type )
      return -3;
    if(desc1.input_obj[i].object_type != desc2.input_obj[i].object_type )
      return -3;
  }
  for (i=0;i<desc1.nb_obj_output;i++)
  {
    if (desc1.output_obj[i].data_type != desc2.output_obj[i].data_type )
      return -4;
    if(desc1.output_obj[i].object_type != desc2.output_obj[i].object_type )
      return -4;
  }
  if ((desc1.complexity != NULL)&&(desc2.complexity != NULL))
  {
    if (strcmp(desc1.complexity,desc2.complexity) != 0)
      return -5;
  }
  else
    return -5;
  return 1;
}

/*
 * problem_cmp_format()
 */
#ifdef __STDC__
int problem_cmp_format(problem_desc desc1,problem_desc desc2)
#else
int problem_cmp_format(desc1,desc2)
problem_desc desc1,desc2;
#endif
{
  int i;

  if (desc1.nb_formats != desc2.nb_formats)
    return -1;
  for (i=0;i<desc1.nb_formats;i++)
  {
    int j;

    if (desc1.formats[i].nb_args != desc2.formats[i].nb_args)
      return -1;
    for(j=0;j<desc1.formats[i].nb_args;j++)
    {
      if (arg_list_cmp(desc1.formats[i].args_desc[j],
                       desc2.formats[i].args_desc[j]) < 0)
        return -1;
    }
  }
  return 1;
}

/*
 * arg_list_cmp()
 */
#ifdef __STDC__
int arg_list_cmp(char *s1,char *s2)
#else
int arg_list_cmp(s1,s2)
char *s1,*s2;
#endif
{
  char *buf;
  char *token;

  buf = (char *)strdup(s1);

  token = (char *)strtok(buf," \t,");
  while(token != NULL)
  {
    if (strstr(s2,token) == NULL)
      return -1;
    token = (char *)strtok(NULL," \t,");
  }
  free(buf);
  return 1;
}

/*
 * power()
 *
 * compute the power of a to b
 */
#ifdef __STDC__
int power(int a,int b)
#else
int power(a,b)
int a;
int b;
#endif
{
  int i;
  int r=1;

  for (i=0;i<b;i++)
    r = r * a;
  return r;
}

/*
 * read_line()
 *  Reads a line from a file
 */

#ifdef __STDC__
char *read_line(FILE *infile,char *line,int *linecount)
#else
char *read_line(infile,line,linecount)
FILE *infile;
char *line;
int *linecount;
#endif
{
  char *tmp = NULL;

  do
  {
    tmp = fgets(line,MAX_LINE_LENGTH,infile);
    if (tmp == NULL)
      break;
    (*linecount)++;
  } while((tmp[0] == '\n')||(tmp[0] == '#'));
  return tmp;
}

