/*****************************************************************/
/*	socket_util.c                                            */
/*	Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/* int establish_socket()                                        */
/* int connect_to_agent()                                        */
/* int connect_to_server()                                       */
/* int connect_to_socket()                                       */
/*****************************************************************/

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include "server.h"
#include "shared_prot.h"
#include "bfunc.h"

#define HOSTNAME_LENGTH 256

/*
 *  Creation of an INET STREAM socket given :
 *     - the desired port number
 */

#ifdef __STDC__
int establish_socket(int *port)
#else
int establish_socket(port)
int *port;
#endif
{
  int desc;                       
  struct sockaddr_in addr;        
  int addr_length;               
  char hostname[HOSTNAME_LENGTH];
  struct hostent *hp;
  int option_value = 1;    

  if((desc =socket(AF_INET,SOCK_STREAM,0)) == -1)
  {
#ifdef VIEW
    perror("create_socket_INET() socket()");
#endif
    return -1;
  }
  if (setsockopt(desc,SOL_SOCKET,SO_REUSEADDR,(char *)(&option_value),sizeof(int)) == -1)
  {
#ifdef VIEW
    perror("create_socket_INET() setsockopt()");
#endif
    return -1;
  }
  BZERO((char *)&addr,sizeof(addr));
  addr.sin_port = htons(*port);
  if (gethostname(hostname,HOSTNAME_LENGTH) == -1)
  {
#ifdef VIEW
    perror("create_socket_INET() gethostname()");
#endif
    close(desc);
    return -1;
  }
  hp = gethostbyname(hostname);
  if (hp == NULL)
  {
#ifdef VIEW
    perror("create_socket_INET() gethostbyname()");
#endif
    close(desc);
    return -1;
  }
  BCOPY(hp->h_addr,&addr.sin_addr,hp->h_length);
  addr.sin_family = AF_INET;
#ifdef DEBUG
  fprintf(stderr,"Binding on port %d\n",*port);
#endif
  if(bind(desc,(struct sockaddr *)&addr,sizeof(addr)))
  {
#ifdef VIEW
    perror("create_socket_INET() bind()");
#endif
    return -1;
  }
  addr_length = sizeof(addr);
  if ((getsockname(desc,(struct sockaddr *)&addr,&addr_length)) == -1)
  {
#ifdef VIEW
    perror("create_socket_INET() getsockname()");
#endif
    close(desc);
    return -1;
  }
  *port = ntohs(addr.sin_port);
  return(desc);
}

/*
 * connect_to_agent()
 *
 *   returns -3 if could not create socket
 *   returns -2 if host could not be located
 *   returns -1 if connection could not be established
 */
#ifdef __STDC__
int connect_to_agent(char *hostname,int IPaddr)
#else
int connect_to_agent(hostname,IPaddr)
char *hostname;
unsigned int IPaddr;
#endif
{
  return connect_to_socket(hostname,IPaddr,AG_PORT);
}


/*
 * connect_to_server()
 *
 *   returns -3 if could not create socket
 *   returns -2 if host could not be located
 *   returns -1 if connection could not be established
 */
#ifdef __STDC__
int connect_to_server(char *hostname,int IPaddr)
#else
int connect_to_server(hostname,IPaddr)
char *hostname;
unsigned int IPaddr;
#endif
{
  return connect_to_socket(hostname,IPaddr,SV_PORT);
}


/*
 * connect_to_socket()
 *
 *   returns -3 if could not create socket
 *   returns -2 if host could not be located
 *   returns -1 if connection could not be established
 */

#ifdef __STDC__
int connect_to_socket(char *hostname,unsigned int IPaddr,int port)
#else
int connect_to_socket(hostname,IPaddr,port)
char *hostname;
unsigned int IPaddr;
int port;
#endif
{
  int sock;
  struct sockaddr_in inet_addr;
  struct hostent *hp;

  if ((sock = socket(AF_INET,SOCK_STREAM,0)) == -1)
  {
    perror("connect_to_socket() socket()");
    return -3;
  }
  BZERO(&inet_addr,sizeof(inet_addr));

  /* Try to do a gethostbyname() */
  hp = gethostbyaddr((char *)&IPaddr,sizeof(unsigned int),AF_INET); 
  if (hp == NULL)
  {
    /* Try to do a gethostbyaddr() */
    hp = gethostbyname(hostname);
    if (hp == NULL)
    {
      close(sock);
      return -2;
    }
  }
  BCOPY(hp->h_addr,&inet_addr.sin_addr,hp->h_length);
  inet_addr.sin_family = AF_INET;
  inet_addr.sin_port = htons(port);
  
#ifdef DEBUG
  fprintf(stderr,"Connecting to %s ...\n",hostname);
#endif
  if (connect(sock,(struct sockaddr *)&inet_addr,sizeof(inet_addr)) == -1)
  {
    close(sock);
    return -1;
  }
  else
    return sock;
}
