/*****************************************************************/
/*      server_util.c                                            */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  set_signals()                                                */
/*  term()                                                       */
/*  clean_up()                                                   */
/*  evilsig()     from PVM                                       */
/*  catch()       from PVM                                       */
/*  biteme()      from PVM                                       */
/*  reap()        from PVM                                       */
/*  my_speed()                                                   */
/*  data_format()                                                */
/*  bailout()                                                    */
/*  distance_request()                                           */
/*  is_there_comp_pid()                                          */
/*  new_comp_pid()                                               */
/*  remove_comp_pid()                                            */
/*****************************************************************/

#include "server.h"
#include "shared_prot.h"

/*
 * set_signals()
 */
#ifdef __STDC__
void set_signals()
#else
void set_signals()
#endif
{
  void reap();
  void catch();
  void evilsig();

#ifndef PLAT_I860 /* this signal interferes with getcube() on I860 */
#ifdef  SYSVSIGNAL
  (void)signal(SIGCLD, reap);
#else
  (void)signal(SIGCHLD, reap);
#endif
#endif /*PLAT_I860*/
  if (signal(SIGINT, SIG_IGN) != SIG_IGN)
    (void)signal(SIGINT, catch);
  if (signal(SIGTERM, SIG_IGN) != SIG_IGN)
    (void)signal(SIGTERM, catch);

  (void)signal(SIGHUP, SIG_IGN);
  (void)signal(SIGPIPE, SIG_IGN);

  (void)signal(SIGILL, evilsig);
  (void)signal(SIGFPE, evilsig);
#ifdef  SIGBUS
  (void)signal(SIGBUS, evilsig);
#endif
  (void)signal(SIGSEGV, evilsig);
#ifdef  SIGSYS
  (void)signal(SIGSYS, evilsig);
#endif

#ifdef SIGDANGER
  (void)signal(SIGDANGER,biteme);
#endif

}

/*
 * term()
 */
#ifdef __STDC__
void term()
#else
void term()
#endif
{
  shutdown(rmt_sock,2);
  close(rmt_sock);
  (void)kill(work_pid,SIGTERM);
  /* Remove the /tmp/NetSolve file */
  if (my_self->number_problems != 0)
  {
    char command_line[64];
    FILE *stream;

    sprintf(command_line,"rm -rf /tmp/Netsolve");
    system(command_line);
  }
  /* Clean up the data structure */
  clean_up(); 
  fprintf(stderr,"\n-----------------\nNetsolve Abort !!\n-----------------\n");
  exit(0);
}

/*
 * clean_up()
 */
#ifdef __STDC__
int clean_up()
#else
int clean_up()
#endif
{
  int i;
  server_item *server;

  for (i=0;i<HASH_SV;i++)
  {
    server = sv_hash_table[i];
    while(server != NULL)
    {
      server_item *tmp;
      tmp = server;
      server = server->next;
      remove_server(tmp);
    } 
  }
  for (i=0;i<HASH_SV;i++)
  {
    server = ag_hash_table[i];
    while(server != NULL)
    {
      server_item *tmp;
      tmp = server;
      server = server->next;
      remove_server(tmp);
    }
  }

  return 0;
}

/*
 * evilsig() : From PVM  Bob Manchek
 */
#ifdef __STDC__
void evilsig(int sig)
#else
void evilsig(sig)
int sig;
#endif
{
  (void)signal(SIGILL, SIG_DFL);
  (void)signal(SIGFPE, SIG_DFL);
#ifdef  SIGBUS
  (void)signal(SIGBUS, SIG_DFL);
#endif
  (void)signal(SIGSEGV, SIG_DFL);
#ifdef  SIGSYS
  (void)signal(SIGSYS, SIG_DFL);
#endif
  (void)signal(SIGINT, SIG_DFL);
  (void)signal(SIGTERM, SIG_DFL);
  (void)signal(SIGHUP, SIG_DFL);
  (void)signal(SIGPIPE, SIG_DFL);
#ifdef  SYSVSIGNAL
  (void)signal(SIGCLD, SIG_DFL);
#else
  (void)signal(SIGCHLD, SIG_DFL);
#endif
#ifdef VIEW
  fprintf(stderr,"evilsig() caugth signal %d\n",sig);
#endif
  term();
}

/*
 * catch() : From PVM  Bob Manchek
 */
#ifdef __STDC__
void catch(int sig)
#else
void catch(sig)
int sig;
#endif
{
  (void)signal(SIGINT, SIG_DFL);
  (void)signal(SIGTERM, SIG_DFL);
#ifdef VIEW
  fprintf(stderr,"catch() caught signal %d\n",sig);
#endif
  term();
}

/*
 * biteme() : Form PVM  Bob Manchek
 */
#ifdef SIGDANGER
#ifdef __STDC__
void biteme(int sig)
#else
void biteme(sig)
int sig;
#endif
{
#ifdef VIEW
  fprintf(stderr,"biteme() caught signal %d\n",sig);
  fprintf(stderr,"Don't panic... it's a RS6K !!!!!\n");
#endif
  (void)signal(SIGDANGER,biteme);
}
#endif /* SIGDANGER */

/* 
 * reap() : Derived from PVM  Bob Manchek
 */

#ifdef __STDC__
void reap(int sig)
#else
void
reap(sig)
        int sig;
#endif
{
        int es = 0;
        int pid;
#ifndef NOWAIT3
#if defined(RUSAGE_SELF)
        struct rusage rus;
#else
        int rus;
#endif
#endif

#ifdef  NOWAIT3
#ifdef  NOWAITPID
        pid = wait(&es);
#else
        pid = waitpid(-1, &es, WNOHANG);
        while (waitpid(-1, &es, WNOHANG) > 0);
#endif
#else   /*NOWAIT3*/
        pid = wait3(&es, WNOHANG,&rus);
        while (wait3(&es, WNOHANG,&rus) > 0);
#endif  /*NOWAIT3*/

#ifdef  SYSVSIGNAL
        (void)signal(SIGCLD, reap);
#endif
}

/*
 * data_format()
 *
 *  return the data format of the host
 */

#ifdef __STDC__
int data_format()
#else
int data_format()
#endif
{
  FILE *stream;
  char *tmp;
  char command_line[256];

  tmp = (char *)getenv("NETSOLVE_ARCH");
  if (my_arch == NULL)
  {
    fprintf(stderr,"The 'NETSOLVE_ARCH' environment variable should be set\n");
    exit(0);
  }
  strcpy(my_arch,tmp);
  return arch_to_format(my_arch);
}

/*
 * bailout()
 *
 *  The server will shutdown ... as cleanly as possible...
 *
 */
#ifdef __STDC__
int bailout()
#else
int bailout()
#endif
{
  int i;
  server_item *server;
  struct sockaddr_in server_INET_add; /* Server INET socket address  */
  struct hostent *hp;                 /* host descriptor             */
  int connection;
  int sock;
  int type;

#ifdef DEBUG
  fprintf(stderr,"ARGHHH!!!!! I'm being murdered..WHY ?? But WHY ??\n");
  fprintf(stderr,"hum... sorry about that.\n");
#endif

  /* Broadcast my death */
  type = I_DIE;
  for (i=0;i<HASH_SV;i++)
  {
    server = sv_hash_table[i];
    while(server != 0)
    {
      if ((server->server.host_status == UP)&&
      (server->server.server_status == RUNNING))
      {
        sock = connect_to_server(server->server.hostname,
                                 server->server.IPaddr);
        if (sock == -3)
          return -1;

        if (sock < 0)
        {
          server = server->next;
          continue;
        }

        /* XDR or RAW */
        if (my_data_format == (server->server.data_format))
          init_send(DATA_RAW,sock);
        else
          init_send(DATA_XDR,sock);
        write_int(&type);
        write_IPaddr(&my_address);
        end_message();
        end_com();
      }
      server = server->next;
    }
  }
  for (i=0;i<HASH_SV;i++)
  {
    server = ag_hash_table[i];
    while(server != 0)
    {
      if ((server->server.host_status == UP)&&
      (server->server.server_status == RUNNING))
      {
        sock = connect_to_server(server->server.hostname,
                                 server->server.IPaddr);
        if (sock == -3)
          return -1;

        if (sock < 0)
        {
          server = server->next;
          continue; 
        }
        /* XDR or RAW */
        if (my_data_format == (server->server.data_format))
          init_send(DATA_RAW,sock);
        else
          init_send(DATA_XDR,sock);
        write_int(&type);
        write_IPaddr(&my_address);
        end_message();
        end_com();
      }
      server = server->next;
    }
  }


  return 1;
}

/*
 * distance_request()
 *
 * If mode == 0, the its a massive request otherwise
 *  it's only a request to (char *)server
 */

#ifdef __STDC__
int distance_request(int mode,server_item *server,int delay)
#else
int distance_request(mode,server,delay)
int mode;
server_item *server;
int delay;
#endif
{
  int connection;
  int bandwidth;
  time_t date;
  FILE *stream;
  char buf1[100];
  char buf2[100];
  char command_line[200];
  int try = 5;
  int i;
  int pid;


  pid = fork();
  if (pid != 0) /* I am the father */
    return pid;

  /* I am the son */

  close(rmt_sock);
  sleep(2);
  if (mode == 1) /* Single request */
  {
    for (i=0;i<try;i++)
    {
      sleep(delay); /* I wait the delay */
      time(&date);
      sprintf(command_line,"%s/bin/%s/pong %s %d",root_path,my_arch,
              server->server.hostname,server->server.IPaddr);
      stream = popen(command_line,"r");
      if (stream == 0)
      {
        perror("distance_request() popen()");
        exit(0);
      }
      while(fgets(buf1,100,stream) == 0);
      while(fgets(buf2,100,stream) == 0);
      pclose(stream);
      connection = atoi(buf1);
      bandwidth  = atoi(buf2);
      if ((connection != -1)&&(bandwidth != -1))
        break;
      sleep(5);
    }
#ifdef DEBUG
    fprintf(stderr,"%s : %d(conn.) %d(rate)\n",server->server.hostname,connection,bandwidth);
#endif
    gen_distance_report(server->server.IPaddr,connection,bandwidth);
    exit(0);
  }
  else /* Global request */
  {
    int j;
    server_item *server;

    for (j=0;j<HASH_SV;j++)
    {
      server = sv_hash_table[j];
      while (server != NULL)
      {
        if (server->number_problems == 0)
        {
          server = server->next;
          continue;
        }
        for (i=0;i<try;i++)
        {
#ifdef DEBUG
          fprintf(stderr,"try # %d\n",i);
#endif
          sleep(delay); /* I wait the delay */
          time(&date);
          sprintf(command_line,"%s/bin/%s/pong %s %d",root_path,my_arch,
                  server->server.hostname,
                  server->server.IPaddr);
          stream = popen(command_line,"r");
          if (stream == 0)
          {
            perror("distance_request() popen()");
            exit(0);
          }
          while(fgets(buf1,100,stream) == 0);
          while(fgets(buf2,100,stream) == 0);
          pclose(stream);
          connection = atoi(buf1);
          bandwidth  = atoi(buf2);
          if ((connection != -1)&&(bandwidth != -1))
            break;
          sleep(5);
        } /* END FOR */
#ifdef DEBUG
        fprintf(stderr,"try # %d returned %d %d\n",i,connection,bandwidth);
#endif
        gen_distance_report(server->server.IPaddr,connection,bandwidth);
        server = server->next;
      } /* END WHILE */
    } /* END FOR */
  }
  exit(0);
}
